//**********************************
//*                                *
//*    ComboListView Component     *
//*    2003 by Derek van Daal      *
//*                                *
//**********************************

unit ComboListView;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls, StdCtrls, ExtCtrls, ImgList, Menus;

const
  WM_AFTER_CREATE = WM_USER + 301; // custom message

type
  TImageMode = (imNoImages, imMainColumnOnly, imAll, imImageStates);
  TSortOrder = (soAscending, soDescending);

  TComboListView = class(TWinControl)
  private
    FHeader : THeaderControl;
    FComboBox : TComboBox;
    FImages : TCustomImageList;
    FImgSortUp : TBitmap;
    FImgSortDown : TBitmap;
    FTempItems : TStrings;

    FHeaderHeight : Integer;
    FItemHeight : Integer;
    FImageSeperator : Char;
    FItemSeperator : Char;
    FTransCol : TColor;
    FImageMode : TImageMode;
    FSortSection : Byte;
    FSortOrder : TSortOrder;
    FSortable : Boolean;

    procedure ChangeSize;
    procedure DrawBorder;
    procedure RunOnce;
    procedure SectionClick(HeaderControl: THeaderControl;
      Section: THeaderSection);
    procedure DrawDefaultUpArrow;
    procedure DrawDefaultDownArrow;
    procedure DrawSection(HeaderControl: THeaderControl;
      Section: THeaderSection; const Rect: TRect; Pressed: Boolean);

    function GetImageIndex(Item, Section : Integer) : Integer;
    function GetItemText(Item, Section : Integer) : String;
    function TxtRect(x1,y1,x2,y2 : integer) : TRect;

    procedure imNoImagesDrawItem(Control: TWinControl; Index: Integer;
      Rect: TRect; State: TOwnerDrawState);
    procedure imMainColumnOnlyDrawItem(Control: TWinControl; Index: Integer;
      Rect: TRect; State: TOwnerDrawState);
    procedure imAllDrawItem(Control: TWinControl; Index: Integer;
      Rect: TRect; State: TOwnerDrawState);
    procedure imImageStatesDrawItem(Control: TWinControl; Index: Integer;
      Rect: TRect; State: TOwnerDrawState);

    procedure SectionResize(Sender: TObject;
      Button: TMouseButton; Shift: TShiftState; X, Y: Integer);

    function GetItems : TStrings;
    procedure SetItems(Value : TStrings);
    procedure SetImageSeperator(Value : Char);
    procedure SetItemSeperator(Value : Char);
    procedure SetImages(Value : TCustomImageList);
    procedure SetTransCol(Value : TColor);
    procedure SetImageMode(Value : TImageMode);
    procedure SetImgSortUp(Value : TBitmap);
    procedure SetImgSortDown(Value : TBitmap);
    procedure SetSortable(Value : Boolean);
    function GetSections : THeaderSections;
    procedure SetSections(Value : THeaderSections);
    procedure SetItemHeight(Value : Integer);
    function GetDropDownCount : Integer;
    procedure SetDropDownCount(Value : Integer);
    function GetCharCase : TEditCharCase;
    procedure SetCharCase(Value : TEditCharCase);
    function GetColor : TColor;
    procedure SetColor(Value : TColor);
    function GetFont : TFont;
    procedure SetFont(Value : TFont);
    function GetHeaderFont : TFont;
    procedure SetHeaderFont(Value : TFont);
    function GetOnChange : TNotifyEvent;
    procedure SetOnChange(Value : TNotifyEvent);
    function GetOnDropDown : TNotifyEvent;
    procedure SetOnDropDown(Value : TNotifyEvent);

    function GetTempItemText(Item, Section : Integer) : String;

    procedure SortItemsDescending(Start, Stop : Integer);
    procedure SortItemsAscending(Start, Stop : Integer);
    procedure SortAscending;
    procedure SortDescending;
    procedure ReselectItem(SelItemText : String);

    procedure WmAfterCreate(var Msg: TMessage); message WM_AFTER_CREATE;
    { Private declarations }
  protected
    procedure WMSize(var Message : TWMSize); message WM_SIZE;
    procedure WMPaint(var Message : TWMPaint); message WM_PAINT;
    procedure WMCreate(var Message : TWMCreate); Message WM_CREATE;

    { Protected declarations }
  public
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;


    { Public declarations }
  published
    property Items : TStrings read GetItems write SetItems;
    property ImgSeperator : Char read FImageSeperator write SetImageSeperator;
    property ItemSeperator : Char read FItemSeperator write SetItemSeperator;
    property Images : TCustomImageList read FImages write SetImages;
    property TransparentColor : TColor read FTransCol write SetTransCol;
    property ImageMode : TImageMode read FImageMode write SetImageMode;
    property ImgSortUp : TBitmap read FImgSortUp write SetImgSortUp;
    property ImgSortDown : TBitmap read FImgSortDown write SetImgSortDown;
    property SortOrder : TSortOrder read FSortOrder write FSortOrder;
    property SortSection : byte read FSortSection write FSortSection;
    property Sortable : Boolean read FSortable write SetSortable;
    property Columns : THeaderSections read GetSections write SetSections;
    property ItemHeight : Integer read FItemHeight write SetItemHeight;
    property DropDownCount : Integer read GetDropDownCount write SetDropDownCount;
    property CharCase : TEditCharCase read GetCharCase write SetCharCase;
    property Color : TColor read GetColor write SetColor;
    property Font : TFont read GetFont write SetFont;
    property HeaderFont : TFont read GetHeaderFont write SetHeaderFont;
    property Enabled;
    property PopupMenu;
    property Visible;
    property OnChange : TNotifyEvent read GetOnChange write SetOnChange;
    property OnDropDown : TNotifyEvent read GetOnDropDown write SetOnDropDown;
    procedure Resort;
    { Published declarations }
  end;

procedure Register;

implementation

{--------------------------------------------------------------------}
procedure Register;
begin
  RegisterComponents('Dvd Components', [TComboListView]);
end;
{--------------------------------------------------------------------}
constructor TComboListView.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  Width := 250;

  FHeader := THeaderControl.Create(Self);
  FComboBox := TComboBox.Create(Self);
  FImgSortUp := TBitmap.Create;
  FImgSortDown := TBitmap.Create;

  FHeader.Parent := Self;
  FHeader.DoubleBuffered := True;
  FHeader.Align := alNone;
  FHeader.Left := 2;
  FHeader.Top := 2;
  FHeader.OnMouseUp := SectionResize;
  FHeader.OnSectionClick := SectionClick;
  FHeader.OnDrawSection := DrawSection;

  FComboBox.Parent := Self;
  FComboBox.DoubleBuffered := True;
  FComboBox.Style := csOwnerDrawFixed;
  FComboBox.ItemHeight := 15;
  FComboBox.SendToBack;
  FComboBox.Left := 0;
  FComboBox.Top := 17;

  FHeaderHeight := FHeader.Height;
  FItemHeight := FComboBox.ItemHeight;

  FImageSeperator := ':';
  FItemSeperator := ';';
  FTransCol := clNone;
  FImageMode := imNoImages;
  FSortOrder := soAscending;
  FSortSection := 0;
  FSortable := False;
end;
{--------------------------------------------------------------------}
destructor TComboListView.Destroy;
begin
  FImgSortUp.Free;
  FImgSortDown.Free;
  FComboBox.Free;
  FHeader.Free;

  inherited Destroy;
end;
{--------------------------------------------------------------------}
procedure TComboListView.ChangeSize;
begin
  FHeader.Width := Width - 4;
  FComboBox.Width := Width;
  Height := FHeaderHeight + FItemHeight + 6;
end;
{--------------------------------------------------------------------}
procedure TComboListView.WMSize(var Message : TWMSize);
begin
  inherited;
  ChangeSize;
end;
{--------------------------------------------------------------------}
procedure TComboListView.DrawBorder;
var
  FCanvas : TCanvas;

begin
  FCanvas := TCanvas.Create;
  FCanvas.Handle := GetDC(Self.Handle);
  FCanvas.Brush.Style := bsClear;

  FCanvas.Pen.Color := cl3DDkShadow;
  FCanvas.MoveTo(1,1);
  FCanvas.LineTo(1,Height-2);
  FCanvas.MoveTo(1,1);
  FCanvas.LineTo(Width-1,1);

  FCanvas.Pen.Color := clBtnShadow;
  FCanvas.MoveTo(0,0);
  FCanvas.LineTo(0,Height);
  FCanvas.MoveTo(0,0);
  FCanvas.LineTo(Width-1,0);

  FCanvas.Pen.Color := clWhite;
  FCanvas.MoveTo(0,Height-1);
  FCanvas.LineTo(Width-1,Height-1);
  FCanvas.LineTo(Width-1,-1);

  FCanvas.Free;
end;
{--------------------------------------------------------------------}
procedure TComboListView.WMPaint(var Message : TWMPaint);
begin
  inherited;
  DrawBorder;
end;
{--------------------------------------------------------------------}
procedure TComboListView.RunOnce;
var
  k : byte;

begin
  SetWindowRgn(FComboBox.Handle, CreateRectRgn(2,2,FComboBox.Width - 2,
               FComboBox.Height - 2), True);

  If FSortable then
    Case FSortOrder of
      soAscending  : SortAscending;
      soDescending : SortDescending;
    end;

  If FHeader.Sections.Count > 0 then
    for k := 0 to FHeader.Sections.Count-1 do
      FHeader.Sections[k].Style := hsOwnerDraw;
end;
{--------------------------------------------------------------------}
procedure TComboListView.SectionClick(HeaderControl: THeaderControl;
  Section: THeaderSection);

begin
  If Sortable then
    begin
      If Section.Index <> FSortSection then
         begin
           FSortSection := Section.Index;
           SortAscending;
         end
      else
         begin
           If SortOrder = soAscending then SortDescending
             else SortAscending;
         end;
    end;
  FHeader.Repaint;
end;
{--------------------------------------------------------------------}
procedure TComboListView.DrawDefaultUpArrow;
begin
  FImgSortup.Width := 0;
  FImgSortUp.Height := 0;
  FImgSortUp.Width := 9;
  FImgSortUp.Height := 13;
  FImgSortUp.Canvas.Pen.Color := clHighLight;
  FImgSortUp.Canvas.MoveTo(1,8);
  FImgSortUp.Canvas.LineTo(8,8);
  FImgSortUp.Canvas.MoveTo(2,7);
  FImgSortUp.Canvas.LineTo(7,7);
  FImgSortUp.Canvas.MoveTo(3,6);
  FImgSortUp.Canvas.LineTo(6,6);
  FImgSortUp.Canvas.MoveTo(4,5);
  FImgSortUp.Canvas.LineTo(4,6);
end;
{--------------------------------------------------------------------}
procedure TComboListView.DrawDefaultDownArrow;
begin
  FImgSortDown.Width := 9;
  FImgSortDown.Height := 4;
  FImgSortDown.Width := 9;
  FImgSortDown.Height := 13;
  FImgSortDown.Canvas.Pen.Color := clHighLight;
  FImgSortDown.Canvas.MoveTo(1,5);
  FImgSortDown.Canvas.LineTo(8,5);
  FImgSortDown.Canvas.MoveTo(2,6);
  FImgSortDown.Canvas.LineTo(7,6);
  FImgSortDown.Canvas.MoveTo(3,7);
  FImgSortDown.Canvas.LineTo(6,7);
  FImgSortDown.Canvas.MoveTo(4,8);
  FImgSortDown.Canvas.LineTo(4,9);
end;
{--------------------------------------------------------------------}
procedure TComboListView.DrawSection(HeaderControl: THeaderControl;
  Section: THeaderSection; const Rect: TRect; Pressed: Boolean);

var
  SourceRect : TRect;
  DestRect : TRect;
  x1,y1,x2,y2 : Integer;

begin
  If FImgSortUp.Width = 0 then DrawDefaultUpArrow;
  If FImgSortDown.Width = 0 then DrawDefaultDownArrow;

  With HeaderControl.Canvas do
    begin
      If Pressed then
        begin
          If SortOrder = soAscending then
            begin
              If Section.Index = FSortSection then
                begin
                  If TextWidth(Section.Text) + 6 > Rect.Left + Rect.Right then
                    TextRect(TxtRect(Rect.Left,Rect.Top,Rect.Right - FimgSortDown.Width - 2,
                             Rect.Bottom), Rect.Left + 7, Rect.Top + 3, Section.Text)
                  else TextOut(Rect.Left + 7, Rect.Top + 3, Section.Text);
                end
              else
                begin
                  If TextWidth(Section.Text) + 6 > Rect.Left + Rect.Right then
                    TextRect(TxtRect(Rect.Left,Rect.Top,Rect.Right, Rect.Bottom),
                             Rect.Left + 7, Rect.Top + 3, Section.Text)
                  else TextOut(Rect.Left + 7, Rect.Top + 3, Section.Text);
                end;
            end
          else
            begin
              If Section.Index = FSortSection then
                begin
                  If TextWidth(Section.Text) + 6 > Rect.Left + Rect.Right then
                    TextRect(TxtRect(Rect.Left,Rect.Top,Rect.Right - FimgSortUp.Width - 2,
                             Rect.Bottom), Rect.Left + 7, Rect.Top + 3, Section.Text)
                  else TextOut(Rect.Left + 7, Rect.Top + 3, Section.Text);
                end
              else
                begin
                  If TextWidth(Section.Text) + 6 > Rect.Left + Rect.Right then
                    TextRect(TxtRect(Rect.Left,Rect.Top,Rect.Right, Rect.Bottom),
                             Rect.Left + 7, Rect.Top + 3, Section.Text)
                  else TextOut(Rect.Left + 7, Rect.Top + 3, Section.Text);
                end;
            end;

          If (Sortable) and (Section.Index = FSortSection) then
            Case FSortOrder of
              soAscending  : begin
                               If Rect.Left + TextWidth(Section.Text) + 12 + FimgSortDown.Width < Rect.Right then
                                 begin
                                   x1 := Rect.Left + TextWidth(Section.Text) + 10;
                                   x2 := Rect.Left + TextWidth(Section.Text) + 10 + FimgSortDown.Width;
                                 end
                               else
                                 begin
                                   If Rect.Right - FImgSortDown.Width - 2 > Rect.Left then
                                     x1 := Rect.Right - FImgSortDown.Width - 2
                                   else
                                     x1 := Rect.Left;
                                   x2 := Rect.Right - 2;
                                 end;

                               y1 := Rect.Top + 2 + 6 - (imgSortDown.Height div 2);
                               y2 := Rect.Top + 2 + 6 - (imgSortDown.Height div 2) + FimgSortDown.Height;

                               SourceRect := TxtRect(0,0,FimgSortDown.Width,FimgSortDown.Height);
                               DestRect := TxtRect(x1+1,y1+1,x2+1,y2+1);
                               BrushCopy(DestRect, FimgSortDown, SourceRect,
                                         imgSortDown.Canvas.Pixels[0,0]);
                             end;
              soDescending : begin
                               If Rect.Left + TextWidth(Section.Text) + 12 + FimgSortUp.Width < Rect.Right then
                                 begin
                                   x1 := Rect.Left + TextWidth(Section.Text) + 10;
                                   x2 := Rect.Left + TextWidth(Section.Text) + 10 + FimgSortUp.Width;
                                 end
                               else
                                 begin
                                   x1 := Rect.Right - FImgSortUp.Width - 2;
                                   x2 := Rect.Right - 2;
                                 end;

                               y1 := Rect.Top + 2 + 6 - (imgSortUp.Height div 2);
                               y2 := Rect.Top + 2 + 6 - (imgSortUp.Height div 2) + FimgSortUp.Height;

                               SourceRect := TxtRect(0,0,FimgSortUp.Width,FimgSortUp.Height);
                               DestRect := TxtRect(x1+1,y1+1,x2+1,y2+1);
                               BrushCopy(DestRect, FimgSortUp, SourceRect,
                                         imgSortUp.Canvas.Pixels[0,0]);
                             end;
            end;
        end
      else
        begin
          If SortOrder = soAscending then
            begin
              If (Section.Index = FSortSection) and (FSortable) then
                begin
                  If TextWidth(Section.Text) + 6 > Rect.Right - Rect.Left then
                    TextRect(TxtRect(Rect.Left,Rect.Top,Rect.Right - FimgSortDown.Width - 2,
                             Rect.Bottom), Rect.Left + 6, Rect.Top + 2, Section.Text)
                  else TextOut(Rect.Left + 6, Rect.Top + 2, Section.Text);
                end
              else
                begin
                  If TextWidth(Section.Text) + 6 > Rect.Right - Rect.Left then
                    TextRect(TxtRect(Rect.Left,Rect.Top,Rect.Right, Rect.Bottom),
                             Rect.Left + 6, Rect.Top + 2, Section.Text)
                  else TextOut(Rect.Left + 6, Rect.Top + 2, Section.Text);
                end;
            end
          else
            begin
              If (Section.Index = FSortSection) and (FSortable) then
                begin
                  If TextWidth(Section.Text) + 6 > Rect.Right - Rect.Left then
                    TextRect(TxtRect(Rect.Left,Rect.Top,Rect.Right - FimgSortUp.Width - 2,
                             Rect.Bottom), Rect.Left + 6, Rect.Top + 2, Section.Text)
                  else TextOut(Rect.Left + 6, Rect.Top + 2, Section.Text);
                end
              else
                begin
                  If TextWidth(Section.Text) + 6 > Rect.Right - Rect.Left then
                    TextRect(TxtRect(Rect.Left,Rect.Top,Rect.Right, Rect.Bottom),
                             Rect.Left + 6, Rect.Top + 2, Section.Text)
                  else TextOut(Rect.Left + 6, Rect.Top + 2, Section.Text);
                end;
            end;

          If (FSortable) and (Section.Index = FSortSection) then
            Case FSortOrder of
              soAscending  : begin
                               If Rect.Left + TextWidth(Section.Text) + 12 + FimgSortDown.Width < Rect.Right then
                                 begin
                                   x1 := Rect.Left + TextWidth(Section.Text) + 10;
                                   x2 := Rect.Left + TextWidth(Section.Text) + 10 + FimgSortDown.Width;
                                 end
                               else
                                 begin
                                   If Rect.Right - FImgSortDown.Width - 2 > Rect.Left then
                                     x1 := Rect.Right - FImgSortDown.Width - 2
                                   else
                                     x1 := Rect.Left;
                                   x2 := Rect.Right - 2;
                                 end;

                               y1 := Rect.Top + 2 + 6 - (imgSortDown.Height div 2);
                               y2 := Rect.Top + 2 + 6 - (imgSortDown.Height div 2) + FimgSortDown.Height;

                               SourceRect := TxtRect(0,0,FimgSortDown.Width,FimgSortDown.Height);
                               DestRect := TxtRect(x1,y1,x2,y2);
                               BrushCopy(DestRect, FimgSortDown, SourceRect,
                                         imgSortDown.Canvas.Pixels[0,0]);
                             end;
              soDescending : begin
                               If Rect.Left + TextWidth(Section.Text) + 12 + FimgSortUp.Width < Rect.Right then
                                 begin
                                   x1 := Rect.Left + TextWidth(Section.Text) + 10;
                                   x2 := Rect.Left + TextWidth(Section.Text) + 10 + FimgSortUp.Width;
                                 end
                               else
                                 begin
                                   x1 := Rect.Right - FImgSortUp.Width - 2;
                                   x2 := Rect.Right - 2;
                                 end;

                               y1 := Rect.Top + 2 + 6 - (imgSortUp.Height div 2);
                               y2 := Rect.Top + 2 + 6 - (imgSortUp.Height div 2) + FimgSortUp.Height;

                               SourceRect := TxtRect(0,0,FimgSortUp.Width,FimgSortUp.Height);
                               DestRect := TxtRect(x1,y1,x2,y2);
                               BrushCopy(DestRect, FimgSortUp, SourceRect,
                                         imgSortUp.Canvas.Pixels[0,0]);
                             end;
            end;
        end;
    end;
end;
{--------------------------------------------------------------------}
function TComboListView.GetImageIndex(Item, Section : Integer) : Integer;
var
  Line : String;
  k : Integer;
  ItemPos : Integer;
  ImgPos : Integer;
  StrResult : ShortString;
  IntResult : Integer;
  Err : Boolean;
  ErrCode : Integer;

begin
  Err := False;
  If (Item >= 0) and (Item < FComboBox.Items.Count) and (Section >= 0) then
    begin
      Inc(Section);
      Line := FComboBox.Items[Item];
      For k := 1 to Section-1 do
        begin
          ItemPos := pos(ItemSeperator, Line);
          If ItemPos > 0 then Delete(Line, 1, ItemPos);
        end;

      ImgPos := Pos(ImgSeperator, Line);
      StrResult := Copy(Line, 1, ImgPos - 1);
      If StrResult <> '' then Val(StrResult, IntResult, ErrCode)
        else Result := -1;
      If ErrCode <> 0 then Err := True;
    end else Err := True;

  If not Err then Result := IntResult
    else Result := -1;
end;
{--------------------------------------------------------------------}
function TComboListView.GetItemText(Item, Section : Integer) : String;
var
  Line : String;
  k : Integer;
  ItemPos : Integer;
  ImgPos : Integer;
  StrResult : ShortString;

begin
  If (Item >= 0) and (Item < FComboBox.Items.Count) and (Section >= 0) then
    begin
      Inc(Section);
      Line := FComboBox.Items[Item];
      For k := 1 to Section-1 do
        begin
          ItemPos := pos(ItemSeperator, Line);
          If ItemPos > 0 then Delete(Line, 1, ItemPos);
        end;

      ImgPos := Pos(ImgSeperator, Line);
      Delete(Line, 1, ImgPos);
      ItemPos := Pos(ItemSeperator, Line);
      StrResult := Copy(Line, 1, ItemPos - 1);
      If StrResult <> '' then Result := StrResult
        else Result := '';
    end else Result := '';
end;
{--------------------------------------------------------------------}
function TComboListView.TxtRect(x1,y1,x2,y2 : integer) : TRect;
begin
  Result := Rect(x1,y1,x2,y2);
end;
{--------------------------------------------------------------------}
procedure TComboListView.imNoImagesDrawItem(Control: TWinControl; Index: Integer;
  Rect: TRect; State: TOwnerDrawState);

var
  FontHeight : Integer;
  TextY : Integer;
  k : Integer;
  ItemWidth : Integer;
  ColumnLeft : Integer;
  Offset : Byte;

begin
  With (Control as TComboBox).Canvas do
    begin
      FontHeight := Round(Font.Size * Font.PixelsPerInch / 72) + 2;
      TextY := Rect.Top + ((Rect.Bottom - Rect.Top) div 2) - FontHeight div 2;

      FillRect(Rect);

      ColumnLeft := 0;
      For k := 0 to FHeader.Sections.Count-1 do
        begin
          If k > 0 then ColumnLeft := ColumnLeft + FHeader.Sections[k-1].Width;
          ItemWidth := TextWidth(GetItemText(Index, k));

          If ItemWidth + 4 <= FHeader.Sections[k].Width then
            begin
              TextOut(ColumnLeft + 4, TextY, GetItemText(Index, k));
            end
          else
            begin
              If FHeader.Sections[k].Width > 10 then
                TextOut(ColumnLeft + FHeader.Sections[k].Width - 7, TextY, '...');
              Offset := 8;
              If FHeader.Sections[k].Width > 10 then
              TextRect(TxtRect(ColumnLeft + 4,
                               Rect.Top,
                               ColumnLeft + FHeader.Sections[k].Width - OffSet,
                               Rect.Bottom),
                               ColumnLeft + 4, TextY, GetItemText(Index, k));
            end;
        end;
    end;
end;
{--------------------------------------------------------------------}
procedure TComboListView.imMainColumnOnlyDrawItem(Control: TWinControl;
  Index: Integer; Rect: TRect; State: TOwnerDrawState);

var
  Bitmap : TBitmap;
  FontHeight : Integer;
  TextY : Integer;
  BitmapY : Integer;
  k : Integer;
  ImgWidth : Integer;
  ItemWidth : Integer;
  ColumnLeft : Integer;
  Offset : Byte;

begin
  Bitmap := TBitmap.Create;
  If FImages <> nil then ImgWidth := FImages.Width
    else ImgWidth := 0;

  With (Control as TComboBox).Canvas do
    begin
      FontHeight := Round(Font.Size * Font.PixelsPerInch / 72) + 2;
      TextY := Rect.Top + ((Rect.Bottom - Rect.Top) div 2) - FontHeight div 2;
      BitmapY := Rect.Top + ((Rect.Bottom - Rect.Top) div 2) - (Bitmap.Height div 2) - 8;

      FillRect(Rect);

      ColumnLeft := 0;
      For k := 0 to FHeader.Sections.Count-1 do
        begin
          If k > 0 then ColumnLeft := ColumnLeft + FHeader.Sections[k-1].Width;
          ItemWidth := ImgWidth + TextWidth(GetItemText(Index, k));

          If k = 0 then
            begin
              If ItemWidth + 8 <= FHeader.Sections[k].Width then
                begin
                  TextOut(ColumnLeft + ImgWidth + 8, TextY, GetItemText(Index, k));
                  If FImages <> nil then
                    FImages.GetBitmap(GetImageIndex(Index, k), Bitmap);
                  Bitmap.Transparent := True;
                  Draw(ColumnLeft + 4, BitmapY, Bitmap);
                  Bitmap.Assign(nil);
                end
              else
                begin
                  If FImages <> nil then
                    FImages.GetBitmap(GetImageIndex(Index, k), Bitmap);
                  Bitmap.Transparent := True;
                  If Bitmap.Width > FHeader.Sections[k].Width-3 then
                    If FHeader.Sections[k].Width-3 > 0 then
                      Bitmap.Width := FHeader.Sections[k].Width-3
                        else Bitmap.Width := 0;

                  If FHeader.Sections[k].Width > 30  then
                    TextOut(ColumnLeft + FHeader.Sections[k].Width - 7, TextY, '...');
                  Offset := 8;
                  If FHeader.Sections[k].Width > 10 then
                  TextRect(TxtRect(ColumnLeft + ImgWidth + 8,
                                   Rect.Top,
                                   ColumnLeft + FHeader.Sections[k].Width - OffSet,
                                   Rect.Bottom),
                                   ColumnLeft + ImgWidth + 8, TextY, GetItemText(Index, k));
                  Draw(ColumnLeft + 4, BitmapY, Bitmap);
                  Bitmap.Assign(nil);
                end;
            end
          else
            begin
              If ItemWidth - 12 <= FHeader.Sections[k].Width then
            begin
              TextOut(ColumnLeft + 4, TextY, GetItemText(Index, k));
            end
          else
            begin
              If FHeader.Sections[k].Width > 10 then
                TextOut(ColumnLeft + FHeader.Sections[k].Width - 7, TextY, '...');
              Offset := 8;
              TextRect(TxtRect(ColumnLeft + 4,
                               Rect.Top,
                               ColumnLeft + FHeader.Sections[k].Width - OffSet,
                               Rect.Bottom),
                               ColumnLeft + 4, TextY, GetItemText(Index, k));
              end;
            end;
        end;
    end;
  Bitmap.Free;
end;
{--------------------------------------------------------------------}
procedure TComboListView.imAllDrawItem(Control: TWinControl; Index: Integer;
  Rect: TRect; State: TOwnerDrawState);

var
  Bitmap : TBitmap;
  FontHeight : Integer;
  TextY : Integer;
  BitmapY : Integer;
  k : Integer;
  ImgWidth : Integer;
  ItemWidth : Integer;
  ColumnLeft : Integer;
  Offset : Byte;

begin
  Bitmap := TBitmap.Create;
  If FImages <> nil then ImgWidth := FImages.Width
    else ImgWidth := 0;

  With (Control as TComboBox).Canvas do
    begin
      FontHeight := Round(Font.Size * Font.PixelsPerInch / 72) + 2;
      TextY := Rect.Top + ((Rect.Bottom - Rect.Top) div 2) - FontHeight div 2;
      BitmapY := Rect.Top + ((Rect.Bottom - Rect.Top) div 2) - (Bitmap.Height div 2) - 8;

      FillRect(Rect);

      ColumnLeft := 0;
      For k := 0 to FHeader.Sections.Count-1 do
        begin
          If k > 0 then ColumnLeft := ColumnLeft + FHeader.Sections[k-1].Width;
          ItemWidth := ImgWidth + TextWidth(GetItemText(Index, k));

          If ItemWidth + 8 <= FHeader.Sections[k].Width then
            begin
              TextOut(ColumnLeft + ImgWidth + 8, TextY, GetItemText(Index, k));
              If FImages <> nil then
                FImages.GetBitmap(GetImageIndex(Index, k), Bitmap);
              Bitmap.Transparent := True;
              Draw(ColumnLeft + 4, BitmapY, Bitmap);
              Bitmap.Assign(nil);
            end
          else
            begin
              If FImages <> nil then
                FImages.GetBitmap(GetImageIndex(Index, k), Bitmap);
              Bitmap.Transparent := True;
              If Bitmap.Width > FHeader.Sections[k].Width-3 then
                If FHeader.Sections[k].Width-3 > 0 then
                  Bitmap.Width := FHeader.Sections[k].Width-3
                    else Bitmap.Width := 0;

              If FHeader.Sections[k].Width > 30  then
                TextOut(ColumnLeft + FHeader.Sections[k].Width - 7, TextY, '...');
              Offset := 8;
              If FHeader.Sections[k].Width > 10 then
              TextRect(TxtRect(ColumnLeft + ImgWidth + 8,
                               Rect.Top,
                               ColumnLeft + FHeader.Sections[k].Width - OffSet,
                               Rect.Bottom),
                               ColumnLeft + ImgWidth + 8, TextY, GetItemText(Index, k));
              Draw(ColumnLeft + 4, BitmapY, Bitmap);
              Bitmap.Assign(nil);
            end;
        end;
    end;
  Bitmap.Free;
end;
{--------------------------------------------------------------------}
procedure TComboListView.imImageStatesDrawItem(Control: TWinControl;
  Index: Integer; Rect: TRect; State: TOwnerDrawState);

var
  Bitmap : TBitmap;
  FontHeight : Integer;
  TextY : Integer;
  BitmapY : Integer;
  k : Integer;
  ImgWidth : Integer;
  ItemWidth : Integer;
  ColumnLeft : Integer;
  Offset : Byte;

begin
  Bitmap := TBitmap.Create;
  If FImages <> nil then ImgWidth := FImages.Width
    else ImgWidth := 0;

  With (Control as TComboBox).Canvas do
    begin
      FontHeight := Round(Font.Size * Font.PixelsPerInch / 72) + 2;
      TextY := Rect.Top + ((Rect.Bottom - Rect.Top) div 2) - FontHeight div 2;
      BitmapY := Rect.Top + ((Rect.Bottom - Rect.Top) div 2) - (Bitmap.Height div 2) - 8;

      FillRect(Rect);

      ColumnLeft := 0;
      For k := 0 to FHeader.Sections.Count-1 do
        begin
          If k > 0 then ColumnLeft := ColumnLeft + FHeader.Sections[k-1].Width;
          ItemWidth := ImgWidth + TextWidth(GetItemText(Index, k));

          If GetImageIndex(Index,k) > -1 then
            begin
              If ItemWidth + 8 <= FHeader.Sections[k].Width then
                begin
                  TextOut(ColumnLeft + ImgWidth + 8, TextY, GetItemText(Index, k));
                  If FImages <> nil then
                    FImages.GetBitmap(GetImageIndex(Index, k), Bitmap);
                  Bitmap.Transparent := True;
                  Draw(ColumnLeft + 4, BitmapY, Bitmap);
                  Bitmap.Assign(nil);
                end
              else
                begin
                  If FImages <> nil then
                    FImages.GetBitmap(GetImageIndex(Index, k), Bitmap);
                  Bitmap.Transparent := True;
                  If Bitmap.Width > FHeader.Sections[k].Width-3 then
                    If FHeader.Sections[k].Width-3 > 0 then
                      Bitmap.Width := FHeader.Sections[k].Width-3
                        else Bitmap.Width := 0;

                  If FHeader.Sections[k].Width > 30  then
                    TextOut(ColumnLeft + FHeader.Sections[k].Width - 7, TextY, '...');
                  Offset := 8;
                  If FHeader.Sections[k].Width > 10 then
                  TextRect(TxtRect(ColumnLeft + ImgWidth + 8,
                                   Rect.Top,
                                   ColumnLeft + FHeader.Sections[k].Width - OffSet,
                                   Rect.Bottom),
                                   ColumnLeft + ImgWidth + 8, TextY, GetItemText(Index, k));
                  Draw(ColumnLeft + 4, BitmapY, Bitmap);
                  Bitmap.Assign(nil);
                end;
            end
          else
            begin
              If ItemWidth - 12 <= FHeader.Sections[k].Width then
            begin
              TextOut(ColumnLeft + 4, TextY, GetItemText(Index, k));
            end
          else
            begin
              If FHeader.Sections[k].Width > 10 then
                TextOut(ColumnLeft + FHeader.Sections[k].Width - 7, TextY, '...');
              Offset := 8;
              If FHeader.Sections[k].Width > 10 then
              TextRect(TxtRect(ColumnLeft + 4,
                               Rect.Top,
                               ColumnLeft + FHeader.Sections[k].Width - OffSet,
                               Rect.Bottom),
                               ColumnLeft + 4, TextY, GetItemText(Index, k));
              end;
            end;
        end;
    end;
  Bitmap.Free;
end;

{--------------------------------------------------------------------}
procedure TComboListView.SectionResize(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);

var
  k : integer;

begin
  If FHeader.Sections.Count > 0 then
    begin
      for k := 0 to FHeader.Sections.Count-1 do
        If FHeader.Sections[k].Left + FHeader.Sections[k].Width >= Width - 4 then
          FHeader.Sections[k].Width := Width - 4 - FHeader.Sections[k].Left;
    end;

  FComboBox.Repaint;
end;
{--------------------------------------------------------------------}
function TComboListView.GetItems : TStrings;
begin
  Result := FComboBox.Items;
end;
{--------------------------------------------------------------------}
procedure TComboListView.SetItems(Value : TStrings);
begin
  FComboBox.Items := Value;
end;
{--------------------------------------------------------------------}
procedure TComboListView.SetImageSeperator(Value : Char);
var
  k : Integer;
  SepFound : Boolean;
  SepPos : Integer;
  TempLine : String;

begin
  If Value <> FItemSeperator then
    begin
      If (FComboBox.Items.Count > 0) and (FImageSeperator <> Value) then
        for k := 0 to FComboBox.Items.Count - 1 do
          begin
            TempLine := FComboBox.Items[k];
            repeat
              SepFound := False;
              SepPos := Pos(FImageSeperator, TempLine);
              If SepPos > 0 then
                begin
                  TempLine[SepPos] := Value;
                  SepFound := True;
                end;
            Until not SepFound;
            FComboBox.Items[k] := TempLine;
          end;
      FImageSeperator := Value;
    end else MessageDlg('Cannot have the same value as ItemSeperator.',
                        mtError, [mbOk], 0);
end;
{--------------------------------------------------------------------}
procedure TComboListView.SetItemSeperator(Value : Char);
var
  k : Integer;
  SepFound : Boolean;
  SepPos : Integer;
  TempLine : String;

begin
  If Value <> FImageSeperator then
    begin
      If (FComboBox.Items.Count > 0) and (FItemSeperator <> Value) then
        for k := 0 to FComboBox.Items.Count - 1 do
          begin
            TempLine := FComboBox.Items[k];
            repeat
              SepFound := False;
              SepPos := Pos(FItemSeperator, TempLine);
              If SepPos > 0 then
                begin
                  TempLine[SepPos] := Value;
                  SepFound := True;
                end;
            Until not SepFound;
            FComboBox.Items[k] := TempLine;
          end;
      FItemSeperator := Value;
    end else MessageDlg('Cannot have the same value as ImgSeperator.',
                        mtError, [mbOk], 0);
end;
{--------------------------------------------------------------------}
procedure TComboListView.SetImages(Value : TCustomImageList);
begin
  FImages := Value;
  If FImages <> nil then FImages.BkColor := FTransCol;
end;
{--------------------------------------------------------------------}
procedure TComboListView.SetTransCol(Value : TColor);
begin
  FTransCol := Value;
  If FImages <> nil then FImages.BkColor := FTransCol;
end;
{--------------------------------------------------------------------}
procedure TComboListView.SetImageMode(Value : TImageMode);
begin
  FImageMode := Value;
  Case FImageMode of
    imNoImages       : FComboBox.OnDrawItem := imNoImagesDrawItem;
    imMainColumnOnly : FComboBox.OnDrawItem := imMainColumnOnlyDrawItem;
    imAll            : FComboBox.OnDrawItem := imAllDrawItem;
    imImageStates    : FComboBox.OnDrawItem := imImageStatesDrawItem;
  end;

end;
{--------------------------------------------------------------------}
procedure TComboListView.SetImgSortUp(Value : TBitmap);
begin
  FImgSortUp.Assign(Value);
  FHeader.Repaint;
end;
{--------------------------------------------------------------------}
procedure TComboListView.SetImgSortDown(Value : TBitmap);
begin
  FImgSortDown.Assign(Value);
  FHeader.Repaint;
end;
{--------------------------------------------------------------------}
procedure TComboListView.SetSortable(Value : Boolean);
begin
  FSortable := Value;
end;
{--------------------------------------------------------------------}
function TComboListView.GetSections : THeaderSections;
begin
  Result := FHeader.Sections;
end;
{--------------------------------------------------------------------}
procedure TComboListView.SetSections(Value : THeaderSections);
begin
  FHeader.Sections.Assign(Value);
end;
{--------------------------------------------------------------------}
procedure TComboListView.SetItemHeight(Value : Integer);
begin
  FComboBox.ItemHeight := Value;
  FItemHeight := Value;
  ChangeSize;
end;
{--------------------------------------------------------------------}
function TComboListView.GetDropDownCount : Integer;
begin
  Result := FComboBox.DropDownCount;
end;
{--------------------------------------------------------------------}
procedure TComboListView.SetDropDownCount(Value : Integer);
begin
  FComboBox.DropDownCount := Value;
end;
{--------------------------------------------------------------------}
function TComboListView.GetCharCase : TEditCharCase;
begin
  Result := FComboBox.CharCase;
end;
{--------------------------------------------------------------------}
procedure TComboListView.SetCharCase(Value : TEditCharCase);
begin
  FcomboBox.CharCase := Value;
end;
{--------------------------------------------------------------------}
function TComboListView.GetColor : TColor;
begin
  Result := FComboBox.Color;
end;
{--------------------------------------------------------------------}
procedure TComboListView.SetColor(Value : TColor);
begin
  FcomboBox.Color := Value;
end;
{--------------------------------------------------------------------}
function TComboListView.GetFont : TFont;
begin
  Result := FComboBox.Font;
end;
{--------------------------------------------------------------------}
procedure TComboListView.SetFont(Value : TFont);
begin
  FcomboBox.Font := Value;
end;
{--------------------------------------------------------------------}
function TComboListView.GetHeaderFont : TFont;
begin
  Result := FHeader.Font;
end;
{--------------------------------------------------------------------}
procedure TComboListView.SetHeaderFont(Value : TFont);
begin
  FHeader.Font := Value;
end;
{--------------------------------------------------------------------}
function TComboListView.GetOnChange : TNotifyEvent;
begin
  Result := FComboBox.OnChange;
end;
{--------------------------------------------------------------------}
procedure TComboListView.SetOnChange(Value : TNotifyEvent);
begin
  FComboBox.OnChange := Value;
end;
{--------------------------------------------------------------------}
function TComboListView.GetOnDropDown : TNotifyEvent;
begin
  Result := FComboBox.OnDropDown;
end;
{--------------------------------------------------------------------}
procedure TComboListView.SetOnDropDown(Value : TNotifyEvent);
begin
  FComboBox.OnDropDown := Value;
end;
{--------------------------------------------------------------------}
function TComboListView.GetTempItemText(Item, Section : Integer) : String;
var
  Line : String;
  k : Integer;
  ItemPos : Integer;
  ImgPos : Integer;
  StrResult : ShortString;

begin
  If (Item >= 0) and (Item < FTempItems.Count) and (Section >= 0) then
    begin
      Inc(Section);
      Line := FTempItems[Item];
      For k := 1 to Section-1 do
        begin
          ItemPos := pos(ItemSeperator, Line);
          If ItemPos > 0 then Delete(Line, 1, ItemPos);
        end;

      ImgPos := Pos(ImgSeperator, Line);
      Delete(Line, 1, ImgPos);
      ItemPos := Pos(ItemSeperator, Line);
      StrResult := Copy(Line, 1, ItemPos - 1);
      If StrResult <> '' then Result := StrResult
        else Result := '';
    end else Result := '';
end;
{--------------------------------------------------------------------}
procedure TComboListView.SortItemsAscending(Start, Stop : Integer);
Var
  Left : Integer;
  Right : Integer;
  Mid : Integer;
  Pivot : String;
  Temp : String;

begin
  Left := Start;
  Right := Stop;
  Mid := (Start + Stop) div 2;

  Pivot := GetTempItemText(Mid, FSortSection);
  Repeat
    While GetTempItemText(Left, FSortSection) < Pivot do inc(Left);
    While Pivot < GetTempItemText(Right, FSortSection) do dec(Right);
    If Left <= Right then
      begin
        Temp := FTempItems[Left];
        FTempItems[Left] := FTempItems[Right];
        FTempItems[Right] := Temp;
        inc(Left);
        dec(Right);
      end;
  Until Left > Right;

  If Start < Right then SortItemsAscending(Start, Right);
  If Left < Stop Then SortItemsAscending(Left, Stop);
end;
{--------------------------------------------------------------------}
procedure TComboListView.SortItemsDescending(Start, Stop : Integer);
Var
  Left : Integer;
  Right : Integer;
  Mid : Integer;
  Pivot : String;
  Temp : String;

begin
  Left := Start;
  Right := Stop;
  Mid := (Start + Stop) div 2;

  Pivot := GetTempItemText(Mid, FSortSection);
  Repeat
    While GetTempItemText(Left, FSortSection) > Pivot do inc(Left);
    While Pivot > GetTempItemText(Right, FSortSection) do dec(Right);
    If Left <= Right then
      begin
        Temp := FTempItems[Left];
        FTempItems[Left] := FTempItems[Right];
        FTempItems[Right] := Temp;
        inc(Left);
        dec(Right);
      end;
  Until Left > Right;

  If Start < Right then SortItemsDescending(Start, Right);
  If Left < Stop Then SortItemsDescending(Left, Stop);
end;
{--------------------------------------------------------------------}
procedure TComboListView.SortAscending;
var
  FSelectedItem : String;

begin
  FSelectedItem := FComboBox.Items[FComboBox.ItemIndex];
  FSortOrder := soAscending;
  FTempItems := TStringList.Create;
  FTempItems.Assign(FComboBox.Items);
  SortItemsAscending(0, FTempItems.Count-1);
  FComboBox.Items.Assign(FTempItems);
  FTempItems.Clear;
  FTempItems.Free;
  ReselectItem(FSelectedItem);
end;
{--------------------------------------------------------------------}
procedure TComboListView.SortDescending;
var
  FSelectedItem : String;

begin
  FSelectedItem := FComboBox.Items[FComboBox.ItemIndex];
  FSortOrder := soDescending;
  FTempItems := TStringList.Create;
  FTempItems.Assign(FComboBox.Items);
  SortItemsDescending(0, FTempItems.Count-1);
  FComboBox.Items.Assign(FTempItems);
  FTempItems.Clear;
  FTempItems.Free;
  ReselectItem(FSelectedItem);
end;
{--------------------------------------------------------------------}
procedure TComboListView.ReselectItem(SelItemText : String);
var
  k : integer;
  found : Boolean;

begin
  if FComboBox.Items.Count > 0 then
    begin
      k := 0;
      Found := False;
      Repeat
        If SelItemText = FComboBox.Items[k] then
          begin
            Found := True;
            FComboBox.ItemIndex := k;
          end;
          inc(k);
      Until (k > FComboBox.Items.Count) or (Found);
    end;
end;
{--------------------------------------------------------------------}
procedure TComboListView.Resort;
begin
  If FSortable then
    Case FSortOrder of
      soAscending  : SortAscending;
      soDescending : SortDescending;
    end;
end;
{--------------------------------------------------------------------}
procedure TComboListView.WMCreate(var Message : TWMCreate);
begin
  inherited;
  PostMessage(Self.Handle, WM_AFTER_CREATE, 0, 0);
end;
{--------------------------------------------------------------------}
procedure TComboListView.WmAfterCreate(var Msg: TMessage);
begin
  RunOnce;
end;
{--------------------------------------------------------------------}
end.
