
{*******************************************************}
{                                                       }
{       4th GUI Library (VCL Edition)                   }
{       Buttons Unit                                    }
{                                                       }
{       Copyright (c) 1996,2003 Sergiy S. Tkachenko     }
{       e-mail: support@4thfebruary.tk                  }
{          Web: www.4thfebruary.tk                      }
{                                                       }
{*******************************************************}

unit FthBtns;

{$I FTHVER.INC}

interface

uses
{$IFDEF WIN32}
  Windows,
{$ELSE}
  WinTypes, WinProcs,
{$ENDIF}
  Messages, SysUtils, Classes, Graphics, Controls, Forms,
{$IFDEF DELPHI4_UP}
  ImgList, ActnList,
{$ENDIF}
  CommCtrl, Buttons, Menus, ExtCtrls;

type
  TFourthStandardButtonSeparatorStyle = (ssVertLines, ssRectangle);

  TFourthStandardButtonOptions = class(TPersistent)
  private
    FAlignment: TAlignment;
    FBackgroundColor: TColor;
    FDisabledFont: TFont;
    FDrawLeftLine: Boolean;
    FDrawRightLine: Boolean;
    FFont: TFont;
    FHotTrackCursor: TCursor;
    FHotTrackFont: TFont;
    FLevel1LightColor: TColor;
    FLevel1ShadowColor: TColor;
    FLevel2LightColor: TColor;
    FLevel2ShadowColor: TColor;
    FLevel3Color: TColor;
    FOnChange: TNotifyEvent;
    FSeparatorStyle: TFourthStandardButtonSeparatorStyle;
    FTransparent: Boolean;
    FVertLineColor: TColor;
    procedure SetAlignment(Value: TAlignment);
    procedure SetBackgroundColor(Value: TColor);
    procedure SetDisabledFont(Value: TFont);
    procedure SetDrawLeftLine(Value: Boolean);
    procedure SetDrawRightLine(Value: Boolean);
    procedure SetFont(Value: TFont);
    procedure SetHotTrackCursor(Value: TCursor);
    procedure SetHotTrackFont(Value: TFont);
    procedure SetLevel1LightColor(Value: TColor);
    procedure SetLevel1ShadowColor(Value: TColor);
    procedure SetLevel2LightColor(Value: TColor);
    procedure SetLevel2ShadowColor(Value: TColor);
    procedure SetLevel3Color(Value: TColor);
    procedure SetSeparatorStyle(Value: TFourthStandardButtonSeparatorStyle);
    procedure SetTransparent(Value: Boolean);
    procedure SetVertLineColor(Value: TColor);
  protected
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    procedure FontChanged(Sender: TObject);
  public
    constructor Create;
  published
    property Alignment: TAlignment read FAlignment write SetAlignment;
    property BackgroundColor: TColor read FBackgroundColor write SetBackgroundColor;
    property DisabledFont: TFont read FDisabledFont write SetDisabledFont;
    property DrawLeftLine: Boolean read FDrawLeftLine write SetDrawLeftLine;
    property DrawRightLine: Boolean read FDrawRightLine write SetDrawRightLine;
    property Font: TFont read FFont write SetFont;
    property HotTrackCursor: TCursor read FHotTrackCursor write SetHotTrackCursor;
    property HotTrackFont: TFont read FHotTrackFont write SetHotTrackFont;
    property Level1LightColor: TColor read FLevel1LightColor write SetLevel1LightColor;
    property Level1ShadowColor: TColor read FLevel1ShadowColor write SetLevel1ShadowColor;
    property Level2LightColor: TColor read FLevel2LightColor write SetLevel2LightColor;
    property Level2ShadowColor: TColor read FLevel2ShadowColor write SetLevel2ShadowColor;
    property Level3Color: TColor read FLevel3Color write SetLevel3Color;
    property SeparatorStyle: TFourthStandardButtonSeparatorStyle read FSeparatorStyle write SetSeparatorStyle;
    property Transparent: Boolean read FTransparent write SetTransparent;
    property VertLineColor: TColor read FVertLineColor write SetVertLineColor;
  end;

  TFourthButtonState = (bsUp, bsDisabled, bsDown, bsExclusive);

  TFourthStandardButton = class(TCustomControl)
  private
    FCaption: TCaption;
    FDragging: Boolean;
    FDropMenu: TPopupMenu;
    FModalResult: TModalResult;
    FMouseInControl: Boolean;
    FOldCursor: TCursor;
    FOptions: TFourthStandardButtonOptions;
    FOnMouseEnter: TNotifyEvent;
    FOnMouseLeave: TNotifyEvent;
    procedure SetCaption(Value: TCaption);
    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure CMSysColorChange(var Message: TMessage); message CM_SYSCOLORCHANGE;
    procedure CMButtonPressed(var Message: TMessage); message CM_BUTTONPRESSED;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
  protected
    FState: TFourthButtonState;
    procedure Loaded; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure OptionsChange(Sender: TObject);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Paint; override;
    procedure Click; override;
  published
{$IFDEF DELPHI4_UP}
    property Action;
{$ENDIF}
    property Align;
{$IFDEF DELPHI4_UP}
    property Anchors;
{$ENDIF}
    property Caption: TCaption read FCaption write SetCaption;
    property Cursor;
    property DragCursor;
{$IFDEF DELPHI4_UP}
    property DragKind;
{$ENDIF}
    property DragMode;
    property DropDownMenu: TPopupMenu read FDropMenu write FDropMenu;
    property Enabled;
    property Height;
    property Hint;
    property Left;
    property ModalResult: TModalResult read FModalResult write FModalResult default 0;
    property Options: TFourthStandardButtonOptions read FOptions write FOptions;
    property ParentShowHint;
    property Name;
    property ShowHint;
    property Top;
    property Tag;
    property Visible;
    property Width;
{$IFDEF DELPHI4_UP}
    property OnCanResize;
{$ENDIF}
    property OnClick;
{$IFDEF DELPHI4_UP}
    property OnConstrainedResize;
{$ENDIF}
{$IFDEF DELPHI5_UP}
    property OnContextPopup;
{$ENDIF}
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
{$IFDEF DELPHI4_UP}
    property OnEndDock;
{$ENDIF}
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseEnter: TNotifyEvent read FOnMouseEnter write FOnMouseEnter;
    property OnMouseLeave: TNotifyEvent read FOnMouseLeave write FOnMouseLeave;
    property OnMouseMove;
    property OnMouseUp;
{$IFDEF DELPHI4_UP}
    property OnResize;
    property OnStartDock;
{$ENDIF}
    property OnStartDrag;
  end;

{ TFourthCaptionLikeButton }

  TFourthCaptionLikeButtonStyle = (clbsNone, clbsClose, clbsHelp, clbsMaximize,
    clbsMinimize, clbsRestore);

  TFourthCaptionLikeButton = class(TCustomControl)
  private
    FDragging: Boolean;
    FStyle: TFourthCaptionLikeButtonStyle;
    procedure SetStyle(Value: TFourthCaptionLikeButtonStyle);
    function GetSystemWidth: Integer;
    function GetSystemHeight: Integer;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure CMSysColorChange(var Message: TMessage); message CM_SYSCOLORCHANGE;
  protected
    FState: TFourthButtonState;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Paint; override;
    procedure SetSystemSize;
  published
    property Enabled;
    property Height;
    property Hint;
    property ParentShowHint;
    property Style: TFourthCaptionLikeButtonStyle read FStyle write SetStyle;
    property ShowHint;
    property Tag;
    property Top;
    property Width;
    property OnClick;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  end;

implementation

{ TFourthStandardButtonOptions }

constructor TFourthStandardButtonOptions.Create;
begin
  FAlignment := taCenter;
  FBackgroundColor := cl3DDkShadow;
  FDisabledFont := TFont.Create;
  FDisabledFont.Color := clBtnShadow;
  FDisabledFont.OnChange := FontChanged;
  FFont := TFont.Create;
  FFont.Color := clBtnHighlight;
  FFont.OnChange := FontChanged;
  FHotTrackCursor := crDefault;
  FHotTrackFont := TFont.Create;
  FHotTrackFont.Color := clBtnText;
  FHotTrackFont.OnChange := FontChanged;
  FLevel1LightColor := {$IFDEF WIN32}cl3DLight{$ELSE}clHighlight{$ENDIF};
  FLevel1ShadowColor := {$IFDEF WIN32}cl3DDkShadow{$ELSE}clBlack{$ENDIF};
  FLevel2LightColor := clBtnHighlight;
  FLevel2ShadowColor := clBtnShadow;
  FLevel3Color := clBtnFace;
  FDrawLeftLine := True;
  FDrawRightLine := True;
  FSeparatorStyle := ssVertLines;
  FTransparent := False;
  FVertLineColor := clBtnFace;
end;

procedure TFourthStandardButtonOptions.SetAlignment(Value: TAlignment);
begin
  if (FAlignment <> Value) then
  begin
    FAlignment := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthStandardButtonOptions.SetBackgroundColor(Value: TColor);
begin
  if (FBackgroundColor <> Value) then
  begin
    FBackgroundColor := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthStandardButtonOptions.SetDisabledFont(Value: TFont);
begin
  if (FDisabledFont <> Value) then
  begin
    FDisabledFont.Assign(Value);
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthStandardButtonOptions.SetDrawLeftLine(Value: Boolean);
begin
  if (FDrawLeftLine <> Value) then
  begin
    FDrawLeftLine := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthStandardButtonOptions.SetDrawRightLine(Value: Boolean);
begin
  if (FDrawRightLine <> Value) then
  begin
    FDrawRightLine := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthStandardButtonOptions.SetFont(Value: TFont);
begin
  if (FFont <> Value) then
  begin
    FFont.Assign(Value);
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthStandardButtonOptions.SetHotTrackCursor(Value: TCursor);
begin
  if (FHotTrackCursor <> Value) then
  begin
    FHotTrackCursor := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthStandardButtonOptions.SetHotTrackFont(Value: TFont);
begin
  if (FHotTrackFont <> Value) then
  begin
    FHotTrackFont.Assign(Value);
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthStandardButtonOptions.SetLevel1LightColor(Value: TColor);
begin
  if (FLevel1LightColor <> Value) then
  begin
    FLevel1LightColor := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthStandardButtonOptions.SetLevel1ShadowColor(Value: TColor);
begin
  if (FLevel1ShadowColor <> Value) then
  begin
    FLevel1ShadowColor := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthStandardButtonOptions.SetLevel2LightColor(Value: TColor);
begin
  if (FLevel2LightColor <> Value) then
  begin
    FLevel2LightColor := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthStandardButtonOptions.SetLevel2ShadowColor(Value: TColor);
begin
  if (FLevel2ShadowColor <> Value) then
  begin
    FLevel2ShadowColor := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthStandardButtonOptions.SetLevel3Color(Value: TColor);
begin
  if (FLevel3Color <> Value) then
  begin
    FLevel3Color := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthStandardButtonOptions.SetTransparent(Value: Boolean);
begin
  if (FTransparent <> Value) then
  begin
    FTransparent := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthStandardButtonOptions.SetSeparatorStyle(Value: TFourthStandardButtonSeparatorStyle);
begin
  if (FSeparatorStyle <> Value) then
  begin
    FSeparatorStyle := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthStandardButtonOptions.SetVertLineColor(Value: TColor);
begin
  if (FVertLineColor <> Value) then
  begin
    FVertLineColor := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthStandardButtonOptions.FontChanged(Sender: TObject);
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

{ TFourthStandardButton }

constructor TFourthStandardButton.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csSetCaption, csCaptureMouse, csOpaque,
    csClickEvents{$IFDEF WIN32}, csReplicatable {$IFDEF DELPHI4_UP},
    csActionClient{$ENDIF}{$ENDIF}];
  FModalResult := mrNone;
  FMouseInControl := False;
  FOldCursor := Cursor;
  FOptions := TFourthStandardButtonOptions.Create;
  FOptions.OnChange := OptionsChange;
  Width := 73;
  Height := 20;
end;

destructor TFourthStandardButton.Destroy;
begin
  FOptions.Free;
  inherited Destroy;
end;

procedure TFourthStandardButton.SetCaption(Value: TCaption);
begin
  if FCaption <> Value then
  begin
    FCaption := Value;
    Invalidate;
  end;
end;

procedure TFourthStandardButton.Paint;
const
  DrawFlags: array[TAlignment] of Integer = (DT_LEFT, DT_RIGHT, DT_CENTER);
var
  PaintRect: TRect;
  FLineHeight, FMarginTop, FMarginBottom: Integer;
begin
  if (not Enabled) then
  begin
    FState := bsDisabled;
    FDragging := False;
  end else
    if (FState = bsDisabled) then
      FState := bsUp;

  PaintRect := Rect(0, 0, Width, Height);
  if not FMouseInControl then
  begin
    if not FOptions.Transparent then
    begin
      Canvas.Brush.Color := FOptions.BackgroundColor;
      Canvas.Brush.Style := bsSolid;
      Canvas.FillRect(PaintRect);
    end else
    begin
      Canvas.Brush.Color := clNone;
      Canvas.Brush.Style := bsClear;
      Canvas.FillRect(PaintRect);
    end;

    FLineHeight := Canvas.TextHeight('QqWwEeRrTtYyUuIiOoPp[{AaSsDdFfGgHhJjKkLlZzXxCcVvBbNnMm');
    FMarginTop := (PaintRect.Bottom - FLineHeight) div 2;
    FMarginBottom := (PaintRect.Bottom - FLineHeight) mod 2;
    if FMarginBottom = 0 then
      FMarginBottom := FMarginTop else
      Inc(FMarginBottom, FMarginTop-1);
    Inc(FMarginTop, 1);

    if FOptions.FSeparatorStyle = ssVertLines then
    begin
      if FOptions.DrawLeftLine then
      begin
        Canvas.Pen.Color := FOptions.VertLineColor;
        Canvas.MoveTo(PaintRect.Left, PaintRect.Top + FMarginTop);
        Canvas.LineTo(PaintRect.Left, PaintRect.Bottom - FMarginBottom);
      end;

      if FOptions.DrawRightLine then
      begin
        Canvas.Pen.Color := FOptions.VertLineColor;
        Canvas.MoveTo(PaintRect.Right-1, PaintRect.Top + FMarginTop);
        Canvas.LineTo(PaintRect.Right-1, PaintRect.Bottom - FMarginBottom);
      end;
    end else
    begin
      Canvas.Pen.Color := FOptions.VertLineColor;
      Canvas.MoveTo(PaintRect.Left, PaintRect.Top);
      Canvas.LineTo(PaintRect.Left, PaintRect.Bottom);

      Canvas.MoveTo(PaintRect.Left, PaintRect.Bottom-1);
      Canvas.LineTo(PaintRect.Right-1, PaintRect.Bottom-1);

      Canvas.MoveTo(PaintRect.Right-1, PaintRect.Bottom-1);
      Canvas.LineTo(PaintRect.Right-1, PaintRect.Top);

      Canvas.MoveTo(PaintRect.Right-1, PaintRect.Top);
      Canvas.LineTo(PaintRect.Left, PaintRect.Top);
    end;

    if FState in [bsDisabled] then
      Canvas.Font.Assign(FOptions.DisabledFont) else
      Canvas.Font.Assign(FOptions.Font);
    InflateRect(PaintRect, -6, -2);
  end else
  begin
    if FState in [bsDisabled] then Exit;
    Canvas.Font.Assign(FOptions.HotTrackFont);
    Canvas.Brush.Color := FOptions.Level3Color;
    Canvas.FillRect(PaintRect);
    if (FState in [bsDown]) then
    begin
      Frame3D(Canvas, PaintRect, FOptions.Level1ShadowColor, FOptions.Level2LightColor, 1);
      Frame3D(Canvas, PaintRect, FOptions.Level2ShadowColor, FOptions.Level1LightColor, 1);
    end else
    begin
      Frame3D(Canvas, PaintRect, FOptions.Level1LightColor, FOptions.Level1ShadowColor, 1);
      Frame3D(Canvas, PaintRect, FOptions.Level2LightColor, FOptions.Level2ShadowColor, 1);
    end;
    InflateRect(PaintRect, -4, 0);
  end;
  SetBkMode(Canvas.Handle, TRANSPARENT);
  if FState in [bsDown] then
  begin
    Inc(PaintRect.Left, 2);
    Inc(PaintRect.Top, 2);
    Inc(PaintRect.Right, 2);
    Inc(PaintRect.Bottom, 2);
  end;

  DrawText(Canvas.Handle, PChar(FCaption), -1, PaintRect,
    DT_SINGLELINE or DrawFlags[FOptions.Alignment] or DT_VCENTER or
    DT_END_ELLIPSIS);
end;

procedure TFourthStandardButton.Loaded;
begin
  inherited Loaded;
  if Enabled then
    FState := bsUp else
    FState := bsDisabled;
end;

procedure TFourthStandardButton.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseDown(Button, Shift, X, Y);
  if (Button = mbLeft) and Enabled then
  begin
    FState := bsDown;
    Invalidate;
  end;
  FDragging := True;
end;

procedure TFourthStandardButton.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  NewState: TFourthButtonState;
begin
  inherited MouseMove(Shift, X, Y);
  if FDragging then
  begin
    NewState := bsUp;
    if (X >= 0) and (X < ClientWidth) and (Y >= 0) and (Y <= ClientHeight) then
      NewState := bsDown;
    if NewState <> FState then
    begin
      FState := NewState;
      Invalidate;
    end;
  end;
end;

procedure TFourthStandardButton.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  DoClick: Boolean;
begin
  inherited MouseUp(Button, Shift, X, Y);
  FState := bsUp;
  Invalidate;
  if FDragging then
  begin
    FDragging := False;
    DoClick := (X >= 0) and (X < ClientWidth) and (Y >= 0) and (Y <= ClientHeight);
    if DoClick then
    begin
      Repaint;
      Click;
    end;
  end;
end;

procedure TFourthStandardButton.CMEnabledChanged(var Message: TMessage);
begin
  if not Enabled then FState := bsDisabled;
  Repaint;
end;

procedure TFourthStandardButton.CMSysColorChange(var Message: TMessage);
begin
  Invalidate;
end;

procedure TFourthStandardButton.CMButtonPressed(var Message: TMessage);
begin
  Invalidate;
end;

procedure TFourthStandardButton.Click;
begin
  inherited;
  if FDropMenu <> nil then
  begin
    FDropMenu.PopupComponent := Self;
    FDropMenu.Popup(GetClientOrigin.x, GetClientOrigin.y + Self.Height);
    Perform(CM_MOUSELEAVE, 0, 0);
  end;
end;

procedure TFourthStandardButton.CMFontChanged(var Message: TMessage);
begin
  inherited;
  Invalidate;
end;

procedure TFourthStandardButton.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  Message.Result := DLGC_WANTALLKEYS;
end;

procedure TFourthStandardButton.CMMouseEnter(var Message:TMessage);
begin
  if not FMouseInControl then
  begin
    FOldCursor := Cursor;
    Cursor := FOptions.HotTrackCursor;
    FMouseInControl := True;
    Invalidate;
  end;
  if Assigned(FOnMouseEnter) then FOnMouseEnter(Self);
end;

procedure TFourthStandardButton.CMMouseLeave(var Message: TMessage);
begin
  if FMouseInControl then
  begin
    Cursor := FOldCursor;
    FMouseInControl := False;
    Invalidate;
  end;
  if Assigned(FOnMouseLeave) then FOnMouseLeave(Self);
end;

procedure TFourthStandardButton.WMEraseBkgnd(var Message: TWMEraseBkgnd);
begin
  if FOptions.Transparent then
    Message.Result := 0 else
    Message.Result := 1;
end;

procedure TFourthStandardButton.OptionsChange(Sender: TObject);
begin
  Invalidate;
end;

{ TFourthCaptionLikeButton }

constructor TFourthCaptionLikeButton.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := [csCaptureMouse, csClickEvents, csOpaque, csReplicatable];
  FState := bsUp;
  FStyle := clbsClose;
  Width := GetSystemWidth;
  Height := GetSystemHeight;
end;

destructor TFourthCaptionLikeButton.Destroy;
begin
  inherited Destroy;
end;

function TFourthCaptionLikeButton.GetSystemWidth: Integer;
begin
  Result := GetSystemMetrics(SM_CXSIZE) - 2;
end;

function TFourthCaptionLikeButton.GetSystemHeight: Integer;
begin
  Result := GetSystemMetrics(SM_CYSIZE) - 4;
end;

procedure TFourthCaptionLikeButton.SetSystemSize;
begin
  Width := GetSystemWidth;
  Height := GetSystemHeight;
end;

procedure TFourthCaptionLikeButton.SetStyle(Value: TFourthCaptionLikeButtonStyle);
begin
  if (FStyle <> Value) then
  begin
    FStyle := Value;
    Invalidate;
  end;
end;

procedure TFourthCaptionLikeButton.Paint;
const
  TypeFlags: array[TFourthCaptionLikeButtonStyle] of Integer = (DFC_BUTTON,
    DFC_CAPTION, DFC_CAPTION, DFC_CAPTION, DFC_CAPTION, DFC_CAPTION);

  StateFlags: array[TFourthCaptionLikeButtonStyle] of Integer = (
    DFCS_BUTTONPUSH, DFCS_CAPTIONCLOSE, DFCS_CAPTIONHELP, DFCS_CAPTIONMAX,
    DFCS_CAPTIONMIN, DFCS_CAPTIONRESTORE);

  SpecFlags: array[TFourthButtonState] of Integer = (0, DFCS_INACTIVE,
    DFCS_PUSHED, 0);
var
  R: TRect;
begin
  if (not Enabled) then
  begin
    FState := bsDisabled;
    FDragging := False;
  end else
    if (FState = bsDisabled) then
      FState := bsUp;
  R := ClientRect;
  DrawFrameControl(Canvas.Handle, R, TypeFlags[FStyle], StateFlags[FStyle] or
    SpecFlags[FState]);
end;

procedure TFourthCaptionLikeButton.MouseDown(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseDown(Button, Shift, X, Y);
  if (Button = mbLeft) and Enabled then
  begin
    FState := bsDown;
    Invalidate;
  end;
  FDragging := True;
end;

procedure TFourthCaptionLikeButton.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  NewState: TFourthButtonState;
begin
  inherited MouseMove(Shift, X, Y);
  if FDragging then
  begin
    NewState := bsUp;
    if (X >= 0) and (X < ClientWidth) and (Y >= 0) and (Y <= ClientHeight) then
      NewState := bsDown;
    if NewState <> FState then
    begin
      FState := NewState;
      Invalidate;
    end;
  end;
end;

procedure TFourthCaptionLikeButton.MouseUp(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  DoClick: Boolean;
begin
  inherited MouseUp(Button, Shift, X, Y);
  FState := bsUp;
  Invalidate;
  if FDragging then
  begin
    FDragging := False;
    DoClick := (X >= 0) and (X < ClientWidth) and (Y >= 0) and (Y <= ClientHeight);
    if DoClick then
    begin
      Repaint;
      Click;
    end;
  end;
end;

procedure TFourthCaptionLikeButton.CMEnabledChanged(var Message: TMessage);
begin
  if not Enabled then FState := bsDisabled;
  Repaint;
end;

procedure TFourthCaptionLikeButton.CMSysColorChange(var Message: TMessage);
begin
  Invalidate;
end;

end.

