
{*******************************************************}
{                                                       }
{       4th GUI Library (VCL Edition)                   }
{       Controls Unit                                   }
{                                                       }
{       Copyright (c) 1996,2003 Sergiy S. Tkachenko     }
{       e-mail: support@4thfebruary.tk                  }
{          Web: www.4thfebruary.tk                      }
{                                                       }
{       Portions (c) Borland Software Corporation       }
{                                                       }
{*******************************************************}

unit FthCtrls;

{$S-,W-,R-,H+,X+}
{$C PRELOAD}
{$I FTHVER.INC}

interface

uses
{$IFDEF WIN32}
  Windows,
{$ELSE}
  WinTypes, WinProcs,
{$ENDIF}
  Messages, SysUtils, Classes, Graphics, Controls, Forms, ExtCtrls;

type
  TFourthPanelColors = class(TPersistent)
  private
    FBackground: TColor;
    FDarkShadow: TColor;
    FHighlight: TColor;
    FOnChange: TNotifyEvent;
    procedure SetBackground(Value: TColor);
    procedure SetDarkShadow(Value: TColor);
    procedure SetHighlight(Value: TColor);
  protected
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  public
    constructor Create;
  published
    property Background: TColor read FBackground write SetBackground;
    property DarkShadow: TColor read FDarkShadow write SetDarkShadow;
    property Highlight: TColor read FHighlight write SetHighlight;
  end;

  TFourthCustomPanel = class(TCustomControl)
  private
    FAutoSizeDocking: Boolean;
    FBevelInner: TPanelBevel;
    FBevelOuter: TPanelBevel;
    FBevelWidth: TBevelWidth;
    FBorderWidth: TBorderWidth;
    FBorderStyle: TBorderStyle;
    FColors: TFourthPanelColors;
    FFullRepaint: Boolean;
    FLocked: Boolean;
    FAlignment: TAlignment;
    procedure CMBorderChanged(var Message: TMessage); message CM_BORDERCHANGED;
    procedure CMTextChanged(var Message: TMessage); message CM_TEXTCHANGED;
    procedure CMCtl3DChanged(var Message: TMessage); message CM_CTL3DCHANGED;
    procedure CMIsToolControl(var Message: TMessage); message CM_ISTOOLCONTROL;
    procedure WMWindowPosChanged(var Message: TWMWindowPosChanged); message WM_WINDOWPOSCHANGED;
    procedure SetAlignment(Value: TAlignment);
    procedure SetBevelInner(Value: TPanelBevel);
    procedure SetBevelOuter(Value: TPanelBevel);
    procedure SetBevelWidth(Value: TBevelWidth);
    procedure SetBorderWidth(Value: TBorderWidth);
    procedure SetBorderStyle(Value: TBorderStyle);
    procedure CMDockClient(var Message: TCMDockClient); message CM_DOCKCLIENT;
  protected
    procedure ColorsChange(Sender: TObject);
    procedure CreateParams(var Params: TCreateParams); override;
    procedure AdjustClientRect(var Rect: TRect); override;
    function CanAutoSize(var NewWidth, NewHeight: Integer): Boolean; override;
    procedure Paint; override;
    property Alignment: TAlignment read FAlignment write SetAlignment default taCenter;
    property BevelInner: TPanelBevel read FBevelInner write SetBevelInner default bvNone;
    property BevelOuter: TPanelBevel read FBevelOuter write SetBevelOuter default bvRaised;
    property BevelWidth: TBevelWidth read FBevelWidth write SetBevelWidth default 1;
    property BorderWidth: TBorderWidth read FBorderWidth write SetBorderWidth default 0;
    property BorderStyle: TBorderStyle read FBorderStyle write SetBorderStyle default bsNone;
    property Colors: TFourthPanelColors read FColors write FColors;
    property FullRepaint: Boolean read FFullRepaint write FFullRepaint default True;
    property Locked: Boolean read FLocked write FLocked default False;
    property ParentColor default False;
  public
    constructor Create(AOwner: TComponent); override;
    function GetControlsAlignment: TAlignment; override;
  end;

  TFourthPanel = class(TFourthCustomPanel)
  public
    property DockManager;
  published
    property Align;
    property Alignment;
    property Anchors;
    property AutoSize;
    property BevelInner;
    property BevelOuter;
    property BevelWidth;
    property BiDiMode;
    property BorderWidth;
    property BorderStyle;
    property Caption;
    property Colors;
    property Constraints;
    property Ctl3D;
    property UseDockManager default True;
    property DockSite;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property FullRepaint;
    property Font;
    property Locked;
    property ParentBiDiMode;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnCanResize;
    property OnClick;
    property OnConstrainedResize;
    property OnContextPopup;
    property OnDockDrop;
    property OnDockOver;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnGetSiteInfo;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnStartDock;
    property OnStartDrag;
    property OnUnDock;
  end;

{ TFourthBevel }

  TBevelStyle = (bsLowered, bsRaised);
  TBevelShape = (bsBox, bsFrame, bsTopLine, bsBottomLine, bsLeftLine,
    bsRightLine, bsSpacer);

  TFourthBevelMargins = class(TPersistent)
  private
    FBottom: Integer;
    FLeft: Integer;
    FRight: Integer;
    FTop: Integer;
    FOnChange: TNotifyEvent;
    procedure SetBottomMargin(Value: Integer);
    procedure SetLeftMargin(Value: Integer);
    procedure SetRightMargin(Value: Integer);
    procedure SetTopMargin(Value: Integer);
  protected
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  public
    constructor Create;
  published
    property Bottom: Integer read FBottom write SetBottomMargin;
    property Left: Integer read FLeft write SetLeftMargin;
    property Right: Integer read FRight write SetRightMargin;
    property Top: Integer read FTop write SetTopMargin;
  end;

  TFourthBevel = class(TGraphicControl)
  private
    FMargins: TFourthBevelMargins;
    FStyle: TBevelStyle;
    FShape: TBevelShape;
    procedure SetStyle(Value: TBevelStyle);
    procedure SetShape(Value: TBevelShape);
  protected
    procedure Paint; override;
    procedure MarginsChange(Sender: TObject);
  public
    constructor Create(AOwner: TComponent); override;
  published
    property Align;
    property Anchors;
    property Constraints;
    property Margins: TFourthBevelMargins read FMargins write FMargins;
    property ParentShowHint;
    property Shape: TBevelShape read FShape write SetShape default bsBox;
    property ShowHint;
    property Style: TBevelStyle read FStyle write SetStyle default bsLowered;
    property Visible;
  end;

implementation

{ TFourthPanelColors }

constructor TFourthPanelColors.Create;
begin
  FBackground := clBtnShadow;
  FDarkShadow := cl3DDkShadow;
  FHighlight := clBtnHighlight;
end;

procedure TFourthPanelColors.SetBackground(Value: TColor);
begin
  if (FBackground <> Value) then
  begin
    FBackground := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthPanelColors.SetDarkShadow(Value: TColor);
begin
  if (FDarkShadow <> Value) then
  begin
    FDarkShadow := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthPanelColors.SetHighlight(Value: TColor);
begin
  if (FHighlight <> Value) then
  begin
    FHighlight := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

{ TFourthPanel }

constructor TFourthCustomPanel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := [csAcceptsControls, csCaptureMouse, csClickEvents,
    csSetCaption, csOpaque, csDoubleClicks, csReplicatable];
  FColors := TFourthPanelColors.Create;
  FColors.OnChange := ColorsChange;
  Font.Color := clBtnHighlight;
  Width := 185;
  Height := 41;
  FAlignment := taCenter;
  BevelOuter := bvLowered;
  BevelWidth := 1;
  FBorderStyle := bsNone;
  FFullRepaint := True;
  UseDockManager := True;
end;

procedure TFourthCustomPanel.CreateParams(var Params: TCreateParams);
const
  BorderStyles: array[TBorderStyle] of DWORD = (0, WS_BORDER);
begin
  inherited CreateParams(Params);
  with Params do
  begin
    Style := Style or BorderStyles[FBorderStyle];
    if NewStyleControls and Ctl3D and (FBorderStyle = bsSingle) then
    begin
      Style := Style and not WS_BORDER;
      ExStyle := ExStyle or WS_EX_CLIENTEDGE;
    end;
    WindowClass.style := WindowClass.style and not (CS_HREDRAW or CS_VREDRAW);
  end;
end;

procedure TFourthCustomPanel.CMBorderChanged(var Message: TMessage);
begin
  inherited;
  Invalidate;
end;

procedure TFourthCustomPanel.CMTextChanged(var Message: TMessage);
begin
  Invalidate;
end;

procedure TFourthCustomPanel.CMCtl3DChanged(var Message: TMessage);
begin
  if NewStyleControls and (FBorderStyle = bsSingle) then RecreateWnd;
  inherited;
end;

procedure TFourthCustomPanel.CMIsToolControl(var Message: TMessage);
begin
  if not FLocked then Message.Result := 1;
end;

procedure TFourthCustomPanel.WMWindowPosChanged(var Message: TWMWindowPosChanged);
var
  BevelPixels: Integer;
  Rect: TRect;
begin
  if FullRepaint or (Caption <> '') then
    Invalidate
  else
  begin
    BevelPixels := BorderWidth;
    if BevelInner <> bvNone then Inc(BevelPixels, BevelWidth);
    if BevelOuter <> bvNone then Inc(BevelPixels, BevelWidth);
    if BevelPixels > 0 then
    begin
      Rect.Right := Width;
      Rect.Bottom := Height;
      if Message.WindowPos^.cx <> Rect.Right then
      begin
        Rect.Top := 0;
        Rect.Left := Rect.Right - BevelPixels - 1;
        InvalidateRect(Handle, @Rect, True);
      end;
      if Message.WindowPos^.cy <> Rect.Bottom then
      begin
        Rect.Left := 0;
        Rect.Top := Rect.Bottom - BevelPixels - 1;
        InvalidateRect(Handle, @Rect, True);
      end;
    end;
  end;
  inherited;
end;

procedure TFourthCustomPanel.Paint;
const
  Alignments: array[TAlignment] of Longint = (DT_LEFT, DT_RIGHT, DT_CENTER);
var
  Rect: TRect;
  TopColor, BottomColor: TColor;
  FontHeight: Integer;
  Flags: Longint;

  procedure AdjustColors(Bevel: TPanelBevel);
  begin
    TopColor := FColors.Highlight;
    if Bevel = bvLowered then TopColor := FColors.DarkShadow;
    BottomColor := FColors.FDarkShadow;
    if Bevel = bvLowered then BottomColor := FColors.Highlight;
  end;

begin
  Rect := GetClientRect;
  if BevelOuter <> bvNone then
  begin
    AdjustColors(BevelOuter);
    Frame3D(Canvas, Rect, TopColor, BottomColor, BevelWidth);
  end;
  Frame3D(Canvas, Rect, Color, Color, BorderWidth);
  if BevelInner <> bvNone then
  begin
    AdjustColors(BevelInner);
    Frame3D(Canvas, Rect, TopColor, BottomColor, BevelWidth);
  end;
  with Canvas do
  begin
    Brush.Color := FColors.Background;
    FillRect(Rect);
    Brush.Style := bsClear;
    Font := Self.Font;
    FontHeight := TextHeight('W');
    with Rect do
    begin
      Top := ((Bottom + Top) - FontHeight) div 2;
      Bottom := Top + FontHeight;
    end;
    Flags := DT_EXPANDTABS or DT_VCENTER or Alignments[FAlignment];
    Flags := DrawTextBiDiModeFlags(Flags);
    DrawText(Handle, PChar(Caption), -1, Rect, Flags);
  end;
end;

procedure TFourthCustomPanel.SetAlignment(Value: TAlignment);
begin
  FAlignment := Value;
  Invalidate;
end;

procedure TFourthCustomPanel.SetBevelInner(Value: TPanelBevel);
begin
  FBevelInner := Value;
  Realign;
  Invalidate;
end;

procedure TFourthCustomPanel.SetBevelOuter(Value: TPanelBevel);
begin
  FBevelOuter := Value;
  Realign;
  Invalidate;
end;

procedure TFourthCustomPanel.SetBevelWidth(Value: TBevelWidth);
begin
  FBevelWidth := Value;
  Realign;
  Invalidate;
end;

procedure TFourthCustomPanel.SetBorderWidth(Value: TBorderWidth);
begin
  FBorderWidth := Value;
  Realign;
  Invalidate;
end;

procedure TFourthCustomPanel.SetBorderStyle(Value: TBorderStyle);
begin
  if (FBorderStyle <> Value) then
  begin
    FBorderStyle := Value;
    RecreateWnd;
  end;
end;

function TFourthCustomPanel.GetControlsAlignment: TAlignment;
begin
  Result := FAlignment;
end;

procedure TFourthCustomPanel.AdjustClientRect(var Rect: TRect);
var
  BevelSize: Integer;
begin
  inherited AdjustClientRect(Rect);
  InflateRect(Rect, -BorderWidth, -BorderWidth);
  BevelSize := 0;
  if BevelOuter <> bvNone then Inc(BevelSize, BevelWidth);
  if BevelInner <> bvNone then Inc(BevelSize, BevelWidth);
  InflateRect(Rect, -BevelSize, -BevelSize);
end;

procedure TFourthCustomPanel.CMDockClient(var Message: TCMDockClient);
var
  R: TRect;
  Dim: Integer;
begin
  if AutoSize then
  begin
    FAutoSizeDocking := True;
    try
      R := Message.DockSource.DockRect;
      case Align of
        alLeft: if Width = 0 then Width := R.Right - R.Left;
        alRight: if Width = 0 then
          begin
            Dim := R.Right - R.Left;
            SetBounds(Left - Dim, Top, Dim, Height);
          end;
        alTop: if Height = 0 then Height := R.Bottom - R.Top;
        alBottom: if Height = 0 then
          begin
            Dim := R.Bottom - R.Top;
            SetBounds(Left, Top - Dim, Width, Dim);
          end;
      end;
      inherited;
      Exit;
    finally
      FAutoSizeDocking := False;
    end;
  end;
  inherited;
end;

function TFourthCustomPanel.CanAutoSize(var NewWidth, NewHeight: Integer): Boolean;
begin
  Result := (not FAutoSizeDocking) and inherited CanAutoSize(NewWidth, NewHeight);
end;

procedure TFourthCustomPanel.ColorsChange(Sender: TObject);
begin
  Invalidate;
end;

{ TFourthBevelMargins }

constructor TFourthBevelMargins.Create;
begin
  FBottom := 0;
  FLeft := 0;
  FRight := 0;
  FTop := 0;
end;

procedure TFourthBevelMargins.SetBottomMargin(Value: Integer);
begin
  if (FBottom <> Value) then
  begin
    FBottom := Value;
    if FBottom < 0 then FBottom := 0;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthBevelMargins.SetLeftMargin(Value: Integer);
begin
  if (FLeft <> Value) then
  begin
    FLeft := Value;
    if FLeft < 0 then FLeft := 0;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthBevelMargins.SetRightMargin(Value: Integer);
begin
  if (FRight <> Value) then
  begin
    FRight := Value;
    if FRight < 0 then FRight := 0;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthBevelMargins.SetTopMargin(Value: Integer);
begin
  if (FTop <> Value) then
  begin
    FTop := Value;
    if FTop < 0 then FTop := 0;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

{ TFourthBevel }

constructor TFourthBevel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csReplicatable];
  FMargins := TFourthBevelMargins.Create;
  FMargins.OnChange := MarginsChange;
  FStyle := bsLowered;
  FShape := bsBox;
  Width := 50;
  Height := 50;
end;

procedure TFourthBevel.SetStyle(Value: TBevelStyle);
begin
  if Value <> FStyle then
  begin
    FStyle := Value;
    Invalidate;
  end;
end;

procedure TFourthBevel.SetShape(Value: TBevelShape);
begin
  if Value <> FShape then
  begin
    FShape := Value;
    Invalidate;
  end;
end;

procedure TFourthBevel.Paint;
const
  XorColor = $00FFD8CE;
var
  Color1, Color2: TColor;
  Temp: TColor;

  procedure BevelRect(const R: TRect);
  begin
    with Canvas do
    begin
      Pen.Color := Color1;
      PolyLine([Point(R.Left, R.Bottom), Point(R.Left, R.Top),
        Point(R.Right, R.Top)]);
      Pen.Color := Color2;
      PolyLine([Point(R.Right, R.Top), Point(R.Right, R.Bottom),
        Point(R.Left, R.Bottom)]);
    end;
  end;

  procedure BevelLine(C: TColor; X1, Y1, X2, Y2: Integer);
  begin
    with Canvas do
    begin
      Pen.Color := C;
      MoveTo(X1, Y1);
      LineTo(X2, Y2);
    end;
  end;

begin
  with Canvas do
  begin
    if (csDesigning in ComponentState) then
    begin
      if (FShape = bsSpacer) then
      begin
        Pen.Style := psDot;
        Pen.Mode := pmXor;
        Pen.Color := XorColor;
        Brush.Style := bsClear;
        Rectangle(0, 0, ClientWidth, ClientHeight);
        Exit;
      end else
      begin
        Pen.Style := psSolid;
        Pen.Mode  := pmCopy;
        Pen.Color := clBlack;
        Brush.Style := bsSolid;
      end;
    end;

    Pen.Width := 1;

    if FStyle = bsLowered then
    begin
      Color1 := clBtnShadow;
      Color2 := clBtnHighlight;
    end else
    begin
      Color1 := clBtnHighlight;
      Color2 := clBtnShadow;
    end;

    case FShape of
      bsBox: BevelRect(Rect(FMargins.Left + 0, FMargins.Top + 0, Width - 1 - FMargins.Right, Height - 1 - FMargins.Bottom));
      bsFrame:
        begin
          Temp := Color1;
          Color1 := Color2;
          BevelRect(Rect(FMargins.Left + 1, FMargins.Top + 1, Width - 1 - FMargins.Right, Height - 1 - FMargins.Bottom));
          Color2 := Temp;
          Color1 := Temp;
          BevelRect(Rect(FMargins.Left + 0, FMargins.Top + 0, Width - 2 - FMargins.Right, Height - 2 - FMargins.Bottom));
        end;
      bsTopLine:
        begin
          BevelLine(Color1, FMargins.Left + 0, FMargins.Top + 0, Width - FMargins.Right, FMargins.Top + 0);
          BevelLine(Color2, FMargins.Left + 0, FMargins.Top + 1, Width - FMargins.Right, FMargins.Top + 1);
        end;
      bsBottomLine:
        begin
          BevelLine(Color1, FMargins.Left + 0, Height - 2 - FMargins.Bottom, Width - FMargins.Right, Height - 2 - FMargins.Bottom);
          BevelLine(Color2, FMargins.Left + 0, Height - 1 - FMargins.Bottom, Width - FMargins.Right, Height - 1 - FMargins.Bottom);
        end;
      bsLeftLine:
        begin
          BevelLine(Color1, FMargins.Left + 0, FMargins.Top + 0, FMargins.Left + 0, Height - FMargins.Bottom);
          BevelLine(Color2, FMargins.Left + 1, FMargins.Top + 0, FMargins.Left + 1, Height - FMargins.Bottom);
        end;
      bsRightLine:
        begin
          BevelLine(Color1, Width - 2 - FMargins.Right, FMargins.Top + 0, Width - 2 - FMargins.Right, Height - FMargins.Bottom);
          BevelLine(Color2, Width - 1 - FMargins.Right, FMargins.Top + 0, Width - 1 - FMargins.Right, Height - FMargins.Bottom);
        end;
    end;
  end;
end;

procedure TFourthBevel.MarginsChange(Sender: TObject);
begin
  Invalidate;
end;

end.
