
{*******************************************************}
{                                                       }
{       4th GUI Library (VCL Edition)                   }
{       Graphics Unit                                   }
{                                                       }
{       Copyright (c) 1996,2003 Sergiy S. Tkachenko     }
{       e-mail: support@4thfebruary.tk                  }
{          Web: www.4thfebruary.tk                      }
{                                                       }
{*******************************************************}

unit FthGraph;

interface

uses Windows, Graphics, Math;

type
  // Describes the mode how to blend pixels.
  TBlendMode = (
    bmConstantAlpha,         // apply given constant alpha
    bmPerPixelAlpha,         // use alpha value of the source pixel
    bmMasterAlpha,           // use alpha value of source pixel and multiply it with the constant alpha value
    bmConstantAlphaAndColor  // blend the destination color with the given constant color und the constant alpha value
  );

procedure Frame3D(Canvas: TCanvas; var Rect: TRect; TopColor, BottomColor: TColor; Width: Integer);
function Get2002Bg(const AColor: TColor): TColor;
function GetXPHotHighlight: TColor;
function GetXPPressedHighlight: TColor;
function LightenColor(AColor: TColor; GlassColor: TColor; Ungray: Boolean): TColor;

implementation

procedure Frame3D(Canvas: TCanvas; var Rect: TRect; TopColor, BottomColor: TColor; Width: Integer);

  procedure DoRect;
  var
    TopRight, BottomLeft: TPoint;
  begin
    with Canvas, Rect do
    begin
      TopRight.X := Right;
      TopRight.Y := Top;
      BottomLeft.X := Left;
      BottomLeft.Y := Bottom;
      Pen.Color := TopColor;
      PolyLine([BottomLeft, TopLeft, TopRight]);
      Pen.Color := BottomColor;
      Dec(BottomLeft.X);
      PolyLine([TopRight, BottomRight, BottomLeft]);
    end;
  end;

begin
  Canvas.Pen.Width := 1;
  Dec(Rect.Bottom); Dec(Rect.Right);
  while Width > 0 do
  begin
    Dec(Width);
    DoRect;
    InflateRect(Rect, -1, -1);
  end;
  Inc(Rect.Bottom); Inc(Rect.Right);
end;

function Get2002Bg(const AColor: TColor): TColor;
var
  R, G, B: Byte;
  BtnFaceRGB: Longint;
begin
  BtnFaceRGB := ColorToRGB(clBtnFace);
  if (GetRValue(BtnFaceRGB) + 7) > 255 then
    R := 255 else
    R := GetRValue(BtnFaceRGB) + 7;
  if (GetGValue(BtnFaceRGB) + 8) > 255 then
    G := 255 else
    G := GetGValue(BtnFaceRGB) + 8;
  if (GetBValue(BtnFaceRGB) + 9) > 255 then
    B := 255 else
    B := GetBValue(BtnFaceRGB) + 9;
  Result := RGB(R, G, B);
end;

function GetXPHotHighlight: TColor;
var
  r, g, b: Double;
  rgb: Longint;
  r_c, g_c, b_c: Double;
  rgb_c: Longint;
begin
  rgb := ColorToRGB(clBtnFace);
  r := rgb and $FF;
  g := (rgb shr 8) and $FF;
  b := (rgb shr 16) and $FF;

  rgb_c := ColorToRGB(clHighlight);
  r_c := rgb_c and $FF;
  g_c := (rgb_c shr 8) and $FF;
  b_c := (rgb_c shr 16) and $FF;

  r := r + (r_c - r) * 0.621;
  g := g + (g_c - g) * 0.621;
  b := b + (b_c - b) * 0.621;

  r := r + (225 - r) * 0.6;
  g := g + (225 - g) * 0.6;
  b := b + (225 - b) * 0.6;

  r := r - (integer(122) - r) * 0.25;
  g := g - (integer(122) - g) * 0.25;
  b := b - (integer(122) - b) * 0.25;

  Result := TColor((Max(Min(Round(b), 255),0) shl 16)
                or (Max(Min(Round(g), 255),0) shl 8)
                or Max(Min(Round(r), 255),0));
end;

function GetXPPressedHighlight: TColor;
var
  r, g, b: Double;
  rgb: Longint;
  r_c, g_c, b_c: Double;
  rgb_c: Longint;
begin
  rgb := ColorToRGB(clBtnFace);
  r := rgb and $FF;
  g := (rgb shr 8) and $FF;
  b := (rgb shr 16) and $FF;

  rgb_c := ColorToRGB(clHighlight);
  r_c := rgb_c and $FF;
  g_c := (rgb_c shr 8) and $FF;
  b_c := (rgb_c shr 16) and $FF;

  r := r + (r_c - r){ * 0.75};
  g := g + (g_c - g){ * 0.75};
  b := b + (b_c - b){ * 0.75};

  r := r + (225 - r) * 0.6;
  g := g + (225 - g) * 0.6;
  b := b + (225 - b) * 0.6;

  r := r - (integer(122) - r) * 0.25;
  g := g - (integer(122) - g) * 0.25;
  b := b - (integer(122) - b) * 0.25;

  Result := TColor((Max(Min(Round(b), 255),0) shl 16)
                or (Max(Min(Round(g), 255),0) shl 8)
                or Max(Min(Round(r), 255),0));
end;

function LightenColor(AColor: TColor; GlassColor: TColor; Ungray: Boolean): TColor;
var
  r, g, b: Double;
  rgb: Longint;
  r_c, g_c, b_c: Double;
  rgb_c: Longint;
begin
  rgb := ColorToRGB(AColor);
  r := rgb and $FF;
  g := (rgb shr 8) and $FF;
  b := (rgb shr 16) and $FF;

  rgb_c := ColorToRGB(GlassColor);
  r_c := rgb_c and $FF;
  g_c := (rgb_c shr 8) and $FF;
  b_c := (rgb_c shr 16) and $FF;

  r := r + (r_c - r) * 0.5;
  g := g + (g_c - g) * 0.5;
  b := b + (b_c - b) * 0.5;

  r := r + (225 - r) * 0.6;
  g := g + (225 - g) * 0.6;
  b := b + (225 - b) * 0.6;

  if Ungray then
  begin
    r := r - (integer(122) - r) * 0.25;
    g := g - (integer(122) - g) * 0.25;
    b := b - (integer(122) - b) * 0.25;
  end;

  Result := TColor((Max(Min(Round(b), 255),0) shl 16)
                or (Max(Min(Round(g), 255),0) shl 8)
                or Max(Min(Round(r), 255),0));
end;

end.
