
{*******************************************************}
{                                                       }
{       4th GUI Library (VCL Edition)                   }
{       Hints Unit                                      }
{                                                       }
{       Copyright (c) 1996,2003 Sergiy S. Tkachenko     }
{       e-mail: support@4thfebruary.tk                  }
{          Web: www.4thfebruary.tk                      }
{                                                       }
{*******************************************************}

unit FthHint;

{$I FTHVER.INC}

interface

uses
{$IFDEF WIN32}
  Windows,
{$ELSE}
  WinTypes, WinProcs,
{$ENDIF}
  Messages, SysUtils, Classes, Graphics, Controls, Forms, Menus, ExtCtrls;

type
  TFourthHintPanelImage = (piError, piInformation, piQuestion, piWarning, piWarning2, piCustom);
  TFourthHintPanelVertAlignment = (vaTop, vaCenter, vaBottom);
  TFourthHintPanelHorzAlignment = (haLeft, haCenter, haRight);  

{ TFourthHintPanelOptions }

  TFourthHintPanelOptions = class(TPersistent)
  private
    FImage: TBitmap;
    FBackgroundColor: TColor;
    FBorderColor: TColor;
    FLeftBorder: Integer;
    FTopBorder: Integer;
    FRightBorder: Integer;    
    FPanelImage: TFourthHintPanelImage;
    FCustomPanelImage: TBitmap;
    FWordWrap: Boolean;
    FTransparentColor: TColor;
    FVertAlignment: TFourthHintPanelVertAlignment;
    FHorzAlignment: TFourthHintPanelHorzAlignment;
    FOnChange: TNotifyEvent;
    procedure SetBackgroundColor(Value: TColor);
    procedure SetBorderColor(Value: TColor);
    procedure SetLeftBorder(Value: Integer);
    procedure SetTopBorder(Value: Integer);
    procedure SetRightBorder(Value: Integer);
    procedure SetPanelImage(Value: TFourthHintPanelImage);
    procedure SetCustomPanelImage(Value: TBitmap);
    procedure SetWordWrap(Value: Boolean);
    procedure SetTransparentColor(Value: TColor);
    procedure SetVertAlignment(Value: TFourthHintPanelVertAlignment);
    procedure SetHorzAlignment(Value: TFourthHintPanelHorzAlignment);
  protected
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  public
    constructor Create;
    destructor Destroy; override;
  published
    property BackgroundColor: TColor read FBackgroundColor write SetBackgroundColor;
    property LeftBorder: Integer read FLeftBorder write SetLeftBorder;
    property TopBorder: Integer read FTopBorder write SetTopBorder;
    property RightBorder: Integer read FRightBorder write SetRightBorder;
    property BorderColor: TColor read FBorderColor write SetBorderColor;
    property CustomPanelImage: TBitmap read FCustomPanelImage write SetCustomPanelImage stored True;
    property HorzAlignment: TFourthHintPanelHorzAlignment read FHorzAlignment write SetHorzAlignment;
    property PanelImage: TFourthHintPanelImage read FPanelImage write SetPanelImage;
    property WordWrap: Boolean read FWordWrap write SetWordWrap;
    property TransparentColor: TColor read FTransparentColor write SetTransparentColor;
    property VertAlignment: TFourthHintPanelVertAlignment read FVertAlignment write SetVertAlignment;
  end;

  TFourthHintPanel = class(TCustomControl)
  private
    FCaption: TCaption;
    FOptions: TFourthHintPanelOptions;
    FOnMouseEnter: TNotifyEvent;
    FOnMouseLeave: TNotifyEvent;
    procedure SetCaption(Value: TCaption);
    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure CMSysColorChange(var Message: TMessage); message CM_SYSCOLORCHANGE;
    procedure CMButtonPressed(var Message: TMessage); message CM_BUTTONPRESSED;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
  protected
    procedure OptionsChange(Sender: TObject);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Paint; override;
    procedure Click; override;
  published
    property Align;
{$IFDEF DELPHI4_UP}
    property Anchors;
{$ENDIF}
    property Caption: TCaption read FCaption write SetCaption;
    property Cursor;
    property DragCursor;
{$IFDEF DELPHI4_UP}
    property DragKind;
{$ENDIF}
    property DragMode;
    property Enabled;
    property Height;
    property Left;
    property Font;
    property Options: TFourthHintPanelOptions read FOptions write FOptions;
    property ParentFont;
    property ParentShowHint;
    property Name;
    property ShowHint;
    property Top;
    property Tag;
    property Visible;
    property Width;
{$IFDEF DELPHI4_UP}
    property OnCanResize;
{$ENDIF}
    property OnClick;
{$IFDEF DELPHI4_UP}
    property OnConstrainedResize;
{$ENDIF}
{$IFDEF DELPHI5_UP}
    property OnContextPopup;
{$ENDIF}
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
{$IFDEF DELPHI4_UP}
    property OnEndDock;
{$ENDIF}
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseEnter: TNotifyEvent read FOnMouseEnter write FOnMouseEnter;
    property OnMouseLeave: TNotifyEvent read FOnMouseLeave write FOnMouseLeave;
    property OnMouseMove;
    property OnMouseUp;
{$IFDEF DELPHI4_UP}
    property OnResize;
    property OnStartDock;
{$ENDIF}
    property OnStartDrag;
  end;

implementation

{$R FTHHINT.RES}

{ TFourthHintPanelOptions }

constructor TFourthHintPanelOptions.Create;
begin
  FImage := TBitmap.Create;

  FBackgroundColor := {$IFDEF WIN32}clInfoBk{$ELSE}clYellow{$ENDIF};
  FBorderColor := clBtnShadow;
  FLeftBorder := 3;
  FTopBorder := 3;
  FRightBorder := 3;

  FCustomPanelImage := TBitmap.Create;

  FPanelImage := piInformation;
  FImage.Handle := LoadBitmap(hInstance, 'TFOURTH_INFORMATION');

  FTransparentColor := clNone;
  FWordWrap := True;

  FVertAlignment := vaTop;
  FHorzAlignment := haLeft;
end;

destructor TFourthHintPanelOptions.Destroy;
begin
  FImage.Free;
  FCustomPanelImage.Free;
  inherited;
end;

procedure TFourthHintPanelOptions.SetBackgroundColor(Value: TColor);
begin
  if FBackgroundColor <> Value then
  begin
    FBackgroundColor := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthHintPanelOptions.SetBorderColor(Value: TColor);
begin
  if FBorderColor <> Value then
  begin
    FBorderColor := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthHintPanelOptions.SetLeftBorder(Value: Integer);
begin
  if FLeftBorder <> Value then
  begin
    FLeftBorder := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthHintPanelOptions.SetTopBorder(Value: Integer);
begin
  if FTopBorder <> Value then
  begin
    FTopBorder := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthHintPanelOptions.SetRightBorder(Value: Integer);
begin
  if FRightBorder <> Value then
  begin
    FRightBorder := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthHintPanelOptions.SetCustomPanelImage(Value: TBitmap);
begin
  if Value <> nil then
  begin
    FCustomPanelImage.Assign(Value);
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthHintPanelOptions.SetPanelImage(Value: TFourthHintPanelImage);
begin
  if FPanelImage <> Value then
  begin
    FPanelImage := Value;
    case FPanelImage of
      piError:       FImage.Handle := LoadBitmap(hInstance, 'TFOURTH_ERROR');
      piInformation: FImage.Handle := LoadBitmap(hInstance, 'TFOURTH_INFORMATION');
      piQuestion:    FImage.Handle := LoadBitmap(hInstance, 'TFOURTH_QUESTION');
      piWarning:     FImage.Handle := LoadBitmap(hInstance, 'TFOURTH_WARNING');
      piWarning2:    FImage.Handle := LoadBitmap(hInstance, 'TFOURTH_WARNING2');
    end;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthHintPanelOptions.SetTransparentColor(Value: TColor);
begin
  if FTransparentColor <> Value then
  begin
    FTransparentColor := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthHintPanelOptions.SetWordWrap(Value: Boolean);
begin
  if FWordWrap <> Value then
  begin
    FWordWrap := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthHintPanelOptions.SetVertAlignment(Value: TFourthHintPanelVertAlignment);
begin
  if FVertAlignment <> Value then
  begin
    FVertAlignment := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthHintPanelOptions.SetHorzAlignment(Value: TFourthHintPanelHorzAlignment);
begin
  if FHorzAlignment <> Value then
  begin
    FHorzAlignment := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

{ TFourthHintPanel }

constructor TFourthHintPanel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csClickEvents, csDoubleClicks, csOpaque,
    {$IFDEF WIN32}csReplicatable,{$ENDIF} csSetCaption];
//  ParentFont := False;
//  Font.Color := clInfoText;
  FOptions := TFourthHintPanelOptions.Create;
  FOptions.OnChange := OptionsChange;
  Width := 171;
  Height := 36;
  FCaption := Self.Name;
end;

destructor TFourthHintPanel.Destroy;
begin
  FOptions.Free;
  inherited Destroy;
end;

procedure TFourthHintPanel.SetCaption(Value: TCaption);
begin
  if FCaption <> Value then
  begin
    FCaption := Value;
    Invalidate;
  end;
end;

procedure TFourthHintPanel.Paint;
const
  DrawWWFlags: array[Boolean] of Integer = (DT_SINGLELINE, DT_WORDBREAK);
  DrawVAFlags: array[TFourthHintPanelVertAlignment] of Integer = (DT_TOP, DT_VCENTER, DT_BOTTOM);
  DrawHAFlags: array[TFourthHintPanelHorzAlignment] of Integer = (DT_LEFT, DT_CENTER, DT_RIGHT);
var
  PaintRect: TRect;
  DrawRect: TRect;
begin
  PaintRect := Rect(0, 0, Width, Height);

  Canvas.Brush.Color := FOptions.FBackgroundColor;
  Canvas.FillRect(PaintRect);

  Canvas.Pen.Color := FOptions.FBorderColor;

  Canvas.MoveTo(PaintRect.Left, PaintRect.Top);
  Canvas.LineTo(PaintRect.Left, PaintRect.Bottom);

  Canvas.MoveTo(PaintRect.Left, PaintRect.Bottom-1);
  Canvas.LineTo(PaintRect.Right-1, PaintRect.Bottom-1);

  Canvas.MoveTo(PaintRect.Right-1, PaintRect.Bottom-1);
  Canvas.LineTo(PaintRect.Right-1, PaintRect.Top);

  Canvas.MoveTo(PaintRect.Right-1, PaintRect.Top);
  Canvas.LineTo(PaintRect.Left, PaintRect.Top);

  Canvas.Font.Assign(Font);

  Inc(PaintRect.Left, 1 + FOptions.FLeftBorder);
  Inc(PaintRect.Top, 1 + FOptions.FTopBorder);
  Dec(PaintRect.Right, 1 + FOptions.FLeftBorder);
  Dec(PaintRect.Bottom, 1+ FOptions.FTopBorder);

  if FOptions.FPanelImage <> piCustom then
  begin
    DrawRect := Rect(PaintRect.Left, PaintRect.Top,
      FOptions.FImage.Width + PaintRect.Left,
      FOptions.FImage.Height + PaintRect.Top);

    Canvas.BrushCopy(DrawRect, FOptions.FImage, Rect(0, 0,
      FOptions.FImage.Width, FOptions.FImage.Height), clFuchsia);

    Inc(PaintRect.Left, FOptions.FImage.Width + FOptions.FRightBorder);
  end else
  begin
    DrawRect := Rect(PaintRect.Left, PaintRect.Top,
      FOptions.FCustomPanelImage.Width + PaintRect.Left,
      FOptions.FCustomPanelImage.Height + PaintRect.Top);

    Canvas.BrushCopy(DrawRect, FOptions.FCustomPanelImage, Rect(0, 0,
      FOptions.FCustomPanelImage.Width, FOptions.FCustomPanelImage.Height),
      FOptions.FTransparentColor);

    Inc(PaintRect.Left, FOptions.FCustomPanelImage.Width + FOptions.FRightBorder);
  end;

  SetBkMode(Canvas.Handle, TRANSPARENT);

  DrawText(Canvas.Handle, PChar(FCaption), -1, PaintRect,
    DrawHAFlags[FOptions.FHorzAlignment] or
    DrawVAFlags[FOptions.FVertAlignment] or
    DrawWWFlags[FOptions.FWordWrap] or DT_EXPANDTABS or DT_END_ELLIPSIS);
end;

procedure TFourthHintPanel.CMEnabledChanged(var Message: TMessage);
begin
  Repaint;
end;

procedure TFourthHintPanel.CMSysColorChange(var Message: TMessage);
begin
  Invalidate;
end;

procedure TFourthHintPanel.CMButtonPressed(var Message: TMessage);
begin
  Invalidate;
end;

procedure TFourthHintPanel.Click;
begin
  inherited;
end;

procedure TFourthHintPanel.CMFontChanged(var Message: TMessage);
begin
  inherited;
  Invalidate;
end;

procedure TFourthHintPanel.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  Message.Result := DLGC_WANTALLKEYS;
end;

procedure TFourthHintPanel.CMMouseEnter(var Message:TMessage);
begin
  if Assigned(FOnMouseEnter) then FOnMouseEnter(Self);
end;

procedure TFourthHintPanel.CMMouseLeave(var Message: TMessage);
begin
  if Assigned(FOnMouseLeave) then FOnMouseLeave(Self);
end;

procedure TFourthHintPanel.OptionsChange(Sender: TObject);
begin
  Invalidate;
end;

end.


