{*******************************************************}
{                                                       }
{       4th GUI Library VCL                             }
{       FourthShortcutBar Editor Unit                   }
{                                                       }
{       Copyright (c) 1996,2003 Sergiy S. Tkachenko     }
{       e-mail: support@4thfebruary.tk                  }
{          Web: www.4thfebruary.tk                      }
{                                                       }
{*******************************************************}

unit FthSBEd;

{$I FTHVER.INC}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  FthSBar, StdCtrls, ComCtrls;

type
  TFourthShortcutBarEditorForm = class(TForm)
    Bar: TFourthShortcutBar;
    gbPanels: TGroupBox;
    edPanels: TEdit;
    btnAddPanel: TButton;
    btnRenamePanel: TButton;
    btnDeletePanel: TButton;
    btnMoveUpPanel: TButton;
    btnMoveDownPanel: TButton;
    gbButtons: TGroupBox;
    edButtons: TEdit;
    btnAddBtn: TButton;
    btnRenameBtn: TButton;
    btnDeleteBtn: TButton;
    btnMoveUpBtn: TButton;
    btnMoveDownBtn: TButton;
    btnOK: TButton;
    btnCancel: TButton;
    edPanelTag: TEdit;
    udPanelTag: TUpDown;
    laPanelTag: TLabel;
    laBtnTag: TLabel;
    edBtnTag: TEdit;
    udBtnTag: TUpDown;
    edImageIndex: TEdit;
    udImageIndex: TUpDown;
    laImageIndex: TLabel;
    cbSmallImages: TCheckBox;
    cbButtons: TComboBox;
    procedure btnAddPanelClick(Sender: TObject);
    procedure btnRenamePanelClick(Sender: TObject);
    procedure btnDeletePanelClick(Sender: TObject);
    procedure btnAddBtnClick(Sender: TObject);
    procedure BarClick(Sender: TObject);
    procedure cbButtonsChange(Sender: TObject);
    procedure btnRenameBtnClick(Sender: TObject);
    procedure btnDeleteBtnClick(Sender: TObject);
    procedure udImageIndexClick(Sender: TObject; Button: TUDBtnType);
    procedure udBtnTagClick(Sender: TObject; Button: TUDBtnType);
    procedure udPanelTagClick(Sender: TObject; Button: TUDBtnType);
    procedure cbSmallImagesClick(Sender: TObject);
  private
    { Private declarations }
  public
    procedure UpdateButtons;
  end;

var
  FourthShortcutBarEditorForm: TFourthShortcutBarEditorForm;

function ShowShortcutBarEditor(ABar: TFourthShortcutBar): Boolean;

implementation

{$R *.DFM}

function ShowShortcutBarEditor(ABar: TFourthShortcutBar): Boolean;
var
  AForm: TFourthShortcutBarEditorForm;
  P, B: Integer;
begin
  Result := False;
  if (ABar.Images = nil) then
  begin
    MessageDlg('You must define ''Images'' property.', mtError, [mbOK], 0);
    Exit;
  end;
  AForm := TFourthShortcutBarEditorForm.Create(nil);
  AForm.Caption := Format('ShortcutBar Editor: %s.%s', [ABar.Parent.Name, ABar.Name]);
  AForm.Bar.Images := ABar.Images;
  AForm.Bar.SmallImages := ABar.SmallImages;
  AForm.Bar.Font.Assign(ABar.Font);
  AForm.cbButtons.Items.Clear;
  AForm.cbButtons.Items.Add('(None)');
  if (ABar.Panels.Count <> 0) then
  begin
    for P := 0 to Pred(ABar.Panels.Count) do
    begin
      with AForm.Bar.Panels.Add do
      begin
        Data := ABar.Panels[P].Data;
        SmallImages := ABar.Panels[P].SmallImages;
        Tag := ABar.Panels[P].Tag;
        Text := ABar.Panels[P].Text;
        if (ABar.Panels[P].Buttons.Count <> 0) then
        begin
          for B := 0 to Pred(ABar.Panels[P].Buttons.Count) do
          begin
            AForm.cbButtons.Items.Add(Format('Button %d', [B]));
            with AForm.Bar.Panels[P].Buttons.Add do
            begin
              Data := ABar.Panels[P].Buttons[B].Data;
              ImageIndex := ABar.Panels[P].Buttons[B].ImageIndex;
              Tag := ABar.Panels[P].Buttons[B].Tag;
              Text := ABar.Panels[P].Buttons[B].Text;
            end;
          end;
        end;
      end;
    end;
  end;
  AForm.udPanelTag.Max := MaxShort;
  AForm.udBtnTag.Max := MaxShort;
  AForm.udImageIndex.Max := MaxShort;
  AForm.UpdateButtons;
  AForm.cbButtons.ItemIndex := 0;
  AForm.ShowModal;
  if (AForm.ModalResult = mrOK) then
  begin
    ABar.Panels.Clear;
    if (AForm.Bar.Panels.Count <> 0) then
    begin
      for P := 0 to Pred(AForm.Bar.Panels.Count) do
      begin
        with ABar.Panels.Add do
        begin
          Data := AForm.Bar.Panels[P].Data;
          SmallImages := AForm.Bar.Panels[P].SmallImages;
          Tag := AForm.Bar.Panels[P].Tag;
          Text := AForm.Bar.Panels[P].Text;
          if (AForm.Bar.Panels[P].Buttons.Count <> 0) then
          begin
            for B := 0 to Pred(AForm.Bar.Panels[P].Buttons.Count) do
            begin
              with ABar.Panels[P].Buttons.Add do
              begin
                Data := AForm.Bar.Panels[P].Buttons[B].Data;
                ImageIndex := AForm.Bar.Panels[P].Buttons[B].ImageIndex;
                Tag := AForm.Bar.Panels[P].Buttons[B].Tag;
                Text := AForm.Bar.Panels[P].Buttons[B].Text;
              end;
            end;
          end;
        end;
      end;
    end;
    Result := True;
  end;
  AForm.Free;
end;

procedure TFourthShortcutBarEditorForm.UpdateButtons;
begin
  btnAddPanel.Enabled := Bar.Panels.Count >= 0;
  btnRenamePanel.Enabled := Bar.Panels.Count > 0;
  btnDeletePanel.Enabled := Bar.Panels.Count > 0;
  btnMoveUpPanel.Enabled := False;
  btnMoveDownPanel.Enabled := False;

  btnAddBtn.Enabled := Bar.Panels.Count > 0;
//  btnRenameBtn.Enabled := False;
//  btnDeleteBtn.Enabled := False;
  btnMoveUpBtn.Enabled := False;
  btnMoveDownBtn.Enabled := False;
end;

procedure TFourthShortcutBarEditorForm.btnAddPanelClick(Sender: TObject);
begin
  if Length(Trim(edPanels.Text)) = 0 then Exit;
  with Bar.Panels.Add do
  begin
    SmallImages := cbSmallImages.Checked;
    Text := edPanels.Text;
    Tag := udPanelTag.Position;
  end;
  UpdateButtons;
end;

procedure TFourthShortcutBarEditorForm.btnRenamePanelClick(
  Sender: TObject);
begin
  if Bar.Panels.Count = 0 then Exit;
  if Length(Trim(edPanels.Text)) = 0 then Exit;
  Bar.Panels[Bar.PanelIndex].Text := edPanels.Text;
  edPanels.Text := '';
  UpdateButtons;
end;

procedure TFourthShortcutBarEditorForm.btnDeletePanelClick(
  Sender: TObject);
begin
  if (Bar.Panels.Count = 0) then Exit;
  if Bar.Panels[Bar.PanelIndex].Data <> nil then Bar.Panels[Bar.PanelIndex].Data := nil;
{$IFNDEF DELPHI4_UP}
  Bar.Panels[Bar.PanelIndex].Free;
{$ELSE}
  Bar.Panels.Delete(Bar.PanelIndex);
{$ENDIF}
  UpdateButtons;
end;

procedure TFourthShortcutBarEditorForm.btnAddBtnClick(Sender: TObject);
begin
  if (Bar.Panels.Count = 0) then Exit;
  with Bar.Panels[Bar.PanelIndex].Buttons.Add do
  begin
    ImageIndex := udImageIndex.Position;
    Tag := udBtnTag.Position;
    Text := edButtons.Text;
    cbButtons.Items.Add(Format('Button %d', [cbButtons.Items.Count-1]));
  end;
  UpdateButtons;
end;

procedure TFourthShortcutBarEditorForm.BarClick(Sender: TObject);
begin
  if Bar.Panels.Count = 0 then Exit;
  if Bar.PanelIndex < 0 then Exit;  
  edPanels.Text := Bar.Panels[Bar.PanelIndex].Text;
  cbSmallImages.Checked := Bar.Panels[Bar.PanelIndex].SmallImages;
  udPanelTag.Position := Bar.Panels[Bar.PanelIndex].Tag;
end;

procedure TFourthShortcutBarEditorForm.cbButtonsChange(Sender: TObject);
begin
  if cbButtons.ItemIndex = -1 then Exit;
  if cbButtons.ItemIndex = 0 then
  begin
    udImageIndex.Position := -1;
    udBtnTag.Position := -1;
    edButtons.Text := '';
  end else
  begin
    udImageIndex.Position := Bar.Panels[Bar.PanelIndex].Buttons[cbButtons.ItemIndex+1].ImageIndex;
    udBtnTag.Position := Bar.Panels[Bar.PanelIndex].Buttons[cbButtons.ItemIndex+1].Tag;
    edButtons.Text := Bar.Panels[Bar.PanelIndex].Buttons[cbButtons.ItemIndex+1].Text;
  end;
  UpdateButtons;
end;

procedure TFourthShortcutBarEditorForm.btnRenameBtnClick(Sender: TObject);
begin
  if (cbButtons.ItemIndex = 0) then Exit;
  with Bar.Panels[Bar.PanelIndex].Buttons[cbButtons.ItemIndex+1] do
  begin
    Text := edButtons.Text;
  end;
  UpdateButtons;
end;

procedure TFourthShortcutBarEditorForm.btnDeleteBtnClick(Sender: TObject);
begin
  if (cbButtons.ItemIndex = 0) then Exit;
  if Bar.Panels[Bar.PanelIndex].Buttons[cbButtons.ItemIndex+1].Data <> nil then
    Bar.Panels[Bar.PanelIndex].Buttons[cbButtons.ItemIndex+1].Data := nil;
{$IFNDEF DELPHI4_UP}
  Bar.Panels[Bar.PanelIndex].Buttons[cbButtons.ItemIndex+1].Free;
{$ELSE}
  Bar.Panels[Bar.PanelIndex].Buttons.Delete(cbButtons.ItemIndex+1);
{$ENDIF}
  cbButtons.Items.Delete(cbButtons.ItemIndex);
  UpdateButtons;
end;

procedure TFourthShortcutBarEditorForm.udImageIndexClick(Sender: TObject;
  Button: TUDBtnType);
begin
  if (cbButtons.ItemIndex = 0) then Exit;
  Bar.Panels[Bar.PanelIndex].Buttons[cbButtons.ItemIndex+1].ImageIndex := udImageIndex.Position;
end;

procedure TFourthShortcutBarEditorForm.udBtnTagClick(Sender: TObject;
  Button: TUDBtnType);
begin
  if (cbButtons.ItemIndex = 0) then Exit;
  Bar.Panels[Bar.PanelIndex].Buttons[cbButtons.ItemIndex+1].Tag := udBtnTag.Position;
end;

procedure TFourthShortcutBarEditorForm.udPanelTagClick(Sender: TObject;
  Button: TUDBtnType);
begin
  if (Bar.Panels.Count = 0) then Exit;
  Bar.Panels[Bar.PanelIndex].Tag := udPanelTag.Position;
end;

procedure TFourthShortcutBarEditorForm.cbSmallImagesClick(Sender: TObject);
begin
  if (Bar.Panels.Count = 0) then
  begin
    cbSmallImages.Checked := False;
    Exit;
  end;
  Bar.Panels[Bar.PanelIndex].SmallImages := cbSmallImages.Checked;
end;

end.
