
{*******************************************************}
{                                                       }
{       4th GUI Library (VCL Edition)                   }
{       Shortcut Bar Unit                               }
{                                                       }
{       Copyright (c) 1996,2003 Sergiy S. Tkachenko     }
{       e-mail: support@4thfebruary.tk                  }
{          Web: www.4thfebruary.tk                      }
{                                                       }
{*******************************************************}

unit FthSBar;

{$I FTHVER.INC}

interface

uses
  Messages, Windows, SysUtils, Classes, Controls, Forms,
{$IFDEF DELPHI4_UP}
  ImgList, ActnList,
{$ENDIF}
  StdCtrls, ExtCtrls, FthGraph, Graphics, ComCtrls;

type

{ TFourthShortcutBarColors }

  TFourthShortcutBarColors = class(TPersistent)
  private
    FBG: TColor;
    FBtnFace: TColor;
    FBtnLight: TColor;
    FBtnShadow: TColor;
    FBtnText: TColor;
    FBtnTextDisabled: TColor;
    FFace: TColor;
    FShadow: TColor;
    FLight: TColor;
    FDarkShadow: TColor;
    FHighlight: TColor;
    FUseDefault: Boolean;
    FOnChange: TNotifyEvent;
    procedure SetBG(Value: TColor);
    procedure SetBtnFace(Value: TColor);
    procedure SetBtnLight(Value: TColor);
    procedure SetBtnShadow(Value: TColor);
    procedure SetBtnText(Value: TColor);
    procedure SetBtnTextDisabled(Value: TColor);
    procedure SetFace(Value: TColor);
    procedure SetShadow(Value: TColor);
    procedure SetLight(Value: TColor);
    procedure SetDarkShadow(Value: TColor);
    procedure SetHighlight(Value: TColor);
    procedure SetUseDefault(Value: Boolean);
  protected
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  public
    constructor Create;
  published
    property Background: TColor read FBG write SetBG default clBtnShadow;
    property BtnFace: TColor read FBtnFace write SetBtnFace default clBtnShadow;
    property BtnLight: TColor read FBtnLight write SetBtnLight default clBtnHighlight;
    property BtnShadow: TColor read FBtnShadow write SetBtnShadow default cl3DDkShadow;
    property BtnText: TColor read FBtnText write SetBtnText default clBtnHighlight;
    property BtnTextDisabled: TColor read FBtnText write SetBtnTextDisabled default clBtnFace;
    property Face: TColor read FFace write SetFace default clBtnFace;
    property Shadow: TColor read FShadow write SetShadow default clBtnShadow;
    property Light: TColor read FLight write SetLight default cl3DLight;
    property DarkShadow: TColor read FDarkShadow write SetDarkShadow default cl3DDkShadow;
    property Highlight: TColor read FHighlight write SetHighlight default clBtnHighlight;
    property UseDefault: Boolean read FUseDefault write SetUseDefault default True;
  end;

{ TFourthShortcutBar }

  TFourthShortcutBarButtonState = (sbbsHot, sbbsDefault, sbbsPressed);

  TFourthShortcutBar = class;
  TFourthShortcutBarPanel = class;
  TFourthShortcutBarButton = class;
  TFourthShortcutBarButtons = class;

{$IFDEF DELPHI4_UP}
  TFourthShortcutBarButtonActionLink = class(TActionLink)
  protected
    FClient: TFourthShortcutBarButton;
    procedure AssignClient(AClient: TObject); override;
    function IsCaptionLinked: Boolean; override;
    function IsEnabledLinked: Boolean; override;
    function IsVisibleLinked: Boolean; override;
    function IsOnExecuteLinked: Boolean; override;
    procedure SetCaption(const Value: string); override;
    procedure SetEnabled(Value: Boolean); override;
    procedure SetVisible(Value: Boolean); override;
    procedure SetOnExecute(Value: TNotifyEvent); override;
  end;

  TFourthShortcutBarButtonActionLinkClass = class of TFourthShortcutBarButtonActionLink;
{$ENDIF}

{ TFourthShortcutBarButton }

  TFourthShortcutBarButton = class(TCollectionItem)
  private
{$IFDEF DELPHI4_UP}
    FActionLink: TFourthShortcutBarButtonActionLink;
{$ENDIF}
    FButtonRect: TRect;
    FData: Pointer;
    FDeleting: Boolean;
    FDisplayRect: TRect;
    FEnabled: Boolean;
    FCaption: TCaption;
    FImageIndex: Integer;
    FOnClick: TNotifyEvent;
    FTag: LongInt;                                              // offered by JH
    FVisible: Boolean;
    procedure SetData(const Value: Pointer);
    procedure SetCaption(const Value: TCaption);
    procedure SetOnClick(const Value: TNotifyEvent);
    procedure SetImageIndex(Value: Integer);
    procedure SetEnabled(Value: Boolean);
  protected
{$IFDEF DELPHI3_UP}
    function GetDisplayName: string; override;
{$ENDIF}
    procedure Click;
{$IFDEF DELPHI4_UP}
    procedure DoActionChange(Sender: TObject);
    procedure ActionChange(Sender: TObject; CheckDefaults: Boolean); dynamic;
    function GetAction: TBasicAction;
    function GetActionLinkClass: TFourthShortcutBarButtonActionLinkClass; dynamic;
    procedure SetAction(Value: TBasicAction);
    procedure SetVisible(Value: Boolean);
    function IsCaptionStored: Boolean;
    function IsEnabledStored: Boolean;
//    function IsHintStored: Boolean;
//    function IsHelpContextStored: Boolean;
    function IsVisibleStored: Boolean;
    function IsOnClickStored: Boolean;
    property ActionLink: TFourthShortcutBarButtonActionLink read FActionLink write FActionLink;
{$ENDIF}
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation);
    property Data: Pointer read FData write SetData;
  published
{$IFDEF DELPHI4_UP}
    property Action: TBasicAction read GetAction write SetAction;
    property Visible: Boolean read FVisible write SetVisible default True;
{$ENDIF}
    property Enabled: Boolean read FEnabled write SetEnabled default True;
    property Caption: TCaption read FCaption write SetCaption;
    property ImageIndex: Integer read FImageIndex write SetImageIndex default -1;
    property Tag: LongInt read FTag write FTag default 0;       // offered by JH
    property OnClick: TNotifyEvent read FOnClick write SetOnClick;
  end;

{ TFourthShortcutBarButtons }

  TFourthShortcutBarButtons = class(TCollection)
  private
    FFourthShortcutBarPanel: TFourthShortcutBarPanel;
    function GetItem(Index: Integer): TFourthShortcutBarButton;
    procedure SetItem(Index: Integer; Value: TFourthShortcutBarButton);
  protected
{$IFDEF DELPHI3_UP}
    function GetOwner: TPersistent; override;
{$ENDIF}
    procedure Update(Item: TCollectionItem); override;
  public
    constructor Create(FourthShortcutBarPanel: TFourthShortcutBarPanel);
    function Add: TFourthShortcutBarButton;
    procedure Delete(Index: Integer); //override;
    property Items[Index: Integer]: TFourthShortcutBarButton read GetItem write SetItem; default;
  end;

{ TFourthShortcutBarPanelActionLink }

{$IFDEF DELPHI4_UP}
  TFourthShortcutBarPanelActionLink = class(TActionLink)
  protected
    FClient: TFourthShortcutBarPanel;
    procedure AssignClient(AClient: TObject); override;
    function IsCaptionLinked: Boolean; override;
    function IsOnExecuteLinked: Boolean; override;
    procedure SetCaption(const Value: string); override;
    procedure SetOnExecute(Value: TNotifyEvent); override;
  end;

  TFourthShortcutBarPanelActionLinkClass = class of TFourthShortcutBarPanelActionLink;
{$ENDIF}

{ TFourthShortcutBarPanel }

  TFourthShortcutBarPanel = class(TCollectionItem)
  private
{$IFDEF DELPHI4_UP}
    FActionLink: TFourthShortcutBarPanelActionLink;
{$ENDIF}
    FFourthShortcutBar: TFourthShortcutBar;
    FData: Pointer;
    FDeleting: Boolean;
    FDisplayRect: TRect;
    FCaption: TCaption;
    FButtons: TFourthShortcutBarButtons;
    FOnClick: TNotifyEvent;
    FSmallImages: Boolean;
    FTag: LongInt;                                              // offered by JH
    procedure SetData(const Value: Pointer);
    procedure SetCaption(const Value: TCaption);
    procedure SetButtons(const Value: TFourthShortcutBarButtons);
    procedure SetOnClick(const Value: TNotifyEvent);
    procedure SetSmallImages(const Value: Boolean);
  protected
{$IFDEF DELPHI3_UP}
    function GetDisplayName: string; override;
{$ENDIF}
    procedure Click;
{$IFDEF DELPHI4_UP}
    procedure DoActionChange(Sender: TObject);
    procedure ActionChange(Sender: TObject; CheckDefaults: Boolean); dynamic;
    function GetAction: TBasicAction;
    function GetActionLinkClass: TFourthShortcutBarPanelActionLinkClass; dynamic;
    procedure SetAction(Value: TBasicAction);
    property ActionLink: TFourthShortcutBarPanelActionLink read FActionLink write FActionLink;
{$ENDIF}
    property Bar: TFourthShortcutBar read FFourthShortcutBar;
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    property Data: Pointer read FData write SetData;
  published
{$IFDEF DELPHI4_UP}
    property Action: TBasicAction read GetAction write SetAction;
{$ENDIF}
    property Buttons: TFourthShortcutBarButtons read FButtons write SetButtons;
    property SmallImages: Boolean read FSmallImages write SetSmallImages;
    property Tag: LongInt read FTag write FTag default 0;       // offered by JH
    property Caption: TCaption read FCaption write SetCaption;
    property OnClick: TNotifyEvent read FOnClick write SetOnClick;
  end;

{ TFourthShortcutBarPanels }

  TFourthShortcutBarPanels = class(TCollection)
  private
    FFourthShortcutBar: TFourthShortcutBar;
    function GetItem(Index: Integer): TFourthShortcutBarPanel;
    procedure SetItem(Index: Integer; Value: TFourthShortcutBarPanel);
  protected
{$IFDEF DELPHI3_UP}
    function GetOwner: TPersistent; override;
{$ENDIF}
    procedure Update(Item: TCollectionItem); override;
  public
    constructor Create(FourthShortcutBar: TFourthShortcutBar);
    function Add: TFourthShortcutBarPanel;
    procedure Delete(Index: Integer);
    procedure Assign(Source: TPersistent); override;
    property Items[Index: Integer]: TFourthShortcutBarPanel read GetItem write SetItem; default;
  end;

{ TFourthShortcutBar }

  TFourthShortcutBarStyle = (sbs2000, sbs2002);

  TDrawPanelEvent = procedure(FourthShortcutBar: TFourthShortcutBar;
    Panel: TFourthShortcutBarPanel; const Rect: TRect) of object;

  TFourthShortcutBar = class(TCustomControl)
  private
    FPanelCursor: TCursor;
    FPanels: TFourthShortcutBarPanels;
    FPanelRect: TRect;        { rectangle for active panel buttons }
    FScrollTimer: TTimer;     { timer for scroll repeat }
    FScrollTop: Integer;      { first visible button }
    FScrollBottom: Integer;   { last visible button }
    FUpR: TRect;              { rect for up scroll button }
    FUpActive: Boolean;       { upscroller visible }
    FUpIsDown: Boolean;       { upscroller depressed }
    FDownR: TRect;            { rect for down scroll button }
    FDownActive: Boolean;     { downscroller visible }
    FDownIsDown: Boolean;     { downscroller depressed }
    FPanelIndex: Integer;     { current panel }
    FTabHeight: Integer;      { height of the clickable tabs }
    FPanelHeight: Integer;    { current height of the buttons area }
    FDownTab: Integer;        { current mousedown Tab }
    FDownButton: Integer;     { current mousedown Button }
    FMoveButton: Integer;     { current mousemove Button }
    FMoveTab: Integer;        { current mousemove Tab }
    FImages: TImagelist;      { large images, caption displayed below image }
    FSmallImages: TImageList; { small images, caption displayed right of image }
    FColors: TFourthShortcutBarColors; { customized colors }
    FStyle: TFourthShortcutBarStyle; { style of look: 2000 or 2002}
    FUseSystemFont: Boolean;
    FSelectedButton: Integer;
    procedure SetPanelCursor(Value: TCursor);
    procedure SetPanels(Value: TFourthShortcutBarPanels);
    procedure SetStyle(Value: TFourthShortcutBarStyle);
    procedure SetUseSystemFont(Value: Boolean);
    procedure UpdatePanelHeader(Index: Integer; State: TFourthShortcutBarButtonState);
    procedure UpdatePanels(UpdateRects: Boolean);
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure SetPanelIndex(const Value: Integer);
    procedure SetTabHeight(const Value: Integer);
    procedure SetImages(const Value: TImagelist);
    procedure UpdatePanel(First: Boolean; Index: Integer);
    procedure UpdateButton(aPanel, aButton: Integer; State: TFourthShortcutBarButtonState);
    procedure SetSmallImages(const Value: TImageList);
    function GetButtonArea(pt: TPoint): Integer;
    procedure DrawDownRaised;
    procedure DrawDownSunken;
    procedure DrawUpRaised;
    procedure DrawUpSunken;
    procedure DrawRaised(R: TRect);
    procedure DrawUpArrow(R: TRect);
    procedure DrawDownArrow(R: TRect);
    procedure DrawSunken(R: TRect);
    procedure TimedScroll(Sender: TObject);
  protected
    procedure CMDesignHitTest(var Msg: TCMDesignHitTest); message CM_DESIGNHITTEST;
    procedure ClickAtPos(APoint: TPoint);
    procedure CreateWnd; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure ColorsChange(Sender: TObject);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure Assign(Source: TPersistent); override;
    procedure ApplyStandardColors;
    function GetPanelAtPos(APoint: TPoint): Integer;
    function GetButtonAtPos(Pt: TPoint): Integer;
  published
    property Align default alLeft;
    property Enabled;
    property Font stored True;
    property Images: TImageList read FImages write SetImages;
    property Hint;
    property Panels: TFourthShortcutBarPanels read FPanels write SetPanels;
    property PanelIndex: Integer read FPanelIndex write SetPanelIndex;
    property Colors: TFourthShortcutBarColors read FColors write FColors;
    property PanelCursor: TCursor read FPanelCursor write SetPanelCursor default {$IFDEF DEFLPHI3_UP}crHandPoint{$ELSE}crDefault{$ENDIF};
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property SelectedButton: Integer read FSelectedButton default -1;    
    property ShowHint;
    property SmallImages: TImageList read FSmallImages write SetSmallImages;
    property Style: TFourthShortcutBarStyle read FStyle write SetStyle;
    property TabHeight: Integer read FTabHeight write SetTabHeight;
    property Visible;
    property UseSystemFont: Boolean read FUseSystemFont write SetUseSystemFont;
    property OnClick;
  {$IFDEF DELPHI5_UP}
    property OnContextPopup;
  {$ENDIF}
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  {$IFDEF DELPHI4_UP}
    property OnResize;
  {$ENDIF}
  end;

implementation

uses Consts;

{ TFourthShortcutBarColors }

constructor TFourthShortcutBarColors.Create;
begin
  FBG := clBtnShadow;
  FBtnFace := clBtnShadow;
  FBtnLight := clBtnHighlight;
  FBtnShadow := cl3DDkShadow;
  FBtnText := clBtnHighlight;
  FBtnTextDisabled := clBtnFace;  
  FFace := clBtnFace;
  FShadow := clBtnShadow;
  FLight := cl3dLight;
  FDarkShadow := cl3dDkShadow;
  FHighlight := clBtnHighlight;
  FUseDefault := True;
end;

procedure TFourthShortcutBarColors.SetBG(Value: TColor);
begin
  if (FBG <> Value) then
  begin
    FBG := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthShortcutBarColors.SetBtnFace(Value: TColor);
begin
  if (FBtnFace <> Value) then
  begin
    FBtnFace := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthShortcutBarColors.SetBtnLight(Value: TColor);
begin
  if (FBtnLight <> Value) then
  begin
    FBtnLight := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthShortcutBarColors.SetBtnShadow(Value: TColor);
begin
  if (FBtnShadow <> Value) then
  begin
    FBtnShadow := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthShortcutBarColors.SetBtnText(Value: TColor);
begin
  if (FBtnText <> Value) then
  begin
    FBtnText := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthShortcutBarColors.SetBtnTextDisabled(Value: TColor);
begin
  if (FBtnTextDisabled <> Value) then
  begin
    FBtnTextDisabled := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthShortcutBarColors.SetFace(Value: TColor);
begin
  if (FFace <> Value) then
  begin
    FFace := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthShortcutBarColors.SetShadow(Value: TColor);
begin
  if (FShadow <> Value) then
  begin
    FShadow := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthShortcutBarColors.SetLight(Value: TColor);
begin
  if (FLight <> Value) then
  begin
    FLight := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthShortcutBarColors.SetDarkShadow(Value: TColor);
begin
  if (FDarkShadow <> Value) then
  begin
    FDarkShadow := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthShortcutBarColors.SetHighlight(Value: TColor);
begin
  if (FHighlight <> Value) then
  begin
    FHighlight := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthShortcutBarColors.SetUseDefault(Value: Boolean);
begin
  if (FUseDefault <> Value) then
  begin
    FUseDefault := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

{ TFourthShortcutBarButtonActionLink }

procedure TFourthShortcutBarButtonActionLink.AssignClient(AClient: TObject);
begin
  FClient := AClient as TFourthShortcutBarButton;
end;

function TFourthShortcutBarButton.GetActionLinkClass: TFourthShortcutBarButtonActionLinkClass;
begin
  Result := TFourthShortcutBarButtonActionLink;
end;

function TFourthShortcutBarButtonActionLink.IsCaptionLinked: Boolean;
begin
  Result := inherited IsCaptionLinked and
    (FClient.Caption = (Action as TCustomAction).Caption);
end;

function TFourthShortcutBarButtonActionLink.IsEnabledLinked: Boolean;
begin
  Result := inherited IsEnabledLinked and
    (FClient.Enabled = (Action as TCustomAction).Enabled);
end;

function TFourthShortcutBarButtonActionLink.IsOnExecuteLinked: Boolean;
begin
  Result := inherited IsOnExecuteLinked and
    (@FClient.OnClick = @Action.OnExecute);
end;

function TFourthShortcutBarButtonActionLink.IsVisibleLinked: Boolean;
begin
  Result := inherited IsVisibleLinked and
    (FClient.Visible = (Action as TCustomAction).Visible);
end;

procedure TFourthShortcutBarButtonActionLink.SetCaption(const Value: string);
begin
  if IsCaptionLinked then FClient.Caption := Value;
end;

procedure TFourthShortcutBarButtonActionLink.SetEnabled(Value: Boolean);
begin
  if IsEnabledLinked then FClient.Enabled := Value;
end;

procedure TFourthShortcutBarButtonActionLink.SetOnExecute(Value: TNotifyEvent);
begin
  if IsOnExecuteLinked then FClient.OnClick := Value;
end;

procedure TFourthShortcutBarButtonActionLink.SetVisible(Value: Boolean);
begin
  if IsVisibleLinked then FClient.Visible := Value;
end;

function TFourthShortcutBarButton.IsCaptionStored: Boolean;
begin
  Result := (ActionLink = nil) or not ActionLink.IsCaptionLinked;
end;

function TFourthShortcutBarButton.IsEnabledStored: Boolean;
begin
  Result := (ActionLink = nil) or not ActionLink.IsEnabledLinked;
end;
{
function TFourthShortcutBarButton.IsHintStored: Boolean;
begin
  Result := (ActionLink = nil) or not ActionLink.IsHintLinked;
end;

function TFourthShortcutBarButton.IsHelpContextStored: Boolean;
begin
  Result := (ActionLink = nil) or not ActionLink.IsHelpContextLinked;
end;
}
function TFourthShortcutBarButton.IsVisibleStored: Boolean;
begin
  Result := (ActionLink = nil) or not ActionLink.IsVisibleLinked;
end;

function TFourthShortcutBarButton.IsOnClickStored: Boolean;
begin
  Result := (ActionLink = nil) or not ActionLink.IsOnExecuteLinked;
end;

{ TFourthShortcutBarPanelActionLink }

procedure TFourthShortcutBarPanelActionLink.AssignClient(AClient: TObject);
begin
  FClient := AClient as TFourthShortcutBarPanel;
end;

function TFourthShortcutBarPanel.GetActionLinkClass: TFourthShortcutBarPanelActionLinkClass;
begin
  Result := TFourthShortcutBarPanelActionLink;
end;

function TFourthShortcutBarPanelActionLink.IsCaptionLinked: Boolean;
begin
  Result := inherited IsCaptionLinked and
    (FClient.Caption = (Action as TCustomAction).Caption);
end;

function TFourthShortcutBarPanelActionLink.IsOnExecuteLinked: Boolean;
begin
  Result := inherited IsOnExecuteLinked and
    (@FClient.OnClick = @Action.OnExecute);
end;

procedure TFourthShortcutBarPanelActionLink.SetCaption(const Value: string);
begin
  if IsCaptionLinked then FClient.Caption := Value;
end;


procedure TFourthShortcutBarPanelActionLink.SetOnExecute(Value: TNotifyEvent);
begin
  if IsOnExecuteLinked then FClient.OnClick := Value;
end;

{ TFourthShortcutBarPanel }

constructor TFourthShortcutBarPanel.Create(Collection: TCollection);
begin
  inherited Create(Collection);
  FDeleting := False;
  FButtons := TFourthShortcutBarButtons.Create(Self);
end;

destructor TFourthShortcutBarPanel.Destroy;
begin
  FDeleting := True;
  Data := nil;
  FButtons.Free;
  inherited;
end;

procedure TFourthShortcutBarPanel.Assign(Source: TPersistent);
begin
  if (Source is TFourthShortcutBarPanel) then
  begin
    Caption := TFourthShortcutBarPanel(Source).Caption;
    Data := TFourthShortcutBarPanel(Source).Data;
    Tag := TFourthShortcutBarPanel(Source).Tag;
  end else
    inherited Assign(Source);
end;

{$IFDEF DELPHI3_UP}
function TFourthShortcutBarPanel.GetDisplayName: string;
begin
  Result := Caption;
  if Result = '' then Result := inherited GetDisplayName;
end;
{$ENDIF}

procedure TFourthShortcutBarPanel.SetData(const Value: Pointer);
begin
  if not FDeleting and (Value <> Data) then
  begin
    FData := Value;
    Changed(False);
  end;
end;

procedure TFourthShortcutBarPanel.SetCaption(const Value: TCaption);
begin
  if (FCaption <> Value) then
  begin
    FCaption := Value;
    Changed(False);
  end;
end;

procedure TFourthShortcutBarPanel.SetButtons(const Value: TFourthShortcutBarButtons);
begin
  FButtons.Assign(Value);
end;

procedure TFourthShortcutBarPanel.SetOnClick(const Value: TNotifyEvent);
begin
  FOnClick := Value;
end;

procedure TFourthShortcutBarPanel.Click;
begin
  if Assigned(OnClick) then OnClick(Self);
end;

procedure TFourthShortcutBarPanel.SetSmallImages(const Value: boolean);
begin
  if (FSmallImages <> Value) then
  begin
    FSmallImages := Value;
    Changed(True);
  end;
end;

function TFourthShortcutBarPanel.GetAction: TBasicAction;
begin
  if ActionLink <> nil then
    Result := ActionLink.Action
  else
    Result := nil;
end;

procedure TFourthShortcutBarPanel.SetAction(Value: TBasicAction);
begin
  if (Value = nil) then
  begin
    if Assigned(ActionLink) then
    begin
      ActionLink.Free;
      ActionLink := nil;
//      Exclude(FControlStyle, csActionClient);
    end;
  end
  else
//    if Assigned(FourthShortcutBar) then
    begin
//      Include(FControlStyle, csActionClient);
      if ActionLink = nil then
        ActionLink := GetActionLinkClass.Create(Self);
      ActionLink.Action := Value;
      ActionLink.OnChange := DoActionChange;
      ActionChange(Value, csLoading in Value.ComponentState);
//      Value.FreeNotification(FourthShortcutBar);
    end;
end;

procedure TFourthShortcutBarPanel.DoActionChange(Sender: TObject);
begin
  if Sender = Action then ActionChange(Sender, False);
end;

procedure TFourthShortcutBarPanel.ActionChange(Sender: TObject; CheckDefaults: boolean);
begin
  if Sender is TCustomAction then
    with TCustomAction(Sender) do
    begin
      if not CheckDefaults or (Self.Caption = '') then
        Self.Caption := Caption;
      if not CheckDefaults or not Assigned(Self.OnClick) then
        Self.OnClick := OnExecute;
    end;
end;

{ TFourthShortcutBarPanels }

constructor TFourthShortcutBarPanels.Create(FourthShortcutBar: TFourthShortcutBar);
begin
  inherited Create(TFourthShortcutBarPanel);
  FFourthShortcutBar := FourthShortcutBar;
end;

function TFourthShortcutBarPanels.Add: TFourthShortcutBarPanel;
begin
  Self.BeginUpdate;
  try
    Result := TFourthShortcutBarPanel(inherited Add);
    Result.FFourthShortcutBar := Self.FFourthShortcutBar;
  finally
    Self.EndUpdate;
  end;
end;

procedure TFourthShortcutBarPanels.Delete(Index: Integer);
begin
  Self.BeginUpdate;
  try
    inherited Delete(Index);
  finally
    Self.EndUpdate;
  end;
  if Index > 0 then
    FFourthShortcutBar.SetPanelIndex(Index-1);
  if (Index = 0) and (Count = 0) then
    FFourthShortcutBar.SetPanelIndex(-1);
  if (Index = 0) and (Count <> 0) then
    FFourthShortcutBar.SetPanelIndex(Index);  
end;

function TFourthShortcutBarPanels.GetItem(Index: Integer): TFourthShortcutBarPanel;
begin
  Result := TFourthShortcutBarPanel(inherited GetItem(Index));
end;

{$IFDEF DELPHI3_UP}
function TFourthShortcutBarPanels.GetOwner: TPersistent;
begin
  Result := FFourthShortcutBar;
end;
{$ENDIF}

procedure TFourthShortcutBarPanels.SetItem(Index: Integer; Value: TFourthShortcutBarPanel);
begin
  Self.BeginUpdate;
  try
    inherited SetItem(Index, Value);
  finally
    Self.EndUpdate;
  end;
end;

procedure TFourthShortcutBarPanels.Update(Item: TCollectionItem);
begin
  if Item <> nil then
    FFourthShortcutBar.UpdatePanelHeader(Item.Index, sbbsDefault) else
    FFourthShortcutBar.UpdatePanels(True);
end;

{ TFourthShortcutBar }

constructor TFourthShortcutBar.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := [csCaptureMouse, csClickEvents{, csDoubleClicks}, csOpaque];
  Canvas.Font := Screen.IconFont;
  Width := 100;
  Align := alLeft;
  FTabHeight := 22;
  FPanelCursor := {$IFDEF DELPHI3_UP}crHandPoint{$ELSE}crDefault{$ENDIF};
  FPanels := TFourthShortcutBarPanels.Create(Self);
  FColors := TFourthShortcutBarColors.Create;
  FColors.OnChange := ColorsChange;
  ParentFont := True;
  FMoveButton := -1;
  FMoveTab := -1;
  FUseSystemFont := True;
end;

destructor TFourthShortcutBar.Destroy;
begin
  FPanels.Free;
  inherited Destroy;
end;

procedure TFourthShortcutBar.CreateWnd;
begin
  inherited CreateWnd;
  UpdatePanels(True);
end;

procedure TFourthShortcutBar.SetPanelCursor(Value: TCursor);
begin
  if (FPanelCursor <> Value) then
  begin
    FPanelCursor := Value;
  end;
end;

procedure TFourthShortcutBar.SetPanels(Value: TFourthShortcutBarPanels);
begin
  FPanels.Assign(Value);
end;

{   }

procedure TFourthShortcutBar.UpdatePanelHeader(Index: Integer;
  State: TFourthShortcutBarButtonState);
var
  S: string;
  R: TRect;
begin
  if HandleAllocated then
  begin
    with Panels[Index] do
    begin
      S := Caption;
      R := FDisplayRect;
      Canvas.Brush.Color := FColors.FFace;
      Canvas.FillRect(R);
      Canvas.Brush.Style := bsClear;

      if FStyle = sbs2000 then
      begin
        case State of
          sbbsPressed:
            begin
              Frame3D(Canvas, R, FColors.FDarkShadow, FColors.FHighlight, 1);
              Frame3D(Canvas, R, FColors.FShadow, FColors.FLight, 1);
            end;
          sbbsDefault:
            begin
              if FPanelIndex = Index then
              begin
                Frame3D(Canvas, R, FColors.FHighlight, FColors.FDarkShadow, 1);
                Inc(R.Right); Dec(R.Left); Dec(R.Top);
                Frame3D(Canvas, R, FColors.FHighlight, FColors.FShadow, 1);
                Dec(R.Right); Inc(R.Left); Inc(R.Top);
              end else
              begin
                Frame3D(Canvas, R, FColors.FHighlight, FColors.FShadow, 1);
                Inc(R.Right); Dec(R.Left); Dec(R.Top);
                Frame3D(Canvas, R, FColors.FHighlight, FColors.FShadow, 1);
                Dec(R.Right); Inc(R.Left); Inc(R.Top);
              end;
            end;
          sbbsHot:
            begin
              Frame3D(Canvas, R, FColors.FHighlight, FColors.FDarkShadow, 1);
              Frame3D(Canvas, R, FColors.FLight, FColors.FShadow, 1);
            end;
        end;
      end;

      if FStyle = sbs2002 then
      begin
        case State of
          sbbsPressed:
            begin
              Frame3D(Canvas, R, FColors.FDarkShadow, FColors.FHighlight, 1);
              Frame3D(Canvas, R, FColors.FShadow, FColors.FLight, 1);
            end;
          sbbsDefault:
            begin
              if FPanelIndex = Index then
                Frame3D(Canvas, R, clBtnHighlight, clBtnShadow, 1) else
                Frame3D(Canvas, R, clBtnHighlight, clBtnShadow, 1);
            end;
          sbbsHot:
            begin
              Frame3D(Canvas, R, FColors.FHighlight, FColors.FDarkShadow, 1);
              Frame3D(Canvas, R, FColors.FLight, FColors.FShadow, 1);
            end;
        end;
      end;

      Dec(R.Right, 3);
      Inc(R.Left, 3);

      if FUseSystemFont then
        Canvas.Font.Color := Screen.IconFont.Color else
        Canvas.Font.Color := Font.Color;
      if Length(S) <> 0 then
        DrawText(Canvas.Handle, @S[1], -1, R, DT_SINGLELINE or DT_CENTER or
          DT_VCENTER or DT_END_ELLIPSIS);
    end;
  end;
end;

procedure TFourthShortcutBar.UpdateButton(aPanel, aButton: Integer;
  State: TFourthShortcutBarButtonState);
var
  R, btnR: TRect;
  fsbButton: TFourthShortcutBarButton;
  img, h, w, dx, dy: Integer;
  s: string;
  small: Boolean;
begin
  if Panels.Count = 0 then Exit;
  if Panels[aPanel].Buttons.Count = 0 then Exit;
  fsbButton := Panels[aPanel].Buttons[aButton];
  R := fsbButton.FDisplayRect;
  btnR := fsbButton.FButtonRect;
  small := Panels[aPanel].SmallImages;
  img := fsbButton.ImageIndex;

  if (FStyle = sbs2000) then { Outlook 2000 }
  begin
    case State of
      sbbsPressed:
        if Panels[aPanel].Buttons[aButton].Enabled then
          Frame3D(Canvas, btnR, FColors.FDarkShadow, FColors.FFace, 1);
      sbbsHot:
        if Panels[aPanel].Buttons[aButton].Enabled then
          Frame3D(Canvas, btnR, FColors.FFace, FColors.FDarkShadow, 1);
      sbbsDefault:
        Frame3D(Canvas, btnR, FColors.FBG, FColors.FBG, 1);
    end;
    if Panels[aPanel].Buttons[aButton].Enabled then
      Canvas.Font.Color := FColors.FBtnText else
      Canvas.Font.Color := FColors.FBtnTextDisabled;
  end;

  if (FStyle = sbs2002) then { Outlook XP }
  begin
    case State of
      sbbsPressed:
        begin
          if Panels[aPanel].Buttons[aButton].Enabled then
          begin
            Canvas.Brush.Color := GetXPPressedHighlight;
            Canvas.FillRect(btnR);
            Frame3D(Canvas, btnR, clHighlight, clHighlight, 1);
          end;
        end;
      sbbsHot:
        begin
          if Panels[aPanel].Buttons[aButton].Enabled then
          begin
            Canvas.Brush.Color := GetXPHotHighlight;
            Canvas.FillRect(btnR);
            Frame3D(Canvas, btnR, clHighlight, clHighlight, 1);
          end;
        end;
      sbbsDefault:
        begin
          Canvas.Brush.Color := Get2002Bg(FColors.Background);
          Canvas.FillRect(btnR);
          Frame3D(Canvas, btnR, Get2002Bg(FColors.Background), Get2002Bg(FColors.Background), 1);
        end;
    end;
    Canvas.Brush.Color := Get2002Bg(FColors.Background);
    if Panels[aPanel].Buttons[aButton].Enabled then
    begin
      if FUseSystemFont then
        Canvas.Font.Color := Screen.IconFont.Color else
        Canvas.Font.Color := FColors.FBtnText;
    end else
      Canvas.Font.Color := FColors.FBtnTextDisabled;
  end; // end XP

  if Small then
  begin
    W := 16;
    H := 16;
    if Assigned(FSmallImages) then
    begin
      W := FSmallImages.Width;
      H := FSmallImages.height;
    end;
  end else
  begin
    W := 32;
    H := 32;
    if Assigned(FImages) then
    begin
      W := FImages.Width;
      H := FImages.Height;
    end;
  end;

  dx := (btnR.Right - BtnR.Left - W) div 2;
  dy := (btnR.Bottom - BtnR.Top - H) div 2;

  if Small then
    FSmallImages.Draw(Canvas, btnR.Left + dx, btnR.Top + dy, img) else
    FImages.Draw(Canvas, btnR.Left + dx, btnR.Top + dy, img);

  S := fsbButton.Caption;

  if Small then
  begin
    R.Left := btnR.Right + 4;
    DrawText(Canvas.Handle, @S[1], -1, R, DT_SINGLELINE or DT_LEFT or DT_END_ELLIPSIS);
  end else
  begin
    R.Top := btnR.Bottom + 2;
    DrawText(Canvas.Handle, @S[1], -1, R, DT_WORDBREAK or DT_CENTER or DT_TOP or DT_END_ELLIPSIS);
  end;
end;

// drawing of scroll buttons
procedure TFourthShortcutBar.DrawRaised(R: TRect);
begin
  with Canvas do
  begin
    Font := Self.Font;
    Brush.Color := FColors.FFace;
    FillRect(R);
    Frame3D(Canvas, R, FColors.FLight, FColors.FDarkShadow, 1);
    Frame3D(Canvas, R, FColors.FHighlight, FColors.FShadow, 1);
    Brush.Style := bsClear;
  end;
end;

procedure TFourthShortcutBar.DrawSunken(R:TRect);
begin
  with Canvas do
  begin
    Font := Self.Font;
    Brush.Color := FColors.FFace;
    FillRect(R);
    Pen.Color := FColors.FDarkShadow;
    MoveTo(R.Right-1,R.top);
    LineTo(R.left,R.top);
    lineto(R.left,R.bottom-1);
    pen.color:= FColors.FLight;
    moveto(R.left,R.bottom);
    lineto(R.right,R.bottom);
    lineto(R.right,R.top);
    pen.color:= FColors.FShadow;
    moveto(R.right-2,R.top+1);
    lineto(R.left+1,R.top+1);
    lineto(R.left+1,R.bottom-2);
    pen.color:= FColors.FFace;
    moveto(R.left+1,R.bottom-1);
    lineto(R.right-1,R.bottom-1);
    lineto(R.right-1,R.top+1);
    brush.style := bsclear;
  end;
end;

procedure TFourthShortcutBar.DrawUpArrow(R: TRect);
begin
  with Canvas do
  begin
    Pen.Color := FColors.FDarkShadow;
    moveto(R.left+5,R.top+11);
    Lineto(R.left+13,R.top+11);
    moveto(R.left+6,R.top+10);
    Lineto(R.left+12,R.top+10);
    moveto(R.left+7,R.top+9);
    Lineto(R.left+11,R.top+9);
    moveto(R.left+8,R.top+8);
    Lineto(R.left+10,R.top+8);
    moveto(R.left+9,R.top+7);
    Lineto(R.left+9,R.top+7);
  end;
end;

procedure TFourthShortcutBar.DrawDownArrow(R: TRect);
begin
  with Canvas do
  begin
    Pen.Color := FColors.FDarkShadow;
    moveto(R.left+5, R.top+8);
    Lineto(R.left+13,R.top+8);
    moveto(R.left+6,R.top+9);
    Lineto(R.left+12,R.top+9);
    moveto(R.left+7,R.top+10);
    Lineto(R.left+11,R.top+10);
    moveto(R.left+8,R.top+11);
    Lineto(R.left+10,R.top+11);
    moveto(R.left+9,R.top+12);
    Lineto(R.left+9,R.top+12);
  end;
end;

procedure TFourthShortcutBar.DrawUpRaised;
begin
  DrawRaised(FUpR);
  DrawUpArrow(FUpR);
end;

procedure TFourthShortcutBar.DrawUpSunken;
begin
  DrawSunken(FUpR);
  DrawUpArrow(FUpR);
end;

procedure TFourthShortcutBar.DrawDownRaised;
begin
  DrawRaised(FDownR);
  DrawDownArrow(FDownR);
end;

procedure TFourthShortcutBar.DrawDownSunken;
begin
  DrawSunken(FDownR);
  DrawDownArrow(FDownR);
end;

procedure TFourthShortcutBar.UpdatePanel(First: Boolean; Index: Integer);
var
  I, Count, ButtonPos, W, H, dx: Integer;
  R, btnR: TRect;
  Small: Boolean;
  ZoneHeight: Integer;
  ZoneR: TRect;
begin
  R := FPanelRect;
  if FStyle = sbs2000 then Canvas.Brush.Color := FColors.FBG;
  if FStyle = sbs2002 then Canvas.Brush.Color := Get2002Bg(FColors.Background);
  Canvas.FillRect(FPanelRect);

  FUpR := Rect(R.Right - 25, R.Top + 5, R.Right - 6, R.Top + 24);
  FUpActive := False;

  FDownR := Rect(R.Right - 25, R.Bottom - 25, R.Right - 6, R.Bottom - 6);
  FDownActive := False;

  Small := Panels[Index].SmallImages;
  if Panels[Index].Buttons = nil then Exit;                    // JH 30-jun-2001
  Count := Panels[Index].Buttons.Count;
  if Count = 0 then Exit;
  ButtonPos := FPanelRect.Top + 2;
  if Small then
  begin
    if Assigned(FSmallImages) then
    begin
      H := FSmallImages.Height + 2;
      W := FSmallImages.Width + 2;
      ZoneHeight := H + 7;
      ZoneR := Rect(R.Left + 2, ButtonPos, R.Right - 2, ButtonPos + ZoneHeight - 1);
      btnR := Rect(ZoneR.Left + 2, ZoneR.Top + 2, ZoneR.Left + 4 + W, ZoneR.Top + 4 + H);
    end else
      Exit;                         // only display buttons when images assigned
  end else
{ if Large Images }
  begin
    if Assigned(FImages) then
    begin
      H := FImages.Height + 4;
      W := FImages.Width + 2;
      ZoneHeight := H + Canvas.TextHeight('hq') + 21;
      ZoneR := Rect(R.Left + 2, ButtonPos, R.Right - 2, ButtonPos + ZoneHeight - 1);
      dx := (R.Right - R.Left - 4 - W) div 2;
      btnR := Rect(ZoneR.left + dx, ZoneR.Top + 4, ZoneR.Left + dx + 4 + W, ZoneR.Top + 6 + H);
    end else
     Exit; // only display buttons when images assigned
  end;
  if First then FScrollTop := 0;
  FScrollBottom := FScrollTop + ((FPanelRect.Bottom - FPanelRect.Top + 1) div ZoneHeight) - 1;
  if FScrollBottom > (Count - 1) then FScrollBottom := Count - 1;
  // set button rects
  if Panels[Index].Buttons.Count > FScrollBottom then
    for I := FScrollTop to FScrollBottom do
    begin
      Panels[Index].Buttons[I].FDisplayRect := ZoneR;
      Panels[Index].Buttons[I].FButtonRect := btnR;
    // next zone
      Inc(ZoneR.Top, ZoneHeight);
      Inc(ZoneR.Bottom, ZoneHeight);
      Inc(btnR.Top, ZoneHeight);
      Inc(btnR.Bottom, Zoneheight);
    end;
  // display buttons
  if Panels[Index].Buttons.Count > FScrollBottom then
    for I := FScrollTop to FScrollBottom do
      UpdateButton(Index, I, sbbsDefault);
    // check display of scroll buttons
//  timed:= assigned(FScrollTimer) and (FScrollTimer.enabled);

  if (FScrollTop > 0) then
  begin
    FUpActive := True;
    if FUpIsDown then
      DrawUpSunken else
      DrawUpRaised;
  end else
    FUpActive := False;

  if FScrollBottom < (Count - 1) then
  begin
    FDownActive := True;
    if FDownIsDown then
      DrawDownSunken  else
      DrawDownRaised;
  end else
   FDownActive := False;
end;

procedure TFourthShortcutBar.UpdatePanels(UpdateRects: Boolean);
const
  MaxPanelCount = 128;
var
  I, Count, PanelPos, ot: Integer;
  R: TRect;
begin
  if HandleAllocated then
  begin
    R := ClientRect;
    if FUseSystemFont then
      Canvas.Font := Screen.IconFont else
      Canvas.Font := Self.Font;
    if FStyle = sbs2000 then Canvas.Brush.Color := FColors.FBG;
    if FStyle = sbs2002 then Canvas.Brush.Color := Get2002Bg(FColors.Background);
    Canvas.FillRect(R);
    if FStyle = sbs2000 then
    begin
      Frame3D(Canvas, R, FColors.FShadow, FColors.FHighlight, 1);
      Frame3D(Canvas, R, FColors.FDarkShadow, FColors.FShadow, 1);
    end else
    begin
       Frame3D(Canvas, R, clBtnShadow, clBtnHighlight, 1);
    end;
    Count := Panels.Count;
    if UpdateRects then
    begin
      if Count > MaxPanelCount then Count := MaxPanelCount;
      if Count = 0 then
      begin
        //
      end else
      begin
        FPanelHeight := Height - 2 - (Count * FTabHeight);
        PanelPos := 1;
        for I := 0 to Count - 1 do
        begin
          Panels[I].FDisplayRect := Rect(1, PanelPos, Width-1, PanelPos+FTabHeight);
          Inc(PanelPos, FTabHeight);
          if I = PanelIndex then Inc(PanelPos, FPanelHeight);
        end;
      end;
    end;

    if (Count > 0) then
    begin
    for I := 0 to (Count - 1) do
      begin
        UpdatePanelHeader(I, sbbsDefault);
        if (I = PanelIndex) then
        begin
          ot := Panels[I].FDisplayRect.Bottom + 1;
          FPanelRect := Rect(2, ot, Width - 3, ot + FPanelHeight - 2);
          UpdatePanel(True, I);
        end;
      end;
    end;

  end;
end;

procedure TFourthShortcutBar.CMFontChanged(var Message: TMessage);
begin
  inherited;
  UpdatePanels(True);
  Invalidate;
end;

procedure TFourthShortcutBar.WMPaint(var Message: TWMPaint);
begin
  UpdatePanels(True);
  inherited;
end;

procedure TFourthShortcutBar.WMSize(var Message: TWMSize);
begin
  { Eat WM_SIZE message to prevent control from doing alignment }
{$IFDEF DELPHI4_UP}
  if not (csLoading in ComponentState) then Resize;
{$ENDIF}
  UpdatePanels(True);
end;

procedure TFourthShortcutBar.CMDesignHitTest(var Msg: TCMDesignHitTest);
const
  CTL_CLICK = 1;
begin
  inherited;
  if (Msg.Keys = CTL_CLICK) then ClickAtPos(Point(Msg.Pos.x, Msg.Pos.y));
end;

procedure TFourthShortcutBar.ClickAtPos(APoint: TPoint);
var
  Index: Integer;
begin
  Index := GetPanelAtPos(APoint);
  if (Index <> -1) then PanelIndex := Index;
end;

function TFourthShortcutBar.GetPanelAtPos(APoint: TPoint): Integer;
var
  I: Integer;
begin
  Result := -1;
  if Panels.Count = 0 then Exit;
  for I := 0 to FPanels.Count - 1 do
  begin
    if PtInRect(FPanels[I].FDisplayRect, APoint) then
    begin
      Result := I;
      Exit;
    end;
  end;
end;

procedure TFourthShortcutBar.SetPanelIndex(const Value: Integer);
begin
  if Panels.Count = 0 then Exit;
  FMoveButton := -1;
  FMoveTab := -1;
  if (FPanelIndex <> Value) then
  begin
    FPanelIndex := Value;
    UpdatePanels(True);
  end;
end;

procedure TFourthShortcutBar.SetTabHeight(const Value: Integer);
begin
  if FTabHeight <> Value then
  begin
   FTabHeight := Value;
   UpdatePanels(True);
  end;
end;

function TFourthShortcutBar.GetButtonArea(Pt: TPoint): Integer;
var
  I: Integer;
begin
  Result := -1;
  if Panels.Count = 0 then Exit;
  if PanelIndex = -1 then Exit;
  if Panels[PanelIndex].Buttons.Count = 0 then Exit;
  for I := FScrollTop to FScrollBottom do
  begin
    try
      if PtInRect(FPanels[PanelIndex].Buttons[I].FDisplayRect, Pt) then
      begin
        Result := I;
        Exit;
      end;
    except
      Beep;
      Continue;
    end;
  end;
end;

function TFourthShortcutBar.GetButtonAtPos(pt: TPoint): Integer;
var
  I: Integer;
begin
  Result := -1;
  if Panels.Count = 0 then Exit;
  if PanelIndex = -1 then Exit;
  if Panels[PanelIndex].Buttons.Count = 0 then Exit;
  for I := FScrollTop to FScrollBottom do
  begin
    if ptInRect(Panels[PanelIndex].Buttons[I].FButtonRect, pt) then
    begin
      Result := I;
      Exit;
    end;
  end;
end;

procedure TFourthShortcutBar.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  Idx: Integer;
  I: Integer;
begin
  Cursor := Self.Cursor;

  if Panels.Count = 0 then Exit;

  Idx := GetPanelAtPos(Point(X, Y));
  FSelectedButton := GetButtonAtPos(Point(X, Y));

{$IFDEF DELPHI3_UP}
  if (Idx <> -1) then
    Cursor := FPanelCursor else
{$ENDIF}
    Cursor := Self.Cursor;

  if (not (ssLeft in Shift)) and (Idx <> FMoveTab) and
    (Panels.Count <> 0) then
  begin
    for I := 0 to Pred(Panels.Count) do
    begin
      if I <> Idx then
        UpdatePanelHeader(I, sbbsDefault) else
        UpdatePanelHeader(I, sbbsHot);
    end;
    FMoveTab := Idx;
  end;

  if not (ssLeft in Shift) then
  begin
    if (PanelIndex <> -1) and (Panels[PanelIndex].Buttons.Count > 0) then
    begin
      Idx := GetButtonArea(Point(X, Y));
      if (Idx <> FMoveButton) then
      begin
        for I := FScrollTop to FScrollBottom do
          if (I <> Idx) then
            UpdateButton(PanelIndex, I, sbbsDefault) else
            UpdateButton(PanelIndex, Idx, sbbsHot);
          FMoveButton := Idx;
      end;
    end;
  end; { if not (ssLeft in Shift) ... }
end;

procedure TFourthShortcutBar.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
var
  Idx: Integer;
begin
  if Panels.Count = 0 then Exit;

  FDownTab := -1;
  FDownButton := -1;

  if (FUpActive = True) then // check if upscroll is pressed
  begin
    if (PtInRect(FUpR, Point(X, Y)) = True) then
    begin
      FUpIsDown := True;
      DrawUpSunken;
      FScrollTimer := TTimer.Create(Self);
      FScrollTimer.Interval := 200;
      FScrollTimer.OnTimer := TimedScroll;
      FScrollTimer.Enabled := True;
      Exit;
    end;
  end;

  if (FDownActive = True) then // check if downscroll is pressed
  begin
    if PtInRect(FDownR, Point(X, Y)) then
    begin
      FDownIsDown := True;
      DrawDownSunken;
      FScrollTimer := TTimer.Create(self);
      FScrollTimer.Interval := 200;
      FScrollTimer.OnTimer := TimedScroll;
      FScrollTimer.Enabled := True;
      Exit;
    end;
  end;

  Idx := GetPanelAtPos(Point(X, Y));

  if (Idx <> -1) then
  begin
    FDownTab := Idx;
    FMoveButton := -1;
    UpdatePanelHeader(FDownTab, sbbsPressed);
  end else
  begin
    Idx := GetButtonAtPos(Point(X, Y));
    if (Idx <> -1) then
    begin
      FDownButton := Idx;
      UpdateButton(PanelIndex, FDownButton, sbbsPressed);
      FSelectedButton := Idx;
    end;
  end;
end;

procedure TFourthShortcutBar.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
var
  Idx: Integer;
begin
  if Panels.Count = 0 then Exit;

  if (FUpIsDown = True) then
  begin
    if Assigned(FScrollTimer) then FScrollTimer.Free;
    FUpIsDown := False;
//    DrawUpRaised;
    if FUpActive then
    begin
      Inc(FScrollTop, -1);
      UpdatePanel(False, PanelIndex);
      Exit;
    end;
  end;

  if (FDownIsDown = True) then
  begin
    if Assigned(FScrollTimer) then FScrollTimer.Free;
    FDownIsDown := False;
//    DrawDownRaised;
    if FDownActive then
    begin
      Inc(FScrollTop, 1);
      UpdatePanel(False, PanelIndex);
      Exit;
    end;
  end;

  if (FDownTab <> -1) then
  begin
    UpdatePanelHeader(FDownTab, sbbsDefault);
    Idx := FDownTab;
    FDownTab := -1;
    PanelIndex := Idx;
    Panels[PanelIndex].Click;
  end else
  if (FDownButton <> -1) then
  begin
    UpdateButton(PanelIndex, FDownButton, sbbsDefault);
    Idx := FDownButton;
    FDownButton := -1;
    Panels[PanelIndex].Buttons[Idx].Click;
  end;

  Idx := GetButtonAtPos(Point(X, Y));
  if (Idx <> -1) then
  begin
    FDownButton := Idx;
    UpdateButton(PanelIndex, FDownButton, sbbsHot);
  end;
end;

procedure TFourthShortcutBar.SetImages(const Value: TImagelist);
begin
  FImages := Value;
end;

procedure TFourthShortcutBar.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  if (Operation = opRemove) and (AComponent = FImages) then
    SetImages(nil);
  if (Operation = opRemove) and (AComponent = FSmallImages) then
    SetSmallImages(nil);
end;

procedure TFourthShortcutBar.SetSmallImages(const Value: TImageList);
begin
  FSmallImages := Value;
end;

procedure TFourthShortcutBar.TimedScroll(Sender: TObject);
begin
  if FUpIsDown then
  begin
    if FUpActive then
    begin
      Inc(FScrollTop, -1);
      UpdatePanel(False, PanelIndex);
    end;
    if not FUpActive then
      if Assigned(FScrollTimer) then
        FScrollTimer.Enabled := False;
  end else
  if FDownIsDown then
  begin
    if FDownActive then
    begin
      Inc(FScrollTop, 1);
      UpdatePanel(False, PanelIndex);
    end;
    if not FDownActive then
      if Assigned(FScrollTimer) then
        FScrollTimer.Enabled := False;
  end else
  if Assigned(FScrollTimer) then
    FScrollTimer.Enabled := False;
end;

procedure TFourthShortcutBar.ColorsChange(Sender: TObject);
begin
  if FColors.UseDefault = True then ApplyStandardColors;
  Invalidate;
end;

{ TFourthShortcutBarButton }

constructor TFourthShortcutBarButton.Create(Collection: TCollection);
begin
  inherited Create(Collection);
  FDeleting := False;
  FImageIndex := -1;
  FEnabled := True;
//  FVisible := True;
end;

destructor TFourthShortcutBarButton.Destroy;
begin
  if Assigned(ActionLink) then
  begin
    FActionLink.Free;
    FActionLink := nil;
  end;  
  FDeleting := True;
  Data := nil;
  inherited;
end;

procedure TFourthShortcutBarButton.Assign(Source: TPersistent);
begin
  if (Source is TFourthShortcutBarButton) then
  begin
    Action := TFourthShortcutBarButton(Source).Action;
    Caption := TFourthShortcutBarButton(Source).Caption;
    Data := TFourthShortcutBarButton(Source).Data;
    Enabled := TFourthShortcutBarButton(Source).Enabled;
    ImageIndex := TFourthShortcutBarButton(Source).ImageIndex;
    Tag := TFourthShortcutBarButton(Source).Tag;
    Visible := TFourthShortcutBarButton(Source).Visible;
    OnClick := TFourthShortcutBarButton(Source).OnClick;
  end else
    inherited Assign(Source);
end;

procedure TFourthShortcutBarButton.Click;
begin
  if Assigned(FOnClick) and (Action <> nil) and (@FOnClick <> @Action.OnExecute) then
    FOnClick(Self) else
  if {not (csDesigning in ComponentState) and }(ActionLink <> nil) and (ActionLink.IsEnabledLinked) then
    ActionLink.Execute else
  if Assigned(FOnClick) then
    FOnClick(Self);
end;

{$IFDEF DELPHI3_UP}
function TFourthShortcutBarButton.GetDisplayName: string;
begin
  Result := Caption;
  if Result = '' then Result := inherited GetDisplayName;
end;
{$ENDIF}

procedure TFourthShortcutBarButton.SetImageIndex(Value: Integer);
begin
  if (FImageIndex <> Value) then
  begin
    FImageIndex := Value;
    Changed(False)
  end;
end;

procedure TFourthShortcutBarButton.SetOnClick(const Value: TNotifyEvent);
begin
  FOnClick := Value;
end;

procedure TFourthShortcutBarButton.SetData(const Value: Pointer);
begin
  if not FDeleting and (Value <> Data) then
  begin
    FData := Value;
    Changed(False);
  end;
end;

procedure TFourthShortcutBarButton.SetCaption(const Value: TCaption);
begin
  if (FCaption <> Value) then
  begin
    FCaption := Value;
    Changed(False);
  end;
end;

procedure TFourthShortcutBarButton.Notification(AComponent: TComponent; Operation: TOperation);
begin
  if Assigned(Action) and (Operation = opRemove) and (Action = AComponent) then
    Action := nil;
end;

function TFourthShortcutBarButton.GetAction: TBasicAction;
begin
  if ActionLink <> nil then
    Result := ActionLink.Action
  else
    Result := nil;
end;

procedure TFourthShortcutBarButton.SetAction(Value: TBasicAction);
begin
  if (Value = nil) then
  begin
    if Assigned(ActionLink) then
    begin
      ActionLink.Free;
      ActionLink := nil;
    end;
  end
  else
//    if Assigned(FourthShortcutBar) then
    begin
      if ActionLink = nil then
        ActionLink := GetActionLinkClass.Create(Self);
      ActionLink.Action := Value;
      ActionLink.OnChange := DoActionChange;
      ActionChange(Value, csLoading in Value.ComponentState);
//      Value.FreeNotification(FourthShortcutBar);
    end;
end;

procedure TFourthShortcutBarButton.DoActionChange(Sender: TObject);
begin
  ActionChange(Sender, False);
end;

procedure TFourthShortcutBarButton.ActionChange(Sender: TObject; CheckDefaults: boolean);
begin
  if Sender is TCustomAction then
    with TCustomAction(Sender) do
    begin
      if not CheckDefaults or (Self.Caption = '') then
        Self.Caption := TCustomAction(Sender).Caption;
      if not CheckDefaults or (Self.Enabled = True) then
        Self.Enabled := TCustomAction(Sender).Enabled;
      if not CheckDefaults or (Self.Visible = True) then
        Self.Visible := TCustomAction(Sender).Visible;
      if not CheckDefaults or not Assigned(Self.OnClick) then
        Self.OnClick := TCustomAction(Sender).OnExecute;
    end;
end;

procedure TFourthShortcutBarButton.SetEnabled(Value: Boolean);
begin
  if (FEnabled <> Value) then
  begin
    FEnabled := Value;
    Changed(True);
  end;
end;
{
procedure TFourthShortcutBarButton.SetVisible(Value: Boolean);
begin
  if FVisible <> Value then
  begin
    FVisible := Value;
    Changed(True);
  end;
end;
}
{ TFourthShortcutBarButtons }

function TFourthShortcutBarButtons.Add: TFourthShortcutBarButton;
begin
  Result := TFourthShortcutBarButton(inherited Add);
end;

procedure TFourthShortcutBarButtons.Delete(Index: Integer);
begin
  Self.BeginUpdate;
  try
    inherited Delete(Index);
  finally
    Self.EndUpdate;
  end;
  Self.FFourthShortcutBarPanel.FFourthShortcutBar.Invalidate;
{
  if Index > 0 then
    FFourthShortcutBar.SetPanelIndex(Index-1);
  if (Index = 0) and (Count = 0) then
    FFourthShortcutBar.SetPanelIndex(-1);
  if (Index = 0) and (Count <> 0) then
    FFourthShortcutBar.SetPanelIndex(Index);
}    
end;

constructor TFourthShortcutBarButtons.Create(FourthShortcutBarPanel: TFourthShortcutBarPanel);
begin
  inherited Create(TFourthShortcutBarButton);
  FFourthShortcutBarPanel := FourthShortcutBarPanel;
end;

function TFourthShortcutBarButtons.GetItem(Index: Integer): TFourthShortcutBarButton;
begin
  Result := TFourthShortcutBarButton(inherited GetItem(Index));
end;

{$IFDEF DELPHI3_UP}
function TFourthShortcutBarButtons.GetOwner: TPersistent;
begin
  Result := FFourthShortcutBarPanel;
end;
{$ENDIF}

procedure TFourthShortcutBarButtons.SetItem(Index: Integer; Value: TFourthShortcutBarButton);
begin
  inherited SetItem(Index, Value);
end;

procedure TFourthShortcutBarButtons.Update(Item: TCollectionItem);
begin
  if (Item <> nil) then FFourthShortcutBarPanel.Changed(False);
end;

procedure TFourthShortcutBarPanels.Assign(Source: TPersistent);
var
  I, B: Integer;
begin
  if (Source is TFourthShortcutBarPanels) then
  begin
    for I := 0 to Pred(TFourthShortcutBarPanels(Source).Count) do
    begin
      with Add do
      begin
        Caption := TFourthShortcutBarPanels(Source).Items[I].Caption;
        Data := TFourthShortcutBarPanels(Source).Items[I].Data;
        SmallImages := TFourthShortcutBarPanels(Source).Items[I].SmallImages;
        Tag := TFourthShortcutBarPanels(Source).Items[I].Tag;
        if TFourthShortcutBarPanels(Source).Items[I].Buttons.Count <> 0 then
        begin
          for B := 0 to Pred((Source as TFourthShortcutBarPanels).Items[I].Buttons.Count) do
          begin
            with (Source as TFourthShortcutBarPanels).Items[I].Buttons.Add do
            begin
              Action := (Source as TFourthShortcutBarPanels).Items[I].Buttons[B].Action;
              Caption := (Source as TFourthShortcutBarPanels).Items[I].Buttons[B].Caption;
              Data := (Source as TFourthShortcutBarPanels).Items[I].Buttons[B].Data;
              Enabled := (Source as TFourthShortcutBarPanels).Items[I].Buttons[B].Enabled;
              ImageIndex := (Source as TFourthShortcutBarPanels).Items[I].Buttons[B].ImageIndex;
              Tag := (Source as TFourthShortcutBarPanels).Items[I].Buttons[B].Tag;
              Visible := (Source as TFourthShortcutBarPanels).Items[I].Buttons[B].Visible;              
            end;
          end;
        end;
      end;
    end;
  end else
    inherited Assign(Source);
end;

procedure TFourthShortcutBar.Assign(Source: TPersistent);
begin
  if (Source is TFourthShortcutBar) then
  begin
    Top := (Source as TFourthShortcutBar).Top;
    Left := (Source as TFourthShortcutBar).Left;
    Height := (Source as TFourthShortcutBar).Height;
    Width := (Source as TFourthShortcutBar).Width;
    Align := (Source as TFourthShortcutBar).Align;
    Images.Assign((Source as TFourthShortcutBar).Images);
    SmallImages.Assign((Source as TFourthShortcutBar).SmallImages);
    Font.Assign((Source as TFourthShortcutBar).Font);
    FPanels.Assign((Source as TFourthShortcutBar).Panels);
    Tag := (Source as TFourthShortcutBar).Tag;
  end else
    inherited Assign(Source);
end;

procedure TFourthShortcutBar.SetStyle(Value: TFourthShortcutBarStyle);
begin
  if (FStyle <> Value) then
  begin
    FStyle := Value;
    ApplyStandardColors;
    Repaint;
  end;
end;

procedure TFourthShortcutBar.SetUseSystemFont(Value: Boolean);
begin
  if (FUseSystemFont <> Value) then
  begin
    FUseSystemFont := Value;
    Repaint;
  end;
end;

procedure TFourthShortcutBar.ApplyStandardColors;
begin
  case FStyle of
    sbs2000:
      begin
        with Self.Colors do
        begin
          Background := clBtnShadow;
          BtnFace := clBtnShadow;
          BtnLight := clBtnHighlight;
          BtnShadow := cl3DDkShadow;
          BtnText := clBtnHighlight;
          BtnTextDisabled := clBtnFace;
          Face := clBtnFace;
          Shadow := clBtnShadow;
          Light := cl3DLight;
          DarkShadow := cl3DDkShadow;
          Highlight := clBtnHighlight;
        end;
      end;
    sbs2002:
      begin
        with Self.Colors do
        begin
          Background := clBtnFace;
          BtnFace := clHighlight;
          BtnLight := clHighlight;
          BtnShadow := clHighlight;
          BtnText := clBtnHighlight;
          BtnTextDisabled := clGrayText;
          Face := clBtnFace;
          Shadow := clBtnShadow;
          Light := cl3DLight;
          DarkShadow := cl3DDkShadow;
          Highlight := clBtnHighlight;
        end;
      end;
    end;
end;

{$IFDEF DELPHI4_UP}
procedure TFourthShortcutBarButton.SetVisible(Value: Boolean);
begin
  //
end;
{$ENDIF}

end.
