
{*******************************************************}
{                                                       }
{       4th GUI Library (VCL Edition)                   }
{       Controls Unit                                   }
{                                                       }
{       Copyright (c) 1996,2003 Sergiy S. Tkachenko     }
{       e-mail: support@4thfebruary.tk                  }
{          Web: www.4thfebruary.tk                      }
{                                                       }
{       Based on the TTabSet component                  }
{       Copyright (C) Borland Software Corporation      }
{                                                       }
{*******************************************************}

unit FthTabs;

{$I FTHVER.INC}

interface

uses
{$IFDEF WIN32}
  Windows,
{$ELSE}
  WinTypes, WinProcs,
{$ENDIF}
{$IFDEF DELPHI4_UP}
  ImgList,
{$ENDIF}
  ExtCtrls, Classes, Graphics, Forms, Controls, Messages;

type
  TScrollBtn = (sbLeft, sbRight);

  TFourthScroller = class(TCustomControl)
  private
    FMin: LongInt;
    FMax: LongInt;
    FPosition: LongInt;
    FOnClick: TNotifyEvent;
    FChange: Integer;
    Bitmap: TBitmap;
    Pressed: Boolean;
    Down: Boolean;
    Current: TScrollBtn;
    pWidth: Integer;
    pHeight: Integer;
    procedure SetMin(Value: LongInt);
    procedure SetMax(Value: LongInt);
    procedure SetPosition(Value: LongInt);
    function CanScrollLeft: Boolean;
    function CanScrollRight: Boolean;
    procedure DoMouseDown(X: Integer);
    procedure WMLButtonDown(var Message: TWMLButtonDown); message WM_LBUTTONDOWN;
    procedure WMLButtonDblClk(var Message: TWMLButtonDblClk); message WM_LBUTTONDBLCLK;
    procedure WMMouseMove(var Message: TWMMouseMove); message WM_MOUSEMOVE;
    procedure WMLButtonUp(var Message: TWMLButtonUp); message WM_LBUTTONUP;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Paint; override;
  published
    property OnClick: TNotifyEvent read FOnClick write FOnClick;
    property Min: LongInt read FMin write SetMin default 0;
    property Max: LongInt read FMax write SetMax default 0;
    property Position: LongInt read FPosition write SetPosition default 0;
    property Change: Integer read FChange write FChange default 1;
  end;

  TFourthTabSet = class;

  TFourthTabList = class(TStringList)
  private
    Tabs: TFourthTabSet;
  public
    procedure Insert(Index: Integer; const S: string); override;
    procedure Delete(Index: Integer); override;
    function Add(const S: string): Integer; override;
    procedure Put(Index: Integer; const S: string); override;
    procedure Clear; override;
    procedure AddStrings(Strings: TStrings); override;
  end;

  TFourthTabSetFlatStyle = (fsNoFlat, fsMiddleFlat, fsFullFlat);

{$IFNDEF DELPHI4_UP}
  TBevelCut = (bvNone, bvLowered, bvRaised, bvSpace);
{$ENDIF}

  TFourthTabSetOptions = class(TPersistent)
  private
    FBGColor: TColor;
    FFlatStyle: TFourthTabSetFlatStyle;
    FOnChange: TNotifyEvent;
    FSeparatorColor: TColor;
    FUnSelColor: TColor;
    FLightColor: TColor;
    FHighlightColor: TColor;
    FShadowColor: TColor;
    FDarkShadowColor: TColor;
    FFaceColor: TColor;
    FBoldSelTab: Boolean;
    FLineOnSide: Boolean;
    FBorderStyle: TBevelCut;
    FRoundedCorners: Boolean;
    procedure SetBGColor(Value: TColor);
    procedure SetFlatStyle(Value: TFourthTabSetFlatStyle);
    procedure SetSeparatorColor(Value: TColor);
    procedure SetUnselectedColor(Value: TColor);
    procedure SetLightColor(Value: TColor);
    procedure SetHighlightColor(Value: TColor);
    procedure SetShadowColor(Value: TColor);
    procedure SetDarkShadowColor(Value: TColor);
    procedure SetFaceColor(Value: TColor);
    procedure SetBoldSelTab(Value: Boolean);
    procedure SetLineOnSide(Value: Boolean);
    procedure SetBorderStyle(Value: TBevelCut);
    procedure SetRoundedCorners(Value: Boolean);
  protected
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  public
    constructor Create;
  published
    property BackgroundColor: TColor read FBGColor write SetBGColor default clBtnShadow;
    property LightColor: TColor read FLightColor write SetLightColor default cl3DLight;
    property HighlightColor: TColor read FHighlightColor write SetHighlightColor default clBtnHighlight;
    property ShadowColor: TColor read FShadowColor write SetShadowColor default clBtnShadow;
    property DarkShadowColor: TColor read FDarkShadowColor write SetDarkShadowColor default cl3DDkShadow;
    property SeparatorColor: TColor read FSeparatorColor write SetSeparatorColor default clBtnFace;
    property UnSelectedColor: TColor read FUnSelColor write SetUnselectedColor default clBtnHighlight;
    property FaceColor: TColor read FFaceColor write SetFaceColor default clBtnFace;

    property BorderStyle: TBevelCut read FBorderStyle write SetBorderStyle default bvNone;
    property FlatStyle: TFourthTabSetFlatStyle read FFlatStyle write SetFlatStyle default fsMiddleFlat;
    property BoldSelectedTab: Boolean read FBoldSelTab write SetBoldSelTab;
    property SeparatorOnSide: Boolean read FLineOnSide write SetLineOnSide;
    property RoundedCorners: Boolean read FRoundedCorners write SetRoundedCorners;
  end;

  TFourthTabSetMargins = class(TPersistent)                        //
  private                                                          //
    FStartMargin: Integer;                                         //
    FEndMargin: Integer;                                           //
    FTopEdge: Integer;                                             // xu[qian], 18/09
    FTextTopMargin: Integer;                                       //
    FTextBottomMargin: Integer;                                    //
    FTextLeftMargin: Integer;                                      //
    FTextRightMargin: Integer;                                     //
    FImgLeftMargin: Integer;
    FImgRightMargin: Integer;
    FImgTopMargin: Integer;
    FOnChange: TNotifyEvent;                                       //
    procedure SetStartMargin(Value: Integer);                      //
    procedure SetEndMargin(Value: Integer);                        //
    procedure SetTopEdge(Value: Integer);                          //
    procedure SetTextTopMargin(Value: Integer);                    // xu[qian], 18/09
    procedure SetTextBottomMargin(Value: Integer);                 //
    procedure SetTextLeftMargin(Value: Integer);                   //
    procedure SetTextRightMargin(Value: Integer);                  //
    procedure SetImgLeftMargin(Value: Integer);
    procedure SetImgRightMargin(Value: Integer);
    procedure SetImgTopMargin(Value: Integer);
  protected                                                        //
    property OnChange: TNotifyEvent read FOnChange write FOnChange;// xu[qian], 18/09
  public                                                           //
    constructor Create;
  published
    property StartMargin: Integer read FStartMargin write SetStartMargin default 5;
    property EndMargin: Integer read FEndMargin write SetEndMargin default 5;
    property EdgeHeight: Integer read FTopEdge write SetTopEdge default 1;
    property TextTopMargin: Integer read FTextTopMargin write SetTextTopMargin default 3;
    property TextBottomMargin: Integer read FTextBottomMargin write SetTextBottomMargin default 3;
    property TextLeftMargin: Integer read FTextLeftMargin write SetTextLeftMargin default 2;
    property TextRightMargin: Integer read FTextRightMargin write SetTextRightMargin default 0;
    property ImgLeftMargin: Integer read FImgLeftMargin write SetImgLeftMargin default 1;
    property ImgRightMargin: Integer read FImgRightMargin write SetImgRightMargin default 2;
    property ImgTopMargin: Integer read FImgTopMargin write SetImgTopMargin default 5;
  end;

  TEdgePart = (epSelectedLeft, epUnselectedLeft, epSelectedRight,
               epUnselectedRight);

  TEdgeType = (etNone, etFirstIsSel, etFirstNotSel, etLastIsSel, etLastNotSel,
               etNotSelToSel, etSelToNotSel, etNotSelToNotSel);

  TFourthTabStyle = (tsStandard, tsOwnerDraw);

  TFourthTabSetOrientation = (toBottom, toTop);

  TMeasureTabEvent = procedure(Sender: TObject; Index: Integer; var TabWidth: Integer) of object;
  TDrawTabEvent = procedure(Sender: TObject; TabCanvas: TCanvas; R: TRect; Index: Integer; Selected: Boolean) of object;
  TTabChangeEvent = procedure(Sender: TObject; NewTab: Integer; var AllowChange: Boolean) of object;

  TFourthTabSet = class(TCustomControl)
  private
    FOptions: TFourthTabSetOptions;
    FMargins: TFourthTabSetMargins;   // xu[qian], 18/09
    FTabs: TStrings;
    FIndexes: TStrings;
    FTabIndex: Integer;
    FFirstIndex: Integer;
    FVisibleTabs: Integer;
    FAutoScroll: Boolean;
    FStyle: TFourthTabStyle;
    FOwnerDrawHeight: Integer;
    FOnMeasureTab: TMeasureTabEvent;
    FOnDrawTab: TDrawTabEvent;
    FOnChange: TTabChangeEvent;
    TabPositions: TList;
    FTabHeight: Integer;
    FScroller: TFourthScroller;
    FDoFix: Boolean;
    FDisabledTabs: TStrings;
    FOrientation: TFourthTabSetOrientation;
    FScrollChangeTab: Boolean;
    FOnScrollClick: TNotifyEvent;
    FImages: TImageList;
    FHotImages: TImageList;                           // xu[qian], 12/10
    FTabHints: TStrings;                              //
    FSavedHint: string;                               //
    FUseSavedHint: Boolean;                           //
    FMouseOverTab: Integer;                           //
    procedure SetTabHints(Value: TStrings);           //
    procedure SetHotImages(Value: TImageList);        //
    procedure SetImages(Value: TImageList);
    procedure SetAutoScroll(Value: Boolean);
    procedure SetTabIndex(Value: Integer);
    procedure SetFirstIndex(Value: Integer);
    procedure SetTabList(Value: TStrings);
    procedure SetIndexes(Value: TStrings);
    procedure SetTabStyle(Value: TFourthTabStyle);
    procedure SetTabHeight(Value: Integer);
    procedure SetDisabledTabList(const Value: TStrings);
    procedure SetOrientation(Value: TFourthTabSetOrientation);
    procedure SetScrollChangeTab(Value: Boolean);
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure WMNCHitTest(var message: TWMNCHitTest); message WM_NCHITTEST;      // xu[qian], 12/10
    function CalcTabPositions(Start, Stop: Integer; Canvas: TCanvas; First: Integer): Integer;
    procedure CreateScroller;
    procedure FixTabPos;
    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure CMDialogChar(var Message: TCMDialogChar); message CM_DIALOGCHAR;
    procedure CMDesignHitTest(var Message: TCMDesignHitTest); message CM_DESIGNHITTEST;
    procedure ScrollClick(Sender: TObject);
    procedure ReadIntData(Reader: TReader);
    procedure ReadBoolData(Reader: TReader);
  protected
    procedure CreateParams(var Params: TCreateParams); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure Paint; override;
    procedure DrawTab(TabCanvas: TCanvas; R: TRect; Index: Integer; Selected: Boolean); virtual;
    function CanChange(NewIndex: Integer): Boolean;
    procedure Change; dynamic;
    procedure GetChildren(Proc: TGetChildProc{$IFDEF DELPHI3_UP}; Root: TComponent{$ENDIF}); override;
    procedure MeasureTab(Index: Integer; var TabWidth: Integer); virtual;
    procedure DefineProperties(Filer: TFiler); override;
    function TabEnabled(Index: Integer): Boolean;
    function GetImageIndex(Index: Integer): Integer;
    procedure OptionsChange(Sender: TObject);
    procedure MarginsChange(Sender: TObject);
    procedure DisplayTabHint(TabIndex: Integer);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    function ItemAtPos(Pos: TPoint): Integer;
    function ItemRect(Item: Integer): TRect;
    procedure SelectNext(Direction: Boolean);
    property Canvas;
    property FirstIndex: Integer read FFirstIndex write SetFirstIndex default 0;
  published
    property Align;
{$IFDEF DELPHI4_UP}
    property Anchors;
{$ENDIF}
    property AutoScroll: Boolean read FAutoScroll write SetAutoScroll default True;
{$IFDEF DELPHI4_UP}
    property Constraints;
{$ENDIF}
    property DragCursor;
{$IFDEF DELPHI4_UP}
    property DragKind;
{$ENDIF}
    property DragMode;
    property Enabled;
    property DisabledTabs: TStrings read FDisabledTabs write SetDisabledTabList;
    property Font;
    property Images: TImageList read FImages write SetImages;
    property HotImages: TImageList read FHotImages write SetHotImages;     // xu[qian], 12/10
    property TabHints: TStrings read FTabHints write SetTabHints;          //
    property Options: TFourthTabSetOptions read FOptions write FOptions;
    property Margins: TFourthTabSetMargins read FMargins write FMargins;   // xu[qian], 18/09
    property Orientation: TFourthTabSetOrientation read FOrientation write SetOrientation default toBottom;
    property ParentShowHint;
    property PopupMenu;
    property ScrollChangeTab: Boolean read FScrollChangeTab write SetScrollChangeTab;
    property ShowHint;
    property Style: TFourthTabStyle read FStyle write SetTabStyle default tsStandard;
    property TabHeight: Integer read FOwnerDrawHeight write SetTabHeight default 20;
    property Tabs: TStrings read FTabs write SetTabList;
    property Indexes: TStrings read FIndexes write SetIndexes;
    property TabIndex: Integer read FTabIndex write SetTabIndex default -1;
//    property TabStop;
    property Visible;
    property VisibleTabs: Integer read FVisibleTabs;
    property OnClick;
    property OnChange: TTabChangeEvent read FOnChange write FOnChange;
    property OnDragDrop;
    property OnDragOver;
    property OnDrawTab: TDrawTabEvent read FOnDrawTab write FOnDrawTab;
{$IFDEF DELPHI4_UP}
    property OnEndDock;
{$ENDIF}
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnMeasureTab: TMeasureTabEvent read FOnMeasureTab write FOnMeasureTab;
    property OnScrollClick: TNotifyEvent read FOnScrollClick write FOnScrollClick;
{$IFDEF DELPHI4_UP}
    property OnStartDock;
{$ENDIF}
    property OnStartDrag;
  end;

implementation

uses
  Consts, SysUtils;

{$R FthTabs}

const
  EdgeWidth = 9;
  HalfEdgeWidth = EdgeWidth div 2;

type
  TTabPos = record
    Size, StartPos: Word;
  end;

{ TFourthScroller }

constructor TFourthScroller.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csOpaque];

  Bitmap := TBitmap.Create;
  pWidth := 26;
  pHeight := 12;
  FMin := 0;
  FMax := 0;
  FPosition := 0;
  FChange := 1;
end;

destructor TFourthScroller.Destroy;
begin
  Bitmap.Free;

  inherited Destroy;
end;

procedure TFourthScroller.Paint;
begin
  with Canvas do
  begin
    { fill the canvas with bgcolor }
    Brush.Color := TFourthTabSet(Parent).FOptions.FBgColor;      // xu[qian], 03/10
    FillRect(Rect(0, 0, Width, Height));                         //

    { paint left button }
    if CanScrollLeft then
    begin
      if Down and (Current = sbLeft) then
{$IFDEF DELPHI3_UP}
        Bitmap.Handle := CreateGrayMappedRes(HInstance, 'FTHSBLEFTDN') else
        Bitmap.Handle := CreateGrayMappedRes(HInstance, 'FTHSBLEFT');
{$ELSE}
        Bitmap.Handle := LoadBitmap(HInstance, 'FTHSBLEFTDN') else
        Bitmap.Handle := LoadBitmap(HInstance, 'FTHSBLEFT');
{$ENDIF}
    end else
{$IFDEF DELPHI3_UP}
      Bitmap.Handle := CreateGrayMappedRes(HInstance, 'FTHSBLEFTDIS');
{$ELSE}
      Bitmap.Handle := LoadBitmap(HInstance, 'FTHSBLEFTDIS');
{$ENDIF}
    Draw(0, 0, Bitmap);

    { paint right button }
    if CanScrollRight then
    begin
      if Down and (Current = sbRight) then
{$IFDEF DELPHI3_UP}
        Bitmap.Handle := CreateGrayMappedRes(HInstance, 'FTHSBRIGHTDN') else
        Bitmap.Handle := CreateGrayMappedRes(HInstance, 'FTHSBRIGHT');
{$ELSE}
        Bitmap.Handle := LoadBitmap(HInstance, 'FTHSBRIGHTDN') else
        Bitmap.Handle := LoadBitmap(HInstance, 'FTHSBRIGHT');
{$ENDIF}
    end else
{$IFDEF DELPHI3_UP}
      Bitmap.Handle := CreateGrayMappedRes(HInstance, 'FTHSBRIGHTDIS');
{$ELSE}
      Bitmap.Handle := LoadBitmap(HInstance, 'FTHSBRIGHTDIS');
{$ENDIF}
    Draw((pWidth div 2), 0, Bitmap);
  end;
end;

procedure TFourthScroller.WMSize(var Message: TWMSize);
begin
  inherited;
  Width := pWidth;
  Height := pHeight;
end;

procedure TFourthScroller.SetMin(Value: Longint);
begin
  if Value < FMax then FMin := Value;
end;

procedure TFourthScroller.SetMax(Value: Longint);
begin
  if Value > FMin then FMax := Value;
end;

procedure TFourthScroller.SetPosition(Value: Longint);
begin
  if Value <> FPosition then
  begin
    if Value < Min then Value := Min;
    if Value > Max then Value := Max;
    FPosition := Value;
    Invalidate;
    if Assigned(FOnClick) then
      FOnClick(Self);
  end;
end;

function TFourthScroller.CanScrollLeft: Boolean;
begin
  Result := Position > Min;
end;

function TFourthScroller.CanScrollRight: Boolean;
begin
  Result := Position < Max;
end;

procedure TFourthScroller.DoMouseDown(X: Integer);
begin
  if X < pWidth div 2 then
    Current := sbLeft else
    Current := sbRight;
  case Current of
    sbLeft:  if not CanScrollLeft then Exit;
    sbRight: if not CanScrollRight then Exit;
  end;
  Pressed := True;
  Down := True;
  Invalidate;
  SetCapture(Handle);
end;

procedure TFourthScroller.WMLButtonDown(var Message: TWMLButtonDown);
begin
  DoMouseDown(Message.XPos);
end;

procedure TFourthScroller.WMLButtonDblClk(var Message: TWMLButtonDblClk);
begin
  DoMouseDown(Message.XPos);
end;

procedure TFourthScroller.WMMouseMove(var Message: TWMMouseMove);
var
  P: TPoint;
  R: TRect;
begin
  if Pressed then
  begin
    P := Point(Message.XPos, Message.YPos);
    R := Rect(0, 0, pWidth div 2, pHeight);
    if Current = sbRight then OffsetRect(R, pWidth div 2, 0);
    if PtInRect(R, P) <> Down then
    begin
      Down := not Down;
      Invalidate;
    end;
  end;
end;

procedure TFourthScroller.WMLButtonUp(var Message: TWMLButtonUp);
var
  NewPos: Longint;
begin
  ReleaseCapture;
  Pressed := False;

  if Down then
  begin
    Down := False;
    NewPos := Position;
    case Current of
      sbLeft: Dec(NewPos, Change);
      sbRight: Inc(NewPos, Change);
    end;
    Position := NewPos;
  end;
end;

{ TFourthTabList }

function TFourthTabList.Add(const S: string): Integer;
begin
  Result := inherited Add(S);

  if Tabs <> nil then
  with Tabs do
  begin
    if (FTabIndex = -1) and (FTabs.Count > 0) then FTabIndex := 0;   // xu[qian], 03/10
    Invalidate;
  end;
end;

procedure TFourthTabList.Insert(Index: Integer; const S: string);
begin
//  if Index = -1 then Index := 0;         // xu[qian], 03/10
  inherited Insert(Index, S);

  if Tabs <> nil then
  with Tabs do
  begin
    if Index <= FTabIndex then Inc(FTabIndex);
    if (FTabIndex = -1) and (FTabs.Count > 0) then FTabIndex := 0;  // xu[qian], 03/10
    Invalidate;
  end;
end;

procedure TFourthTabList.Delete(Index: Integer);
var
  OldIndex: Integer;
begin
  if Index = -1 then Exit;                      // xu[qian], 03/10

  OldIndex := Tabs.TabIndex;                    // xu[qian], 05/11
  if OldIndex >= Index then                     //
  begin                                         //
    if OldIndex = 0 then                        //
      Tabs.FTabIndex := 0 else                  //
      Tabs.FTabIndex := OldIndex - 1;           //
  end;                                          //

  inherited Delete(Index);

  if Count = 0 then Tabs.FTabIndex := -1;       // xu[qian], 05/11
  Tabs.Invalidate;
end;

procedure TFourthTabList.Put(Index: Integer; const S: string);
begin
  inherited Put(Index, S);

  if Tabs <> nil then Tabs.Invalidate;
end;

procedure TFourthTabList.Clear;
begin
  inherited Clear;

  Tabs.FTabIndex := -1;
  Tabs.Invalidate;
end;

procedure TFourthTabList.AddStrings(Strings: TStrings);
begin
  SendMessage(Tabs.Handle, WM_SETREDRAW, 0, 0);
  inherited AddStrings(Strings);
  SendMessage(Tabs.Handle, WM_SETREDRAW, 1, 0);

  Tabs.Invalidate;
end;

{ TFourthTabSetOptions }

constructor TFourthTabSetOptions.Create;
begin
  FBGColor := clBtnShadow;
  FFlatStyle := fsMiddleFlat;
  FSeparatorColor := clBtnFace;
  FUnSelColor := clBtnHighlight;
  FLightColor := cl3DLight;
  FHighlightColor := clBtnHighlight;
  FShadowColor := clBtnShadow;
  FDarkShadowColor := cl3DDkShadow;
  FFaceColor := clBtnFace;
  FBoldSelTab := True;
  FLineOnSide := True;
  FBorderStyle := bvNone;
  FRoundedCorners := True;
end;

procedure TFourthTabSetOptions.SetBGColor(Value: TColor);
begin
  if (FBGColor <> Value) then
  begin
    FBGColor := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthTabSetOptions.SetBorderStyle(Value: TBevelCut);
begin
  if (FBorderStyle <> Value) then
  begin
    FBorderStyle := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthTabSetOptions.SetFlatStyle(Value: TFourthTabSetFlatStyle);
begin
  if (FFlatStyle <> Value) then
  begin
    FFlatStyle := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthTabSetOptions.SetSeparatorColor(Value: TColor);
begin
  if (FSeparatorColor <> Value) then
  begin
    FSeparatorColor := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthTabSetOptions.SetUnselectedColor(Value: TColor);
begin
  if (FUnSelColor <> Value) then
  begin
    FUnSelColor := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthTabSetOptions.SetLightColor(Value: TColor);
begin
  if (FLightColor <> Value) then
  begin
    FLightColor := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthTabSetOptions.SetHighlightColor(Value: TColor);
begin
  if (FHighlightColor <> Value) then
  begin
    FHighlightColor := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthTabSetOptions.SetShadowColor(Value: TColor);
begin
  if (FShadowColor <> Value) then
  begin
    FShadowColor := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthTabSetOptions.SetDarkShadowColor(Value: TColor);
begin
  if (FDarkShadowColor <> Value) then
  begin
    FDarkShadowColor := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthTabSetOptions.SetFaceColor(Value: TColor);
begin
  if (FFaceColor <> Value) then
  begin
    FFaceColor := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthTabSetOptions.SetBoldSelTab(Value: Boolean);
begin
  if (FBoldSelTab <> Value) then
  begin
    FBoldSelTab := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthTabSetOptions.SetLineOnSide(Value: Boolean);
begin
  if (FLineOnSide <> Value) then
  begin
    FLineOnSide := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthTabSetOptions.SetRoundedCorners(Value: Boolean);
begin
  if (FRoundedCorners <> Value) then
  begin
    FRoundedCorners := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

{ TFourthTabSetMargins }

constructor TFourthTabSetMargins.Create;
begin
  FStartMargin := 5;      // xu[qian], 18/09
  FEndMargin := 5;        //

  FTopEdge := 1;          // xu[qian], 18/09

  FTextTopMargin := 3;    //
  FTextBottomMargin := 3; // xu[qian], 18/09
  FTextLeftMargin := 2;   //
  FTextRightMargin := 0;  //

  FImgLeftMargin := 1;
  FImgRightMargin := 2;
  FImgTopMargin := 5;
end;

procedure TFourthTabSetMargins.SetStartMargin(Value: Integer);
begin
  if (FStartMargin <> Value)  then
  begin
    FStartMargin := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthTabSetMargins.SetEndMargin(Value: Integer);
begin
  if (FEndMargin <> Value) then
  begin
    FEndMargin := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthTabSetMargins.SetTopEdge(Value: Integer);
begin
  if (FTopEdge <> Value)  then
  begin
    FTopEdge := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthTabSetMargins.SetTextTopMargin(Value: Integer);
begin
  if (FTextTopMargin <> Value) then
  begin
    FTextTopMargin := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthTabSetMargins.SetTextBottomMargin(Value: Integer);
begin
  if (FTextBottomMargin <> Value) then
  begin
    FTextBottomMargin := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthTabSetMargins.SetTextLeftMargin(Value: Integer);
begin
  if (FTextLeftMargin <> Value) then
  begin
    FTextLeftMargin := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthTabSetMargins.SetTextRightMargin(Value: Integer);
begin
  if (FTextRightMargin <> Value)  then
  begin
    FTextRightMargin := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthTabSetMargins.SetImgTopMargin(Value: Integer);
begin
  if (FImgTopMargin <> Value) then
  begin
    FImgTopMargin := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthTabSetMargins.SetImgLeftMargin(Value: Integer);
begin
  if (FImgLeftMargin <> Value) then
  begin
    FImgLeftMargin := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TFourthTabSetMargins.SetImgRightMargin(Value: Integer);
begin
  if (FImgRightMargin <> Value)  then
  begin
    FImgRightMargin := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

{ TFourthTabSet }

constructor TFourthTabSet.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  ControlStyle := [csCaptureMouse, csDoubleClicks, csOpaque];
  Width := 185;
  Height := 21;

  ParentFont := True;

  FOptions := TFourthTabSetOptions.Create;
  FOptions.OnChange := OptionsChange;

  FMargins := TFourthTabSetMargins.Create;     // xu[qian], 18/09
  FMargins.OnChange := MarginsChange;

  FTabHints := TStringList.Create;             // xu[qian], 12/10
  FMouseOverTab := -1;                         //
  FUseSavedHint := False;                      //

  TabPositions := TList.Create;
//  FTabHeight := Height - (FTopEdge + FBottomEdge);  //  better: define it, when drawing.

  FTabs := TFourthTabList.Create;
  TFourthTabList(FTabs).Tabs := Self;
  FDisabledTabs := TStringList.Create;
  FIndexes := TStringList.Create;

  CreateScroller;

  FTabIndex := -1;
  FFirstIndex := 0;
  FVisibleTabs := 0; { set by draw routine }
  { initialize default values }
  FAutoScroll := True;
  FStyle := tsStandard;
  FOwnerDrawHeight := 20;

  FScrollChangeTab := False;
end;

procedure TFourthTabSet.OptionsChange(Sender: TObject);
begin
  Invalidate;
end;

procedure TFourthTabSet.MarginsChange(Sender: TObject);
begin
  Invalidate;
end;

procedure TFourthTabSet.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params.WindowClass do
    Style := Style and not (CS_VREDRAW or CS_HREDRAW);
end;

procedure TFourthTabSet.CreateScroller;
begin
  FScroller := TFourthScroller.Create(Self);
  with FScroller do
  begin
    Parent := Self;
    Top := 5;
    Min := 0;
    Max := 0;
    Position := 0;
    Visible := False;
    OnClick := ScrollClick;
  end;
end;

destructor TFourthTabSet.Destroy;
begin
  FTabHints.Free;          // xu[qian], 12/10
  FTabs.Free;
  FDisabledTabs.Free;
  FIndexes.Free;
  TabPositions.Free;
  inherited Destroy;
end;

procedure TFourthTabSet.ScrollClick(Sender: TObject);
begin
  FirstIndex := TFourthScroller(Sender).Position;
  if not (csLoading in ComponentState) then
    if Assigned(FOnScrollClick) then FOnScrollClick(Self);
  if FScrollChangeTab then
  begin
    TabIndex := TFourthScroller(Sender).Position;
  end;
end;

function TFourthTabSet.ItemAtPos(Pos: TPoint): Integer;            // xu[qian], 21/Nov
var
  TabPos: TTabPos;
  I: Integer;
begin
  Result := -1;
  if not (Pos.Y in [0..ClientHeight]) then Exit;
  for I := 0 to TabPositions.Count - 1 do                       
  begin
    Pointer(TabPos) := TabPositions[I];
    if (Pos.X - TabPos.StartPos in [0..TabPos.Size]) then
    begin
      if (FOrientation = toBottom) and (Pos.Y in [0..FTabHeight]) then
      begin
        Result := I + FScroller.Position;
        Exit;
      end else
      if (Pos.Y in [ClientHeight - FTabHeight..FTabHeight]) then
      begin
        Result := I + FScroller.Position;             
        Exit;
      end;
    end;
  end;
end;

function TFourthTabSet.ItemRect(Item: Integer): TRect;
var
  TabPos: TTabPos;
begin
  if (TabPositions.Count > 0) and (Item >= 0) and (Item < TabPositions.Count) then
  begin
    Pointer(TabPos) := TabPositions[Item];
    Result := Rect(TabPos.StartPos, 0, TabPos.StartPos + TabPos.Size, FTabHeight);
    InflateRect(Result, 1, -2);
  end else
    Result := Rect(0, 0, 0, 0);
end;

procedure TFourthTabSet.Paint;
const
  FSepartorMargin = 3;
var
  MemBitmap: TBitmap;
  TabStart, LastTabPos: Integer;
  TabPos: TTabPos;
  Tab: Integer;
  IsFirst, IsSelected: Boolean;
  IsFirstTab, IsLastTab: Boolean;
  R, BmpR: TRect;
  Loop: Integer;
  OldFontStyle: TFontStyles;
  FTopEdge2, FTopEdge3: Integer;       // xu[qian], 03/10
  aImages: TImageList;                 // xu[qian], 12/10
begin
  if not HandleAllocated then Exit;

  if Style = tsOwnerDraw then                                         // xu[qian], 18/09
    FTabHeight := FOwnerDrawHeight else                               //
    FTabHeight := FMargins.FTextTopMargin + FMargins.FTextBottomMargin - Font.Height + 3;

  if not FUseSavedHint then FSavedHint := Hint;                       // xu[qian], 12/10

  MemBitmap := TBitmap.Create;
  try
    if FOptions.BorderStyle = bvNone then
    begin
      MemBitmap.Width := ClientWidth;
      MemBitmap.Height := ClientHeight;
    end else
    begin
      MemBitmap.Width := ClientWidth - 2;
      MemBitmap.Height := ClientHeight - 1;
    end;

    MemBitmap.Canvas.Font := Self.Canvas.Font;

    TabStart := FMargins.StartMargin + EdgeWidth; { where does first text appear }
    LastTabPos := Width - FMargins.EndMargin; { tabs draw until this position }
    if (FOptions.FBorderStyle = bvNone) or (FOptions.FBorderStyle = bvSpace) then
      FScroller.Left := Width - FScroller.Width - 2 else
      FScroller.Left := Width - FScroller.Width - 4;
    FTopEdge3 := FMargins.FTopEdge + 1;                                         // xu[qian], 03/10
    FScroller.Top := (Height - FTopEdge3 - FScroller.Height) div 2 + FTopEdge3; //
    FTopEdge2 := MemBitmap.Height  - FMargins.FTopEdge;                         // xu[qian], 18/09

  { do initial calculations for how many tabs are visible }
    FVisibleTabs := CalcTabPositions(TabStart, LastTabPos, MemBitmap.Canvas, FirstIndex);

     { enable the scroller if FAutoScroll = True and not all tabs are visible }

    if not FScrollChangeTab then
    begin
      if AutoScroll and (FVisibleTabs < Tabs.Count) then
      begin
        Dec(LastTabPos, FScroller.Width - 4);
      { recalc the tab positions }
        FVisibleTabs := CalcTabPositions(TabStart, LastTabPos, MemBitmap.Canvas, FirstIndex);

      { set the scroller's range }
        FScroller.Visible := True;
        ShowWindow(FScroller.Handle, SW_SHOW);
        FScroller.Min := 0;
        FScroller.Max := Tabs.Count - VisibleTabs;
        FScroller.Position := FirstIndex;
      end else
      if VisibleTabs >= Tabs.Count then
      begin
        FScroller.Visible := False;
        ShowWindow(FScroller.Handle, SW_HIDE);
      end;
    end else
    begin
      FScroller.Visible := True;
      ShowWindow(FScroller.Handle, SW_SHOW);
      FScroller.Min := 0;
      FScroller.Max := Tabs.Count - 1;
      FScroller.Position := FirstIndex;
    end;

    if FDoFix then
    begin
      FixTabPos;
      FVisibleTabs := CalcTabPositions(TabStart, LastTabPos, MemBitmap.Canvas, FirstIndex);
    end;
    FDoFix := False;

    { draw background of tab area }
    with MemBitmap.Canvas do
    begin
      Brush.Color := FOptions.FBgColor;
      FillRect(Rect(0, 0, MemBitmap.Width, MemBitmap.Height));

      Pen.Color := FOptions.FFaceColor;
      if FOrientation = toBottom then
      begin
        for Loop := 0 to FMargins.FTopEdge - 1 do
        begin
          MoveTo(0, Loop);
          LineTo(MemBitmap.Width, Loop);
        end;
      end else
      begin
        for Loop := 0 to FMargins.FTopEdge - 1 do
        begin
          MoveTo(0, Height - 1 - Loop);
          LineTo(MemBitmap.Width, Height - 1 - Loop);
        end;
      end;

      if FOrientation = toBottom then
      begin
        Pen.Color := FOptions.DarkShadowColor;
        MoveTo(0, FMargins.FTopEdge);
        LineTo(MemBitmap.Width, FMargins.FTopEdge);
      end else
      begin
        Pen.Color := FOptions.HighlightColor;
        MoveTo(0, FTopEdge2);
        LineTo(MemBitmap.Width, FTopEdge2);
      end;

      if FOptions.FlatStyle = fsNoFlat then
      begin
        if FOrientation = toBottom then
        begin
          Pen.Color := FOptions.ShadowColor;
          MoveTo(0, FMargins.FTopEdge - 1);     // xu[qian], 18/09
          LineTo(MemBitmap.Width, FMargins.FTopEdge - 1);
        end else
        begin
          Pen.Color := FOptions.FLightColor;
          MoveTo(0, FTopEdge2 + 1);
          LineTo(MemBitmap.Width, FTopEdge2 + 1);
        end;
      end;
    end;

    for Tab := 0 to TabPositions.Count - 1 do
    begin
      if not TabEnabled(Tab + FirstIndex) then Continue;

      Pointer(TabPos) := TabPositions[Tab];

      IsFirst := (Tab = 0);
      IsSelected := Tab + FirstIndex = TabIndex;
      IsFirstTab := IsFirst and (FScroller.Position = 0);            //
      IsLastTab := (Tab = TabPositions.Count - 1) and                // xu[qian], 03/10
                   (FScroller.Position + Tab = Tabs.Count - 1);      //


      if FOrientation = toBottom then
        R := Rect(TabPos.StartPos - HalfEdgeWidth, FMargins.FTopEdge - 1,
                  TabPos.StartPos + HalfEdgeWidth + TabPos.Size,
                  FMargins.FTopEdge - 1 + FTabHeight)  // xu[qian], 18/09
      else
        R := Rect(TabPos.StartPos - HalfEdgeWidth, FTopEdge2 + 1 - FTabHeight,
                  TabPos.StartPos + HalfEdgeWidth + TabPos.Size,
                  FTopEdge2 + 1);

      with MemBitmap.Canvas do
      begin
        if IsSelected then
        begin
          Brush.Color := FOptions.FaceColor;
          FillRect(R);

          if FOrientation = toBottom then
             R := Rect(TabPos.StartPos, FMargins.FTopEdge,
                       TabPos.StartPos + TabPos.Size, FMargins.FTopEdge - 1 +
                       FTabHeight) else
             R := Rect(TabPos.StartPos, FTopEdge2 + 1 - FTabHeight,
                       TabPos.StartPos + TabPos.Size, FTopEdge2);

          Font.Color := Self.Font.Color;

          if FOptions.FBoldSelTab then                                // xu[qian], 18/09
          begin
            OldFontStyle := Self.Font.Style;
            Font.Style := Self.Font.Style + [fsBold];
          end;

          if IsSelected and Assigned(FHotImages) then                 // xu[qian], 12/10
            aImages := FHotImages else                                //
            aImages := FImages;                                       //
          if Assigned(aImages) then
          begin
            if GetImageIndex(Tab + FirstIndex) <> -1 then
            begin
              aImages.Draw(MemBitmap.Canvas, R.Left + FMargins.FImgLeftMargin,
                R.Top + FMargins.FImgTopMargin, GetImageIndex(Tab + FirstIndex));
              Inc(R.Left, aImages.Width + FMargins.FImgRightMargin);
            end;
          end;
          //
          Inc(R.Top, FMargins.FTextTopMargin);
          Inc(R.Left, FMargins.FTextLeftMargin);
          //
          Brush.Style := bsClear;
          DrawText(Handle, PChar(Tabs[Tab + FirstIndex]), Length(Tabs[Tab + FirstIndex]), R, DT_CENTER);

          if FOptions.FBoldSelTab then Font.Style := OldFontStyle;    // xu[qian], 18/09

          Pen.Color := FOptions.HighlightColor; { Draw the left level1 }
          if FOrientation = toBottom then
          begin
            if FOptions.FlatStyle = fsNoFlat then
              MoveTo(TabPos.StartPos - HalfEdgeWidth, FMargins.FTopEdge - 1) else
              MoveTo(TabPos.StartPos - HalfEdgeWidth, FMargins.FTopEdge);
            if FOptions.FRoundedCorners then
              LineTo(TabPos.StartPos - HalfEdgeWidth, FTabHeight + FMargins.FTopEdge - 1) else
              LineTo(TabPos.StartPos - HalfEdgeWidth, FTabHeight + FMargins.FTopEdge);
          end else
          begin // toTop
            if FOptions.FRoundedCorners then
            begin
              MoveTo(TabPos.StartPos - HalfEdgeWidth, FTopEdge2 + 2 - FTabHeight);
              LineTo(TabPos.StartPos - HalfEdgeWidth, FTopEdge2 + 1);
            end else
            begin
              MoveTo(TabPos.StartPos - HalfEdgeWidth, FTopEdge2 + 1 - FTabHeight);
              LineTo(TabPos.StartPos - HalfEdgeWidth, FTopEdge2 + 1);
            end;
          end;

          if FOptions.FlatStyle = fsNoFlat then
          begin
            Pen.Color := FOptions.LightColor; { Draw the left level1 }
            if FOrientation = toBottom then
            begin
              MoveTo(TabPos.StartPos - HalfEdgeWidth + 1, FMargins.FTopEdge - 1);
              LineTo(TabPos.StartPos - HalfEdgeWidth + 1, FTabHeight + FMargins.FTopEdge - 1);
            end else
            begin
              MoveTo(TabPos.StartPos - HalfEdgeWidth + 1, FTopEdge2 + 1 - FTabHeight);
              LineTo(TabPos.StartPos - HalfEdgeWidth + 1, FTopEdge2 + 1);
            end;
          end;

          Pen.Color := FOptions.DarkShadowColor; { Draw the bottom-right __| }
          if FOrientation = toBottom then
          begin
            if FOptions.FRoundedCorners then
            begin
              MoveTo(TabPos.StartPos + HalfEdgeWidth + TabPos.Size, FMargins.FTopEdge);
              LineTo(TabPos.StartPos + HalfEdgeWidth + TabPos.Size, FTabHeight + FMargins.FTopEdge - 1);

              MoveTo(TabPos.StartPos - HalfEdgeWidth + 1, FTabHeight + FMargins.FTopEdge - 1);
              LineTo(TabPos.StartPos + HalfEdgeWidth + TabPos.Size, FTabHeight + FMargins.FTopEdge - 1);
            end else
            begin
              MoveTo(TabPos.StartPos + HalfEdgeWidth + TabPos.Size, FMargins.FTopEdge);
              LineTo(TabPos.StartPos + HalfEdgeWidth + TabPos.Size, FTabHeight + FMargins.FTopEdge);

              MoveTo(TabPos.StartPos - HalfEdgeWidth, FTabHeight + FMargins.FTopEdge - 1);
              LineTo(TabPos.StartPos + HalfEdgeWidth + TabPos.Size, FTabHeight + FMargins.FTopEdge - 1);
            end;
          end else
          begin // toTop
            if FOptions.FRoundedCorners then
            begin
              MoveTo(TabPos.StartPos + HalfEdgeWidth + TabPos.Size, FTopEdge2 + 2 - FTabHeight);
              LineTo(TabPos.StartPos + HalfEdgeWidth + TabPos.Size, FTopEdge2);
            end else
            begin
              MoveTo(TabPos.StartPos + HalfEdgeWidth + TabPos.Size, FTopEdge2 + 1 - FTabHeight);
              LineTo(TabPos.StartPos + HalfEdgeWidth + TabPos.Size, FTopEdge2);
            end;

            Pen.Color := FOptions.FHighlightColor;
            MoveTo(TabPos.StartPos - HalfEdgeWidth + 1, FTopEdge2 + 1 - FTabHeight);
            LineTo(TabPos.StartPos + HalfEdgeWidth + TabPos.Size{ + 1}, FTopEdge2 + 1 - FTabHeight);
          end;

          if (FOptions.FlatStyle = fsNoFlat) or
             (FOptions.FlatStyle = fsMiddleFlat) then
          begin
            if FOrientation = toBottom then { Draw the bottom __ }
            begin
              Pen.Color := FOptions.FShadowColor;
              if FOptions.FRoundedCorners then
              begin
                MoveTo(TabPos.StartPos - HalfEdgeWidth, FTabHeight + FMargins.FTopEdge - 2);
                LineTo(TabPos.StartPos + TabPos.Size + HalfEdgeWidth, FTabHeight + FMargins.FTopEdge - 2);
              end else
              begin
                MoveTo(TabPos.StartPos - HalfEdgeWidth + 1, FTabHeight + FMargins.FTopEdge - 2);
                LineTo(TabPos.StartPos + TabPos.Size + HalfEdgeWidth, FTabHeight + FMargins.FTopEdge - 2);
              end;
            end else
            begin
              Pen.Color := FOptions.FLightColor;
              MoveTo(TabPos.StartPos - HalfEdgeWidth + 1, FTopEdge2 + 2 - FTabHeight);
              LineTo(TabPos.StartPos + HalfEdgeWidth + TabPos.Size, FTopEdge2 + 2 - FTabHeight);
            end;
          end;

          if (FOptions.FlatStyle = fsNoFlat) then
          begin
            Pen.Color := FOptions.ShadowColor;
            if FOrientation = toBottom then
            begin
              MoveTo(TabPos.StartPos + TabPos.Size + HalfEdgeWidth - 1, FMargins.FTopEdge - 1);
              LineTo(TabPos.StartPos + TabPos.Size + HalfEdgeWidth - 1, FTabHeight + FMargins.FTopEdge - 1);
            end else
            begin
              MoveTo(TabPos.StartPos + TabPos.Size + HalfEdgeWidth - 1, FTopEdge2 + 3 - FTabHeight);
              LineTo(TabPos.StartPos + TabPos.Size + HalfEdgeWidth - 1, FTopEdge2 + 1);
            end;
         end;

        end else
        begin
          if FOrientation = toBottom then
            R := Rect(TabPos.StartPos, FMargins.FTopEdge,
                      TabPos.StartPos + TabPos.Size, FTabHeight + FMargins.FTopEdge - 1)
          else
            R := Rect(TabPos.StartPos, FTopEdge2 + 1 - FTabHeight,
                      TabPos.StartPos + TabPos.Size, FTopEdge2);

          Brush.Style := bsClear;

          if TabEnabled(Tab + FirstIndex) then
             Font.Color := FOptions.UnSelectedColor else
             Font.Color := Self.Font.Color;

          if IsSelected and Assigned(FHotImages) then     // xu[qian], 12/10
            aImages := FHotImages else                    //
            aImages := FImages;                           //
          if Assigned(aImages) then
          begin
            if GetImageIndex(Tab + FirstIndex) <> -1 then
            begin
              aImages.Draw(MemBitmap.Canvas, R.Left + FMargins.FImgLeftMargin,
                R.Top + FMargins.FImgTopMargin, GetImageIndex(Tab + FirstIndex));
              Inc(R.Left, aImages.Width + FMargins.FImgRightMargin);
            end;
          end;

          Inc(R.Top, FMargins.FTextTopMargin);
          Inc(R.Left, FMargins.FTextLeftMargin);
          //
          DrawText(Handle, PChar(Tabs[Tab + FirstIndex]), Length(Tabs[Tab + FirstIndex]), R, DT_CENTER or DT_VCENTER);

          Pen.Color := FOptions.SeparatorColor; // Separators

          if (IsFirst and not IsFirstTab) or                                //
             (IsFirstTab and FOptions.FLineOnSide) then                     // xu[qian], 03/10
          begin
            if FOrientation = toBottom then
            begin
              MoveTo(TabPos.StartPos - HalfEdgeWidth - 1, FMargins.FTopEdge + FSepartorMargin);                  //
              LineTo(TabPos.StartPos - HalfEdgeWidth - 1, FMargins.FTopEdge + FTabHeight - FSepartorMargin);     //
            end else
            begin
              MoveTo(TabPos.StartPos - HalfEdgeWidth - 1, FTopEdge2 - FSepartorMargin);                          //
              LineTo(TabPos.StartPos - HalfEdgeWidth - 1, FTopEdge2 - FTabHeight + FSepartorMargin);             //
            end;
          end;

          if not (not FOptions.FLineOnSide and IsLastTab) then              // xu[qian], 03/10
          begin
            if FOrientation = toBottom then
            begin
              MoveTo(TabPos.StartPos + TabPos.Size + HalfEdgeWidth + 1, FMargins.FTopEdge + FSepartorMargin);                   //
              LineTo(TabPos.StartPos + TabPos.Size + HalfEdgeWidth + 1, FMargins.FTopEdge + FTabHeight - FSepartorMargin);      //
            end else
            begin
              MoveTo(TabPos.StartPos + TabPos.Size + HalfEdgeWidth + 1, FTopEdge2 - FSepartorMargin);                           //
              LineTo(TabPos.StartPos + TabPos.Size + HalfEdgeWidth + 1, FTopEdge2 - FTabHeight + FSepartorMargin);              //
            end;
          end;

          if FOrientation = toBottom then // Border
          begin
            Pen.Color := FOptions.DarkShadowColor;
            MoveTo(TabPos.StartPos, FMargins.FTopEdge);
            LineTo(TabPos.StartPos + TabPos.Size, FMargins.FTopEdge);
          end else
          begin
            Pen.Color := FOptions.FHighlightColor;
            MoveTo(TabPos.StartPos, FTopEdge2);
            LineTo(TabPos.StartPos + TabPos.Size, FTopEdge2);
          end;

        end;
      end;
    end;

  { draw onto the screen }
    if (not (FOptions.BorderStyle = bvLowered)) and
     (not (FOptions.BorderStyle = bvRaised)) then
    begin
      Canvas.Draw(0, 0, MemBitmap);
    end else
    begin
      BmpR := ClientRect;
      if FOrientation = toBottom then
        Inc(BmpR.Top, FMargins.EdgeHeight) else
        Dec(BmpR.Bottom, FMargins.EdgeHeight-1);

      if FOptions.BorderStyle = bvLowered then
        Frame3D(Canvas, BmpR, FOptions.DarkShadowColor,
        FOptions.HighlightColor, 1);

      if FOptions.BorderStyle = bvRaised then
        Frame3D(Canvas, BmpR, FOptions.HighlightColor,
        FOptions.DarkShadowColor, 1);

      if FOrientation = toBottom then
        Canvas.Draw(1, 0, MemBitmap) else
        Canvas.Draw(1, 1, MemBitmap);
    end;
  finally
    MemBitmap.Free;
  end;
end;

procedure TFourthTabSet.FixTabPos;
var
  FLastVisibleTab: Integer;

  function GetRightSide: Integer;
  begin
    Result := Width - FMargins.EndMargin;
    if AutoScroll and (FVisibleTabs < Tabs.Count - 1) then
      Dec(Result, FScroller.Width + 4);
  end;

  function ReverseCalcNumTabs(Start, Stop: Integer; Canvas: TCanvas;
    Last: Integer): Integer;
  var
    W: Integer;
  begin
    if HandleAllocated then
    begin
      Result := Last;
      while (Start >= Stop) and (Result >= 0) do
        with Canvas do
        begin
          W := TextWidth(Tabs[Result]);
          if (FStyle = tsOwnerDraw) then MeasureTab(Result, W);
          Dec(Start, W + EdgeWidth); { next usable position }
          if Start >= Stop then Dec(Result);
        end;
      if (Start < Stop) or (Result < 0) then Inc(Result);
    end
    else
      Result := FFirstIndex;
  end;

begin
  if Tabs.Count > 0 then
  begin
    FLastVisibleTab := FFirstIndex + FVisibleTabs - 1;
    if FTabIndex > FLastVisibleTab then
      FFirstIndex := ReverseCalcNumTabs(GetRightSide, FMargins.StartMargin + EdgeWidth, Canvas, FTabIndex)
    else if (FTabIndex >= 0) and (FTabIndex < FFirstIndex) then
      FFirstIndex := FTabIndex;
  end;
end;

procedure TFourthTabSet.SetAutoScroll(Value: Boolean);
begin
  if (FAutoScroll <> Value) then
  begin
    FAutoScroll := Value;
    FScroller.Visible := False;
    ShowWindow(FScroller.Handle, SW_HIDE);
    Invalidate;
  end;
end;

function TFourthTabSet.CanChange(NewIndex: Integer): Boolean;
begin
  if TabEnabled(NewIndex) then
  begin
    Result := True;
    if Assigned(FOnChange) then
      FOnChange(Self, NewIndex, Result);
  end else
    Result := False;
end;

procedure TFourthTabSet.Change;
var
  CanBeChanged: Boolean;
begin
  CanBeChanged := True;
//  if (TabIndex = -1) and (Tabs.Count > 0) then TabIndex := 0;    // xu[qian], 03/10
  if Assigned(FOnChange) then FOnChange(Self, TabIndex, CanBeChanged);
end;

procedure TFourthTabSet.SetTabIndex(Value: Integer);
var
   NewValue: Integer;
   Found : Boolean;
begin
  if Value <> FTabIndex then
  begin
    if (Value < -1) or (Value >= Tabs.Count) then
{$IFDEF DELPHI3_UP}
      raise Exception.Create(SInvalidTabIndex);
{$ELSE}
      raise Exception.Create(LoadStr(SInvalidTabIndex));
{$ENDIF}

    if CanChange(Value) then
    begin
      FTabIndex := Value;
      FixTabPos;
      Click;
      Invalidate;
    end else
    begin
       Found := False;
       NewValue := Value + 1;
       while NewValue <> Value do
       begin
          if NewValue >= fTabs.count then
             NewValue := 0;
          if (NewValue < fTabs.count) and (not TabEnabled(NewValue)) then
             Inc(NewValue)
          else
          begin
             Found := True;
             Break;
          end
       end;
       if Found and CanChange(NewValue) then
       begin
         FTabIndex := NewValue;
         FixTabPos;
         Click;
         Invalidate;
       end;
    end;
  end;
end;

procedure TFourthTabSet.SelectNext(Direction: Boolean);
var
  NewIndex: Integer;
begin
  if Tabs.Count > 1 then
  begin
    NewIndex := TabIndex;
    if Direction then
      Inc(NewIndex) else
      Dec(NewIndex);
    if NewIndex = Tabs.Count then
      NewIndex := 0 else
      if NewIndex < 0 then
        NewIndex := Tabs.Count - 1;
    SetTabIndex(NewIndex);
  end;
end;

procedure TFourthTabSet.SetFirstIndex(Value: Integer);
begin
  if (Value >= 0) and (Value < Tabs.Count) then
  begin
    FFirstIndex := Value;
    Invalidate;
  end;
end;

procedure TFourthTabSet.SetTabList(Value: TStrings);
begin
  FTabs.Assign(Value);
  FTabIndex := -1;
  if FTabs.Count > 0 then
    TabIndex := 0
  else
    Invalidate;
end;

procedure TFourthTabSet.SetTabStyle(Value: TFourthTabStyle);
begin
  if Value <> FStyle then
  begin
    FStyle := Value;
    Invalidate;
  end;
end;

procedure TFourthTabSet.SetTabHeight(Value: Integer);
var
  SaveHeight: Integer;
begin
  if Value <> FOwnerDrawHeight then
  begin
    SaveHeight := FOwnerDrawHeight;
    try
      FOwnerDrawHeight := Value;
      FTabHeight := Value;
      Invalidate;
    except
      FOwnerDrawHeight := SaveHeight;
      FTabHeight := SaveHeight;
      raise;
    end;
  end;
end;

procedure TFourthTabSet.DrawTab(TabCanvas: TCanvas; R: TRect; Index: Integer;
  Selected: Boolean);
begin
  if Assigned(FOnDrawTab) then
    FOnDrawTab(Self, TabCanvas, R, Index, Selected);
end;

procedure TFourthTabSet.GetChildren(Proc: TGetChildProc{$IFDEF DELPHI3_UP}; Root: TComponent{$ENDIF});
begin
  //
end;

procedure TFourthTabSet.MeasureTab(Index: Integer; var TabWidth: Integer);
begin
  if Assigned(FOnMeasureTab) then
    FOnMeasureTab(Self, Index, TabWidth);
end;

procedure TFourthTabSet.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  TabPos: TTabPos;
  I: Integer;
  Extra: Integer;
  MinLeft: Integer;
  MaxRight: Integer;
begin
  if (Button = mbLeft) and (Y <= FTabHeight) then
  begin
    if (Y < FTabHeight div 2) then
      Extra := EdgeWidth div 3 else
      Extra := EdgeWidth div 2;

    for I := 0 to TabPositions.Count - 1 do
    begin
      Pointer(TabPos) := TabPositions[I];
      MinLeft := TabPos.StartPos - Extra;
      MaxRight := TabPos.StartPos + TabPos.Size + Extra;
      if (X >= MinLeft) and (X <= MaxRight) and TabEnabled(FirstIndex + I) then
      begin
        SetTabIndex(FirstIndex + I);
        Break;
      end;
    end;
  end;

  inherited MouseDown(Button, Shift, X, Y);                // xu[qian], 21/Nov
end;

procedure TFourthTabSet.WMSize(var Message: TWMSize);
var
  NumVisTabs, LastTabPos: Integer;

  function CalcNumTabs(Start, Stop: Integer; Canvas: TCanvas;
    First: Integer): Integer;
  var
    W: Integer;
  begin
    Result := First;
    while (Start < Stop) and (Result < Tabs.Count) do
      with Canvas do
      begin
        W := TextWidth(Tabs[Result]);
        if (FStyle = tsOwnerDraw) then MeasureTab(Result, W);
        Inc(Start, W + EdgeWidth); { next usable position }
        if Start <= Stop then Inc(Result);
      end;
  end;

begin
  inherited;
  if Tabs.Count > 1 then
  begin
    LastTabPos := Width - FMargins.EndMargin;
    NumVisTabs := CalcNumTabs(FMargins.StartMargin + EdgeWidth, LastTabPos, Canvas, 0);
    if (FTabIndex = Tabs.Count) or (NumVisTabs > FVisibleTabs) or
      (NumVisTabs = Tabs.Count) then FirstIndex := Tabs.Count - NumVisTabs;
    FDoFix := True;
  end;
  Invalidate;
end;

procedure TFourthTabSet.CMFontChanged(var Message: TMessage);
begin
  inherited;
  Canvas.Font := Font;
  Invalidate;
end;

procedure TFourthTabSet.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  Message.Result := DLGC_WANTALLKEYS;
end;

procedure TFourthTabSet.CMDialogChar(var Message: TCMDialogChar);
var
  I: Integer;
begin
  for I := 0 to FTabs.Count - 1 do
  begin
    if IsAccel(Message.CharCode, FTabs[I]) then
    begin
      Message.Result := 1;
      if FTabIndex <> I then
        SetTabIndex(I);
      Exit;
    end;
  end;
  inherited;
end;

procedure TFourthTabSet.DefineProperties(Filer: TFiler);
begin
  { Can be removed after version 1.0 }
  if Filer is TReader then inherited DefineProperties(Filer);
  Filer.DefineProperty('TabOrder', ReadIntData, nil, False);
  Filer.DefineProperty('TabStop', ReadBoolData, nil, False);
end;

procedure TFourthTabSet.ReadIntData(Reader: TReader);
begin
  Reader.ReadInteger;
end;

procedure TFourthTabSet.ReadBoolData(Reader: TReader);
begin
  Reader.ReadBoolean;
end;

procedure TFourthTabSet.SetDisabledTabList(const Value: TStrings);
begin
  FDisabledTabs.Assign(Value);
  Invalidate;
end;

function TFourthTabSet.TabEnabled(index: integer): Boolean;
begin
  Result := FDisabledTabs.IndexOf(FTabs[Index]) = -1;
end;

function TFourthTabSet.CalcTabPositions(Start, Stop: Integer; Canvas: TCanvas; First: Integer): Integer;
var
  Index: Integer;
  TabPos: TTabPos;
  W: Integer;
  //
  OldFontStyle: TFontStyles;                          // xu[qian], 18/09
begin
  TabPositions.Count := 0; { erase all previously cached data }
  Index := First;
  while (Start < Stop) and (Index < Tabs.Count) do
  begin
    with Canvas do
    begin
      if FOptions.FBoldSelTab then                    //
      begin                                           //
        OldFontStyle := Font.Style;                   // xu[qian], 18/09
        Font.Style := Font.Style + [fsBold];          //
      end;                                            //

      if TabEnabled(Index) then
      begin
         TabPos.StartPos := Start;
         W := TextWidth(Tabs[Index]) + FMargins.FTextLeftMargin + FMargins.FTextRightMargin;   // xu[qian], 18/09

         { Owner }
         if (FStyle = tsOwnerDraw) then MeasureTab(Index, W);

         if Assigned(FImages) then
           TabPos.Size := W + FImages.Width + 4 else
           TabPos.Size := W;
         Inc(Start, TabPos.Size + EdgeWidth); { next usable position }
      end;

      if FOptions.FBoldSelTab then Font.Style := OldFontStyle;        // xu[qian], 18/09

      if Start <= Stop then
      begin
        TabPositions.Add(Pointer(TabPos)); { add to list }
        Inc(Index);
      end;
    end;
  end;
  Result := Index - First;
end;

procedure TFourthTabSet.SetOrientation(Value: TFourthTabSetOrientation);
begin
  if (FOrientation <> Value) then
  begin
    FOrientation := Value;
    Invalidate;
  end;
end;

procedure TFourthTabSet.CMDesignHitTest(var Message: TCMDesignHitTest);
begin
  inherited;
  if (Message.Keys = 1) then
    SetTabIndex(ItemAtPos(Point(Message.Pos.x, Message.Pos.y)));
end;

procedure TFourthTabSet.SetImages(Value: TImagelist);
begin
  FImages := Value;
  Invalidate;
end;

procedure TFourthTabSet.SetHotImages(Value: TImagelist);        // xu[qian], 12/10
begin
  FHotImages := Value;
  Invalidate;
end;

procedure TFourthTabSet.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  if (Operation = opRemove) and (AComponent = FImages) then     // xu[qian], 12/10
    SetImages(nil);                                             // i don't know, how to do
end;

function TFourthTabSet.GetImageIndex(Index: Integer): Integer;
begin
  Result := -1;
  if (Index < FIndexes.Count) then
  try
    Result := StrToInt(FIndexes[Index]);
  except
    Result := -1;
  end;
end;

procedure TFourthTabSet.SetIndexes(Value: TStrings);
begin
  FIndexes.Assign(Value);
  Invalidate;
end;

procedure TFourthTabSet.DisplayTabHint(TabIndex: Integer);           // xu[qian], 13/10
begin
  Application.CancelHint;
  if (TabIndex in [0..TabHints.Count - 1]) then
  begin
    Hint := Trim(TabHints.Strings[TabIndex])
  end else
  if TabIndex = -1 then
    Hint := '' else
    Hint := FSavedHint;
end;

procedure TFourthTabSet.SetTabHints(Value: TStrings);                // xu[qian], 12/10
begin
  FTabHints.Assign(Value);
end;

procedure TFourthTabSet.WMNCHitTest(var Message: TWMNCHitTest);      // xu[qian], 13/10
var
  Result: Integer;
begin
  inherited;
  if not (csDesigning in ComponentState) and ShowHint then
  begin
    Result := ItemAtPos(ScreenToClient(Point(Message.XPos, Message.YPos)));
    if (Result <> FMouseOverTab) and (TabHints.Count > 0) then
    begin
      DisplayTabHint(Result);
      FMouseOverTab := Result;
    end;
  end else
    FMouseOverTab := -1;
end;

procedure TFourthTabSet.SetScrollChangeTab(Value: Boolean);
begin
  if (FScrollChangeTab <> Value) then
  begin
    FScrollChangeTab := Value;
    Invalidate;
  end;
end;

end.

