//**********************************************************
//*                  Copyright Information                 *
//*                                                        *
//*              ImgListBox by Derek van Daal              *
//*                                                        *
//* This component is absolutely freeware which means that *
//* you can do whatever you like with it :) If you make    *
//* any changes, please e-mail me the source at:           *
//* derekvandaal@hotmail.com                               *
//**********************************************************
unit ImgListBox;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ImgList;

type
  TImgListBox = class(TListBox)
  private
    FImages : TCustomImageList;
    FImageIndex : Array of Integer;
    FLastImage : Integer;
    FTransCol : TColor;

    procedure DrawItem(Control: TWinControl; Index: Integer;
      Rect: TRect; State: TOwnerDrawState);
    procedure SetImages(Value : TCustomImageList);
    procedure SetTransCol(Value : TColor);
    { Private declarations }
  protected
    { Protected declarations }
  public
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;

    procedure SetImageIndex(Item : Integer; Index : Integer);
    { Public declarations }
  published
    property Images: TCustomImageList read FImages write SetImages;
    property TransparentColor : TColor read FTransCol write SetTransCol;
    { Published declarations }
  end;

procedure Register;

implementation

{--------------------------------------------------------------------}
procedure Register;
begin
  RegisterComponents('Additional', [TImgListBox]);
end;
{--------------------------------------------------------------------}
constructor TImgListBox.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);

  OnDrawItem := DrawItem;
  Style := lbOwnerDrawFixed;
  ItemHeight := 18;

  FLastImage := -1;
  SetLength(FImageIndex, 0);

  FTransCol := clNone;
end;
{--------------------------------------------------------------------}
destructor TImgListBox.Destroy;
begin

  inherited Destroy;
end;
{--------------------------------------------------------------------}
procedure TImgListBox.SetImageIndex(Item : Integer; Index : Integer);
var
  k : Integer;

begin
  If Item > FLastImage then
    begin
      FLastImage := Item;
      SetLength(FImageIndex, Item + 1);
      If FImages <> nil then
        For k := 0 to FLastImage do
          begin
            If (FImageIndex[k] < 0) or (FImageIndex[k] > FImages.Count-1) then
              FImageIndex[k] := -1;
          end;
    end;
  FImageIndex[Item] := Index;
end;
{--------------------------------------------------------------------}
procedure TImgListBox.DrawItem(Control: TWinControl; Index: Integer;
  Rect: TRect; State: TOwnerDrawState);

var
  Bitmap : TBitmap;
  FontHeight : Byte;

begin
  Bitmap := TBitmap.Create;
  With (Control as TListBox).Canvas do
    begin
      Pen.Color := clWindowText;
      FillRect(Rect);
      FontHeight := Round(Font.Size * Font.PixelsPerInch / 72) + 2;
      If FImages <> nil then
        begin
          TextOut(FImages.Width + 4,
            Rect.Top + ((Rect.Bottom - Rect.Top) div 2) - (FontHeight div 2),
            Self.Items[Index]);

          If Index <= FLastImage then
            begin
              FImages.GetBitmap(FImageIndex[Index], Bitmap);
              Bitmap.Transparent := True;
              Draw(2, Rect.Top + ((Rect.Bottom - Rect.Top) div 2) - (Bitmap.Height div 2), Bitmap);
            end;
        end
      else
        begin
          TextOut(3, Rect.Top + ((Rect.Bottom - Rect.Top) div 2) - (FontHeight div 2),
            Self.Items[Index]);
        end;
    end;

  Bitmap.Free;
end;
{--------------------------------------------------------------------}
procedure TImgListBox.SetImages(Value : TCustomImageList);
begin
  FImages := value;
  If FImages <> nil then FImages.BkColor := FTransCol;
end;
{--------------------------------------------------------------------}
procedure TImgListBox.SetTransCol(Value : TColor);
begin
  FTransCol := Value;
  If FImages <> nil then FImages.BkColor := FTransCol;
end;
{--------------------------------------------------------------------}


end.
