{ Author
  \*********************************************************
  
  TmsgArray.pas 1.01
  
  (c) 2003, AGG Software, http://www.aggsoft.com,
  info@aggsoft.com
  
  All rights reserved.
  
  \*********************************************************
  
  Description
  Array of messages and messages boxes
  
  Conditions
  The source code is given as is. The author is not responsible
  for any possible damage done due to the use of this code. The
  component can be freely used in any application. The complete
  \source code remains property of the author and may not be
  distributed, published, given or sold in any form as such. No
  parts of the source code can be included in any other
  component or application without written authorization of the
  \author.
  
  
                                                                }
unit MsgArray;
{$G+}
(*$IMPORTEDDATA ON*)

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls,
  Forms, StdCtrls ,ExtCtrls;

const

  { Predefined dialog buttons captions }
  SMsgDlgYes = '&Yes';
  SMsgDlgNo = '&No';
  SMsgDlgOK = 'OK';
  SMsgDlgCancel = 'Cancel';
  SMsgDlgHelp = '&Help';
  SMsgDlgHelpNone = 'No help available';
  SMsgDlgHelpHelp = 'Help';
  SMsgDlgAbort = '&Abort';
  SMsgDlgRetry = '&Retry';
  SMsgDlgIgnore = '&Ignore';
  SMsgDlgAll = '&All';
  SMsgDlgNoToAll = 'N&o to All';
  SMsgDlgYesToAll = 'Yes to &All';
  SMsgDlgUnknown = 'Unknown';

type

  { Exception class of this component. For example will be raised
    if a message does not exist.                                  }
  EMsgArrayError = class(Exception);


  { Event class for showing and closing dialog box }
  TMessageEvent=procedure (Sender:TObject; MessageIdx:integer; MessageID:string) of object;
  { Event class on new button creation }
  TMessageButtonCreate=procedure (Sender:TObject; MessageIdx:integer; MessageID:string; Button:TButton) of object;

  TMsgArrayType = (mtWarning, mtError, mtInformation, mtConfirmation, mtCustom);
  { Buttons modal results }
  TMsgArrayBtn = (mbYes, mbNo, mbOK, mbCancel, mbAbort, mbRetry, mbIgnore,
    mbAll, mbNoToAll, mbYesToAll, mbHelp, mbUnknown);
  TMsgArrayButtons = set of TMsgArrayBtn;

  { Message dialogs icons types }
  TMessageIcon = (icon_NONE,icon_APPLICATION,icon_ASTERISK,
                  icon_ERROR,icon_EXCLAMATION,icon_HAND,
                  icon_INFORMATION,icon_QUESTION,icon_WARNING,
                  icon_WINLOGO,icon_USER);
  {
  TMsgStrinList = class(TStringList)
  protected
   procedure SetTextStr(const Value: string); override;
  end;
  }
  TMsgArray = class(TComponent)
  private
    { Private declarations }
   FOwner : TComponent;
   FMessages:TStringList;
   FMessagesText:TStringList;
   FMessagesCaption:TStringList;
   FDontShowAgainText:TStringList;
   FButtonCaptions:TStringList;
   FAgainCheckCaption:string;
   FInformationCaption:string;
   FWarningCaption:string;
   FErrorCaption:string;
   FQuestionCaption:string;
   FOnHelp:TMessageEvent;
   FOnClose:TMessageEvent;
   FOnShow:TMessageEvent;
   FOnButtonCreate:TMessageButtonCreate;
   FIamHere:boolean;
   procedure SetStringsList(Index:integer; Value: TStringList);
   procedure SetDontShowMsgAgain(Index: Integer; Value:boolean);
   function GetDontShowMsgAgain(Index: Integer):boolean;
   { This method will be automatically executed after changes in
     messages. For internal use only.                            }
   procedure UpdateMessagesTextAndCaption;
   { This method will be automatically executed after changes in
     messages captions, messages text arrays. For internal use
     \only.                                                      }
   procedure UpdateMessages;
   procedure Update(Sender:TObject);
   function GetParam(str:string; r:string; idx:integer):string;
   function GetAveCharSize(Canvas: TCanvas): TPoint;
   procedure FormShow(Sender:TObject);
   procedure HelpClick(Sender:TObject);
   procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
   function ParamFromID(ParamIndex:integer; MsgID:string):string;
   {
     Convert text identifer to message icon type
   }
   function TextToIconType(Text:string; var IconIdx:integer):TMessageIcon;
   {
     Convert text identifer to button type
   }
   function TextToButtonType(Text:string):TMsgArrayBtn;
   function TextFromID(const MsgID:string):string;
   function CaptionFromID(const MsgID:string):string;
  protected
  public
   { Public declarations }
   constructor Create(AOwner: TComponent); override;
   destructor Destroy; override;
   {
     Change all "Show again" state, to other value. Messages or messages boxes with
     'D' state will be ignored.
   }
   procedure SetDontShowAgainState(State:boolean=true);
   function ShowModal(Index:integer; IgnoreShowAgain:boolean=false):TModalResult;overload;
   {
     <B>function</B> ShowModal(<B>Index</B>: integer; <B>const</B> Args: <B>array</B> <B>of</B>
     <B>const</B>): TModalResult; <B>overload</B>;

     <B>function</B> ShowModal(MsgID: <B>string</B>): TModalResult; <B>overload</B>;

     <B>function</B> ShowModal(MsgID: <B>string</B>; <B>const</B> Args: <B>array</B> <B>of</B>
     <B>const</B>): TModalResult; <B>overload</B>;
   }
   function ShowModal(Index:integer; const Args: array of const; IgnoreShowAgain:boolean):TModalResult;overload;
   function ShowModal(MsgID:string; IgnoreShowAgain:boolean=false):TModalResult;overload;
   function ShowModal(MsgID:string; const Args: array of const; IgnoreShowAgain:boolean):TModalResult;overload;
   { "Show again" current state of messages array. All items in
     this array automatically update after changing in the <LINK TMsgArray.MessagesArray, MessagesArray>.
     Also <LINK TMsgArray.MessagesArray, MessagesArray> will be
     updated after changes in this item                                                                   }
   property DontShowAgain[Index: Integer]:boolean read GetDontShowMsgAgain write SetDontShowMsgAgain;
   { Messages text array indexed by <LINK TMsgArray.MessagesArray, messages ID>.
     Read only.                                                                  }
   property MsgText[const MsgID:string]: string read TextFromID;
   { Messages caption array indexed by <LINK TMsgArray.MessagesArray, messages ID>.
     Read only.                                                                     }
   property MsgCaption[const MsgID:string]: string read CaptionFromID;
  published
   { Default caption for "Show again" checkbox }
   property AgainCheckCaption:string read FAgainCheckCaption write FAgainCheckCaption;
   { Messages array. All items in this array must be in the
     following format:

     Message ID||Message text||Message caption||Message
     buttons||Show again flag||Message dialog icon

     Where:

         * Message ID is string with message unique identifier.
           (Example: MSG_HELLO);
         * Message text is message dialog text (Example: Hello!);
         * Message caption is message dialog caption (Example:
           Info). If message caption have been omitted, then caption
           will be changed to default value of this dialog type;
         * Message buttons is string value describing buttons type
           in the dialog window. Possible constants: 'YES', 'NO', 'OK',
           'CANCEL', 'ABORT', 'RETRY', 'IGNORE', 'ALL', 'NOTOALL',
           'YESTOALL', 'HELP'. Some buttons in the dialog will be
           appeared if you combine some buttons identifiers. For example
           'OK+CANCEL';
         * Show again flag is one of the following characters: Y -
           flag is checked, N - flag isn't checked, D - don't show
           checked flag;
         * Message dialog icon is one of the following constants:
           'APPLICATION','ASTERISK', 'ERROR','EXCLAMATION','HAND',
           'INFORMATION','QUESTION','WARNING', 'WINLOGO'. This constants
           are identical for MessageBox Win32 API function parameters.   }
   property MessagesArray:TStringList index 0 read FMessages write SetStringsList;
   { Messages text array. All items in this array automatically
     update after changing in the <LINK TMsgArray.MessagesArray, MessagesArray>.
     Also <LINK TMsgArray.MessagesArray, MessagesArray> will be
     updated after changes in this item                                          }
   property MessagesText:TStringList index 1 read FMessagesText write SetStringsList;
   { Messages caption array. All items in this array automatically
     update after changing in the <LINK TMsgArray.MessagesArray, MessagesArray>.
     Also <LINK TMsgArray.MessagesArray, MessagesArray> will be
     updated after changes in this item                                          }
   property MessagesCaption:TStringList index 2 read FMessagesCaption write SetStringsList;
   { This published property you can use for store "Show again"
     current state of messages array to registry or INI file. All
     items in this array automatically update after changing in
     the MessagesArray. Also MessagesArray will be updated after
     changes in this item                                         }
   property DontShowAgainText:TStringList index 3 read FDontShowAgainText write SetStringsList;
   { Dialog buttons captions. On create will be loaded with
     default values in English language.

     After changing hold new values                         }
   property ButtonCaptions:TStringList index 4 read FButtonCaptions write SetStringsList;
   { Default caption for information dialogs }
   property InformationCaption:string read FInformationCaption write FInformationCaption;
   { Default caption for warnings dialogs }
   property WarningCaption:string read FWarningCaption write FWarningCaption;
   { Default caption for error dialogs }
   property ErrorCaption:string read FErrorCaption write FErrorCaption;
   { This event will be generated on Help button click. Here you
     can call help windows.                                      }
   property OnHelp:TMessageEvent read FOnHelp write FOnHelp;
   { This event will be generated before closing dialog windows.
     As event parametrs you receive: message identifier (<LINK TMsgArray.MessagesArray, MsgID>)
     and message index (MessageIdx) in the <LINK TMsgArray.MessagesArray, MessagesArray>.       }
   property OnClose:TMessageEvent read FOnClose write FOnClose;
   { This event will be generated before showing dialog windows.
     As event parametrs you receive: message identifier (<LINK TMsgArray.MessagesArray, MsgID>)
     and message index (MessageIdx) in the <LINK TMsgArray.MessagesArray, MessagesArray>.       }
   property OnShow:TMessageEvent read FOnShow write FOnShow;
   { This event will be generated after new button creation in the
     dialog windows. Here you can change button caption and other
     button parameters. As event parametrs you receive: message
     identifier (<LINK TMsgArray.MessagesArray, MsgID>) and
     message index (MessageIdx) in the <LINK TMsgArray.MessagesArray, MessagesArray>. }
   property OnButtonCreate:TMessageButtonCreate read FOnButtonCreate write FOnButtonCreate;
  end;

const

  MAX_PARAMS=7;
  {
    Indexes of message box parameters in the Messages array
    string

    MSG_ID_IDX - message indentifier index;

    MSG_ICON_IDX - message box icon index;

    MSG_SHOW_AGAIN_IDX - show again index;
    
    MSG_TEXT_IDX - message text index;
    
    MSG_CAPTION_IDX- message caption index;
    
    MSG_BUTTON_IDX - message buttons index.                
  }
  MSG_ID_IDX=0;
  MSG_TEXT_IDX=1;
  MSG_CAPTION_IDX=2;
  MSG_BUTTON_IDX=3;
  MSG_SHOW_AGAIN_IDX=4;
  MSG_ICON_IDX=5;
  MSG_DEFAULT_RESULT_IDX=6;

  { Message box button string identifiers. }
  ButtonNames: array[TMsgArrayBtn] of string = (
    'YES', 'NO', 'OK', 'CANCEL', 'ABORT', 'RETRY', 'IGNORE', 'ALL', 'NOTOALL',
    'YESTOALL', 'HELP', #0);

procedure Register;

implementation

const

  mcHorzMargin = 8;
  mcVertMargin = 6;
  mcHorzSpacing = 10;
  mcVertSpacing = 10;
  mcButtonWidth = 50;
  mcButtonHeight = 14;
  mcButtonSpacing = 4;

  { Message box icons string identifiers. }
  TMessageIconStr:array [0..9] of string  =
                 ('','APPLICATION','ASTERISK',
                  'ERROR','EXCLAMATION','HAND',
                  'INFORMATION','QUESTION','WARNING',
                  'WINLOGO');

  ModalResults: array[TMsgArrayBtn] of Integer = (
    mrYes, mrNo, mrOk, mrCancel, mrAbort, mrRetry, mrIgnore, mrAll, mrNoToAll,
    mrYesToAll, 0, mrNone);

var
  { Predefined buttons caption array }
  BtnCaptions: array[TMsgArrayBtn] of string = (
    SMsgDlgYes, SMsgDlgNo, SMsgDlgOK, SMsgDlgCancel, SMsgDlgAbort,
    SMsgDlgRetry, SMsgDlgIgnore, SMsgDlgAll, SMsgDlgNoToAll, SMsgDlgYesToAll,
    SMsgDlgHelp, SMsgDlgUnknown);

constructor TMsgArray.Create(AOwner: TComponent);
var B:TMsgArrayBtn;
    s:string;
begin
  inherited;
  FIamHere:=false;
  FOwner := AOwner;
  FMessages:=TStringList.Create;
  FMessagesText:=TStringList.Create;
  FMessagesCaption:=TStringList.Create;
  FDontShowAgainText:=TStringList.Create;
  FButtonCaptions:=TStringList.Create;
  FAgainCheckCaption:='Do not ask me again';
  FInformationCaption:='Information';
  FWarningCaption:='Warning';
  FErrorCaption:='Error';
  FQuestionCaption:='Question';
  for B := Low(TMsgArrayBtn) to High(TMsgArrayBtn) do
   begin
    s:=BtnCaptions[B];
    FButtonCaptions.Add(s);
   end;
  UpdateMessagesTextAndCaption;
  FMessages.OnChange:=Update;
  FMessagesText.OnChange:=Update;
  FMessagesCaption.OnChange:=Update;
  FDontShowAgainText.OnChange:=Update;
end;

destructor TMsgArray.Destroy;
begin
  FreeAndNil(FMessagesCaption);
  FreeAndNil(FMessagesText);
  FreeAndNil(FMessages);
  FreeAndNil(FDontShowAgainText);
  FreeAndNil(FButtonCaptions);
  inherited;
end;

function TMsgArray.GetParam(str:string; r:string; idx:integer):string;
var i,k:integer;
    stemp:string;
begin
  Result:='';
  stemp:=str;
  for i:=1 to idx do
    begin
     k:=pos(r,stemp);
     if (k=0)then k:=length(stemp)+1;
     Result:=copy(stemp,1,k-1);
     delete(stemp,1,k+length(r)-1);
    end;
  Result:=Trim(Result);
end;

procedure TMsgArray.SetStringsList(Index:integer; Value: TStringList);
begin
  case index of
   0:begin
      FMessages.Assign(Value);
      UpdateMessagesTextAndCaption;
     end;
   1:begin
      FMessagesText.Assign(Value);
      UpdateMessages;
     end;
   2:begin
      FMessagesCaption.Assign(Value);
      UpdateMessages;
     end;
   3:begin
      FDontShowAgainText.Assign(Value);
      UpdateMessages;
     end;
   4:FButtonCaptions.Assign(Value);
   end;
end;

procedure TMsgArray.SetDontShowMsgAgain(Index: Integer; Value:boolean);
var state:boolean;
    j:integer;
    sl:TStringList;
    s:string;
begin
 if Index>=FMessages.Count then exit;
 state:= GetDontShowMsgAgain(Index);
 if state<>value then
  begin
   sl:=TStringList.Create;
   sl.Text:=stringreplace(FMessages[Index], '||',#13#10,[rfReplaceAll]);
   for j:=sl.Count to MAX_PARAMS do sl.Add('');
   if state then sl[MSG_SHOW_AGAIN_IDX]:='Y' else sl[MSG_SHOW_AGAIN_IDX]:='N';
   for j:=0 to MAX_PARAMS-1 do s:=s+sl[j]+'||';
   FMessages[Index]:=s;
   FreeAndNil(sl);
   UpdateMessagesTextAndCaption;
  end;
end;

function TMsgArray.GetDontShowMsgAgain(Index: Integer):boolean;
begin
 Result:=true;
 if Index>=FMessages.Count then exit;
 Result:= getParam(FMessages[Index],'||',MSG_SHOW_AGAIN_IDX+1) = 'Y';
end;

procedure TMsgArray.UpdateMessagesTextAndCaption;
var i,j:integer;
    sl:TStringList;
begin
 sl:=TStringList.Create;
 FMessagesText.BeginUpdate;
 FMessagesText.Clear;
 FMessagesCaption.BeginUpdate;
 FMessagesCaption.Clear;
 FDontShowAgainText.BeginUpdate;
 FDontShowAgainText.Clear;
 for i:=0 to FMessages.Count-1 do
  begin
   sl.Text:=stringreplace(FMessages[i], '||',#13#10,[rfReplaceAll]);
   for j:=sl.Count to MAX_PARAMS do sl.Add('');
   FMessagesText.Add(sl[MSG_TEXT_IDX]);
   FMessagesCaption.Add(sl[MSG_CAPTION_IDX]);
   if sl[MSG_SHOW_AGAIN_IDX]='N' then FDontShowAgainText.Add('N')
   else if sl[MSG_SHOW_AGAIN_IDX]='Y' then FDontShowAgainText.Add('Y')
   else FDontShowAgainText.Add('D');
  end;
 FMessagesText.EndUpdate;
 FMessagesCaption.EndUpdate;
 FDontShowAgainText.EndUpdate;
 FreeAndNil(sl);
end;

procedure TMsgArray.UpdateMessages;
var i,j:integer;
    sl:TStringList;
    s:string;
begin
 sl:=TStringList.Create;
 FMessages.BeginUpdate;
 for i:=0 to FMessages.Count-1 do
  begin
   sl.Text:=stringreplace(FMessages[i], '||',#13#10,[rfReplaceAll]);
   for j:=sl.Count to MAX_PARAMS do sl.Add('');
   if (i<FMessagesText.Count) then
    sl[MSG_TEXT_IDX]:=FMessagesText[i];
   if (i<FMessagesCaption.Count) then
    sl[MSG_CAPTION_IDX]:=FMessagesCaption[i];
   if (i<FDontShowAgainText.Count) then
    if FDontShowAgainText[i]='N' then sl[MSG_SHOW_AGAIN_IDX]:='N'
    else if FDontShowAgainText[i]='Y' then sl[MSG_SHOW_AGAIN_IDX]:='Y'
    else sl[MSG_SHOW_AGAIN_IDX]:='D';
   s:='';
   for j:=0 to MAX_PARAMS-1 do s:=s+sl[j]+'||';
   FMessages[i]:=s;
  end;
 FMessages.EndUpdate;
 FreeAndNil(sl);
end;

procedure TMsgArray.Update(Sender:TObject);
begin
 if FIamHere then exit;
 FIamHere:=true;
 if Sender=FMessages then UpdateMessagesTextAndCaption
 else UpdateMessages;
 FIamHere:=false;
end;

function TMsgArray.GetAveCharSize(Canvas: TCanvas): TPoint;
var
  I: Integer;
  Buffer: array[0..51] of Char;
begin
  for I := 0 to 25 do Buffer[I] := Chr(I + Ord('A'));
  for I := 0 to 25 do Buffer[I + 26] := Chr(I + Ord('a'));
  GetTextExtentPoint(Canvas.Handle, Buffer, 52, TSize(Result));
  Result.X := Result.X div 52;
end;

function TMsgArray.TextToIconType(Text:string; var IconIdx:integer):TMessageIcon;
var s:string;
    i:integer;
begin
 s:=UpperCase(trim(Text));
 Result:=icon_NONE;
 if s='' then exit;
 for i:=1 to High(TMessageIconStr) do
  if TMessageIconStr[i]=s then
   begin
    Result:=TMessageIcon(i);
    exit;
   end;
 for i:=1 to length(s) do
  if not (s[i] in ['0'..'9']) then exit;
 IconIdx:=StrToInt(s);
 Result:=icon_USER;
end;

function TMsgArray.TextToButtonType(Text:string):TMsgArrayBtn;
var s:string;
    B:TMsgArrayBtn;
begin
 s:=UpperCase(trim(Text));
 Result:=mbUnknown;
 if s='' then exit;
 for B := Low(TMsgArrayBtn) to High(TMsgArrayBtn) do
  if ButtonNames[B]=s then
   begin
    Result:=B;
    exit;
   end;
end;

{
  Call this method to display modal dialog window. You can call messages by index
  \or messages ID, with or without formatting parameters.
}
function TMsgArray.ShowModal(Index:integer; const Args: array of const; IgnoreShowAgain:boolean):TModalResult;
var s,
    MsgText,MsgCaption,MsgState,MsgButton,MsgIcon,
    MsgID,MsgDefaultResult:string;
    AllButtonsWidth,i,IconIndex,BtnIdx:integer;
    Form: TForm;
    Prompt: TLabel;
    IconImg: TImage;
    Checkbox: TCheckbox;
    DialogUnits: TPoint;
    ImageOffsetX,ImageOffsetY,TextHeight,OneLineTextWidth,
    TextRowHeight,TextRows,TextWidth,ALeft, ATop:longint;
    TopBorder,LeftBorder,ButtonTop,
    ButtonWidth, ButtonHeight, ButtonSpacing: Integer;
    Btn:TButton;
    BtnList:TStringList;
    B:TMsgArrayBtn;
    MessageIconType:TMessageIcon;
begin
 if (Index>=FMessages.Count)or(Index<0) then
  begin
   raise EMsgArrayError.Create(Format('Invalid message index [%d]',[Index]));
   exit;
  end;
 s:=FMessages[Index];
 MsgID:=GetParam(s,'||',MSG_ID_IDX+1);
 MsgText:=GetParam(s,'||',MSG_TEXT_IDX+1);
 MsgText:=StringReplace(MsgText,'%0D',#13,[rfReplaceAll]);
 MsgText:=StringReplace(MsgText,'%0A',#10,[rfReplaceAll]);
 if Length(Args)>0 then
  MsgText:=Format(MsgText, Args);
 MsgCaption:=GetParam(s,'||',MSG_CAPTION_IDX+1);
 MsgState:=GetParam(s,'||',MSG_SHOW_AGAIN_IDX+1);
 MsgButton:=GetParam(s,'||',MSG_BUTTON_IDX+1);
 if MsgButton='' then MsgButton:='OK';
 MsgDefaultResult:=GetParam(s,'||',MSG_DEFAULT_RESULT_IDX+1);
 if MsgDefaultResult='' then MsgDefaultResult:=GetParam(MsgButton,'+',1);
 MsgIcon:=GetParam(s,'||',MSG_ICON_IDX+1);
 if (MsgState='')or(not (MsgState[1] in ['D','Y','N'])) then MsgState:='D';
 if (MsgState='Y')and(not IgnoreShowAgain) then
  begin
   B:=TextToButtonType(MsgDefaultResult);
   Result:=ModalResults[B];
   exit;
  end;
 Form := TForm.Create(Application);
 IconImg:=nil;
 with Form do
  try
    Tag:=Index;
    OnShow:=FormShow;
    OnCloseQuery:=FormCloseQuery;
    ImageOffsetX:=0;ImageOffsetY:=0;
    Canvas.Font := Font;
    DialogUnits := GetAveCharSize(Canvas);
    BorderStyle := bsDialog;
    Position := poScreenCenter;
    MessageIconType:=TextToIconType(MsgIcon, IconIndex);
    if MsgCaption='' then
     if MessageIconType in [icon_INFORMATION,icon_ASTERISK] then MsgCaption:=FInformationCaption
     else if MessageIconType in [icon_EXCLAMATION,icon_WARNING] then MsgCaption:=FWarningCaption
     else if MessageIconType in [icon_ERROR,icon_HAND] then MsgCaption:=FErrorCaption
     else if MessageIconType in [icon_QUESTION] then MsgCaption:=FQuestionCaption
     else if MessageIconType = icon_APPLICATION then MsgCaption:=Application.Title;
    Caption := MsgCaption;

    LeftBorder:=MulDiv(mcHorzMargin, DialogUnits.X, 4);
    TopBorder:=MulDiv(mcVertMargin, DialogUnits.Y, 8);
    Prompt := TLabel.Create(Form);
    Prompt.AutoSize:=true;
    Prompt.Caption:=MsgText;
    TextRowHeight:=Canvas.TextHeight(MsgText);
    OneLineTextWidth:=Canvas.TextWidth(MsgText)+2*LeftBorder;
    TextHeight:=Prompt.Height;
    TextWidth:=Prompt.Width+2*LeftBorder;
    if MessageIconType<>icon_NONE then
     begin
      IconImg:=TImage.Create(Form);
      with IconImg do
       begin
        IconImg.Parent := Form;
        Transparent:=true;
        AutoSize:=true;
        Left:=LeftBorder;
        Top:=TopBorder;
        case MessageIconType of
         icon_APPLICATION:IconImg.Picture.Icon.Handle:=LoadIcon(0, IDI_APPLICATION);
         icon_ASTERISK:IconImg.Picture.Icon.Handle:=LoadIcon(0, IDI_ASTERISK);
         icon_ERROR:IconImg.Picture.Icon.Handle:=LoadIcon(0, IDI_ERROR);
         icon_EXCLAMATION:IconImg.Picture.Icon.Handle:=LoadIcon(0, IDI_EXCLAMATION);
         icon_HAND:IconImg.Picture.Icon.Handle:=LoadIcon(0, IDI_HAND);
         icon_INFORMATION:IconImg.Picture.Icon.Handle:=LoadIcon(0, IDI_INFORMATION);
         icon_QUESTION:IconImg.Picture.Icon.Handle:=LoadIcon(0, IDI_QUESTION);
         icon_WARNING:IconImg.Picture.Icon.Handle:=LoadIcon(0, IDI_WARNING);
         icon_WINLOGO:IconImg.Picture.Icon.Handle:=LoadIcon(0, IDI_WINLOGO);
         icon_USER:;
        end;
        ImageOffsetX:=Width+LeftBorder;
        ImageOffsetY:=Top+trunc(IconImg.Height/2)-TextHeight;
       end;
     end;
    ButtonWidth := MulDiv(mcButtonWidth, DialogUnits.X, 4);
    ButtonHeight := MulDiv(mcButtonHeight, DialogUnits.Y, 8);
    BtnList:=TStringList.Create;
    BtnList.Text:=StringReplace(MsgButton, '+', #13, [rfReplaceAll]);
    ButtonSpacing := MulDiv(mcButtonSpacing, DialogUnits.X, 4);
    AllButtonsWidth := BtnList.Count * ButtonWidth +
                       (BtnList.Count-1) * ButtonSpacing;
    AutoSize:=false;
    ClientWidth := TextWidth+ImageOffsetX;

    if((AllButtonsWidth+2*LeftBorder)>ClientWidth) then
     ClientWidth := AllButtonsWidth+2*LeftBorder;

    with Prompt do
    begin
      Parent := Form;
      WordWrap:=true;
      AutoSize := false;
      Caption := MsgText;
      ALeft:=ImageOffsetX+LeftBorder;
      if ImageOffsetY>0 then ATop:=ImageOffsetY else ATop:=TopBorder;
      Left:=ALeft;
      Top:=ATop;
//      AutoSize := true;
    end;
    Checkbox := TCheckbox.Create(Form);
    with Checkbox do
    begin
      Visible:=MsgState<>'D';
      Checked:=MsgState='Y';
      Parent := Form;
      Caption := FAgainCheckCaption;
      Left := Prompt.Left;
      Width := Canvas.TextWidth(FAgainCheckCaption)+2*LeftBorder;
      if not Visible then
       Top := Prompt.Top+Prompt.Height-Height
      else
       begin
        Top := Prompt.Top+Prompt.Height+TopBorder;
        if(Width>ClientWidth) then
         ClientWidth := Width;
       end;
    end;

    ButtonTop := Checkbox.Top+Checkbox.Height+TopBorder;
    if (IconImg<>nil)then
     begin
      if ButtonTop<(IconImg.Top+IconImg.Height+TopBorder) then
       ButtonTop:=IconImg.Top+IconImg.Height+TopBorder;
     end;
    ClientHeight := ButtonTop+ButtonHeight+TopBorder;
    for i:=0 to BtnList.Count-1 do
     begin
      B:=TextToButtonType(trim(BtnList[i]));
      if B=mbUnknown then continue;
      Btn:=TButton.Create(Form);
      with Btn do
       begin
        Parent := Form;
        Tag:=Index;
        Caption := 'OK';
        ModalResult := mrOk;
        BtnIdx:=ord(B);
        if BtnIdx<FButtonCaptions.Count then
         Caption := FButtonCaptions[BtnIdx];
        ModalResult := ModalResults[B];
        if B=mbHelp then OnClick:=HelpClick;
        ALeft := trunc((Form.ClientWidth-AllButtonsWidth)/2)+i*ButtonWidth+i*ButtonSpacing;
        SetBounds(ALeft, ButtonTop, ButtonWidth, ButtonHeight);
        if Form.ActiveControl=nil then
         begin
          Default := True;
          Form.ActiveControl:=Btn;
         end;
        if Assigned(FOnButtonCreate) then
         FOnButtonCreate(Self, Index, MsgID, Btn);
       end;
     end;
    FreeAndNil(BtnList);
    Result := ShowModal;
    if (Result<>mrNone)and(CheckBox.Visible)then
     begin
      if CheckBox.Checked then MsgState:='Y' else MsgState:='N';
      for i:=FDontShowAgainText.Count to Index do FDontShowAgainText.Add('');
      if FDontShowAgainText[Index]<>MsgState then FDontShowAgainText[Index]:=MsgState;
     end;
  finally
    Form.Free;
  end;
end;

function TMsgArray.ShowModal(Index:integer; IgnoreShowAgain:boolean=false):TModalResult;
begin
 Result:=ShowModal(Index,[],IgnoreShowAgain);
end;

function TMsgArray.ShowModal(MsgID:string; IgnoreShowAgain:boolean=false):TModalResult;
var i,Index:integer;
begin
 Index:=-1;
 for i:=0 to FMessages.Count-1 do
  if GetParam(FMessages[i],'||',MSG_ID_IDX+1)=MsgID then
   begin
    Index:=i;
    break;
   end;
 Result:=ShowModal(Index,[],IgnoreShowAgain);
end;


function TMsgArray.ShowModal(MsgID:string; const Args: array of const; IgnoreShowAgain:boolean):TModalResult;
var i,Index:integer;
begin
 Index:=-1;
 for i:=0 to FMessages.Count-1 do
  if GetParam(FMessages[i],'||',MSG_ID_IDX+1)=MsgID then
   begin
    Index:=i;
    break;
   end;
 Result:=ShowModal(Index, Args, IgnoreShowAgain);
end;

function TMsgArray.ParamFromID(ParamIndex:integer; MsgID:string):string;
var i,Index:integer;
begin
 Result:='';
 Index:=-1;
 for i:=0 to FMessages.Count-1 do
  if GetParam(FMessages[i],'||',MSG_ID_IDX+1)=MsgID then
   begin
    Index:=i;
    Result:=GetParam(FMessages[i],'||',ParamIndex+1);
    Result:=StringReplace(Result,'%0D',#13,[rfReplaceAll]);
    Result:=StringReplace(Result,'%0A',#10,[rfReplaceAll]);
    break;
   end;
 if Index<0 then
  raise EMsgArrayError.Create(Format('Invalid message ID [%s]',[MsgID]));
end;

function TMsgArray.TextFromID(const MsgID:string):string;
begin
 Result:=ParamFromID(MSG_TEXT_IDX, MsgID);
end;

function TMsgArray.CaptionFromID(const MsgID:string):string;
begin
 Result:=ParamFromID(MSG_CAPTION_IDX, MsgID);
end;

procedure TMsgArray.SetDontShowAgainState(State:boolean=true);
var i:integer;
    c,state_str:char;
    UpdateNeed:boolean;
begin
 UpdateNeed:=false;
 FDontShowAgainText.BeginUpdate;
 if State then state_str:='Y' else state_str:='N';
 for i:=0 to FDontShowAgainText.Count-1 do
  begin
   if FDontShowAgainText[i]='N' then c:='N'
   else if FDontShowAgainText[i]='Y' then c:='Y'
   else c:='D';

   if (c<>state_str)and(c<>'D')then
    begin
     UpdateNeed:=true;
     FDontShowAgainText[i]:=state_str;
    end;
  end;
 FDontShowAgainText.EndUpdate;
 if UpdateNeed then UpdateMessages;
end;

procedure TMsgArray.FormShow(Sender:TObject);
var MessageID:string;
    MessageIdx:integer;
begin
 if Assigned(FOnShow) then
  begin
   MessageID:='';
   MessageIdx:=TComponent(Sender).Tag;
   if MessageIdx<FMessages.Count then
     MessageID:=GetParam(FMessages[MessageIdx],'||',MSG_ID_IDX+1);
   FOnShow(Self, MessageIdx, MessageID);
  end;
end;

procedure TMsgArray.HelpClick(Sender:TObject);
var MessageID:string;
    MessageIdx:integer;
begin
 if Assigned(FOnHelp) then
  begin
   MessageID:='';
   MessageIdx:=TComponent(Sender).Tag;
   if MessageIdx<FMessages.Count then
     MessageID:=GetParam(FMessages[MessageIdx],'||',MSG_ID_IDX+1);
   FOnHelp(Self, MessageIdx, MessageID);
  end;
end;

procedure TMsgArray.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
var MessageID:string;
    MessageIdx:integer;
begin
 if Assigned(FOnClose) then
  begin
   MessageID:='';
   MessageIdx:=TComponent(Sender).Tag;
   if MessageIdx<FMessages.Count then
     MessageID:=GetParam(FMessages[MessageIdx],'||',MSG_ID_IDX+1);
   FOnClose(Self, MessageIdx, MessageID);
  end;
end;
{
procedure TMsgStrinList.SetTextStr(const Value: string);
begin
 Changing;
 inherited SetTextStr(Value);
 Changed;
end;
}

{ Register component in the Delphi IDE }
procedure Register;
begin
  RegisterComponents('System', [TMsgArray]);
end;

end.
