#ifndef FRONTA_H
  #define FRONTA_H

  #include <iostream.h>

  template<class T> class CPrioFronta {
	private : 
		T *m_Pole;
		int *m_Priority;
		int m_Velikost;	// maximalni velikost (delka) fronty
		int m_Zacatek, m_Konec;	// kde je zacatek a konec fronty
		int m_Pocet;	// pocet prvku prave se nachazejicich ve fronte
			
	public :
		CPrioFronta(int _Velikost);
		~CPrioFronta();

		bool Vloz(const T& _Vklad, int _iPriorita);
		bool Vyjmi(T& _Vyber);
		
		bool JePrazdna() { return ((0 == m_Pocet) ? true : false); }
		bool JePlna() { return ((m_Pocet == m_Velikost) ? true : false); }

		void Vypis();
  };
	
  template<class T> CPrioFronta<T>::CPrioFronta(int _Velikost)
  {
	  m_Velikost = -1;
	  m_Pocet = 0;
	  m_Zacatek = 0;
	  m_Konec = -1;	// abychom v metode Vloz dostali spravny konec po prvnim vlozeni
	  
	  m_Pole = new T[_Velikost];


	  if(m_Pole)
	  {
		  // Vytvorime pole priorit pro prvky
		  m_Priority = new int[_Velikost];

		  if(m_Priority)
		  {
		  	  // pokud se podarilo nastavime velikost
			  m_Velikost = _Velikost;

			  // vynulujeme pole priorit
			  for(int i = 0; i < _Velikost; i++)
			  {
				  m_Priority[i] = 30000;
			  }
		  }
		  else
		  {
			  // jinak smazeme druhe pole
			  delete[] m_Pole;
		  }
	  }
  }

  template<class T> CPrioFronta<T>::~CPrioFronta()
  {
	  if(m_Priority)
	  {
		  delete[] m_Priority;
		  m_Priority = NULL;
	  }

	  if(m_Pole)
	  {
		  delete[] m_Pole;
		  m_Pole = NULL;
		  m_Velikost = 0;
	  }
  }

  template<class T> bool CPrioFronta<T>::Vloz(const T& _Vklad, int _iPriorita)
  {
	  bool bNavrat = false;

	  if(m_Pole && m_Priority)
	  {
		  if(!JePlna())
		  {
			  if(m_Konec == (m_Velikost - 1))
			  {
				  // pokud konec dosahne konce pole, pak koncime na prvnim prvku pole
				  m_Konec = 0;
			  }
			  else
			  {
				  m_Konec++;
			  }

			  // hledame misto pro novy prvek

			  int zarazpoz = m_Konec;
			  int poz;
			  
			  // nasledujici promenna znaci jestli je index konce pred zacatkem, musime overit i prazdnost fronty
			  bool bPrekroucena = (m_Konec < m_Zacatek);

			  while((zarazpoz > m_Zacatek) || bPrekroucena)
			  {
				  poz = zarazpoz - 1;
				  
				  if(poz < 0)
				  {
					  poz = m_Velikost - 1;
				  }

				  if(m_Priority[poz] < _iPriorita)
				  {
					  // posuneme prvek s nizsi prioritou
					  m_Pole[zarazpoz] = m_Pole[poz];
					  m_Priority[zarazpoz] = m_Priority[poz];

					  // a jdeme na dalsi prvek - ovsem kdyz dosahneme zacatku, tak preskocime na konec
					  if(zarazpoz == 0)
					  {
						  zarazpoz = m_Velikost - 1;
						  bPrekroucena = false;	// uz je zacatek pred koncem
					  }
					  else
					  {
						  zarazpoz--;
					  }
				  }
				  else
				  {
					  break;	// prerusime cyklus
				  }
			  }

			  // na nalezene misto ulozime novy prvek
			  m_Pole[zarazpoz] = _Vklad;
			  m_Priority[zarazpoz] = _iPriorita;	// ulozime prioritu
			  m_Pocet++;	// pribyl prvek

			  bNavrat = true;
		  }
	  }

	  return bNavrat;
  }

  template<class T> bool CPrioFronta<T>::Vyjmi(T& _Vyber)
  {
	  bool bNavrat = false;

	  if(m_Pole && m_Priority)
	  {
		  if(!JePrazdna())
		  {
			  _Vyber = m_Pole[m_Zacatek];
			  
			  if(m_Zacatek == (m_Velikost - 1))
			  {
				  // pokud zacatek dosahne konce pole, pak zaciname zase na prvnim prvku
				  m_Zacatek = 0;
			  }
			  else
			  {
				  // posun na novy zacatek
				  m_Zacatek++;
			  }

			  m_Pocet--;	// ubyl prvek
			  
			  bNavrat = true;
		  }
	  }

	  return bNavrat;
  }

  template <class T> void CPrioFronta<T>::Vypis()
  {
	  if(m_Pole)
	  {
		  if(!JePrazdna())
		  {
			  cout << "Obsahuje " << m_Pocet << " objektu (volnych " << m_Velikost - (m_Pocet) << ")" << endl;
			  for(int i = 0; i < m_Velikost; i++)
			  {
				  cout << m_Pole[i] << " s prioritou " << m_Priority[i] << endl;
			  }
		  }
		  else
		  {
			  cout << "Fronta je prazdna" << endl;
		  }
	  }
  }

#endif
