unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, DB, ADODB, Grids, DBGrids, ExtCtrls, DBCtrls, AdoConEd, StdCtrls,
  IniFiles, ComCtrls;

type
  TForm1 = class(TForm)
    DBGrid1: TDBGrid;
    DataSource1: TDataSource;
    ADOConnection1: TADOConnection;
    ADODataSet1: TADODataSet;
    DBNavigator1: TDBNavigator;
    StatusBar1: TStatusBar;
    Button1: TButton;
    Button2: TButton;
    Button3: TButton;
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure Button3Click(Sender: TObject);
    procedure ADOConnection1Disconnect(Connection: TADOConnection;
      var EventStatus: TEventStatus);
    procedure ADOConnection1ConnectComplete(Connection: TADOConnection;
      const Error: Error; var EventStatus: TEventStatus);
  private
    { Private declarations }
  public
    { Public declarations }
    function BuildADOConnectString(ADOConnection: TADOConnection): string;
    procedure ConfigADODatabase;
    function ConnectADODatabase: Boolean;
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

{Most of this code comes from code snippets which were provided to me courtesy of
 John Wester of Group W. Here are some notes he also gave me:
  1.  Remember to include AdoConEd in your "uses" clause.
  2.  If your user selects OLE DB Provider from the list that gets displayed when you
      click Build in the dialog presented by the call to EditConnectionString you'll
      then get a dialog that will present servers running MS-SQL and then the db's that
      they are serving up.
  3.  This has been tested with MDAC 2.8 only so I suggest that you have that version or above
      on your system.
 Note: This codes uses the customers table from the example Northwind database which comes
 with either the MSDE or MSSQL.

 This example courtesy of Bob Dalton from Digital Logistics. Please do not send me questions on this code.
 I provide NO tech support for this example! I just don't have the time....
}

//courtesy of John Wester

function TForm1.BuildADOConnectString(ADOConnection: TADOConnection): string;
begin
  result := ADOConnection.ConnectionString;
  ADOConnection1.Close;
  // calling EditConnectionString displays the dialog, and
  // sets the ConnectString property to the resulting value.
  if EditConnectionString(ADOConnection) then result := ADOConnection.ConnectionString;
end;

//courtesy of John Wester

procedure TForm1.ConfigADODatabase;
var
  IniFile: TIniFile;
  ConnectString: string;
  ADOIniName: string;
  TheIniFileName: string;
begin
  TheIniFileName := 'ADOTest.ini';
  ADOIniName := IncludeTrailingPathDelimiter(ExtractFilePath(paramstr(0))) + TheIniFileName;
  IniFile := TIniFile.Create(ADOIniName);
  try
    ConnectString := BuildADOConnectString(ADOConnection1);
    IniFile.WriteString('ADODATABASE', 'CONNECTSTRING', ConnectString);
  finally
    IniFile.Free;
  end;
end;

//courtesy of John Wester and modified by Bob Dalton

function TForm1.ConnectADODatabase: Boolean;
var
  IniFile: TIniFile;
  ADOIniName: string;
  TheIniFileName: string;
  ConnectString: string;
begin
  result := False;
  StatusBar1.Panels[1].text := 'Not Connected';
  TheIniFileName := 'ADOTest.ini';
  ADOIniName := IncludeTrailingPathDelimiter(ExtractFilePath(paramstr(0))) + TheIniFileName;
  IniFile := TIniFile.Create(ADOIniName);
  try
    ConnectString := IniFile.ReadString('ADODATABASE', 'CONNECTSTRING', '');
    ADOConnection1.ConnectionString := ConnectString;
  finally
    IniFile.Free;
  end;

  try
    if ConnectString <> '' then
    begin
      ADOConnection1.Connected := True;
      ADODataSet1.Active := True;
      DBNavigator1.Enabled := True;
      StatusBar1.Panels[1].text := 'Connected';
      result := True;
    end
    else
    begin
      if MessageDlg('Cannot connect to ADO database. Configure ' +
        'ADO Connection?', mtError, [mbYes, mbNo], 0) = mrYes then
      begin
        ConfigADODatabase;
        ConnectADODatabase;
      end
      else
      begin
        Result := False;
      end;
    end;
  except
    if MessageDlg('Cannot connect to ADO database. Configure ' +
      'ADO Connection?', mtError, [mbYes, mbNo], 0) = mrYes then
    begin
      ConfigADODatabase;
      ConnectADODatabase;
    end
    else
      Result := False;
  end;
end;

procedure TForm1.Button1Click(Sender: TObject);
begin
  ConnectADODatabase;
end;

procedure TForm1.Button2Click(Sender: TObject);
begin
  ADODataSet1.Active := False;
  ADOConnection1.Close;
  StatusBar1.Panels[1].text := 'Not Connected'
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  StatusBar1.Panels[1].text := 'Not Connected';
end;

procedure TForm1.Button3Click(Sender: TObject);
var
  IniFile: TIniFile;
  ConnectString: string;
  ADOIniName: string;
  TheIniFileName: string;
begin
  if MessageDlg('Editing the connection will delete any existing MSSQL server connection information. Proceed?',
    mtConfirmation, [mbYes, mbNo], 0) = mrYes then
  begin
    TheIniFileName := 'ADOTest.ini';
    ADOIniName := IncludeTrailingPathDelimiter(ExtractFilePath(paramstr(0))) + TheIniFileName;
    IniFile := TIniFile.Create(ADOIniName);
    try
      ConnectString := '';
      IniFile.WriteString('ADODATABASE', 'CONNECTSTRING', ConnectString);
    finally
      IniFile.Free;
    end;
    if ADODataSet1.Active = True then ADODataSet1.Active := False;
    ADOConnection1.Close;
    ADOConnection1.ConnectionString := '';
    StatusBar1.Panels[1].text := 'Not Connected';
    ConnectADODatabase;
  end;
end;

procedure TForm1.ADOConnection1Disconnect(Connection: TADOConnection;
  var EventStatus: TEventStatus);
begin
  if EventStatus = esOK then showmessage('Disconnected Successfully');
end;

procedure TForm1.ADOConnection1ConnectComplete(Connection: TADOConnection;
  const Error: Error; var EventStatus: TEventStatus);
begin
  if EventStatus = esOK then showmessage('Connected Successfully');
end;

end.

