{*******************************************************}
{                                                       }
{         Copyright (c) FINANCES Ltd                    }
{         2003   ver. 1.1                                 }
{                                                       }
{         www.finances-ltd.narod.ru                     }
{         finances-ltd@narod.ru                         }
{                                                       }
{*******************************************************}


{
TzIniData has the same basic functions as TIniFile and many additional methods for stored compressed data in binary file, such as:
Propertes:
- BOF, EOF
- RecordCount
- RecNo
- Active
- Sorted
- FileName
- FormSave
- FileNameAutoSet
- Password

Methods:
- IsEmpty
- NotEmpty
- Rename
- CopyRecord
- EmptyTable
- RecordExist
- Search
- GoToRec
- Delete
- Save(ItemName: string; ms: TStream)
- Save(ItemName: Integer; ms: TStream)
- Save(ItemName: string; Text: string)
- Save(ItemName: string; Value: Boolean)
- Save(ItemName: string; Value: Integer)
- Save(ItemName: string; Value: DWord)
- Save(ItemName: string; Value: Double)
- Save(ItemName: string; Value: TFont)
- Save(ItemName: string; Objct: TObject)   // TTreeView, TListView, TStrings...
- Save(aForm: TCustomForm)
- SaveFromFile(ItemName, FileName: string)
- SaveToFile(ItemName, FileName: string)
- SaveToStream(ItemName: string; OutStream: TStream)
- Load(ItemName: string; ms: TStream)
- Load(ItemName: string; Objct: TObject)
- Load(ItemName: string; var Text: string)
- Load(ItemName: string; Value: TFont)
- Load(aForm: TCustomForm)
- LoadToStream(ItemName: string = '')
- AsString(ItemName: string; Default: string = '')
- AsBool(ItemName: string; Default: boolean = False)
- AsInteger(ItemName: string; Default: Integer = 0)
- AsInt(ItemName: string; Default: Integer = 0)
- AsFloat(ItemName: string; Default: Double = 0)
- AsDate(ItemName: string; Default: Double = 0)
- AsStream(ItemName: string; Value: TStream)
- AsStream(ItemName: string)
- AsStrings(ItemName: string; var VarStrings: TStrings)
- AsStrings(ItemName: string; var VarStrings: TStringList)

Events:
- BeforeOpen
- AfterOpen
- BeforeClose
}

{$DEFINE INCLUDE_COMCTRL}
{$DEFINE INCLUDE_GRAPHIC}
{$DEFINE INCLUDE_HTML}

unit zIniData;

interface

uses
  Windows, Messages, SysUtils, Classes, Forms, ShlObj,
  ExtCtrls,
  {$IFDEF INCLUDE_GRAPHIC}
  Graphics,
  Jpeg,
  {$ENDIF}
  {$IFDEF INCLUDE_HTML}
  OleCtrls,
  SHDocVw,
  ActiveX,
  {$ENDIF}
  {$IFDEF INCLUDE_COMCTRL}
  ComCtrls,
  {$ENDIF}
  LHARoutines;

type

  TzSearchOpt = (zsoFromTop, zsoForward, zsoBack);
  type TzRecType = (zrtUnnoun,
                    zrtStrings,
                    zrtStringList,
                    zrtBitmap,
                    zrtIcon,
                    zrtJPEGImage,
                    zrtTreeView,
                    zrtListView,
                    zrtHtml);

  TRecInfo = record
    RecName: Integer;
    StartPos: Integer;
    RecLength: Integer;
  end;
  PRecInfo=^TRecInfo;

  TDataFileHeader = record
    GenHeader: Array[1..5] of Char;
    Created: TDateTime;
  end;

  TRefreshStateHeader = record
    StartData: Integer;
    RecCount: Integer;
    LastModified: TDateTime;
    Msg_Num: Integer;
    Password: Integer;
  end;

  TzStrStream = class( TMemoryStream )
  public
    constructor Create;
    destructor Destroy; override;

    procedure SetPointer(Ptr: Pointer; Size: Integer ); virtual;
  end;

  TFormSaveFormat   = (fsfNone, fsfSize, fsfPosition, fsfBoth);
  TFormSavePlace    = (fspNone, fspWinDir, fspAppDir, fspAppDataDir);
  EzIniDataError = class(Exception);

  TzIniData = class(TComponent)
  private
    { Private declarations }
    OldWndProc: TFarProc;
    NewWndProc: Pointer;
    FTableName: string;
    FStream: THandleStream;
    FzStrStream: TzStrStream;

    FRecInfoSize : Integer;
    FRecInfo : TRecInfo;

    FDataFileHeaderSize : Integer;
    FDataFileHeader : TDataFileHeader;

    FRefreshStateHeaderSize: Integer;
    FRefreshStateHeader: TRefreshStateHeader;
    FReadOnly,       // For use in Network
    FCanCompress: Boolean;

//    NewRecType: TzRecType;

    FRecNo,
    FMsgNum: Integer;
    FIsOpen: Boolean;
    FSorted: Boolean;

    FBeforeOpen,
    FAfterOpen,
    FBeforeClose,
    FAfterClose: TNotifyEvent;

    FBOF, FEOF, CheckMode: Boolean;
    FFormSave: TFormSaveFormat;
    FFileNameAutoSet: TFormSavePlace;

    procedure _ReadRecord(RecNum:Integer); overload;
    procedure _ReadRecord(Buffer:PChar; IntRecNum:Integer); overload;

    procedure _WriteRecord(RecNum:Integer); overload;
    procedure _WriteRecord(Buffer:PChar; IntRecNum:Integer); overload;

    procedure SetActive(Value: Boolean);
    procedure SetTableName(Value: string);
    function GetRecordCount: Integer;
    procedure SetRecNo(RecNum:Integer);
    function GetRecNo: Integer;
    procedure IncMsgNum;
    procedure IncRecCount;
    function GetRecLength: Integer;

    function _FoundItem(ItemName: Integer): Boolean;

    function _CompareRecords(Rec1,Rec2:Integer):Integer;
    procedure _SwapRecords(Rec1,REc2:Integer);
    function AllocRecordBuffer: PChar;

    procedure DeleteOnlyData;
    procedure SaveOnlyData(ItemName: Integer; ms: TStream);

    function GetObjectType(Objct: TObject): TzRecType;
    function ReadObjectToStream(Objct: TObject; var ms: TMemoryStream): Boolean;

    procedure HookWndProc(var Msg: TMessage);
    procedure DoSavePlacement(aForm: TCustomForm = nil);
    procedure LoadPlacement(aForm: TCustomForm = nil);
    procedure SetReadOnly(Value: Boolean);

  protected
    procedure Loaded; override;
    { Protected declarations }
    procedure WriteHeader;
    procedure ReadHeader;

    procedure WriteRefreshState;
    procedure ReadRefreshState;

    procedure _Next;
    procedure _Prior;
    procedure _First;
    procedure _Last;

    function _FSearch(ItemName: Integer): Integer;
    function _Search(ItemName: Integer; SearchDirection: TzSearchOpt = zsoFromTop): Integer;

    procedure _Delete;

    function _AddLast(ItemName: Integer; ms: TStream): Boolean;
    function _Insert(ItemName: Integer; ms: TStream): Boolean; overload;
    function _Replace(ms: TStream; ItemName: Integer): Boolean; overload;

    function _InsertBlankRecord(ItemName: Integer): Boolean;

    procedure CheckName(ItemName: string);
    procedure MoveBlock(FromPos, Delta, ASize: Integer);

    {$IFDEF INCLUDE_COMCTRL}
    procedure _SaveListView(LV: TListView; var ms: TMemoryStream);
    procedure _LoadListView(LV: TListView; ms: TMemoryStream);
    {$ENDIF}

  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    property BOF: Boolean read FBOF;
    property EOF: Boolean read FEOF;
    property RecordCount: Integer read GetRecordCount;
    property RecNo: Integer read GetRecNo write SetRecNo;

    procedure CreateTable;
    procedure Open;
    procedure Close;
    procedure First;
    procedure Last;
    procedure Next;
    procedure Prior;
    procedure Sort;
    function Size: Int64;
    function Created: TDateTime;
    function LastModified: TDateTime;
    function CRC(pStr: string): Integer;

    function  IsEmpty: Boolean;
    function  NotEmpty: Boolean;

    function Save(ItemName: string; ms: TStream): Boolean; overload;
    function Save(ItemName: Integer; ms: TStream): Boolean; overload;
    function Save(ms: TStream): Boolean; overload;
    function Save(ItemName: string; Text: string): Boolean; overload;
    function Save(ItemName: string; Value: Boolean): Boolean; overload;
    function Save(ItemName: string; Value: Integer): Boolean; overload;
//    function Save(ItemName: string; Value: Ordinal): Boolean; overload;
    function Save(ItemName: string; Value: DWord): Boolean; overload;
    function Save(ItemName: string; Value: Double): Boolean; overload;
    function Save(ItemName: string; Value: TFont): Boolean; overload;
    function Save(ItemName: string; Objct: TObject): Boolean; overload;
    function Save(ItemName: Integer; Objct: TObject): Boolean; overload;
    procedure Save(aForm: TCustomForm); overload;
    function SaveFromFile(ItemName, FileName: string): Boolean;
    function SaveToFile(ItemName, FileName: string): Boolean; overload;
    function SaveToFile(FileName: string): Boolean; overload;
    function SaveToStream(ItemName: string; OutStream: TStream): Boolean;

    function Load(ItemName: string; ms: TStream): Boolean; overload;
    function Load(ItemName: Integer; ms: TStream): Boolean; overload;
    function Load(ms: TStream): Boolean; overload;
    function Load(ItemName: string; Objct: TObject): Boolean; overload;
    function Load(ItemName: Integer; Objct: TObject): Boolean; overload;
    function Load(ItemName: string; var Text: string): Boolean; overload;
    function Load(ItemName: string; Value: TFont): Boolean; overload;
    procedure Load(aForm: TCustomForm); overload;
    function LoadToStream(ItemName: string = ''): TzStrStream;

    function Rename(OldName, NewName: string): Boolean;
    function CopyRecord(SourceName, TargetName: string): Boolean;

    procedure CheckActive;
    procedure CheckInActive;

    procedure Commit;
    procedure EmptyTable;

    function  GoToRec: Integer; overload;
    function  GoToRec(ItemName: Integer): Boolean; overload;
    function  GoToRec(ItemName: string): Boolean; overload;

    function  RecordExist(ItemName: Integer): Boolean; overload;
    function  RecordExist(ItemName: string): Boolean; overload;

    function  FSearch(ItemName: Integer): Integer; overload;
    function  FSearch(ItemName: string): Integer; overload;
    function  Search(ItemName: Integer; SearchDirection: TzSearchOpt = zsoFromTop): boolean; overload;
    function  Search(ItemName: string; SearchDirection: TzSearchOpt = zsoFromTop): boolean; overload;

    function AsString(ItemName: string; Default: string = ''): string;
    function AsBool(ItemName: string; Default: boolean = False): boolean;
    function AsInteger(ItemName: string; Default: Integer = 0): Integer; overload;
    function AsInt(ItemName: string; Default: Integer = 0): Integer; overload;
    function AsFloat(ItemName: string; Default: Double = 0): Double;
    function AsDate(ItemName: string; Default: Double = 0): TDateTime;
    function AsStream(ItemName: string; Value: TStream): Boolean; overload;
    function AsStream(ItemName: string): TStream; overload;
    function AsStrings(ItemName: string; var VarStrings: TStrings): Boolean; overload;
    function AsStrings(ItemName: string; var VarStrings: TStringList): Boolean; overload;

    function Delete(ItemName: Integer = 0): Boolean; overload;
    function Delete(ItemName: string): Boolean; overload;

    property Password: Integer read FRefreshStateHeader.Password;
    //property RecordType: TzRecType read FRecInfo.RecType;

  published
    { Published declarations }
    property ReadOnly: Boolean read FReadOnly write SetReadOnly Default False;
    property Active: Boolean read FIsOpen write SetActive Default False;
    property Sorted: Boolean read FSorted write FSorted Default True;
    property FileName: string read FTableName write SetTableName;
    property FormSave: TFormSaveFormat read FFormSave write FFormSave default fsfBoth;
    property FileNameAutoSet: TFormSavePlace read FFileNameAutoSet write FFileNameAutoSet default fspAppDataDir;

    property BeforeOpen: TNotifyEvent read FBeforeOpen write FBeforeOpen;
    property AfterOpen: TNotifyEvent read FAfterOpen write FAfterOpen;
    property BeforeClose: TNotifyEvent read FBeforeClose write FBeforeClose;
    property AfterClose: TNotifyEvent read FAfterClose write FAfterClose;
  end;

const
  COMPRESSED_FILE_EXT = '.zip;.lha;.lzh;.arj;.cab;.tar;.rar;.zoo;.ace;.bh;.jar';
  ERROR_SAVE_STREAM = ' - ERROR save to stream';
  INVALIDE_RECORD_NAME = 'Invalide record name: ''''';
  RECORD_NOT_EXISTS = 'Record by name ''%s'' not exists';
  CANT_WRITE_READONLY_MODE = 'Can''t write ReadOnly mode opened file:';

procedure Register;

implementation

function IntToFloat(Value: Integer): Double;
begin
  Result :=
    EncodeDate(
      LongRec(Value).Hi shr 9 + 1902,
      LongRec(Value).Hi shr 5 and 15,
      LongRec(Value).Hi and 31) +
    EncodeTime(
      LongRec(Value).Lo shr 11,
      LongRec(Value).Lo shr 5 and 63,
      LongRec(Value).Lo and 31 shl 1, 0);
end;

function TzIniData.CRC(pStr: string): Integer;
const
  Table: array[0..255] of DWORD =
   ($00000000, $77073096, $EE0E612C, $990951BA, $076DC419, $706AF48F, $E963A535, $9E6495A3,
    $0EDB8832, $79DCB8A4, $E0D5E91E, $97D2D988, $09B64C2B, $7EB17CBD, $E7B82D07, $90BF1D91,
    $1DB71064, $6AB020F2, $F3B97148, $84BE41DE, $1ADAD47D, $6DDDE4EB, $F4D4B551, $83D385C7,
    $136C9856, $646BA8C0, $FD62F97A, $8A65C9EC, $14015C4F, $63066CD9, $FA0F3D63, $8D080DF5,
    $3B6E20C8, $4C69105E, $D56041E4, $A2677172, $3C03E4D1, $4B04D447, $D20D85FD, $A50AB56B,
    $35B5A8FA, $42B2986C, $DBBBC9D6, $ACBCF940, $32D86CE3, $45DF5C75, $DCD60DCF, $ABD13D59,
    $26D930AC, $51DE003A, $C8D75180, $BFD06116, $21B4F4B5, $56B3C423, $CFBA9599, $B8BDA50F,
    $2802B89E, $5F058808, $C60CD9B2, $B10BE924, $2F6F7C87, $58684C11, $C1611DAB, $B6662D3D,
    $76DC4190, $01DB7106, $98D220BC, $EFD5102A, $71B18589, $06B6B51F, $9FBFE4A5, $E8B8D433,
    $7807C9A2, $0F00F934, $9609A88E, $E10E9818, $7F6A0DBB, $086D3D2D, $91646C97, $E6635C01,
    $6B6B51F4, $1C6C6162, $856530D8, $F262004E, $6C0695ED, $1B01A57B, $8208F4C1, $F50FC457,
    $65B0D9C6, $12B7E950, $8BBEB8EA, $FCB9887C, $62DD1DDF, $15DA2D49, $8CD37CF3, $FBD44C65,
    $4DB26158, $3AB551CE, $A3BC0074, $D4BB30E2, $4ADFA541, $3DD895D7, $A4D1C46D, $D3D6F4FB,
    $4369E96A, $346ED9FC, $AD678846, $DA60B8D0, $44042D73, $33031DE5, $AA0A4C5F, $DD0D7CC9,
    $5005713C, $270241AA, $BE0B1010, $C90C2086, $5768B525, $206F85B3, $B966D409, $CE61E49F,
    $5EDEF90E, $29D9C998, $B0D09822, $C7D7A8B4, $59B33D17, $2EB40D81, $B7BD5C3B, $C0BA6CAD,
    $EDB88320, $9ABFB3B6, $03B6E20C, $74B1D29A, $EAD54739, $9DD277AF, $04DB2615, $73DC1683,
    $E3630B12, $94643B84, $0D6D6A3E, $7A6A5AA8, $E40ECF0B, $9309FF9D, $0A00AE27, $7D079EB1,
    $F00F9344, $8708A3D2, $1E01F268, $6906C2FE, $F762575D, $806567CB, $196C3671, $6E6B06E7,
    $FED41B76, $89D32BE0, $10DA7A5A, $67DD4ACC, $F9B9DF6F, $8EBEEFF9, $17B7BE43, $60B08ED5,
    $D6D6A3E8, $A1D1937E, $38D8C2C4, $4FDFF252, $D1BB67F1, $A6BC5767, $3FB506DD, $48B2364B,
    $D80D2BDA, $AF0A1B4C, $36034AF6, $41047A60, $DF60EFC3, $A867DF55, $316E8EEF, $4669BE79,
    $CB61B38C, $BC66831A, $256FD2A0, $5268E236, $CC0C7795, $BB0B4703, $220216B9, $5505262F,
    $C5BA3BBE, $B2BD0B28, $2BB45A92, $5CB36A04, $C2D7FFA7, $B5D0CF31, $2CD99E8B, $5BDEAE1D,
    $9B64C2B0, $EC63F226, $756AA39C, $026D930A, $9C0906A9, $EB0E363F, $72076785, $05005713,
    $95BF4A82, $E2B87A14, $7BB12BAE, $0CB61B38, $92D28E9B, $E5D5BE0D, $7CDCEFB7, $0BDBDF21,
    $86D3D2D4, $F1D4E242, $68DDB3F8, $1FDA836E, $81BE16CD, $F6B9265B, $6FB077E1, $18B74777,
    $88085AE6, $FF0F6A70, $66063BCA, $11010B5C, $8F659EFF, $F862AE69, $616BFFD3, $166CCF45,
    $A00AE278, $D70DD2EE, $4E048354, $3903B3C2, $A7672661, $D06016F7, $4969474D, $3E6E77DB,
    $AED16A4A, $D9D65ADC, $40DF0B66, $37D83BF0, $A9BCAE53, $DEBB9EC5, $47B2CF7F, $30B5FFE9,
    $BDBDF21C, $CABAC28A, $53B39330, $24B4A3A6, $BAD03605, $CDD70693, $54DE5729, $23D967BF,
    $B3667A2E, $C4614AB8, $5D681B02, $2A6F2B94, $B40BBE37, $C30C8EA1, $5A05DF1B, $2D02EF8D);

var i: Word;
begin
  Result := $ffffffff;
  pStr := UpperCase(pStr);
  for i := 1 to Length(pStr) do
    Result := (Result shr 8) xor Table[Ord(pStr[i]) xor (Result and $000000FF)];
  Result := not Result;
end;

{$IFDEF INCLUDE_HTML}
procedure WB_Load(WB: TWebBrowser; SL: TStrings);
var ms: TMemoryStream;
begin
  WB.Stop;
  
  {while WB.ReadyState <> READYSTATE_COMPLETE do
  begin
    Application.ProcessMessages;
    Sleep(0);
  end;}

  if Assigned(WB.Document) then
  begin
    ms := TMemoryStream.Create;
    try
      SL.SaveToStream(ms);
      ms.Seek(0, 0);
      (WB.Document as IPersistStreamInit).Load(TStreamAdapter.Create(ms));
    finally
      WB.Tag := 0;                      // Flag for reload
      ms.Free;
    end;
  end
  else WB.Navigate('about:blank');
end;
{$ENDIF}

{ TzzStream }
constructor TzStrStream.Create;
begin
  inherited Create;
  Clear;
end;

destructor TzStrStream.Destroy;
begin
  inherited Destroy;
end;

procedure TzStrStream.SetPointer( Ptr: Pointer; Size: Integer );
begin
  inherited SetPointer( Ptr, Size );
end;
{End TzzStream }

constructor TzIniData.Create(AOwner: TComponent);
//var i: Integer;
begin
  inherited Create(AOwner);

  FRefreshStateHeaderSize := SizeOf(TRefreshStateHeader);
  FDataFileHeaderSize := SizeOf(TDataFileHeader);

  FRecInfoSize := SizeOf(TRecInfo);

  FTableName := '';
  FMsgNum := 0;
  FIsOpen := False;
  FEOF := True;
  FBOF := True;
  CheckMode := False;
  FzStrStream := TzStrStream.Create;
  FSorted := True;
  FFormSave := fsfBoth;
  FFileNameAutoSet := fspAppDataDir;

  {if Assigned(Application.MainForm) and ((Owner as TForm) <> Application.MainForm) then
    with Application.MainForm do
      for i := 0 to ComponentCount - 1 do
        if (Components[I] is TzIniData) then
           with (Components[I] as TzIniData) do begin
             FFormSave := FormSave;
             FTableName := FileName;
             FFileNameAutoSet := FileNameAutoSet;
             Break;
           end;}

  OldWndProc := nil;
  NewWndProc := nil;
end;

destructor TzIniData.Destroy;
begin
  FzStrStream.Free;
  if (Owner <> nil) and Assigned(OldWndProc) then
    SetWindowLong((Owner as TForm).Handle, GWL_WNDPROC, LongInt(OldWndProc));
  if Assigned(NewWndProc) then FreeObjectInstance(NewWndProc);
  if FIsOpen then Close; //InternalClose;
  inherited Destroy;
end;

procedure TzIniData.CreateTable;
//var i: Integer;
begin
  CheckInActive;
  FStream := TFileStream.Create (FTableName, fmCreate or fmShareDenyNone);
  FEOF := True;
  FBOF := True;

  try
    FillChar(FDataFileHeader, FDataFileHeaderSize, 0);
    with FDataFileHeader do begin
      GenHeader := 'TzID'+^Z;
      Created := Now;
    end;
    WriteHeader;

    FillChar(FRefreshStateHeader, FRefreshStateHeaderSize, 0);
    with FRefreshStateHeader do begin
      StartData := FDataFileHeaderSize + FRefreshStateHeaderSize;
      RecCount := 0;
      LastModified := Now;
      Msg_Num := 0;
      Password := 0;
    end;
    WriteRefreshState;

  finally
    FStream.Free;
  end;
end;

procedure TzIniData.Open;
//var ms: TMemoryStream;
begin
  if FIsOpen or (FTableName = '') then Exit;
  if Assigned(FBeforeOpen) then BeforeOpen(Self);

  if not FileExists (FTableName) then CreateTable;
//    raise EzzTableError.Create ('Open: Table file not found - ' + FTableName);

  if FReadOnly then
    FStream := TFileStream.Create(FTableName, fmOpenRead + fmShareDenyNone)
  else FStream := TFileStream.Create(FTableName, fmOpenReadWrite + fmShareDenyNone);

  if FStream.Size < FDataFileHeaderSize + FRefreshStateHeaderSize then begin
    FStream.Free;
    raise EzIniDataError.Create ('zIniData Open: File is corrapted - ' + FTableName);
  end;

  ReadHeader;
  if FDataFileHeader.GenHeader <> 'TzID'+^Z then begin
    FStream.Free;
    raise EzIniDataError.Create ('zIniData Open: Wrong file format - ' + FTableName);
  end;
  ReadRefreshState;
  FMsgNum :=  FRefreshStateHeader.Msg_Num;
  FIsOpen := True;

  FEOF := RecordCount = 0;
  FBOF := True;

  FRecNo := 0;
  _ReadRecord(0);
  if Assigned(FAfterOpen) then AfterOpen(Self);
end;

procedure TzIniData.Close;
var Handle: Integer;
begin
  if FIsOpen then begin
    if Assigned(FBeforeClose) then BeforeClose(Self);
    if not (csDestroying in ComponentState) then
      DoSavePlacement;
    FStream.Free;
    FIsOpen := False;
    try
      if not FReadOnly then begin
        Handle := FileOpen(FTableName, fmOpenReadWrite + fmShareDenyNone);
        FileSetDate(Handle, DateTimeToFileDate(FRefreshStateHeader.LastModified));
        FileClose(Handle);
      end;
    except
    end;
    if Assigned(FAfterClose) then AfterClose(Self);
  end;
end;

procedure TzIniData.CheckActive;
var i: Integer;
begin
  if CheckMode then Exit;

  if not FIsOpen then
    raise EzIniDataError.Create ('Table not in Open mode - ' + FTableName)
  else begin
    CheckMode := True;
    try
      ReadRefreshState;
      if FMsgNum =  FRefreshStateHeader.Msg_Num then Exit;

      i := FRecInfo.RecName;
      _ReadRecord(FRecNo);
      if FRecInfo.RecName = i then Exit;

      if Sorted then
        i := _FSearch(i)
      else i := _Search(i);

      if i > 0 then
        _ReadRecord(i - 1);

      FMsgNum :=  FRefreshStateHeader.Msg_Num;
    finally
      CheckMode := False;
    end;
  end;
end;

procedure TzIniData.CheckInActive;
begin
  if FIsOpen then
    raise EzIniDataError.Create ('Table in Open mode - ' + FTableName);
end;

procedure TzIniData.WriteHeader;
begin
  FStream.Position := 0;
  FStream.WriteBuffer(fDataFileHeader, fDataFileHeaderSize);
end;

procedure TzIniData.ReadHeader;
begin
  FStream.Position := 0;
  FStream.ReadBuffer(fDataFileHeader, fDataFileHeaderSize);
end;

procedure TzIniData.WriteRefreshState;
begin
  FRefreshStateHeader.LastModified := Now;
  FStream.Position := fDataFileHeaderSize;
  FStream.WriteBuffer(FRefreshStateHeader, FRefreshStateHeaderSize);
end;

procedure TzIniData.ReadRefreshState;
begin
  FStream.Position := fDataFileHeaderSize;
  FStream.ReadBuffer(FRefreshStateHeader, FRefreshStateHeaderSize);
end;

procedure TzIniData.SetActive(Value: Boolean);
begin
  if Value and not FIsOpen then Open
  else if not Value and FIsOpen then Close;
end;

procedure TzIniData.SetTableName(Value: string);
begin
  if (Value <> FTableName) or ((FTableName <> '') and not FIsOpen) or
           ((FTableName = '') and FIsOpen) then begin
    FTableName := Value;
    if FIsOpen then Close;
    if (Value <> '') and (not (csDesigning in ComponentState)) then begin
      CheckInActive;
      Open;
    end;
  end;
end;

function TzIniData.GetRecordCount: Integer;
begin
  CheckActive;
  Result := FRefreshStateHeader.RecCount;
end;

procedure TzIniData.SetRecNo(RecNum:Integer);
begin
  if (RecordCount > 0) and (RecNum > 0) and (RecNum <= RecordCount) and (RecNo <> RecNum) then begin
    FRecNo := RecNum - 1;
    _ReadRecord(FRecNo);
  end;
end;

function TzIniData.GetRecNo: Integer;
begin
  CheckActive;
  Result := FRecNo + 1;
end;

Procedure TzIniData._ReadRecord(RecNum:Integer);
begin
  if RecordCount > 0 then try
    FRecNo := RecNum;
    FStream.Position := FRefreshStateHeader.StartData + (FRecInfoSize * RecNum);
    FStream.ReadBuffer(FRecInfo, FRecInfoSize);
  except
  end;
end;

/////////////////////////     SAVE      /////////////////////////////

function TzIniData.Save(ItemName: string; ms: TStream): Boolean;
var i: Integer;
begin
  Result := False;
  if ms = nil then Exit;
  CheckActive;
  CheckName(ItemName);
  i := CRC(UpperCase(ItemName));
  if ms.Size = 0 then begin
    if GoToRec(i) then Delete;
    Exit;
  end;
  Result := Save(i, ms);
end;

function TzIniData.Save(ItemName: Integer; ms: TStream): Boolean;
begin
  if FReadOnly then
    raise EzIniDataError.Create (CANT_WRITE_READONLY_MODE + #13 + FTableName);

  if RecordExist(ItemName) then
    Result := _Replace(ms, ItemName)
  else if Sorted and (RecordCount > 0) then
    Result := _Insert(ItemName, ms)
  else Result := _AddLast(ItemName,  ms);
end;

function TzIniData.Save(ms: TStream): Boolean;
begin
  Result := _Replace(ms, GoToRec);
end;

function TzIniData.Save(ItemName: string; Objct: TObject): Boolean;
var ms: TMemoryStream;
begin
  Result := False;
  CheckActive;
  CheckName(ItemName);
  ms := TMemoryStream.Create;

  try
    if ReadObjectToStream(Objct, ms) then
      Result := Save(ItemName, ms)
    else
      raise EzIniDataError.Create (ItemName + ERROR_SAVE_STREAM);
  finally
    ms.Free;
  end;
end;

function TzIniData.Save(ItemName: Integer; Objct: TObject): Boolean;
var ms: TMemoryStream;
begin
  Result := False;
  CheckActive;
  ms := TMemoryStream.Create;

  try
    if ReadObjectToStream(Objct, ms) then
      Result := Save(ItemName, ms)
    else
      raise EzIniDataError.Create (ERROR_SAVE_STREAM);
  finally
    ms.Free;
  end;
end;

function TzIniData.Save(ItemName: string; Text: string): Boolean;
var ms: TMemoryStream;
begin
  CheckActive;
  CheckName(ItemName);
  ms := TMemoryStream.Create;

  try
    ms.WriteBuffer(Pointer(Text)^, Length(Text));
    Result := Save(ItemName, ms);
  finally
    ms.Free;
  end;
end;

function TzIniData.Save(ItemName: string; Value: Boolean): Boolean;
var AddMode: Boolean;
begin
  CheckActive;
  if FReadOnly then
    raise EzIniDataError.Create (CANT_WRITE_READONLY_MODE + #13 + FTableName);

  CheckName(ItemName);
  AddMode := _InsertBlankRecord(CRC(UpperCase(ItemName)));
  with FRecInfo do RecLength := Integer(Value);
  _WriteRecord(FRecNo);
  if AddMode then IncRecCount;
  Result := True;
  IncMsgNum;
end;

function TzIniData.Save(ItemName: string; Value: Integer): Boolean;
var AddMode: Boolean;
begin
  CheckActive;
  if FReadOnly then
    raise EzIniDataError.Create (CANT_WRITE_READONLY_MODE + #13 + FTableName);

  CheckName(ItemName);
  AddMode := _InsertBlankRecord(CRC(UpperCase(ItemName)));
  with FRecInfo do RecLength := Value;
  _WriteRecord(FRecNo);
  if AddMode then IncRecCount;
  Result := True;
  IncMsgNum;
end;

{function TzIniData.Save(ItemName: string; Value: Ordinal): Boolean;
begin

end;}

function TzIniData.Save(ItemName: string; Value: DWord): Boolean;
begin
  Result := Save(ItemName, Integer(Value));
end;

function TzIniData.Save(ItemName: string; Value: Double): Boolean;

  function FloatToInt: Integer;
  var
    Year, Month, Day, Hour, Min, Sec, MSec: Word;
  begin
    DecodeDate(Value, Year, Month, Day);
    DecodeTime(Value, Hour, Min, Sec, MSec);
    LongRec(Result).Lo := (Sec shr 1) or (Min shl 5) or (Hour shl 11);
    LongRec(Result).Hi := Day or (Month shl 5) or ((Year - 1980) shl 9);
  end;

begin
  Result := Save(ItemName, FloatToInt);
end;

function TzIniData.Save(ItemName: string; Value: TFont): Boolean;
begin
  Result := False;
  if not Assigned(Value) then Exit;
  Result := Save(ItemName,
       Value.Name + ',' +
       IntToStr(Value.CharSet) + ',' +
       IntToStr(Value.Color) + ',' +
       IntToStr(Value.Size) + ',' +
       IntToStr(Byte(Value.Style)));
end;

procedure TzIniData.Save(aForm: TCustomForm);
begin
  DoSavePlacement(aForm);
end;

function TzIniData.SaveFromFile(ItemName, FileName: string): Boolean;
var fs: TFileStream;
    s: string;
begin
  CheckActive;
  if FReadOnly then
    raise EzIniDataError.Create (CANT_WRITE_READONLY_MODE + #13 + FTableName);

  CheckName(ItemName);
  if not FileExists(FileName) then
    raise EzIniDataError.Create ('File not found:' + #13 + FileName);

  s := LowerCase(ExtractFileExt(FileName));
  FCanCompress := (s = '') or (Pos(s + ';', COMPRESSED_FILE_EXT) = 0);
  fs := TFileStream.Create(FileName, fmOpenRead + fmShareDenyNone);
  try
    Result := Save(ItemName, fs);
  finally
    fs.Free;
  end;
end;

function  TzIniData.SaveToFile(ItemName, FileName: string): Boolean;
var InStream: TFileStream;
begin
  Result := False;
  CheckActive;
  CheckName(ItemName);
  if (RecordCount = 0) or (FileName = '') or (not GoToRec(ItemName)) or
     (not RecordExist(ItemName)) or
     (FileExists(FileName) and (not DeleteFile(FileName))) then Exit;

//  if not GoToRec(ItemName) then FSearch(ItemName);

  try
    try
      InStream := TFileStream.Create (FileName, fmCreate);
      Result := Load(ItemName, InStream);
    finally
      InStream.Free;
    end;
  except
  end
end;

function TzIniData.SaveToFile(FileName: string): Boolean;
var InStream: TFileStream;
begin
  Result := False;
  CheckActive;
  if FReadOnly then
    raise EzIniDataError.Create (CANT_WRITE_READONLY_MODE + #13 + FTableName);

  if (RecordCount = 0) or (FileName = '') or
     (FileExists(FileName) and (not DeleteFile(FileName))) then Exit;

  try
    InStream := TFileStream.Create (FileName, fmCreate);
    try
      Result := Load(InStream);
    finally
      InStream.Free;
    end;
  except
  end
end;

function TzIniData.SaveToStream(ItemName: string; OutStream: TStream): Boolean;
begin
  Result := False;
  CheckActive;
  CheckName(ItemName);
  if (RecordCount = 0) or (not GoToRec(ItemName)) then Exit;
  try
    Result := Load(ItemName, OutStream);
  except
  end
end;

//////////////////////////////////////////////////////////////////// SAVE

function TzIniData._AddLast(ItemName: Integer; ms: TStream): Boolean;
var ps: TMemoryStream;
    m: Integer;
begin
  ps := TMemoryStream.Create;
  ms.Position := 0;
  try
    m := 1;    // not Compressed
    ms.Position := 0;
    if not (ms is TFileStream) or FCanCompress then begin
      LHACompress(ms, ps);
      if ps.Size >= ms.Size then begin
        ps.Clear;
        ps.CopyFrom(ms, 0);
      end
      else m := -1;             // compressed
    end
    else ps.CopyFrom(ms, 0);

    // Move Records info
    MoveBlock(FRefreshStateHeader.StartData, ps.Size,
              RecordCount * FRecInfoSize);
    // Write new Item Contents
    FStream.Position := FRefreshStateHeader.StartData;
    with FRecInfo do begin
      RecName := ItemName;
      StartPos := FStream.Position;
      RecLength := m * ps.Size;
    end;
    FStream.CopyFrom(ps, 0);

    // Set of begining of Records info
    Inc(FRefreshStateHeader.StartData, ps.Size);
    FRecNo := FRefreshStateHeader.RecCount;

    // Write Record info
    _WriteRecord(FRecNo);
    Result := True;
    IncRecCount;
  finally
    ps.Free;
    IncMsgNum;
  end;
end;

function TzIniData._Insert(ItemName: Integer; ms: TStream): Boolean;
var ps: TMemoryStream;
    i, m: Integer;
begin
  CheckActive;

  if Sorted then begin
    if RecordCount = 0 then i := 0
    else i := _FSearch(ItemName);
    if i = 0 then begin
      Result := _AddLast(ItemName, ms);
      Exit;
    end
    else if i > 0 then begin
      _Next;
      if FEof then begin
        Result := _AddLast(ItemName, ms);
        Exit;
      end;
    end
    else _ReadRecord(Abs(i) - 1);
  end;

  ps := TMemoryStream.Create;
  ms.Position := 0;

  try
    m := 1;    // not Compressed
    ms.Position := 0;
    if not (ms is TFileStream) or FCanCompress then begin
      LHACompress(ms, ps);
      if ps.Size >= ms.Size then begin
        ps.Clear;
        ps.CopyFrom(ms, 0);
      end
      else m := -1;             // compressed
    end
    else ps.CopyFrom(ms, 0);

    // Move Record data
    MoveBlock(FRefreshStateHeader.StartData, ps.Size,
              FRefreshStateHeader.RecCount * FRecInfoSize);
    // Write new Item Info
    FStream.Position := FRefreshStateHeader.StartData;
    with FRecInfo do begin
      RecName := ItemName;
      StartPos := FStream.Position;
      RecLength := m * ps.Size;
    end;
    FStream.CopyFrom(ps, 0);
    Inc(FRefreshStateHeader.StartData, ps.Size);
    ps.Clear;
    WriteRefreshState;

    // Move Records info
    MoveBlock(FRefreshStateHeader.StartData + FRecNo * FRecInfoSize, FRecInfoSize,
             (FRefreshStateHeader.RecCount - FRecNo) * FRecInfoSize);

    _WriteRecord(FRecNo);
    Result := True;
    IncRecCount;
  finally
    ps.Free;
    IncMsgNum;
  end;
end;

function TzIniData._Replace(ms: TStream; ItemName: Integer): Boolean;
var i: Integer;
begin
  Result := False;
  CheckActive;
  if (RecordCount = 0) then Exit;
  if (ItemName <> 0) and (ItemName <> GoToRec) then begin
    i := _FSearch(ItemName);
    if i < 1 then
      raise EzIniDataError.Create (Format('ERROR replace: Record by name %u not Exists ', [ItemName]));
    _ReadRecord(i - 1);
  end;

  SaveOnlyData(ItemName, ms);
  Result := True;
  IncMsgNum;
end;

function TzIniData.Rename(OldName, NewName: string): Boolean;
var i, _start, _Length, NewItem: Integer;
begin
  Result := False;
  CheckActive;
  if FReadOnly then
    raise EzIniDataError.Create (CANT_WRITE_READONLY_MODE + #13 + FTableName);

  CheckName(OldName);
  CheckName(NewName);
  if not RecordExist(OldName) then Exit;
    //raise EzIniDataError.Create (Format(RECORD_NOT_EXISTS, [OldName]));
  if RecordExist(NewName) then
    raise EzIniDataError.Create (Format('Record by name ''%s'' already exists', [NewName]));
  NewItem := CRC(UpperCase(NewName));

  if not GoToRec(OldName) then Exit;
  _start := FRecInfo.StartPos;
  _Length := FRecInfo.RecLength;
  i := RecNo;

  if not Sorted or (i = Abs(_FSearch(NewItem))) then begin
    FRecInfo.RecName := NewItem;
    _WriteRecord(FRecNo);
  end
  else begin
    // Delete Record info
    MoveBlock(FRefreshStateHeader.StartData + i * FRecInfoSize, -FRecInfoSize,
           (FRefreshStateHeader.RecCount - i) * FRecInfoSize);
    FStream.Size := FRefreshStateHeader.StartData + FRefreshStateHeader.RecCount * FRecInfoSize;
    i := _FSearch(NewItem);
    with FRecInfo do begin
      RecName := NewItem;
      StartPos := _start;
      RecLength := _Length;
    end;
    if i = 0 then begin
      FRecNo := FRefreshStateHeader.RecCount - 1;
      _WriteRecord(FRecNo);
    end
    else if i < 0 then begin
      FRecNo := Abs(i) - 1;
      // Move Records info
      MoveBlock(FRefreshStateHeader.StartData + FRecNo * FRecInfoSize, FRecInfoSize,
           (FRefreshStateHeader.RecCount - FRecNo) * FRecInfoSize);
      _WriteRecord(FRecNo);
    end;
  end;
  Result := True;
  IncMsgNum;
end;

function TzIniData.CopyRecord(SourceName, TargetName: string): Boolean;
begin
  Result := False;
  CheckActive;
  CheckName(SourceName);
  CheckName(TargetName);
  if not RecordExist(SourceName) then
    raise EzIniDataError.Create (Format(RECORD_NOT_EXISTS, [SourceName]));
  if Load(SourceName, FzStrStream) then begin
    Save(TargetName, FzStrStream);
    FzStrStream.Size := 0;
    Result := True;
  end;
end;

procedure TzIniData.MoveBlock(FromPos, Delta, ASize: Integer);
var mss: TMemoryStream;
    s, t: Integer;
begin
  CheckActive;
  if FReadOnly then
    raise EzIniDataError.Create (CANT_WRITE_READONLY_MODE + #13 + FTableName);

  mss := TMemoryStream.Create;
  if (Delta = 0) or (ASize = 0) then Exit;
  try
    t := 0;
    while t < ASize do begin
      if ASize - t > 500000 then s := 500000
      else s := ASize - t;
      FStream.Position := FromPos + t;
      mss.CopyFrom(FStream, s);
      FStream.Position := FromPos + Delta + t;
      FStream.CopyFrom(mss, 0);
      mss.Clear;
      Inc(t, s);
    end;
  finally
    mss.Free;
  end;
end;

Procedure TzIniData._WriteRecord(RecNum:Integer);
begin
  if FReadOnly then
    raise EzIniDataError.Create (CANT_WRITE_READONLY_MODE + #13 + FTableName);

  try
    if (RecNum >= 0) or (RecNum < RecordCount) then begin
      FStream.Position := FRefreshStateHeader.StartData + (FRecInfoSize * RecNum);
      FStream.WriteBuffer(FRecInfo, FRecInfoSize);
    end;
  except
  end;
end;

procedure TzIniData._WriteRecord(Buffer:PChar; IntRecNum:Integer);
begin
  if FReadOnly then
    raise EzIniDataError.Create (CANT_WRITE_READONLY_MODE + #13 + FTableName);

  try
    if (IntRecNum >= 0) or (IntRecNum < RecordCount) then begin
      FStream.Position := FRefreshStateHeader.StartData + (FRecInfoSize * IntRecNum);
      FStream.WriteBuffer(Buffer^, FRecInfoSize);
    end;
  except
  end;
end;

procedure TzIniData.IncMsgNum;
begin
  ReadRefreshState;
  if FReadOnly then Exit;
  if FRefreshStateHeader.Msg_Num + 10 >= MAXINT then
    FRefreshStateHeader.Msg_Num := 0
  else Inc(FRefreshStateHeader.Msg_Num);
  FMsgNum :=  FRefreshStateHeader.Msg_Num;
  WriteRefreshState;
  Commit;
end;

procedure TzIniData.Commit;
begin
  CheckActive;
  if FReadOnly then Exit;
  FlushFileBuffers(FStream.Handle);
end;

procedure TzIniData.EmptyTable;
begin
  CheckActive;
  if FReadOnly then
    raise EzIniDataError.Create (CANT_WRITE_READONLY_MODE + #13 + FTableName);

  FRefreshStateHeader.StartData := FDataFileHeaderSize + FRefreshStateHeaderSize;
  FStream.Size := FRefreshStateHeader.StartData;
  FRefreshStateHeader.RecCount := 0;
  WriteRefreshState;
  FEOF := True;
  FBOF := True;
  FRecNo := 0;
  IncMsgNum;
end;

procedure TzIniData.First;
begin
  CheckActive;
  if RecordCount > 0 then begin
    if FRecNo > 0 then _First;
  end
  else _First;
end;

procedure TzIniData._First;
begin
  CheckActive;
//  FStream.position := FRefreshStateHeader.StartData;
  if RecordCount > 0 then begin
    if FRecNo > 0 then begin
      _ReadRecord(0);
      FEOF := False;
      FBOF := True;
    end;
  end
  else begin
    FillChar(FRecInfo, FRecInfoSize, 0);
    FEOF := True;
    FBOF := True;
  end;
end;

procedure TzIniData.Last;
begin
  CheckActive;
  if RecordCount > 0 then begin
    if FRecNo < RecordCount - 1 then _Last;
  end
  else _Last;
end;

procedure TzIniData._Last;
begin
  CheckActive;
  if RecordCount > 0 then begin
    if FRecNo < RecordCount - 1 then begin
      _ReadRecord(RecordCount - 1);
      FEOF := True;
      FBOF := False;
    end;
  end
  else begin
    FRecNo := 0;
    FillChar(FRecInfo, FRecInfoSize, 0);
    FEOF := True;
    FBOF := True;
  end;
end;

procedure TzIniData.Next;
begin
  CheckActive;
  if RecordCount - 1 > FRecNo then _Next
  else FEOF := True;
end;

procedure TzIniData._Next;
begin
  CheckActive;
  if RecordCount - 1 > FRecNo then begin
    _ReadRecord(FRecNo + 1);
    FEOF := False;
    FBOF := False;
  end
  else FEOF := True;
end;

procedure TzIniData.Prior;
begin
  CheckActive;
  if FRecNo > 0 then _Prior
  else FBOF := True;
end;

procedure TzIniData._Prior;
begin
  CheckActive;
  if FRecNo > 0 then begin
    _ReadRecord(FRecNo - 1);
    FEOF := False;
    FBOF := False;
  end
  else FBOF := True;
end;

////////////////////     LOAD      ////////////////////////////////////

function TzIniData.Load(ItemName: string; ms: TStream): Boolean;
var i: Integer;
//    ps: TMemoryStream;
//    s: string;
begin
  CheckActive;
  CheckName(ItemName);
  Result := False;
  i := CRC(UpperCase(ItemName));
  if GoToRec <> i then begin
    i := _FSearch(i);
    if i < 1 then Exit;
    _ReadRecord(i - 1);
  end;
  Result := Load(ms);
end;

function TzIniData.Load(ItemName: Integer; ms: TStream): Boolean;
var i: Integer;
//    ps: TMemoryStream;
//    s: string;
begin
  CheckActive;
  Result := False;
  if GoToRec <> ItemName then begin
    i := _FSearch(ItemName);
    if i < 1 then Exit;
    _ReadRecord(i - 1);
  end;
  Result := Load(ms);
end;

function TzIniData.Load(ms: TStream): Boolean;
var i: Integer;
    ps: TMemoryStream;
    s: string;
begin
  CheckActive;
  Result := False;
  if RecordCount = 0 then Exit;

  ps := TMemoryStream.Create;
  try
    if (GetRecLength > 0) and (FRecInfo.StartPos > 0) and
       (FRecInfo.StartPos + GetRecLength < FStream.Size) then begin
      try
        FStream.Position := FRecInfo.StartPos;
        ps.CopyFrom(FStream, GetRecLength);
        if FRecInfo.RecLength < 0 then     // if minus - compressed
          LHAExpand(ps, ms)
        else if (Created < EncodeDate(2003, 05, 1)) and (ps.Size > 5) then begin
          ps.Position := 0;
          SetString(s, nil, 5);
          ps.Read(Pointer(s)^, 5);
          i := Ord(s[5]);
          if (i < 32) and not (i in [10, 13]) then
            LHAExpand(ps, ms)
          else ms.CopyFrom(ps, 0);
        end
        else ms.CopyFrom(ps, 0);
      except
      end;
      ms.Position := 0;
      Result := True;
    end;
  finally
    ps.Free;
  end;
end;

function TzIniData.Load(ItemName: string; Objct: TObject): Boolean;
var i: Integer;
    ms: TMemoryStream;
    SL: TStrings;
    NewRecType: TzRecType;
begin
  Result := False;
  CheckActive;
  CheckName(ItemName);

  NewRecType := GetObjectType(Objct);
  case NewRecType of
    zrtStrings, zrtStringList: (Objct as TStrings).Clear;
    {$IFDEF INCLUDE_COMCTRL}
    zrtListView: (Objct as TListView).Items.Clear;
    {$ENDIF}
  end;

  if ItemName <> '' then begin
    i := CRC(UpperCase(ItemName));
    if GoToRec <> i then begin
      i := _FSearch(i);
      if i < 1 then Exit;
      _ReadRecord(i - 1);
    end;
  end;

  ms := TMemoryStream.Create;

  try
    Load(ms);
    case NewRecType of
      zrtStrings, zrtStringList: (Objct as TStrings).LoadFromStream(ms);
      {$IFDEF INCLUDE_COMCTRL}
      zrtTreeView: (Objct as TCustomTreeView).LoadFromStream(ms);
      zrtListView: _LoadListView((Objct as TListView), ms);
      {$ENDIF}
      {$IFDEF INCLUDE_GRAPHIC}
      zrtBitmap: (Objct as TBitmap).LoadFromStream(ms);
      zrtIcon: (Objct as TIcon).LoadFromStream(ms);
      zrtJPEGImage: (Objct as TJPEGImage).LoadFromStream(ms);
      {$ENDIF}
      {$IFDEF INCLUDE_HTML}
      zrtHtml: begin
                 SL := TStringList.Create;
                 try
                   SL.LoadFromStream(ms);
                   WB_Load((Objct as TWebBrowser), SL);
                 finally
                   SL.Free;
                 end;
               end;
      {$ENDIF}
      else Exit;
    end;
    Result := True;
  finally
    ms.Free;
  end;
end;

function TzIniData.Load(ItemName: Integer; Objct: TObject): Boolean;
begin
  case GetObjectType(Objct) of
    zrtStrings, zrtStringList: (Objct as TStrings).Clear;
    {$IFDEF INCLUDE_COMCTRL}
    zrtListView: (Objct as TListView).Items.Clear;
    {$ENDIF}
  end;
  if GoToRec(ItemName) then Result := Load('', Objct);
end;

function TzIniData.Load(ItemName: string; var Text: string): Boolean;
var ms: TMemoryStream;
    i: Integer;
begin
  Result := False;
  CheckActive;
  CheckName(ItemName);
  i := CRC(UpperCase(ItemName));
  if GoToRec <> i then begin
    i := _FSearch(i);
    if i < 1 then Exit;
    _ReadRecord(i - 1);
  end;

  ms := TMemoryStream.Create;

  try
    Result := Load(ItemName, ms);
    SetLength(Text, ms.Size);
    ms.ReadBuffer(Pointer(Text)^, ms.Size);
  finally
    ms.Free;
  end;
end;

function TzIniData.AsString(ItemName: string; Default: string = ''): string;
var ms: TMemoryStream;
    i: Integer;
begin
  Result := Default;
  CheckActive;
  CheckName(ItemName);
  i := CRC(UpperCase(ItemName));
  if GoToRec <> i then begin
    i := _FSearch(i);
    if i < 1 then Exit;
    _ReadRecord(i - 1);
  end;

  if FRecInfo.StartPos > 0 then begin
    ms := TMemoryStream.Create;
    try
      if not Load(ItemName, ms) then Exit;
      ms.Position := 0;
      SetString(Result, nil, ms.Size);
      ms.Read(Pointer(Result)^, ms.Size);
    finally
      ms.Free;
    end;
  end
  else Result := IntToStr(FRecInfo.RecLength);
end;

function TzIniData.GoToRec: Integer;
begin
  CheckActive;
  if RecordCount > 0 then Result := FRecInfo.RecName
  else Result := 1;
end;

function TzIniData.GoToRec(ItemName: Integer): Boolean;
var NewRec: Integer;
begin
  if FSorted then
    NewRec := _FSearch(ItemName)
  else NewRec := _Search(ItemName);
  Result := NewRec > 0;
  if Result then RecNo := NewRec;
end;

function TzIniData.GoToRec(ItemName: string): Boolean;
begin
  Result := GoToRec(CRC(UpperCase(ItemName)));
end;

function TzIniData.FSearch(ItemName: Integer): Integer;
begin
  Result := _FSearch(ItemName);
  if Abs(Result) <> RecNo then begin
    if Result <> 0 then
      RecNo := Abs(Result)
    else Last;
  end;
end;

function TzIniData.FSearch(ItemName: string): Integer;
begin
  Result := FSearch(CRC(UpperCase(ItemName)));
end;

// 0 - Append
// 1..RecordCount - Founded
// -1..-RecordCount - not Founded and RecNo to Insert
function TzIniData._FSearch(ItemName: Integer): Integer;

  function __FindFirst(Rec_Num, RecStart: Integer): Integer;
  begin
    while Rec_Num >= RecStart do begin
      _ReadRecord(Rec_Num);
      if FRecInfo.RecName = ItemName then Dec(Rec_Num)
      else break;
    end;
    if Rec_Num < RecStart then Result := RecStart + 1
    else Result := Rec_Num + 2;
  end;

var Rec_Num, RecStart, RecEnd, OldRec: Integer;
begin
  Result := 0;
  OldRec := FRecNo;

  try
    RecEnd := RecordCount - 1;
    RecStart := 0;

    if FRecInfo.RecName = ItemName then begin
      Result := __FindFirst(FRecNo - 1, RecStart);
      exit;
    end;

    _ReadRecord(RecStart);
    if FRecInfo.RecName > ItemName then begin
      Result := -1;
      exit;
    end
    else if FRecInfo.RecName = ItemName then begin
      Result := 1;
      exit;
    end;

    _ReadRecord(RecEnd);
    if FRecInfo.RecName < ItemName then begin
      Result := 0;
      exit;
    end
    else if FRecInfo.RecName = ItemName then begin
      Result := __FindFirst(RecEnd, RecStart);
      exit;
    end;

    while RecStart < RecEnd do begin
      if RecStart + 10 >= RecEnd then begin
        if RecStart = 0 then Rec_Num := 1
        else Rec_Num := RecStart;
        while Rec_Num <= RecEnd do begin
          _ReadRecord(Rec_Num);
          if FRecInfo.RecName < ItemName then Inc(Rec_Num)
          else break;
        end;
        if FRecInfo.RecName = ItemName then
          Result := Rec_Num + 1
        else Result := -(Rec_Num + 1);
        exit;
      end;

      Rec_Num := (RecEnd + RecStart) div 2;
      _ReadRecord(Rec_Num);
      if FRecInfo.RecName = ItemName then begin
        Result := __FindFirst(Rec_Num - 1, RecStart);
        exit;
      end;
      if FRecInfo.RecName > ItemName then RecEnd := Rec_Num
      else RecStart := Rec_Num;
    end;
  finally
    _ReadRecord(OldRec);
  end;
end;

function TzIniData.Delete(ItemName: Integer = 0): Boolean;
begin
  Result := False;
  CheckActive;
  if FReadOnly then
    raise EzIniDataError.Create (CANT_WRITE_READONLY_MODE + #13 + FTableName);

  if (RecordCount = 0) or ((ItemName <> 0) and not GoToRec(ItemName)) then Exit;
  _Delete;
  Result := True;
//  if RecordCount = 0 then
end;

function TzIniData.Delete(ItemName: string): Boolean;
begin
  Result := Delete(CRC(UpperCase(ItemName)));
end;

procedure TzIniData._Delete;
var i, RecLen, StartPos: Integer;
begin
  CheckActive;
  if FReadOnly then
    raise EzIniDataError.Create (CANT_WRITE_READONLY_MODE + #13 + FTableName);

  RecLen := GetRecLength;
  StartPos := FRecInfo.StartPos;
  i := RecNo;
  // Delete Record Contents
  if StartPos > 0 then begin
    MoveBlock(FRecInfo.StartPos + RecLen, -RecLen,
              FStream.Size - (FRecInfo.StartPos + RecLen));

    Dec(FRefreshStateHeader.StartData, RecLen);
  end;
  Dec(FRefreshStateHeader.RecCount);
  WriteRefreshState;

  // Delete Record info
  MoveBlock(FRefreshStateHeader.StartData + i * FRecInfoSize, -FRecInfoSize,
           (FRefreshStateHeader.RecCount - FRecNo) * FRecInfoSize);
  FStream.Size := FRefreshStateHeader.StartData + FRefreshStateHeader.RecCount * FRecInfoSize;

  if FRefreshStateHeader.RecCount = 0 then begin
    FRecNo := 0;
    FRecInfo.RecName := 0;
    FRecInfo.StartPos := 0;
    FRecInfo.RecLength := 0;
  end
  else if FRecNo > FRefreshStateHeader.RecCount - 1 then
    FRecNo := FRefreshStateHeader.RecCount - 1;

  // Update start positions
  FStream.Position := FRefreshStateHeader.StartData;
  if StartPos > 0 then
    for i := 0 to Pred(FRefreshStateHeader.RecCount) do begin
      FStream.ReadBuffer(FRecInfo, FRecInfoSize);
      if FRecInfo.StartPos > StartPos then begin
        Dec(FRecInfo.StartPos, RecLen);
        FStream.Position := FStream.Position - FRecInfoSize;
        FStream.WriteBuffer(FRecInfo, FRecInfoSize);
      end;
    end;

  // Set current record
  _ReadRecord(FRecNo);
  IncMsgNum;
end;

procedure TzIniData.Sort;

  procedure QSort;
  { uNElem - number of elements to sort }

    procedure qSortHelp(pivotP: Integer; nElem: word);
    label
      TailRecursion,
      qBreak;
    var
      leftP, rightP, pivotEnd, pivotTemp, leftTemp: word;
      lNum: Integer;
      retval: integer;
    begin
      TailRecursion:
        if (nElem <= 2) then
          begin
            if (nElem = 2) then

              begin
                rightP := pivotP +1;
                if (_CompareRecords(pivotP, rightP) > 0) then
                  _SwapRecords(pivotP, rightP);
              end;
            exit;
          end;
        rightP := (nElem -1) + pivotP;
        leftP :=  (nElem shr 1) + pivotP;
        { sort pivot, left, and right elements for "median of 3" }
        if (_CompareRecords(leftP, rightP) > 0) then _SwapRecords(leftP, rightP);
        if (_CompareRecords(leftP, pivotP) > 0) then _SwapRecords(leftP, pivotP)

        else if (_CompareRecords(pivotP, rightP) > 0) then _SwapRecords(pivotP, rightP);
        if (nElem = 3) then
          begin
            _SwapRecords(pivotP, leftP);
            exit;
          end;
        { now for the classic Horae algorithm }
        pivotEnd := pivotP + 1;
        leftP := pivotEnd;
        repeat
          retval := _CompareRecords(leftP, pivotP);
          while (retval <= 0) do
            begin
              if (retval = 0) then
                begin
                  _SwapRecords(LeftP, PivotEnd);
                  Inc(PivotEnd);
                end;
              if (leftP < rightP) then
                Inc(leftP)
              else
                goto qBreak;
              retval := _CompareRecords(leftP, pivotP);
            end; {while}
          while (leftP < rightP) do
            begin

              retval := _CompareRecords(pivotP, rightP);
              if (retval < 0) then
                Dec(rightP)
              else
                begin
                  _SwapRecords(leftP, rightP);
                  if (retval <> 0) then
                    begin
                      Inc(leftP);
                      Dec(rightP);
                    end;
                  break;
                end;
            end; {while}

        until (leftP >= rightP);
      qBreak:
        if (_CompareRecords(leftP, pivotP) <= 0) then Inc(leftP);
        leftTemp := leftP -1;
        pivotTemp := pivotP;
        while ((pivotTemp < pivotEnd) and (leftTemp >= pivotEnd)) do
          begin
            _SwapRecords(pivotTemp, leftTemp);
            Inc(pivotTemp);
            Dec(leftTemp);
          end; {while}
        lNum := (leftP - pivotEnd);
        nElem := ((nElem + pivotP) -leftP);

        if (nElem < lNum) then
          begin
            qSortHelp(leftP, nElem);
            nElem := lNum;
          end
        else
          begin
            qSortHelp(pivotP, lNum);
            pivotP := leftP;
          end;
        goto TailRecursion;
      end; {qSortHelp }

  begin
    if (RecordCount < 2) then  exit; { nothing to sort }
    qSortHelp(1, RecordCount);
  end; { QSort }


begin
  CheckActive;
  if FReadOnly then
    raise EzIniDataError.Create (CANT_WRITE_READONLY_MODE + #13 + FTableName);

  try
    QSort;
    First;
  finally
    IncMsgNum;
  end;
end;

// Compare two records.  Returns -1 if REC1 < REC2, 0 if REC1 = REC2, or
// 1 if REC1 > REC2.
Function TzIniData._CompareRecords(Rec1,Rec2:Integer):Integer; FAR;
{-Compare the records Rec1, Rec2 and return -1 if Rec1 < Rec2, 0 if Rec1 = Rec2,
  1 if Rec1 > Rec2 }
//VAR
//  IX : Integer;

  Function CompareHelper(Rec1,Rec2:Integer):Integer;
  VAR
    IKey1, IKey2 : Integer;
  BEGIN
    //k for IntTable
          _ReadRecord(Rec1-1);
          IKey1 := GoToRec;
          _ReadRecord(Rec2-1);
          IKey2 := GoToRec;
          if IKey1 < IKey2 then
            Result := -1
          else
            if IKey1 > IKey2 then
              Result := 1
            else
              Result := 0;
  END;

BEGIN

  Result := CompareHelper(Rec1,Rec2);
  {IX := 0;
  REPEAT // Loop through all available sortfields until not equal or no more sort fiels.
    Result := CompareHelper(SortFields[IX],Rec1,Rec2);
    Inc(IX);
  UNTIL (Result <> 0) or (IX > High(SortFields));}
END;

Procedure TzIniData._SwapRecords(Rec1,REc2:Integer);
VAR
  Buffer1, Buffer2 : PChar;
//  Bookmark1, BOokmark2 : TBookmarkFlag;
BEGIN
  Rec1 := Rec1 - 1;
  Rec2 := Rec2 - 1;
  if Rec1 < 0 then Exit;
  if Rec2 < 0 then Exit;
  Buffer1 := AllocRecordBuffer;
  Buffer2 := AllocRecordBuffer;
  _ReadRecord(Buffer1,Rec1);
  _ReadRecord(Buffer2,Rec2);
//  Bookmark1 := GetBookmarkFlag(Buffer1);
//  Bookmark2 := GetBookmarkFlag(Buffer2);
//  SetBookmarkFlag(Buffer1,Bookmark2);
//  SetBookmarkFlag(Buffer2,Bookmark1);
  _WriteRecord(Buffer1,Rec2);
  _WriteRecord(Buffer2,Rec1);
  FreeMem(Buffer1);
  FreeMem(Buffer2);
END;

function TzIniData.AllocRecordBuffer: PChar;
begin
  GetMem(Result, FRecInfoSize);
     if Result<>nil then FillChar(Result^, FRecInfoSize, 0);
end;

procedure TzIniData._ReadRecord(Buffer:PChar; IntRecNum:Integer);
begin
  if RecordCount > 0 then try
    FRecNo := IntRecNum;
    FStream.Position := FRefreshStateHeader.StartData + (FRecInfoSize * IntRecNum);
    FStream.ReadBuffer(Buffer^, FRecInfoSize);
  except
  end;
end;

function TzIniData.Search(ItemName: Integer; SearchDirection: TzSearchOpt = zsoFromTop): boolean;
var i: Integer;
begin
  i := _Search(ItemName, SearchDirection);
  Result := i > 0;
  if Result then begin
    FRecNo := i - 1;
    _ReadRecord(FRecNo);
  end;
end;

function TzIniData.Search(ItemName: string; SearchDirection: TzSearchOpt = zsoFromTop): boolean;
begin
  Result := Search(CRC(UpperCase(ItemName)), SearchDirection);
end;

function TzIniData._Search(ItemName: Integer; SearchDirection: TzSearchOpt = zsoFromTop): Integer;

  function CheckFind(RNum: Integer): Boolean;
  begin
    Result := False;
    _ReadRecord(RNum);
    // Check key record equal to record.
    if (GoToRec = ItemName) then begin
      Result := True;
      FRecNo := RNum;
    end;
  end;

var
   i, k, f, OldRecNo: integer;
begin
  // Loop for all records.
  Result := 0;
  if (RecordCount = 0) then Exit;
  OldRecNo := FRecNo;

  case SearchDirection of
    zsoFromTop:
       begin // from top
         f := 0;
         k := RecordCount - 1;
         for i := f to k do if CheckFind(i) then begin
           Result := i + 1;
           break;
         end;
       end;
    zsoForward:
       begin  // from current record to end
         f := FRecNo + 1;
         k := RecordCount - 1;
         for i := f to k do if CheckFind(i) then begin
           Result := i + 1;
           break;
         end;
       end;
    zsoBack:
       begin  // from current record to top
         f := FRecNo - 1;
         k := 0;
         for i := f downto k do if CheckFind(i) then begin
           Result := i + 1;
           break;
         end;
       end;
    else Exit;
  end;

  FRecNo := OldRecNo;
  _ReadRecord(FRecNo);
end;

procedure TzIniData.DeleteOnlyData;
var i, RecLen, StartPos, OldRec: Integer;
begin
  CheckActive;
  if (RecordCount = 0) or (FRecInfo.StartPos = 0) or (FRecInfo.RecLength = 0) then Exit;
  RecLen := GetRecLength;
  StartPos := FRecInfo.StartPos;
  OldRec := FRecNo;

  // Delete Record Contents
  MoveBlock(FRecInfo.StartPos + GetRecLength, -RecLen,
            FStream.Size - (FRecInfo.StartPos + RecLen));

  FStream.Size := FStream.Size - RecLen;

  Dec(FRefreshStateHeader.StartData, RecLen);
  WriteRefreshState;

  // Update start positions
  FStream.Position := FRefreshStateHeader.StartData;
  for i := 0 to Pred(RecordCount) do begin
    _ReadRecord(i);
    if (FRecInfo.StartPos > StartPos) then begin
      Dec(FRecInfo.StartPos, RecLen);
      FStream.Position := FStream.Position - FRecInfoSize;
      FStream.WriteBuffer(FRecInfo, FRecInfoSize);
    end;
  end;

  _ReadRecord(OldRec);
  with FRecInfo do begin
    StartPos := 0;
    RecLength := 0;
  end;
  _WriteRecord(FRecNo);
  IncMsgNum;
end;

procedure TzIniData.SaveOnlyData(ItemName: Integer; ms: TStream);
var ps: TMemoryStream;
    i, m: Integer;
begin
  if GoToRec <> ItemName then begin
    i := _FSearch(ItemName);
    if i < 1 then
      raise EzIniDataError.Create(Format('Record by name %u not exists', [ItemName]));
    _ReadRecord(i - 1);
  end;

  ps := TMemoryStream.Create;
  ms.Position := 0;
  try
    m := 1;    // not Compressed
    ms.Position := 0;
    if not (ms is TFileStream) or FCanCompress then begin
      LHACompress(ms, ps);
      if ps.Size >= ms.Size then begin
        ps.Clear;
        ps.CopyFrom(ms, 0);
      end
      else m := -1;             // compressed
    end
    else ps.CopyFrom(ms, 0);

    if ps.Size <> GetRecLength then begin
      DeleteOnlyData;
      // Move Records info
      MoveBlock(FRefreshStateHeader.StartData, ps.Size,
                RecordCount * FRecInfoSize);
      // Write new Item Contents
      FStream.Position := FRefreshStateHeader.StartData;
      with FRecInfo do begin
        RecName := ItemName;
        StartPos := FStream.Position;
        RecLength := m * ps.Size;
      end;

      FStream.CopyFrom(ps, 0);

      // Set of begining of Records info
      Inc(FRefreshStateHeader.StartData, ps.Size);
      WriteRefreshState;

      // Write Record info
      _WriteRecord(FRecNo);
    end
    else begin
      FStream.Position := FRecInfo.StartPos;
      FStream.CopyFrom(ps, 0);
    end;
  finally
    ps.Free;
    IncMsgNum;
  end;
end;

function TzIniData.ReadObjectToStream(Objct: TObject; var ms: TMemoryStream): Boolean;

  {$IFDEF INCLUDE_HTML}
  function WB_GetHTMLCode(WebBrowser: TWebBrowser; var ms: TMemoryStream): boolean;
  var
    ps: IPersistStreamInit;
    sa: IStream;
    s: string;
  begin
    ps := WebBrowser.Document as IPersistStreamInit;
    s := '';
    try
      sa := TStreamAdapter.Create(ms, soReference) as IStream;
      Result := Succeeded(ps.Save(sa, True));
    finally
    end;
  end;
  {$ENDIF}

begin
  Result := False;
  try
    case GetObjectType(Objct) of
      zrtStringList: (Objct as TStringList).SaveToStream(ms);
      zrtStrings: (Objct as TStrings).SaveToStream(ms);
      {$IFDEF INCLUDE_COMCTRL}
      zrtTreeView: (Objct as TCustomTreeView).SaveToStream(ms);
      zrtListView: if (Objct as TListView).Items.Count > 0 then
                     _SaveListView((Objct as TListView), ms);
      {$ENDIF}
      {$IFDEF INCLUDE_GRAPHIC}
      zrtBitmap: (Objct as TBitmap).SaveToStream(ms);
      zrtIcon: (Objct as TIcon).SaveToStream(ms);
      zrtJPEGImage: (Objct as TJPEGImage).SaveToStream(ms);
      {$ENDIF}
      {$IFDEF INCLUDE_HTML}
      zrtHtml: WB_GetHTMLCode(Objct as TWebBrowser, ms);
      {$ENDIF}
    end;
    Result := True;
  except
  end;
end;

function TzIniData.GetObjectType(Objct: TObject): TzRecType;
begin
  Result := zrtUnnoun;
  if (Objct is TStringList) then               // sumetime need
    Result := zrtStringList
  else if (Objct is TStrings) then
    Result := zrtStrings
  {$IFDEF INCLUDE_COMCTRL}
  else if (Objct is TCustomTreeView) then
    Result := zrtTreeView
  else if (Objct is TListView) then
    Result := zrtListView
  {$ENDIF}
  {$IFDEF INCLUDE_GRAPHIC}
  else if (Objct is TBitmap) then
    Result := zrtBitmap
  else if (Objct is TIcon) then
    Result := zrtIcon
  else if (Objct is TJPEGImage) then
    Result := zrtJPEGImage
  {$ENDIF}
  {$IFDEF INCLUDE_HTML}
  else if (Objct is TWebBrowser) then
    Result := zrtHtml
  {$ENDIF}
end;

function TzIniData.RecordExist(ItemName: Integer): Boolean;
begin
  if RecordCount = 0 then
    Result := False
  else if FSorted then
    Result := _FSearch(ItemName) > 0
  else Result := _Search(ItemName) > 0;
end;

function TzIniData.RecordExist(ItemName: string): Boolean;
begin
  Result := RecordExist(CRC(UpperCase(ItemName)));
end;

procedure TzIniData.CheckName(ItemName: string);
begin
  if (ItemName = '') and (RecordCount = 0) then
    raise EzIniDataError.Create(INVALIDE_RECORD_NAME);
end;

// True - New Record Added
// False - Record Exist
function TzIniData._InsertBlankRecord(ItemName: Integer): Boolean;
var i: Integer;
begin
  Result := False;
  CheckActive;

  if (ItemName = 0) then begin
    if (RecordCount = 0) then
      raise EzIniDataError.Create(INVALIDE_RECORD_NAME);
    ItemName := GoToRec;  // if Replace current record
  end;

  if (ItemName <> GoToRec) then begin
    if not RecordExist(ItemName) then begin
      Result := True;
      if Sorted then begin
        i := _FSearch(ItemName);
        if i = 0 then
          FRecNo := RecordCount
        else begin
          FRecNo := Abs(i) - 1;
          MoveBlock(FRefreshStateHeader.StartData + FRecNo * FRecInfoSize, FRecInfoSize,
                   (FRefreshStateHeader.RecCount - FRecNo) * FRecInfoSize);
        end
      end
      else FRecNo := RecordCount;
      with FRecInfo do begin
        RecName := ItemName;
        StartPos := 0;
        RecLength := 0;
      end;
    end
    else begin
      if Sorted then FSearch(ItemName)
      else Search(ItemName, zsoFromTop);
      if (FRecInfo.StartPos > 0) and (GetRecLength > 0) then begin
        DeleteOnlyData;
        with FRecInfo do begin
          StartPos := 0;
          RecLength := 0;
        end;
      end;
    end;
  end;
//  with FRecInfo do begin
//  end;
  _WriteRecord(FRecNo);
end;

function TzIniData.AsBool(ItemName: string; Default: boolean = False): boolean;
begin
  Result := Default;
  CheckActive;
  CheckName(ItemName);
  if _FoundItem(CRC(UpperCase(ItemName))) then
    Result := Boolean(FRecInfo.RecLength);
end;

function TzIniData.AsInteger(ItemName: string; Default: Integer = 0): Integer;
begin
  Result := Default;
  CheckActive;
  CheckName(ItemName);
  if _FoundItem(CRC(UpperCase(ItemName))) then
    Result := FRecInfo.RecLength;
end;

function TzIniData.AsInt(ItemName: string; Default: Integer = 0): Integer;
begin
  Result := AsInteger(ItemName, Default);
end;

function TzIniData.AsFloat(ItemName: string; Default: Double = 0): Double;
begin
  Result := Default;
  CheckActive;
  CheckName(ItemName);
  if _FoundItem(CRC(UpperCase(ItemName))) then
    Result := IntToFloat(AsInteger(ItemName));
end;

function TzIniData.AsDate(ItemName: string; Default: Double = 0): TDateTime;
begin
  Result := TDateTime(AsFloat(ItemName, Default));
end;

function TzIniData.AsStream(ItemName: string; Value: TStream): Boolean;
begin
  Result := Load(ItemName, Value);
end;

function TzIniData.AsStream(ItemName: string): TStream;
begin
  FzStrStream.Clear;
  Load(ItemName, FzStrStream);
  Result := FzStrStream;
end;

function TzIniData.AsStrings(ItemName: string; var VarStrings: TStrings): Boolean;
begin
  VarStrings.Clear;
  VarStrings.LoadFromStream(AsStream(ItemName));
  Result := True;
  FzStrStream.Clear;
end;

function TzIniData.AsStrings(ItemName: string; var VarStrings: TStringList): Boolean;
begin
  VarStrings.Clear;
  VarStrings.LoadFromStream(AsStream(ItemName));
  Result := True;
  FzStrStream.Clear;
end;

{$IFDEF INCLUDE_COMCTRL}
procedure TzIniData._SaveListView(LV: TListView; var ms: TMemoryStream);
var  idxItem,idxSub{,IdxImage} : integer;
     pText     : pChar;
     sText     : String;
     W,ItemCount,SubCount : word;
     MySignature : Array [0..2] of char;
begin
  CheckActive;
  if FReadOnly then
    raise EzIniDataError.Create (CANT_WRITE_READONLY_MODE + #13 + FTableName);

  //Initialization
  ItemCount:=0;
  SubCount:=0;

  MySignature := 'LVS'; //  ListViewStream
  ms.Clear();
  try
    ms.Write(MySignature, sizeof(MySignature));

    with LV do begin
      if Items.Count = 0 then ItemCount := 0         // List is empty
      else ItemCount := Items.Count;
      ms.Write(ItemCount, Sizeof(ItemCount));
      for idxItem := 0 to Pred(ItemCount) do
        with items[idxItem] do begin
          //Save subitems count
          SubCount := SubItems.Count;
          ms.Write(SubCount, Sizeof(SubCount));

          //Save ImageIndex                          // I don't need
          //IdxImage := ImageIndex;                  // You can uncomment
          //ms.Write(IdxImage, Sizeof(IdxImage));

          //Save Caption
          sText := Caption;
          w := length(sText);
          pText := StrAlloc( Length( sText ) + 1);
          StrPLCopy(pText,sText,Length( sText ) );
          ms.Write(w, sizeof(w));
          ms.Write(pText^,w);
          StrDispose(pText);

          for idxSub:=0 to SubItems.Count - 1 do begin //Save Item's subitems
            sText := SubItems[idxSub];
            w := length(sText);
            pText := StrAlloc( Length( sText ) + 1);
            StrPLCopy(pText,sText,Length( sText ) );
            ms.Write(w, sizeof(w));
            ms.Write(pText^,w);
            StrDispose(pText);
          end;
        end;
      end;
  finally
  end;
end;

procedure TzIniData._LoadListView(LV: TListView; ms: TMemoryStream);
var
   IdxItem,IdxSubItem{, IdxImage} : Integer;
   W,ItemCount,SubCount:Word;
   pText:pchar;
   PTemp:pChar;
   MySignature :  Array [0..2] of Char;
begin
  CheckActive;

  with LV do begin

    Items.BeginUpdate;
    Items.Clear;

    try

      ItemCount:=0;
      SubCount:=0;

      ms.Read(MySignature, sizeof(MySignature));
      if MySignature <> 'LVS' then Exit;
      ms.Read(ItemCount, sizeof(ItemCount));

      For idxItem := 0 to Pred(ItemCount) do
        with Items.Add do begin
          // Read SubCount
          ms.Read( SubCount,sizeof( SubCount ) );

          // Read imageindex                                // I don't need
          //ms.Read( IdxImage,sizeof( IdxImage ) );         // You can uncomment
          //ImageIndex := IdxImage;

          // Read the Caption
          ms.Read(w,SizeOf(w));
          pText:=StrAlloc(w + 1);
          pTemp:=StrAlloc(w + 1);
          ms.Read(pTemp^,W);
          StrLCopy(pText,pTemp,W);
          Caption := StrPas(pText);
          StrDispose(pTemp);
          StrDispose(pText);

          for idxSubItem := 0 to Pred(SubCount) do begin
            ms.Read(w, SizeOf(w));
            pText:=StrAlloc(w + 1);
            pTemp:=StrAlloc(w + 1);
            ms.Read(pTemp^, W);
            StrLCopy(pText,pTemp, W);
            Items[idxItem].SubItems.Add( StrPas(pText) );
            StrDispose(pTemp);
            StrDispose(pText);
          end;
        end;

    finally
      Items.EndUpdate;
    end;
  end;
end;
{$ENDIF}

function TzIniData.IsEmpty: Boolean;
begin
  CheckActive;
  Result := FRefreshStateHeader.RecCount = 0;
end;

function TzIniData.NotEmpty: Boolean;
begin
  CheckActive;
  Result := FRefreshStateHeader.RecCount > 0;
end;

function TzIniData.Load(ItemName: string; Value: TFont): Boolean;
var s, Data: string;
    i: Integer;
begin
  Result := False;
  s := AsString(ItemName, '');
  if s = '' then Exit;
  try
    i := Pos(',', s);
    if i > 0 then
    begin
      {Name}
      Data := Trim(Copy(s, 1, i-1));
      if Data <> '' then
        Value.Name := Data;
      System.Delete(s, 1, i);
      i := Pos(',', s);
      if i > 0 then
      begin
        {CharSet}
        Data := Trim(Copy(s, 1, i-1));
        if Data <> '' then
          Value.Charset := TFontCharSet(StrToIntDef(Data, Value.Charset));
        System.Delete(s, 1, i);
        i := Pos(',', s);
        if i > 0 then
        begin
          {Color}
          Data := Trim(Copy(s, 1, i-1));
          if Data <> '' then
            Value.Color := TColor(StrToIntDef(Data, Value.Color));
          System.Delete(s, 1, i);
          i := Pos(',', s);
          if i > 0 then
          begin
           {Size}
           Data := Trim(Copy(s, 1, i-1));
           if Data <> '' then
             Value.Size := StrToIntDef(Data, Value.Size);
           System.Delete(s, 1, i);
           {Style}
           Data := Trim(s);
           if Data <> '' then
             Value.Style := TFontStyles(Byte(StrToIntDef(Data, Byte(Value.Style))));
          end
        end
      end;
      Result := True;
    end;
  except
    Result := False;
  end;
end;

procedure TzIniData.Load(aForm: TCustomForm);
begin
  LoadPlacement(aForm);
end;

function TzIniData.LoadToStream(ItemName: string = ''): TzStrStream;
begin
  CheckActive;
  FzStrStream.Clear;
  Result := nil;
  if (RecordCount = 0) then Exit;
  if ItemName = '' then begin
    if Load(FzStrStream) then
      Result := FzStrStream;
  end
  else if Load(ItemName, FzStrStream) then
    Result := FzStrStream;
end;

procedure TzIniData.IncRecCount;
begin
  if FReadOnly then Exit;

  Inc(FRefreshStateHeader.RecCount);
  WriteRefreshState;
  IncMsgNum;
end;

function TzIniData._FoundItem(ItemName: Integer): Boolean;
begin
  CheckActive;
  if ItemName = 0 then Result := RecordCount > 0
  else if (ItemName <> GoToRec) then
    Result := GoToRec(ItemName)
  else Result := True;
end;

function TzIniData.Size: Int64;
begin
  Result := 0;
  CheckActive;
  Result := FStream.Size;
end;

function TzIniData.Created: TDateTime;
begin
  Result := 0;
  CheckActive;
  Result := FDataFileHeader.Created;
end;

function TzIniData.LastModified: TDateTime;
begin
  Result := 0;
  CheckActive;
  Result := FRefreshStateHeader.LastModified;
end;

procedure TzIniData.HookWndProc(var Msg: TMessage);
begin
  with Msg do begin
    Result := CallWindowProc(OldWndProc, (Owner as TForm).Handle, Msg,
      wParam, lParam);
    case Msg of
    WM_DESTROY: if not (csDesigning in ComponentState) then
                  DoSavePlacement;
    end;
  end;
end;

procedure TzIniData.DoSavePlacement(aForm: TCustomForm = nil);
var s, n: string;
    SL: TStrings;
begin
  if not FIsOpen then Exit;
  if not Assigned(aForm) then begin
    aForm := (Owner as TForm);
    if (aForm.WindowState <> wsNormal) or
       (FFormSave = fsfNone) then Exit;
  end;

  if FReadOnly then
    raise EzIniDataError.Create (CANT_WRITE_READONLY_MODE + #13 + FTableName);

  SL := TStringList.Create;
  try
    with aForm do begin
      if Left < 0 then
        n := '0'
      else n := IntToStr(Left);

      SL.Add(n);
      if Top < 0 then
        n := '0'
      else n := IntToStr(Top);
      SL.Add(n);
      SL.Add(IntToStr(Width));
      SL.Add(IntToStr(Height));
      Save(ClassName, SL);
    end;
  finally
    SL.Free;
  end;
end;

procedure TzIniData.LoadPlacement(aForm: TCustomForm = nil);
var s: string;
    Left, Top, Width, Height: Integer;
    SL: TStrings;
begin
  if not Assigned(aForm) then begin
    aForm := (Owner as TForm);
    if not FIsOpen or (FFormSave = fsfNone) or not FileExists(FTableName) then Exit;
  end
  else if not FIsOpen or not FileExists(FTableName) then Exit;

  SL := TStringList.Create;
  try
    Load(aForm.ClassName, SL);
    if SL.Count < 4 then Exit;

    if (FFormSave <> fsfSize) or (aForm <> (Owner as TForm)) then begin
      Left := StrToInt(SL[0]);
      Top := StrToInt(SL[1]);
    end
    else begin
      Left := aForm.Left;
      Top := aForm.Top;
    end;

    if (FFormSave <> fsfPosition) or (aForm <> (Owner as TForm)) then begin
      Width := StrToInt(SL[2]);
      Height := StrToInt(SL[3]);
    end
    else begin
      Width := aForm.Width;
      Height := aForm.Height;
    end;
    aForm.SetBounds(Left, Top, Width, Height);
  finally
    SL.Free;
  end;
end;

procedure TzIniData.Loaded;
  function GetAppDataDir: string;
   var
    ppidl : PItemIDList;
    szPath: array[0..MAX_PATH - 1] of Char;
  begin
     SHGetSpecialFolderLocation(Application.Handle, CSIDL_APPDATA	, ppidl);
     SHGetPathFromIDList(ppidl, szPath);
     Result := String(szPath);
  end;
var szStr : Array [0..255] of Char;
    s: string;
begin
  if (csDesigning in ComponentState) then Exit;
  if FTableName = '' then
    s := ChangeFileExt(ExtractFileName(Application.ExeName), '.ini')
  else s := FTableName;
  if (FTableName = '') and (FFileNameAutoSet <> fspNone) then begin
    {if Assigned(Application.MainForm) and ((Owner as TForm) <> Application.MainForm) then
      with Application.MainForm do
        for i := 0 to ComponentCount - 1 do
          if (Components[I] is TzIniData) then
             with (Components[I] as TzIniData) do begin
               FFormSave := FormSave;
               FTableName := FileName;
               FFileNameAutoSet := FileNameAutoSet;
               Break;
             end;
    else}
    if FFileNameAutoSet = fspAppDataDir then
      FTableName := GetAppDataDir + '\' + s
    else if FFileNameAutoSet = fspAppDir then
      FTableName := ExtractFilePath(Application.ExeName) + s
    else if (GetWindowsDirectory(@szStr,255) <> 0) then
      FTableName := StrPas(szStr) + '\' + s
    else Exit;
    SetTableName(FTableName);
    OldWndProc := TFarProc(GetWindowLong((Owner as TForm).Handle, GWL_WNDPROC));
    NewWndProc := MakeObjectInstance(HookWndProc);
    SetWindowLong((Owner as TForm).Handle, GWL_WNDPROC, LongInt(NewWndProc));
    LoadPlacement;
  end;
end;

function TzIniData.GetRecLength: Integer;
begin
  Result := Abs(FRecInfo.RecLength);
end;

procedure TzIniData.SetReadOnly(Value: Boolean);
begin
  if FIsOpen then
    raise EzIniDataError.Create ('Can''t change ReadOnly property for opened file:' +
                                 #13 + FTableName);
  if FReadOnly <> Value then FReadOnly := Value;
end;

procedure Register;
begin
  RegisterComponents('zPage', [TzIniData]);
end;

end.
