{*******************************************************}
{                                                       }
{         Copyright (c) FINANCES Ltd                    }
{         2003   ver. 2                                 }
{                                                       }
{         www.finances-ltd.narod.ru                     }
{         finances-ltd@narod.ru                         }
{                                                       }
{*******************************************************}

// zTable is compact but powerful, reliable and easy-to-use BDE alternative database engine


unit zTable;

interface

uses SysUtils, Classes, Db, DbCommon, extctrls, TypInfo, Dialogs, Windows, Forms, Variants, Controls;

{$ifdef VER100}      // Delphi 3
{$define LEVEL3}
{$define DELPHI}
{$endif}

{$ifdef VER110}      // C++ Builder 3
{$define LEVEL3}
{$define BCB}
{$endif}

{$ifdef VER120}      // Delphi 4
{$define LEVEL4}
{$define DELPHI}
{$endif}

{$ifdef VER125}      // C++ Builder 4
{$define LEVEL4}
{$define BCB}
{$endif}

//***********************************************************************

const MAX_ACCESS = 10;

type
    EzTableError = class(Exception);

    TRecInfo=record
        RecordNo: integer;
        BookmarkFlag: TBookmarkFlag;
    end;
    PRecInfo=^TRecInfo;

type TTabState = (tsWait, tsBusy, tsLock);

type
  TRefreshStateHeader = record
    ModifiedNum: SmallInt;
    UserConfirm: SmallInt;
    State: TTabState;
    Msg_Num: Integer;
    Computer_Count: Integer;
    Stored_Int: Integer;
  end;

type
  TNetRecord = record
    LockRecNum: Integer;
    LockValue: array [0..4] of Char;
    FieldIndex: Byte;
    Busy: Boolean;
    ComputerName: array [0..24] of Char;
    UserName: array [0..24] of Char;
    TableUserName: array [0..24] of Char;
    EntryDate: TDateTime;
  end;

type
  TDataFileHeader = record
    GenHeader : Array[1..4] of Char;
    RecordSize : Integer;
    NumberFields : Integer;
    Compressed: Boolean;
    LastModified: TDateTime;
    Created: TDateTime;
    Modified: Boolean;
    StartData: Integer;
    Password: string[10];
    TableKind: Word;
    Options: string[10];
  end;

type
  PBDescribField = ^TBDescribField;
  TBDescribField = packed Record
    Name : Array[0..27] of Char;
    DataType : TFieldType;
    DataSize : Integer;
    Size : Integer;
    Precision : Integer;
    Alignment: TAlignment;
    DisplayWidth: Integer;
    DisplayLabel: Array[0..50] of Char;
    ReadOnly: Boolean;
    Tag: Integer;
    Visible: Boolean;
    EditMask: Array[0..20] of Char;
    Required: Boolean;
    DefaultExpression: Array[0..20] of Char;
    DisplayFormat: Array[0..20] of Char;
    EditFormat: Array[0..20] of Char;
    MaxValue: Double;
    MinValue: Double;
    DisplayValues: Array[0..20] of Char;
    Transliterate: Boolean;
    Currency: Boolean;
  end;

  TzTableSaveFlag = (mtfSaveData, mtfSaveCalculated,mtfSaveLookup,mtfSaveNonVisible,mtfSaveBlobs,mtfSaveDef,mtfSkipRest);
  TzTableSaveFlags = set of TzTableSaveFlag;

  TzTableCompareOption = (mtcoDescending,mtcoCaseInsensitive,mtcoPartialKey);
  TzTableCompareOptions = set of TzTableCompareOption;

  {TkbmBlob = class
  private
     FLength:cardinal;
     FBuffer:PChar;
  public
     constructor Create;
     destructor Destroy; override;

     procedure LoadFromStream(stream:TMemoryStream);
     procedure SaveToStream(stream:TMemoryStream);
     property Length:cardinal read FLength write FLength;
     property Buffer:Pchar read FBuffer write FBuffer;
  end;
  PkbmBlob=^TkbmBlob;

  TkbmOnLoadRecord = procedure(DataSet:TDataSet; var Accept:boolean) of object;
  TkbmOnLoadField = procedure(DataSet:TDataSet; FieldNo:integer; Field:TField) of object;
  TkbmOnCompress = procedure(UnCompressedStream,CompressedStream:TStream) of object;
  TkbmOnDeCompress = procedure(CompressedStream,DeCompressedStream:TStream) of object;}
  TFOnStructChangedEvent = procedure (Sender: TDataSet; OldFieldDefs: TFieldDefs) of object;

  TzTable = class(TDataSet)
  private
    FStream: THandleStream; // the physical table
    FTableName: string; // table path and file name

    FDataFileHeaderSize : Integer;
    FDataFileHeader : TDataFileHeader;

    FRefreshStateHeaderSize: Integer;
    FNetRecordSize: Integer;
    FRefreshStateHeader: TRefreshStateHeader;
    FNetRecord: TNetRecord;
    FTableUserName: string;

    FExclusive: Boolean;
    FSortByField: string;
    FTemporal: Boolean;
    FTmpFile: string;
    FFiltered: Boolean;
    FFilterText: string;
    Ar0: array of Byte;
    Ar1: array of Byte;
    Ar2: array of String[2];
    Ar3: array of Variant;
    FFilterCount: Integer;
    FModified: Boolean;
    FOnStoredIntChange: TDataSetNotifyEvent;
    FOnUserCountChange: TDataSetNotifyEvent;
    FBeforeCreate: TDataSetNotifyEvent;
    FRecordCount: Integer;
    FBeforeRefreshValue: Variant;
    FEnableRecMove: Boolean;
    FEnableSort: Boolean;
    FAscend: Word;
    FOnStructChanged: TFOnStructChangedEvent;
    FSelfRefresh: Boolean;
    FTableKind: Word;
    FPassword, FOptions: string;
    FStoredInt: Integer;
    FAfterMove: TDataSetNotifyEvent;

        FIsOpen:                                Boolean;
        FRecNo:                                 integer;
        FFilterBuffer:                          PChar;
        FAutoIncField:                          TField;
        FAutoIncMax,
        FBufferSize,
        FStartRecInfo,
        FStartCalculated:integer;
        FRecordSize:                            integer;
        FFieldOfs:                              array [0..255] of integer;
        FReadOnly:                              boolean;
//        FBlobs:                                 boolean;
        FRecalcOnIndex:                         boolean;
        FKeyOptions:                            TzTableCompareOptions;

//        FOnLoadRecord:                          TkbmOnLoadRecord;
//        FOnLoadField:                           TkbmOnLoadField;

//        FOnCompressSave:                        TkbmOnCompress;
//        FOnDecompressLoad:                      TkbmOnDecompress;

//        FOnCompressBlobStream:                  TkbmOnCompress;
//        FOnDecompressBlobStream:                TkbmOnDecompress;

    procedure SetStoredInt(Value: Integer);
    procedure SetTableKind(Value: Word);
    procedure SetPassword(Value: string);
    procedure SetOptions(Value: string);
    procedure SetSelfRefresh(Value: Boolean);
    function GetUserCount: Byte;
    function GetModifiedNum: SmallInt;
    function GetState: TTabState;
    function GetComputerCount: Integer;
    procedure OnChangeTimer(Sender: TObject);
    procedure SetRefreshDelay(Value: Integer);
    function GetRefreshDelay: Integer;
    procedure _AppendRecord(Buffer: PChar);
    procedure _ReadRecord(Buffer: PChar; IntRecNum:Integer);
    procedure _WriteRecord(Buffer:PChar; IntRecNum:Integer);
    function GetTotalCountCount: Integer;
    procedure SetExclusive(Value: Boolean);
    procedure SetTemporal(Value:Boolean);

    function GetActiveRecordBuffer:         PChar;

    function FilterRecord(Buffer: PChar):   Boolean;
    procedure _InternalFirst;
    procedure _InternalLast;
    function  _InternalNext:boolean;
    function  _InternalPrior:boolean;

    function  IsFiltered:boolean;
    procedure SetModified(Value: Boolean);
    procedure SetCompressed(Value: Boolean);
    function GetCompressed: Boolean;
    function GetModified: Boolean;
    function GetCreated: TDateTime;
    function GetLastModified: TDateTime;
    procedure IncreaseUserConfirm;
    Function _CompareRecords(SortFields:Array of String;Rec1,Rec2:Integer):Integer;
    procedure _SwapRecords(Rec1,REc2:Integer);

    procedure ReadNetRecord(RecNum: Integer = -1);
    procedure WriteNetRecord(RecNum: Integer = -1);
    function FNetWorkHeaderSize: Integer;
    procedure InsertSpace(SpaceLength: Integer);
    procedure SetTableUserName(Value: string);
    function GetMyComputerName: string;
    function GetMyLoginName: string;

  protected
    BefRecNo: Integer;
    FChangeTimer: TTimer;
    ModifiedNum, UserConfirm: SmallInt;
    MyNum: SmallInt;
    MyLock: TTabState;

    procedure InternalOpen; override;
    procedure InternalClose; override;
    procedure InternalFirst;override;
    procedure InternalLast;override;

    procedure InternalDelete; override;
    procedure InternalInitRecord(Buffer: PChar); override;
    procedure InternalPost; override;
    procedure InternalAddRecord(Buffer: Pointer; Append: Boolean); override;

    procedure InternalInitFieldDefs; override;
    procedure InternalSetToRecord(Buffer: PChar); override;

    procedure InternalRefresh; override;

    procedure DoBeforeClose; override;
    procedure DoAfterOpen; override;
    procedure DoOnNewRecord; override;

    function IsCursorOpen: Boolean; override;
    function GetCanModify: Boolean; override;
    function GetRecordSize: Word;override;
    function GetRecordCount: integer;override;

    function AllocRecordBuffer: PChar; override;
    procedure FreeRecordBuffer(var Buffer: PChar); override;
    procedure PopulateField(Buffer:PChar;Field:TField;AValue:Variant);
    procedure PopulateRecord(Buffer:PChar;Fields:string;Values:variant);

    function GetFieldSize(FieldType:TFieldType; Size:longint):longint;
    function GetFieldPointer(Buffer:PChar; Field:TField):PChar;

    procedure SetFieldData(Field: TField; Buffer: Pointer);override;

    function GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult; override;
    function FindRecord(Restart, GoForward: Boolean): Boolean; override;

    function GetRecNo: integer;override;
    procedure SetRecNo(Value: integer);override;

    function GetBookmarkFlag(Buffer: PChar): TBookmarkFlag; override;
    procedure SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); override;
    procedure GetBookmarkData(Buffer: PChar; Data: Pointer); override;
    procedure SetBookmarkData(Buffer: PChar; Data: Pointer); override;
    procedure InternalGotoBookmark(Bookmark: Pointer); override;

    function BCDToCurr(BCD: Pointer; var Curr: Currency): Boolean; //override;
    function CurrToBCD(const Curr: Currency; BCD: Pointer; Precision, Decimals: Integer): Boolean; //override;

    procedure InternalHandleException; override;

//        function CopyRecords(Source,Destination:TDataSet; Count:longint):longint;
    procedure SaveFieldsProperties;

    function LocateRecord(const KeyFields: string; const KeyValues: Variant;
                              Options: TLocateOptions; SyncCursor: Boolean): Boolean;

    procedure OpenCursor(InfoQuery: Boolean); override;
    procedure CloseCursor; override;
    procedure CreateTmpTable;
    function  GetTempDirectory: string;
    procedure SetFilterText(const Value: string); override;
    function  FilterByFilter: Boolean;
    procedure SetFiltered(Value: Boolean); override;
    procedure DoBeforeOpen; override;
    procedure DoBeforePost; override;
    procedure DoAfterPost; override;
    procedure DoBeforeRefresh; override;
    procedure DoAfterRefresh; override;
    procedure DoBeforeEdit; override;
    procedure DoBeforeDelete; override;
    procedure DoBeforeInsert; override;
    procedure DoAfterDelete; override;
    procedure DoChanges;
    function SetTabState(Value: TTabState): Boolean;

  public
    MsgNum: Integer;
    ErrorNum: Integer;

    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure CreateTable;
    procedure EmptyTable;
    function IsSequenced:Boolean; override;
    function MoveRecord(Source, Destination: Integer): Boolean;
    function ChangeRecords(Source, Destination: Integer): Boolean;
    function MoveCurRecord(Destination:Longint):Boolean;
    function Lookup(const KeyFields: string; const KeyValues: Variant; const ResultFields: string): Variant; override;
    function Locate(const KeyFields: string; const KeyValues: Variant; Options: TLocateOptions): Boolean; override;
    function GetFieldData(Field: TField; Buffer: Pointer): Boolean; override;
    procedure Refresh;

    property TotalCount: Integer read GetTotalCountCount;
    procedure WriteRefreshState;
    procedure WriteHeader;
    procedure ReadHeader;
    procedure CheckLock;
    function LSearch(FindValue : Variant): Integer;
    function FSearch(FindValue : Variant): Integer;
    function Search(AField: TField; SearchValue: Variant;
                         FromBegining: Boolean = False): boolean; overload;
    function Search(Farr: array of TField; SearchValue: array of Variant;
                         FromBegining: Boolean): boolean; overload;
    function Search(Farr: array of TField; SearchValue: array of Variant): boolean; overload;
    function SearchNum(AField: TField; SearchValue: Variant;
                         SearchDirection: SmallInt): boolean;
    // 0 - From Bigening
    // 1 - Up
    // -1 - Down
    function SearchStr(AField: TField; SearchValue: string; SearchDirection: SmallInt;
                       WholeWord, MatchCase, ExactSearch: Boolean): boolean;
    function InsEdit(AValue: Variant): Integer;
    function CopyRecord(Destination: Integer): Boolean;
    function DeleteRecord(RecN: Integer): Boolean;
    procedure DeleteRecords(AField: TField; AValue: Variant);
    function  NotEmpty: Boolean;
    function  IsEmpty: Boolean;
    procedure Sort(AField: TField);

    function Lock: Boolean;
    function UnLock: Boolean;

    function LockByValue(FldIndex: Integer; Value: string): Boolean;
    procedure UnLockByValue;
    procedure ClearLocks;
    procedure SortTable(SortFields : Array of String; Increment: Boolean = True);
    procedure Commit;

    procedure GetComputerList(SL: TStrings);
    procedure GetUserList(SL: TStrings);
    procedure GetTableUserList(SL: TStrings);
    procedure GetUserListWithEntryDate(SL: TStrings);
    procedure GetTableUserListWithEntryDate(SL: TStrings);
    procedure GetTableUserListWithComputerName(SL: TStrings);

    property UserCount: Byte read GetUserCount;
    property TabState: TTabState read GetState;
    property StoredInt: Integer read FStoredInt write SetStoredInt;
    property ComputerCount: Integer read GetComputerCount;
    property Compressed: Boolean read GetCompressed write SetCompressed Default False;
    property Modified: Boolean read GetModified write SetModified Default False;
    property Created: TDateTime read GetCreated;
    property LastModified: TDateTime read GetLastModified;
    property MyComputerName: string read GetMyComputerName;
    property MyLoginName: string read GetMyLoginName;

  published
    property Active;
    property Filtered: Boolean read FFiltered write SetFiltered default False;
    property ReadOnly:boolean read FReadOnly write FReadOnly default false;
//        property OnLoadRecord:TkbmOnLoadRecord read FOnLoadRecord write FOnLoadRecord;
//        property OnLoadField:TkbmOnLoadField read FOnLoadField write FOnLoadField;
//        property OnCompressSave:TkbmOnCompress read FOnCompressSave write FOnCompressSave;
//        property OnDecompressLoad:TkbmOnDecompress read FOnDecompressLoad write FOnDecompressLoad;
    property BeforeOpen;
    property AfterOpen;
    property BeforeClose;
    property AfterClose;
    property BeforeInsert;
    property AfterInsert;
    property BeforeEdit;
    property AfterEdit;
    property BeforePost;
    property AfterPost;
    property BeforeCancel;
    property AfterCancel;
    property BeforeDelete;
    property AfterDelete;
    property BeforeScroll;
    property AfterScroll;
    property BeforeRefresh{: TDataSetNotifyEvent read FBeforeRefresh write FBeforeRefresh};
    property AfterRefresh{: TDataSetNotifyEvent read FAfterRefresh write FAfterRefresh};
    property OnCalcFields;
    property OnDeleteError;
    property OnEditError;
    property OnFilterRecord;
    property OnNewRecord;
    property OnPostError;
    property TableName: string read FTableName write FTableName;
    property SortByField: string read FSortByField write FSortByField;
    property EnableRecMove: Boolean read FEnableRecMove write FEnableRecMove default False;
    property EnableSort: Boolean read FEnableSort write FEnableSort default False;
    property Exclusive: Boolean read FExclusive write SetExclusive default False;
    property Temporal: Boolean read FTemporal write SetTemporal;
    property Filter: string read FFilterText write SetFilterText;
    property OnStructChanged: TFOnStructChangedEvent read FOnStructChanged write FOnStructChanged;
    property RefreshDelay: Integer read GetRefreshDelay write SetRefreshDelay;
    property SelfRefresh: Boolean read FSelfRefresh write SetSelfRefresh;
    property TableKind: Word read FTableKind write SetTableKind;
    property Password: string read FPassword write SetPassword;
    property Options: string read FOptions write SetOptions;
    property OnStoredIntChange: TDataSetNotifyEvent read FOnStoredIntChange write FOnStoredIntChange;
    property OnUserCountChange: TDataSetNotifyEvent read FOnUserCountChange write FOnUserCountChange;
    property BeforeCreate: TDataSetNotifyEvent read FBeforeCreate write FBeforeCreate;
    property AfterMove: TDataSetNotifyEvent read FAfterMove write FAfterMove;
    property TableUserName: string read FTableUserName write SetTableUserName;
  end;

const
  ToUpperChars: array[0..255] of Char =
    (#$00,#$01,#$02,#$03,#$04,#$05,#$06,#$07,#$08,#$09,#$0A,#$0B,#$0C,#$0D,#$0E,#$0F,
     #$10,#$11,#$12,#$13,#$14,#$15,#$16,#$17,#$18,#$19,#$1A,#$1B,#$1C,#$1D,#$1E,#$1F,
     #$20,#$21,#$22,#$23,#$24,#$25,#$26,#$27,#$28,#$29,#$2A,#$2B,#$2C,#$2D,#$2E,#$2F,
     #$30,#$31,#$32,#$33,#$34,#$35,#$36,#$37,#$38,#$39,#$3A,#$3B,#$3C,#$3D,#$3E,#$3F,
     #$40,#$41,#$42,#$43,#$44,#$45,#$46,#$47,#$48,#$49,#$4A,#$4B,#$4C,#$4D,#$4E,#$4F,
     #$50,#$51,#$52,#$53,#$54,#$55,#$56,#$57,#$58,#$59,#$5A,#$5B,#$5C,#$5D,#$5E,#$5F,
     #$60,#$41,#$42,#$43,#$44,#$45,#$46,#$47,#$48,#$49,#$4A,#$4B,#$4C,#$4D,#$4E,#$4F,
     #$50,#$51,#$52,#$53,#$54,#$55,#$56,#$57,#$58,#$59,#$5A,#$7B,#$7C,#$7D,#$7E,#$7F,
     #$80,#$81,#$82,#$81,#$84,#$85,#$86,#$87,#$88,#$89,#$8A,#$8B,#$8C,#$8D,#$8E,#$8F,
     #$80,#$91,#$92,#$93,#$94,#$95,#$96,#$97,#$98,#$99,#$8A,#$9B,#$8C,#$8D,#$8E,#$8F,
     #$A0,#$A1,#$A1,#$A3,#$A4,#$A5,#$A6,#$A7,#$A8,#$A9,#$AA,#$AB,#$AC,#$AD,#$AE,#$AF,
     #$B0,#$B1,#$B2,#$B2,#$A5,#$B5,#$B6,#$B7,#$A8,#$B9,#$AA,#$BB,#$A3,#$BD,#$BD,#$AF,
     #$C0,#$C1,#$C2,#$C3,#$C4,#$C5,#$C6,#$C7,#$C8,#$C9,#$CA,#$CB,#$CC,#$CD,#$CE,#$CF,
     #$D0,#$D1,#$D2,#$D3,#$D4,#$D5,#$D6,#$D7,#$D8,#$D9,#$DA,#$DB,#$DC,#$DD,#$DE,#$DF,
     #$C0,#$C1,#$C2,#$C3,#$C4,#$C5,#$C6,#$C7,#$C8,#$C9,#$CA,#$CB,#$CC,#$CD,#$CE,#$CF,
     #$D0,#$D1,#$D2,#$D3,#$D4,#$D5,#$D6,#$D7,#$D8,#$D9,#$DA,#$DB,#$DC,#$DD,#$DE,#$DF);

function Q_PosText(const FindString, SourceString: string; StartPos: Integer): Integer;
function Q_PosStr(const FindString, SourceString: string; StartPos: Integer): Integer;
function Q_StrScan(const S: string; Ch: Char; StartPos: Integer = 1): Integer;

procedure Register;

implementation
//{$R zTable.RES}

const
  kbmSupportedFieldTypes=[ftString,ftSmallint,ftInteger,ftLargeInt,ftWord,ftBoolean,ftFloat,ftCurrency,
                          ftDate,ftTime,ftDateTime,ftAutoInc,ftBCD,ftBlob,ftMemo,ftGraphic,ftFmtMemo,
                          ftParadoxOle,ftDBaseOle,ftTypedBinary,ftBytes, ftVarBytes
{$ifndef LEVEL3}
                          ,ftFixedChar,ftWideString,ftLargeInt
{$endif}
                          ];

  kbmStringTypes=[ftString,ftMemo,ftFmtMemo
{$ifndef LEVEL3}
                          ,ftFixedChar,ftWideString
{$endif}
                          ];

//  kbmBlobTypes=[ftBlob,ftMemo,ftGraphic,ftFmtMemo,ftParadoxOle,ftDBaseOle,ftTypedBinary,ftBytes,ftVarBytes];

{$ifdef LEVEL3}
  FieldTypeNames: array[TFieldType] of string = (
    'Unknown', 'String', 'SmallInt', 'Integer', 'Word', 'Boolean', 'Float',
    'Currency', 'BCD', 'Date', 'Time', 'DateTime', 'Bytes', 'VarBytes',
    'AutoInc', 'Blob', 'Memo', 'Graphic', 'FmtMemo', 'ParadoxOle',
    'dBaseOle', 'TypedBinary', 'Cursor');
{$endif}

//  kbmNonBlobTypes=kbmSupportedFieldTypes - kbmBlobTypes;

  // Table definition magic words.
  kbmTableDefMagicStart = '@@TABLEDEF START@@';
  kbmTableDefMagicEnd = '@@TABLEDEF END@@';

  // Binary file magic word.
  kbmBinaryMagic = '@@BINARY@@';

// Code special characters (LF,CR,%,#0)
// CR (#13) -> %c
// LF (#10) -> %n
// #0 -> %0
// % -> %%
{function StringToCodedString(const Source:string):string;
var
   i,j:integer;
   l:integer;
begin
     // Count CR/LF.
     l:=0;
     for i:=1 to length(Source) do
         if Source[i] in [#13,#10,'%',#0] then inc(l);

     // If no special characters, return the original string.
     if l=0 then
     begin
          Result:=Source;
          exit;
     end;

     // If any special characters, make room for them.
     SetLength(Result,length(Source)+l);

     // Code special characters.
     j:=1;
     for i:=1 to length(Source) do
         case Source[i] of
              #13: begin
                        Result[j]:='%'; inc(j);
                        Result[j]:='c'; inc(j);
                   end;
              #10: begin
                        Result[j]:='%'; inc(j);
                        Result[j]:='n'; inc(j);
                   end;
              #0:  begin
                        Result[j]:='%'; inc(j);
                        Result[j]:='0'; inc(j);
                   end;
              '%': begin
                        Result[j]:='%'; inc(j);
                        Result[j]:='%'; inc(j);
                   end;
              else
                   begin
                        Result[j]:=Source[i];
                        inc(j);
                   end;
         end;
end;

// Decode special characters (LF,CR,%,#0)
// %c -> CR (#13)
// %n -> LF (#10)
// %% -> %
// %0 -> #0
function CodedStringToString(const Source:string):string;
var
   i,j:integer;
begin
     SetLength(Result,length(Source));

     // Code special characters.
     i:=1;
     j:=1;
     while true do
     begin
          if i>length(Source) then break;
          if Source[i]='%' then
          begin
               inc(i);
               case Source[i] of
                    'c': Result[j]:=#13;
                    'n': Result[j]:=#10;
                    '%': Result[j]:='%';
                    '0': Result[j]:=#0;
               end;
               inc(j);
          end
          else
          begin
               Result[j]:=Source[i];
               inc(j);
          end;
          inc(i);
     end;

     // Cut result string to right length.
     if i<>j then SetLength(Result,j-1);
end;

// Code a string as BASE 64.
function StringToBase64(const Source:string):string;
var
   Act: Word;
   Bits,I,P,Len: Integer;
begin
     Bits:=0;
     Len:=(Length(Source)*4+2) div 3;
     if Len>0 then
     begin
          SetLength(Result,Len);
	  P:=1;
	  Act:=0;
	  for I:=1 to Pred(Len) do
          begin
	       if Bits<6 then
               begin
	            Act:=(Act shr 6) or (Ord(Source[P]) shl Bits);
		    Inc(P);
		    Inc(Bits,2);
               end
               else
               begin
	            Dec(Bits,6);
		    Act:=Act shr 6;
               end;
	       Result[I]:=Char(Act and 63+32);
          end;
	  Result[Len]:=Char(Act shr 6+32);
     end;
end;

// Decode BASE64 string.
function Base64ToString(const S: string): string;
var
   Act: Word;
   Bits,I,P,Len: Integer;
begin
     Len:=(Length(S)*3) div 4;
     SetLength(Result,Len);
     Bits:=0;
     Act:=0;
     P:=1;
     for I:=1 to system.Length(S) do
     begin
          Act:=Act or (Ord(S[I])-32) shl Bits;
	  if Bits>=2 then
          begin
	       Result[P]:=Char(Act and $FF);
	       Inc(P);
	       Act:=Act shr 8;
	       Dec(Bits,2);
          end
          else
	      Inc(Bits,6);
     end;
end;}

// Compare two records.
{function TzTable._InternalCompareRecords(FieldList:TList; Record1,Record2:PChar; Options:TzTableCompareOptions): Integer;
var
   p1,p2:PChar;
   fld:TField;
   i:integer;
begin
     Result:=0;

     // Loop through all indexfields, left to right.
     for i:=0 to FieldList.Count-1 do
     begin
          fld:=TField(FieldList[i]);

          // Get data for specified field for the two records.
          p1:=GetFieldPointer(Record1,fld);
          p2:=GetFieldPointer(Record2,fld);

          // Check if both not null.
          if (p1[0]<>#0) and (p2[0]<>#0) then
          begin
               // Skip null flag.
               inc(p1);
               inc(p2);

               // Compare the fields.
               Result:=CompareFields(p1,p2,fld.DataType,(mtcoPartialKey in Options),(mtcoCaseInsensitive in Options));
          end
          else if Boolean(p1[0]) then Result:=1
          else if Boolean(p2[0]) then Result:=-1;
          if Result<>0 then break;
     end;
end;}

constructor TzTable.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
     FRecalcOnIndex:=false;

  FRefreshStateHeaderSize := SizeOf(TRefreshStateHeader);
  FDataFileHeaderSize := SizeOf(TDataFileHeader);
  FNetRecordSize := SizeOf(TNetRecord);

  FFilterText := '';
  FFilterCount := -1;
  FModified := False;
  BefRecNo := 0;
  FSelfRefresh := True;
  FPassWord := '';
  FTableKind := 0;
  FOptions := '';
  MsgNum := 0;
  ErrorNum := 0;
  FStoredInt := 0;

  FChangeTimer := TTimer.Create(Self);
  FChangeTimer.Enabled := False;
  FChangeTimer.Interval := 500;
  FChangeTimer.OnTimer := OnChangeTimer;
end;

destructor TzTable.Destroy;
begin
  if FIsOpen then Close; //InternalClose;
  Ar0 := nil;
  Ar1 := nil;
  Ar2 := nil;
  Ar3 := nil;
  FChangeTimer.Free;

  if FTemporal and (FTmpFile <>'') then
    try
      SysUtils.DeleteFile(TableName);
    except
     raise;
    end;
  inherited Destroy;
end;

// Copy masterfields to detail table if a master/detail relation.
procedure TzTable.DoOnNewRecord;
begin
  inherited DoOnNewRecord;

  // Update autoinc if such a field is defined.
  if Assigned(FAutoIncField) then
    FAutoIncField.AsInteger:=FAutoIncMax+1;
end;

function TzTable.GetFieldSize(FieldType:TFieldType; Size:longint):longint;
begin
     case FieldType of
          ftFixedChar,ftWideString,
          ftString:             Result:=Size+1;
          ftSmallInt:           Result:=SizeOf(SmallInt);
          ftInteger:            Result:=SizeOf(Integer);
          ftLargeInt:           Result:=SizeOf(Int64);
          ftWord:               Result:=SizeOf(Word);
          ftBoolean:            Result:=SizeOf(WordBool);
          ftFloat:              Result:=SizeOf(Double);
          ftCurrency:           Result:=SizeOf(Double);
          ftDate:               Result:=SizeOf(Integer);
          ftTime:               Result:=SizeOf(Integer);
          ftDateTime:           Result:=SizeOf(Double);
          ftAutoInc:            Result:=SizeOf(Integer);
{          ftBlob:               Result:=SizeOf(PkbmBlob);
          ftMemo:               Result:=SizeOf(PkbmBlob);
          ftGraphic:            Result:=SizeOf(PkbmBlob);
          ftFmtMemo:            Result:=SizeOf(PkbmBlob);
          ftParadoxOle:         Result:=SizeOf(PkbmBlob);
          ftDBaseOle:           Result:=SizeOf(PkbmBlob);
          ftTypedBinary:        Result:=SizeOf(PkbmBlob);
          ftBytes:              Result:=SizeOf(PkbmBlob);
          ftVarBytes:           Result:=SizeOf(PkbmBlob);}
          ftBCD:                Result:=34;
     else
          Result:=0;
     end;
end;

procedure TzTable.CreateTable;
var i, Offs, Ix:Integer;
    DescribF : TBDescribField;
begin
  if not FTemporal and FileExists(TableName) then
    raise EzTableError.Create (TableName + ': Allready exists');

  CheckInactive;

  // If no fielddefs existing, use the previously defined fields.
  if (FieldDefs.Count = 0) and (Fields.Count > 0) then
      for i:=0 to FieldCount-1 do
          with Fields[i] do
               if FieldKind = fkData then
                  FieldDefs.Add(FieldName, DataType, Size, Required);

  if (FieldDefs.Count = 0) and Assigned(FBeforeCreate) then
    FBeforeCreate(Self);

  if FieldDefs.Count = 0 then
    raise EzTableError.Create (FTableName + ' Open: FieldDefs.Count = 0');

  FStream := TFileStream.Create (FTableName, fmCreate or fmShareDenyNone);

  try
    FillChar(FRefreshStateHeader, FRefreshStateHeaderSize, 0);
    with FRefreshStateHeader do begin
      ModifiedNum := 0;
      State := tsWait;
      UserConfirm := 0;
      Msg_Num := 0;
      Computer_Count := 1;
      Stored_Int := 0;
    end;
    WriteRefreshState;

    FillChar(FNetRecord, FNetRecordSize, 0);
    with FNetRecord do begin
      LockRecNum := -1;
      LockValue := '';
      FieldIndex := 255;
      Busy := False;
    end;
    WriteNetRecord(0);

    FillChar(FDataFileHeader, FDataFileHeaderSize, 0);
    with FDataFileHeader do begin
      GenHeader := 'zTB'+^Z;
      RecordSize := 0; // filled later
      NumberFields := FieldDefs.Count;
      Compressed := False;
      Created := Now;
      LastModified := Now;
      Modified := False;
      StartData := -1;
      Password := FPassWord;
      TableKind := FTableKind;
      Options := FOptions;
    end;
    WriteHeader;

    Offs := 0;
    for Ix := 0 to FieldDefs.Count-1 do
      begin
        with FieldDefs.Items[Ix] do begin
          FillChar(DescribF, SizeOf(DescribF),#0);
          DescribF.DataType := DataType;
          StrCopy(DescribF.Name, PChar(FieldDefs.Items[Ix].Name));
          DescribF.Size := Size;
          DescribF.Precision:=Precision;
          if FieldCount = 0 then begin
            if DataType in [ftSmallInt, ftWord, ftInteger, ftLargeInt, ftFloat, ftCurrency] then begin
              DescribF.Alignment := taRightJustify;
              case DataType of
                ftFloat, ftLargeInt, ftCurrency: DescribF.DisplayWidth := 14;
                ftInteger: DescribF.DisplayWidth := 10;
                else DescribF.DisplayWidth := 5;
              end;
            end
            else begin
              DescribF.Alignment := taLeftJustify;
              case DataType of
                ftString: DescribF.DisplayWidth := Size;
                ftDateTime: DescribF.DisplayWidth := 20;
                else DescribF.DisplayWidth := 10;
              end;
            end;

            StrCopy(DescribF.DisplayLabel, PChar(FieldDefs.Items[Ix].Name));
            DescribF.ReadOnly := False;
            DescribF.Tag := 0;
            DescribF.Visible := True;
          end
          else begin
            if Fields[Ix].Name = '' then
              Fields[Ix].Name := Name + FieldDefs.Items[Ix].Name;
            DescribF.Alignment := Fields[Ix].Alignment;
            DescribF.DisplayWidth := Fields[Ix].DisplayWidth;
            StrCopy(DescribF.DisplayLabel, PChar(Fields[Ix].DisplayLabel));
            DescribF.ReadOnly := Fields[Ix].ReadOnly;
            DescribF.Tag := Fields[Ix].Tag;
            DescribF.Visible := Fields[Ix].Visible;
            StrCopy(DescribF.EditMask, PChar(Fields[Ix].EditMask));
            DescribF.Required := Fields[Ix].Required;
            StrCopy(DescribF.DefaultExpression, PChar(Fields[Ix].DefaultExpression));
          end;

          case DataType of
            ftString:
              DescribF.DataSize := Size + 1;
            ftBoolean, ftSmallInt, ftWord:
              DescribF.DataSize := 2;
            ftInteger, ftDate, ftTime:
              DescribF.DataSize := 4;
            ftFloat, ftLargeInt, ftCurrency, ftDateTime:
              DescribF.DataSize := 8;
          else
            raise EzTableError.Create ('InitFieldsDefs: Unsupported field type');
          end;
          Inc(Offs,DescribF.DataSize + 1);
          FStream.Write(DescribF,SizeOf(DescribF));
        end;
      end;
    fDataFileHeader.StartData := FStream.Position;
    fDataFileHeader.RecordSize := Offs;
    FRecordSize := Offs;
    WriteHeader;

    if not Temporal and (FieldCount > 0) then SaveFieldsProperties;
  finally
    FStream.Free;
  end;
end;

procedure TzTable.InternalOpen;

  function GetNetUserName: string;
  var
    UserName: PChar;
    Count: longword;
  begin
    {retrieve the required size of the user name buffer}
    Count := 0;
    GetUserName(nil,Count);

    {allocate memory for the user name}
    Username := StrAlloc(Count);

    {retrieve the user name}
    if GetUserName(UserName,count) then
      result:= StrPas(UserName)
    else
      result:='Unknown';

    {dispose of allocated memory}
    StrDispose(UserName)
  end;

  function GetCompName: string;
  var
    ComputerName   :array[0..MAX_COMPUTERNAME_LENGTH+1] of char;  // holds the name
    Size           :Longword;
  begin
    {initialize the computer name size variable}
    Size := MAX_COMPUTERNAME_LENGTH+1;

    {retrieve the computer name}
    if GetComputerName(ComputerName, Size) then
      result:= StrPas(Computername)
    else result:='Unknown';
  end;

  procedure FillMyNetRecord(u, s: string);
  begin
    with FNetRecord do begin
      LockRecNum := -1;
      LockValue := '';
      FieldIndex := 255;
      Busy := True;
      StrCopy(ComputerName, PChar(u));
      StrCopy(UserName, PChar(s));
      StrCopy(TableUserName, PChar(FTableUserName));
      EntryDate := Now;
    end
  end;

var
   i, cc: integer;
   sz: integer;
   DescribF : TBDescribField;
   s, u: string;
begin
  if not FileExists (FTableName) then
    raise EzTableError.Create ('Open: Table file not found - ' + FTableName);
  FStream := TFileStream.Create (FTableName, fmOpenReadWrite + fmShareDenyNone);

  cc := ComputerCount;
  if cc = 0 then begin   // for support old version only
    cc := 10;
    FStream.Seek(0, soFromBeginning);
    FRefreshStateHeader.Computer_Count := 10;
    WriteRefreshState;
    FStream.Position := FRefreshStateHeaderSize;
    InsertSpace(FNetRecordSize * 10 - 120);
    with FNetRecord do begin
      LockRecNum := -1;
      LockValue := '     ';
      FieldIndex := 255;
      Busy := False;
      ComputerName := '';
      UserName := '';
      TableUserName := '';
      EntryDate := Now;
    end;
    for i := 0 to 9 do WriteNetRecord(i);
    ReadHeader;
    FDataFileHeader.StartData := FDataFileHeader.StartData + FNetRecordSize * 10 - 120;
    FStream.Seek(FRefreshStateHeaderSize + FRefreshStateHeader.Computer_Count * FNetRecordSize, soFromBeginning);
    FStream.WriteBuffer(fDataFileHeader, fDataFileHeaderSize);
  end;

  FStream.Seek(FRefreshStateHeaderSize + FNetWorkHeaderSize, soFromBeginning);
  FStream.ReadBuffer(FDataFileHeader, FDataFileHeaderSize);

  Assert(FDataFileHeader.NumberFields > 0, Name + ' - Invalid number of fields.');

//  if not FTemporal then
    InternalInitFieldDefs;

  if FRecordSize = -1 then Exit;

  if DefaultFields and (Fields.Count = 0) then CreateFields;

  if FieldDefs.Count = 0 then
    raise EzTableError.Create (FTableName + ' Open: FieldDefs.Count = 0');

  FRecordSize:=0;

  FStream.Seek(FRefreshStateHeaderSize + FNetWorkHeaderSize + fDataFileHeaderSize,
               soFromBeginning);

     for i:=0 to FieldDefs.Count - 1 do
         with FieldDefs[i] do begin
              // Check if field type supported.
              if not (DataType in kbmSupportedFieldTypes) then
                 raise EzTableError.Create('Fielddef '+
                         Name
                         +' Datatype '
                         + inttostr(integer(DataType))
                         +' not supported.');
              FFieldOfs[i]:=FRecordSize;

              //if fDataFileHeader.VersionNumber > 1 then begin
                FStream.Read(DescribF,SizeOf(DescribF));
                if Fields[i].Name = '' then begin
                  Fields[i].Name := Name;
                  Fields[i].Alignment := DescribF.Alignment;
                  Fields[i].DisplayWidth := DescribF.DisplayWidth;
                  Fields[i].DisplayLabel := DescribF.DisplayLabel;
                  Fields[i].ReadOnly := DescribF.ReadOnly;
                  Fields[i].Tag := DescribF.Tag;
                  Fields[i].Visible := DescribF.Visible;
                end;
                Fields[i].DefaultExpression := DescribF.DefaultExpression;
                Fields[i].EditMask := DescribF.EditMask;
              //end;
              {else begin
                if Fields[i].Name = '' then Fields[i].Name := Name;
//                Fields[i].Name := Name;
              //end;}

              // Look for fieldsize.
              sz:=GetFieldSize(DataType,Size);
              inc(sz);   // 1.st byte is boolean flag for Null or not.
              inc(FRecordSize, sz);
         end;

     BindFields(True);
     FRecNo := -1;
     BookmarkSize := sizeof(Integer);
     FStartRecInfo := FRecordSize;
     FStartCalculated := FStartRecInfo+SizeOf(TRecInfo);
     FBufferSize := FRecordSize + Sizeof(TRecInfo) + CalcFieldsSize;

  FModified := fDataFileHeader.Modified;
  FRecordCount := TotalCount;
  FFilterCount := -1;

  ModifiedNum := GetModifiedNum;
  FStoredInt :=  FRefreshStateHeader.Stored_Int;

  MyNum := -1;
  MyLock := tsWait;
  u := GetCompName;
  s := GetNetUserName;

  for i := 0 to cc - 1 do begin
    ReadNetRecord(i);
    if (MyNum = -1) and not FNetRecord.Busy then begin
      MyNum := i;
      FillMyNetRecord(u, s);
    end
    else FNetRecord.Busy := False;
    WriteNetRecord(i);
  end;
  if MyNum = -1 then begin // Increase network computer count
    FStream.Position := FRefreshStateHeaderSize + FNetWorkHeaderSize;
    InsertSpace(FNetRecordSize);
    FDataFileHeader.StartData := FDataFileHeader.StartData + FNetRecordSize;
    FRefreshStateHeader.Computer_Count := FRefreshStateHeader.Computer_Count + 1;
    FStream.Seek(FRefreshStateHeaderSize + FRefreshStateHeader.Computer_Count * FNetRecordSize, soFromBeginning);
    FStream.WriteBuffer(fDataFileHeader, fDataFileHeaderSize);
    MyNum := FRefreshStateHeader.Computer_Count - 1;
    FillMyNetRecord(u, s);
    WriteNetRecord;
  end;
  IncreaseUserConfirm;

//  Assert(MyNum > 0, 'Invalid NetWork user number.');       // ????????

  if FSelfRefresh then FChangeTimer.Enabled := True;

  if (UserCount = 1) and (TabState <> tsWait) then begin
    FRefreshStateHeader.State := tsWait;
    WriteRefreshState;
  end;

  FIsOpen:=True;
end;

procedure TzTable.ReadNetRecord(RecNum: Integer = -1);
begin
  if RecNum = -1 then RecNum := MyNum;
  FStream.Position := FRefreshStateHeaderSize + RecNum * FNetRecordSize;
  FStream.ReadBuffer(FNetRecord, FNetRecordSize);
end;

procedure TzTable.WriteNetRecord(RecNum: Integer = -1);
begin
  if RecNum = -1 then RecNum := MyNum;
  FStream.Position := FRefreshStateHeaderSize + RecNum * FNetRecordSize;
  FStream.WriteBuffer(FNetRecord, FNetRecordSize);
  Commit;
  UpDateCursorPos;
end;

function TzTable.FNetWorkHeaderSize: Integer;
begin
  Result := ComputerCount * FNetRecordSize;
end;

procedure TzTable.InternalClose;
var OpenCount, Handle: Integer;
    LastDate: TDateTime;
begin
  if (MyLock <> tsWait) then begin       // Unlock table before close
    FRefreshStateHeader.State := tsWait;
    WriteRefreshState;
  end;

  FChangeTimer.Enabled := False;

  LastDate := Now;
  if FIsOpen then begin
    if not Temporal and (FieldCount > 0) then SaveFieldsProperties;
    ReadNetRecord;
    FNetRecord.Busy := False;
    WriteNetRecord;
    IncreaseUserConfirm;
    OpenCount := GetUserCount;
    LastDate := LastModified;
    if Assigned(OnUserCountChange) then FOnUserCountChange(Self);
  end
  else OpenCount := -1954;
  MyNum := 0;
  BindFields(False);
  if DefaultFields then DestroyFields;
  FieldDefs.Clear;

  FStream.Free;
  FIsOpen:=False;
  if OpenCount = 0 then begin
    try
      Handle := FileOpen(FTableName, fmOpenReadWrite + fmShareDenyNone);
      FileSetDate(Handle, DateTimeToFileDate(LastDate));
      FileClose(Handle);
    except
    end;
  end;
end;

procedure TzTable.InternalInitFieldDefs;
var i:integer;
    DescribF : TBDescribField;
    OldRecSize : Integer;
    OldFieldDefs: TFieldDefs;
begin
  OldFieldDefs := TFieldDefs.Create(Self);
  with OldFieldDefs do begin
    FStream.Seek(FRefreshStateHeaderSize + FNetWorkHeaderSize + fDataFileHeaderSize,
                 soFromBeginning);
    Assert(fDataFileHeader.NumberFields > 0,'Invalid number of fields.');

    OldRecSize := 0;
    if (fDataFileHeader.NumberFields > 0) then begin
      for i := 0 to fDataFileHeader.NumberFields - 1 do begin
        FStream.Read(DescribF,SizeOf(DescribF));
        OldFieldDefs.Add(DescribF.Name, DescribF.DataType,
                         DescribF.Size, False);
        Inc(OldRecSize, DescribF.DataSize);
      end;
    end;
    Inc(OldRecSize, fDataFileHeader.NumberFields);
  end;

  if (FieldDefs.Count = 0) and (Fields.Count > 0) then
    for i := 0 to FieldCount - 1 do with Fields[i] do
      if FieldKind = fkData then
        FieldDefs.Add(FieldName, DataType, Size, Required);

  if FieldDefs.Count = 0 then begin
    FieldDefs.Assign(OldFieldDefs);
    fRecordSize := OldRecSize;
  end
  else begin
    FRecordSize := 0;
    for i := 0 to FieldDefs.Count - 1 do
       Inc(FRecordSize, GetFieldSize(FieldDefs[i].DataType, FieldDefs[i].Size) + 1);
  end;

  if ((OldFieldDefs.Count = 0) or (OldFieldDefs.Count <> FieldDefs.Count)) or
           (OldRecSize <> FRecordSize) then begin

    if Assigned(OnStructChanged) then begin
      InternalClose;
      OnStructChanged(Self, OldFieldDefs);
    end
    else begin
      ShowMessage('ERROR '+Name+ #10#13+
            'OldFieldDefs.Count ' + IntToStr(OldFieldDefs.Count)+#10#13+
            'FieldDefs.Count ' + IntToStr(FieldDefs.Count)+#10#13+
            'OldRecSize ' + IntToStr(OldRecSize)+#10#13+
            'FRecordSize ' + IntToStr(FRecordSize));
      InternalClose;
      raise EzTableError.Create ('ERROR ' + FTableName + ': Structure changed');
    end;
  end;
end;

function TzTable.GetActiveRecordBuffer:  PChar;
begin
     case State of
          dsBrowse:        if IsEmpty then
                              Result := nil
                           else
                              Result := ActiveBuffer;
          dsCalcFields:    Result := CalcBuffer;
          dsFilter:        Result:=FFilterBuffer;
          dsEdit,dsInsert: Result:=ActiveBuffer;
          dsNewValue,dsOldValue,dsCurValue: Result:=ActiveBuffer;
{$IFDEF LEVEL4}
          dsBlockRead:     Result := ActiveBuffer;
{$ENDIF}
     else
          Result:=nil;
     end;
end;

function TzTable.GetFieldPointer(Buffer:PChar; Field:TField):PChar;
begin
     Result:=Buffer;
     if Buffer=nil then exit;
     if Field.FieldNo<=0 then
        inc(Result,FStartCalculated+Field.Offset)
     else
        inc(Result,FFieldOfs[Field.FieldNo-1]);
end;

// Result is data in the buffer and a boolean return (true=not null, false=is null).
function TzTable.GetFieldData(Field: TField; Buffer: Pointer): Boolean;
var
   SourceBuffer: PChar;
begin
     Result:=False;
     if not FIsOpen then exit;
     SourceBuffer:=GetFieldPointer(GetActiveRecordBuffer,Field);
     if SourceBuffer=nil then Exit;

     if Assigned(Buffer) and (Field.DataType in kbmSupportedFieldTypes) then
          Move(SourceBuffer[1], Buffer^, GetFieldSize(Field.DataType,Field.Size));

     Result:=SourceBuffer[0]<>#0;
end;

procedure TzTable.SetFieldData(Field: TField; Buffer: Pointer);
var
   DestinationBuffer: PChar;
   sz:integer;
begin
     if not FIsOpen then exit;
     DestinationBuffer:=GetFieldPointer(GetActiveRecordBuffer,Field);
     if DestinationBuffer=nil then Exit;

     if Field.ReadOnly then DatabaseErrorFmt('Field %s is read only',[Field.DisplayName]);
     if Field.FieldKind in [fkData, fkInternalCalc] then Field.Validate(Buffer);

     // Check if it could change the sort order.
     sz:=GetFieldSize(Field.DataType,Field.Size);
     {if not ((Buffer<>nil) and (Boolean(DestinationBuffer[0]))) then FSorted:=false
     else} if (Field.DataType in [{$ifndef LEVEL3}ftFixedChar,ftWideString,{$endif} ftString]) then
        begin
             //l1:=strlen(Buffer);
             //l2:=strlen(PChar(@DestinationBuffer[1]));
        end;
     //else if not CompareMem(Buffer,PChar(@DestinationBuffer[1]),sz) then FSorted:=false;

     // Copy the data to the buffer.
     if LongBool(Buffer) then DestinationBuffer[0]:=#1
     else DestinationBuffer[0]:=#0;

     if Assigned(Buffer) and (Field.DataType in kbmSupportedFieldTypes) then
     begin
          if longbool(Buffer) then
             Move(Buffer^,DestinationBuffer[1],sz)
          else
             FillChar(DestinationBuffer[1],sz, 0);
     end;

     if not (State in [dsCalcFields, dsFilter, dsNewValue]) then
        DataEvent(deFieldChange, Longint(Field));

end;

function TzTable.BCDToCurr(BCD: Pointer; var Curr: Currency): Boolean;
begin
     Move(BCD^, Curr, SizeOf(Currency));
     Result := True;
end;

function TzTable.CurrToBCD(const Curr: Currency; BCD: Pointer; Precision, Decimals: Integer): Boolean;
begin
     Move(Curr, BCD^, SizeOf(Currency));
     Result := True;
end;

function TzTable.IsCursorOpen: Boolean;
begin
     Result:=FIsOpen;
end;

function TzTable.GetCanModify: Boolean;
begin
     Result:=not FReadOnly;
end;

function TzTable.GetRecordSize: Word;
begin
     Result:=FRecordSize;
end;

function TzTable.AllocRecordBuffer: PChar;
begin
     GetMem(Result, FBufferSize);
     if Result<>nil then FillChar(Result^,FBufferSize,0);
end;

procedure TzTable.FreeRecordBuffer(var Buffer: PChar);
begin
     FreeMem(Buffer);
     Buffer:=nil;
end;

{$IFDEF LEVEL4}
procedure TzTable.SetBlockReadSize(Value: Integer);
var
   DoNext: Boolean;
begin
     if Value <> BlockReadSize then
     begin
          if (Value > 0) or (Value < -1) then
          begin
               inherited;
               BlockReadNext;
          end
          else
          begin
               DoNext:=Value=-1;
               Value:=0;
               inherited;

               if DoNext then
                  Next
               else
               begin
                    CursorPosChanged;
                    Resync([]);
               end;
          end;
     end;
end;
{$ENDIF}

// Fill one field with contents of a variant.
procedure TzTable.PopulateField(Buffer:PChar; Field:TField; AValue:Variant);
var
   p:PChar;
   pValue:PChar;

   s:array[0..dsMaxStringSize] of Char;
   si:smallint;
{$ifndef LEVEL3}
   li:LargeInt;
{$endif}
   i:integer;
   w:word;
   d:double;
   dt:TDateTime;
begin
     p:=GetFieldPointer(Buffer,Field);

     // Populate with null?
     if VarIsNull(AValue) then
     begin
          Boolean(p[0]):=false;
          FillChar(p[1],GetFieldSize(Field.DataType, Field.Size), 0);
          exit;
     end;

     // Populate with value.
     Boolean(p[0]):=true;
     with Field do
     begin
          pValue:=nil;
          case DataType of
               ftFixedChar,ftWideString,
               ftString:
                   begin
                        StrLCopy(s,PChar(VarToStr(AValue)),DataSize);
                        if TStringField(Field).Transliterate then
                           DataSet.Translate(s,s,True);
                        pValue:=s;
                   end;

               ftSmallint:
                   begin
                        si:=AValue;
                        pValue:=PChar(@si);
                   end;
               ftLargeInt:
                   begin
                        li:=trunc(double(AValue));
                        pValue:=PChar(@li);
                   end;
               ftInteger,
               ftDate,
               ftTime,
               ftAutoInc:
                   begin
                        i:=AValue;
                        pValue:=PChar(@i);
                   end;

               ftWord:
                   begin
                        w:=AValue;
                        pValue:=PChar(@w);
                   end;

               ftBoolean: ;

               ftFloat,
               ftCurrency:
                   begin
                        d:=AValue;
                        pValue:=PChar(@d);
                   end;

               ftDateTime:
                   begin
                        dt:=AValue;
                        pValue:=PChar(@dt);
                   end;
          end;
          if pValue<>nil then Move(pValue^,p[1],GetFieldSize(DataType,Size));
     end;
end;

// Fill record with values for specified fields.
procedure TzTable.PopulateRecord(Buffer:PChar; Fields:string; Values:variant);
var
   FieldList:Tlist;
   i:integer;
   n:integer;
begin
     FieldList := TList.Create;
     try
        GetFieldList(FieldList,Fields);

        n:=VarArrayDimCount(Values);
        if n>1 then raise EzTableError.Create('Values variant array has invalid dimension count');
        if (n=0) and (FieldList.count>1) then raise EzTableError.Create('More fields than values');
        if FieldList.Count<1 then raise EzTableError.Create('There must be at least one field');

        // Single value.
        if n=0 then begin
          PopulateField(Buffer,TField(FieldList.Items[0]),Values);
          exit;
        end;

        // Several values.
        for i:=0 to FieldList.Count-1 do begin
          PopulateField(Buffer,TField(FieldList.Items[i]),Values[i]);
        end;
     finally
        FieldList.free;
     end;
end;

procedure TzTable.InternalFirst;
begin
  _InternalFirst;
end;

procedure TzTable.InternalLast;
begin
  _InternalLast;
end;

procedure TzTable._InternalFirst;
begin
  FRecNo := -1;
end;

procedure TzTable._InternalLast;
begin
  FRecNo := TotalCount;
end;

function TzTable._InternalNext:boolean;
begin
  if FRecordCount - 1 > FRecNo then begin
    Inc(FRecNo);
    Result:=true;
  end
  else Result:=false;
end;

function TzTable._InternalPrior:boolean;
begin
  if FrecNo > 0 then begin
    Dec(FRecNo);
    Result:=true;
  end
  else Result:=false;
end;

function TzTable.GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult;
var
   Acceptable: Boolean;
begin
     Result:=grOK;
     Acceptable:=False;
     repeat
         begin
              case GetMode of
                   gmCurrent: begin
                                   if FRecNo>=FRecordCount then Result:=grEOF
                                   else if FRecNo<0 then Result:=grBOF
                                   else Result:=grOk;
                              end;
                   gmNext:    begin
                                   if _InternalNext then Result:=grOK
                                   else Result:=grEOF;
                              end;
                   gmPrior:   begin
                                   if _InternalPrior then Result:=grOK
                                   else Result:=grBOF;
                              end;
              end;
              if Result=grOk then begin
                _ReadRecord(Buffer, FRecNo);
                //fill information part of buffer
                with PRecInfo(Buffer+FStartRecInfo)^ do begin
                  RecordNo := FRecNo;
                  BookmarkFlag := bfCurrent;
                end;

                //fill calc fields part of buffer
                //ClearCalcFields(Buffer);
                //GetCalcFields(Buffer);
                Acceptable:=FilterRecord(Buffer);
                if (GetMode=gmCurrent) and not Acceptable then Result:=grError;
              end
         end;
     until (Result<>grOk) or Acceptable;
end;

function TzTable.FindRecord(Restart, GoForward: Boolean): Boolean;
var
   Status:boolean;
begin
     CheckBrowseMode;
     DoBeforeScroll;
     SetFound(False);
     UpdateCursorPos;
     CursorPosChanged;

     if GoForward then
     begin
          if Restart then _InternalFirst;
          Status := _InternalNext;
     end else
     begin
          if Restart then _InternalLast;
          Status := _InternalPrior;
     end;

     if Status then
     begin
          Resync([rmExact, rmCenter]);
          SetFound(True);
     end;
     Result := Found;
     if Result then DoAfterScroll;
end;

// Is any record filtering applied.
// Could be master/detail or userdefined filter.
function TzTable.IsFiltered:boolean;
begin
     Result:=(Filtered and (Assigned(OnFilterRecord) or (FFilterText <> ''))) {or
             ((FMasterLink <> nil) and Assigned(FMasterLink.DataSource) and (FMasterLink.FieldNames<>'')
              and (FIndexList.Count>0))};
end;

// Filter records in general for
// master/detail and userdefined filter.
function TzTable.FilterRecord(Buffer: PChar): Boolean;
var
   SaveState: TDatasetState;
begin
     Result:=True;
     if not IsFiltered then Exit;

     // Now we will apply the filters on the record.
     SaveState:=SetTempState(dsFilter);
     FFilterBuffer:=Buffer;

     // Call users own filtering if specified.
     if FFilterText <> '' then Result := FilterByFilter;
     if Result and Assigned(OnFilterRecord) then OnFilterRecord(self, Result);

     // Finished filtering.
     RestoreState(SaveState);
end;

procedure TzTable.InternalSetToRecord(Buffer: PChar);
begin
     FRecNo:=PRecInfo(Buffer+FStartRecInfo).RecordNo;
end;

function TzTable.GetRecordCount: integer;
var
   SaveState: TDataSetState;
   SavePosition: integer;
   TempBuffer: PChar;
begin
     if not IsFiltered then Result:=FRecordCount
     else if FFilterCount <> -1 then Result := FFilterCount
     else begin
          Result:=0;
          SaveState:=SetTempState(dsBrowse);
          SavePosition:=FRecNo;
          try
             TempBuffer:=AllocRecordBuffer;
             InternalFirst;
             while GetRecord(TempBuffer,gmNext,True)=grOk do Inc(Result);
          finally
             FFilterCount := Result;
             RestoreState(SaveState);
             FRecNo:=SavePosition;
             FreeRecordBuffer(TempBuffer);
          end;
     end;
end;

function TzTable.GetRecNo: integer;
begin
  UpdateCursorPos;
  if (FRecNo = -1) and (RecordCount > 0) then
    Result := 1
  else Result := FRecNo + 1;
end;

procedure TzTable.SetRecNo(Value: Integer);
var TempBuffer: PChar;
begin
   try
       if notEmpty and (Value > 0) and (Value <= TotalCount) then begin
         if FFiltered then begin
           TempBuffer := AllocRecordBuffer;
           _ReadRecord(TempBuffer, Value - 1);
           if not FilterRecord(TempBuffer) then Exit;
         end;
         FRecNo := Value - 1;
         // refresh the position of the dataset
         Resync([rmCenter]);
       end;
   except
   end;
end;

procedure TzTable.InternalDelete;
var DeltaBuf: PChar;
    n, nt, p: Integer;
begin
  CheckActive;
  if IsEmpty then Exit;
  UpDateCursorPos;

  FRecordCount := TotalCount;

  nt := (FRecordCount - RecNo) * FRecordSize;
  if nt > 500000 then n := 500000
  else n := nt;

  GetMem(DeltaBuf, n);
  try
    p := FDataFileHeader.StartData + FRecordSize * RecNo;
    while nt > 0 do begin
      FStream.Position := p;
      FStream.ReadBuffer(DeltaBuf^, n);
      FStream.Position := p - FRecordSize;
      FStream.WriteBuffer (DeltaBuf^, n);
      Inc(p, n);
      Dec(nt, n);
      if nt > 0 then begin
        if nt > 500000 then n := 500000
        else n := nt;
      end;
    end;
  finally
    FreeMem(DeltaBuf);
  end;

  Dec(FRecordCount);
  FStream.Size := FStream.Size - FRecordSize;
  FFilterCount := -1;

  FModified := True;
  ReadHeader;
  FDataFileHeader.Modified := True;
  FDataFileHeader.LastModified := Now;
  SetCompressed(False);       // Writes header

  UpdateCursorPos;
end;

procedure TzTable.InsertSpace(SpaceLength: Integer);
var DeltaBuf: PChar;
    n, nt, p: Integer;
begin
  nt := FStream.Size - FStream.Position;
  if nt > 500000 then n := 500000
  else n := nt;

  GetMem(DeltaBuf, n);
  try
    p := FStream.Size;
    while nt > 0 do begin
      FStream.Position := p - n;
      FStream.ReadBuffer(DeltaBuf^, n);
      FStream.Position := p - n + SpaceLength;
      FStream.WriteBuffer (DeltaBuf^, n);
      Dec(p, n);
      Dec(nt, n);
      if nt > 0 then begin
        if nt > 500000 then n := 500000
        else n := nt;
      end;
    end;
  finally
    FreeMem(DeltaBuf);
  end;
end;

procedure TzTable.SetTableUserName(Value: string);
begin
  if FTableUserName = Value then Exit;
  FTableUserName := Value;
  if FIsOpen then begin
    ReadNetRecord;
    StrCopy(FNetRecord.TableUserName, PChar(Value));
    WriteNetRecord;
  end;
end;

function TzTable.GetMyComputerName: string;
begin
  ReadNetRecord;
  Result := FNetRecord.ComputerName;
end;

function TzTable.GetMyLoginName: string;
begin
  ReadNetRecord;
  Result := FNetRecord.UserName;
end;

procedure TzTable.InternalInitRecord(Buffer: PChar);
begin
  FillChar(Buffer^,FBufferSize,0);
  PRecInfo(Buffer+FStartRecInfo)^.RecordNo := FRecNo;
end;

procedure TzTable.InternalPost;
type
    TAutoInc=packed record
       Used:Byte;
       Int:Integer;
    end;
var n:integer;
   pAutoInc:^TAutoInc;
begin
  UpDateCursorPos;
  n := 0;

  // If an autoinc field is specified, allways keey track of highest used number.
  if Assigned(FAutoIncField) then begin
    pAutoInc:=Pointer(GetFieldPointer(ActiveBuffer,FAutoIncField));
    if FAutoIncMax<pAutoInc^.Int then FAutoIncMax:=pAutoInc^.Int;
  end;

  if State = dsEdit then
    _WriteRecord (ActiveBuffer, FRecNo)
  else begin
    if GetBookmarkFlag(ActiveBuffer) <> bfEOF then
      n := PRecInfo(ActiveBuffer+FStartRecInfo)^.RecordNo + 1;

    InternalLast;
    _AppendRecord(ActiveBuffer);
    Inc(FRecordCount);
  end;
  FFilterCount := -1;
  BefRecNo := n;
end;

procedure TzTable.SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag);
begin
  PRecInfo(Buffer + FStartRecInfo).BookmarkFlag := Value;
end;

function TzTable.GetBookmarkFlag(Buffer: PChar): TBookmarkFlag;
begin
  Result:=PRecInfo(Buffer+FStartRecInfo).BookmarkFlag;
end;

procedure TzTable.GetBookmarkData(Buffer: PChar; Data: Pointer);
begin
  PInteger(Data)^ := PRecInfo(Buffer + FStartRecInfo).RecordNo;
end;

procedure TzTable.SetBookmarkData(Buffer: PChar; Data: Pointer);
begin
  PRecInfo(Buffer + FStartRecInfo).RecordNo := Pinteger(Data)^;
end;

procedure TzTable.InternalGotoBookmark (Bookmark: Pointer);
var
  ReqBookmark: Integer;
begin
     ReqBookmark := PInteger (Bookmark)^;
     if (ReqBookmark >= 0) and (ReqBookmark < TotalCount) then
        FRecNo := ReqBookmark
     else
        raise EzTableError.Create('Bookmark ' + IntToStr(ReqBookmark) + ' not found');
end;

procedure TzTable.InternalHandleException;
begin
  Application.HandleException(Self);
end;

procedure TzTable.EmptyTable;
begin
  if Active then begin
    if not FTemporal and not FExclusive then
      raise EzTableError.Create ('ERROR Table not open in Exclusive mode: ' + TableName);
    try
      FStream.Size := fDataFileHeader.StartData;
      ClearBuffers;
      FStream.Position := fDataFileHeader.StartData;
      FRecNo := -1;
      DataEvent(deDataSetChange, 0);
    finally
      FRecordCount := 0;
    end;
  end
end;


// Sneak in before the table is closed.
procedure TzTable.DoBeforeClose;
begin
  if State in [dsEdit, dsInsert] then Post;
  inherited;
end;

// Sneak in after the table is opened.
procedure TzTable.DoAfterOpen;
begin
  SetFilterText(FFilterText);
  if FFiltered then Resync([]);
  inherited;
  if Assigned(OnUserCountChange) then FOnUserCountChange(Self);
end;


// Sequential search.
function  TzTable.Search(AField: TField; SearchValue: Variant;
                         FromBegining: Boolean = False): boolean;
var
   i, k, f: integer;
   IsNumeric, IsBoolean, IsString: Boolean;
begin
  if State in [dsEdit, dsInsert] then Post;
  // Loop for all records.
  UpdateCursorPos;
  Result := False;
  if IsEmpty then Exit;
  if FromBegining or BOF then f := 0
  else f := RecNo;
  k := TotalCount-1;
  IsBoolean := False;
  IsString := False;
  IsNumeric := AField.DataType in [ftInteger, ftLargeInt, ftFloat, ftSmallInt, ftCurrency];
  if not IsNumeric then IsBoolean := AField.DataType = ftBoolean;
  if not IsNumeric and not IsBoolean then IsString := AField.DataType = ftString;

  for i := f to k do begin
    _ReadRecord(ActiveBuffer, i);
    // Check filtering.
    if IsFiltered and not FilterRecord(ActiveBuffer) then continue;

    // Check key record equal to record.
    if (AField.Value = SearchValue) or ((AField.Value = Null) and
       (IsNumeric and (SearchValue = 0)) or
       (IsString and (SearchValue = '')) or
       (IsBoolean and (SearchValue = False))) then begin
      Result := True;
      FRecNo := i;
      Resync([rmCenter,rmExact]);
      exit;
    end;
  end;
  if NotEmpty then _ReadRecord(ActiveBuffer, FRecNo);
end;

function TzTable.Search(Farr: array of TField; SearchValue: array of Variant;
                         FromBegining: Boolean): boolean;
var
   i, k, f, j: integer;
   SubResult: Boolean;
begin
  if State in [dsEdit, dsInsert] then Post;
  // Loop for all records.
  UpdateCursorPos;
  Result := False;
  SubResult := False;
  if IsEmpty or (Length(Farr) = 0) or (Length(SearchValue) = 0) then Exit;
  if (Length(Farr) <> Length(SearchValue)) or (Length(Farr) = 0) or (Length(SearchValue) = 0)then
      raise EzTableError.Create ('ERROR Search array parameters');

  if FromBegining or BOF then f := 0
  else f := RecNo;
  k := TotalCount-1;

  for i := f to k do begin
    _ReadRecord(ActiveBuffer, i);
    // Check filtering.
    if IsFiltered and not FilterRecord(ActiveBuffer) then continue;

    // Check key record equal to record.
    for j := 0 to High(Farr) do begin
      SubResult := (Farr[j].Value = SearchValue[j]) or ((Farr[j].Value = Null) and
         ((Farr[j].DataType in [ftInteger, ftLargeInt, ftFloat, ftSmallInt, ftCurrency]) and (SearchValue[j] = 0)) or
         ((Farr[j].DataType = ftString) and (SearchValue[j] = '')) or
         ((Farr[j].DataType = ftBoolean) and (SearchValue[j] = False)));
      if not SubResult then Break;
    end;
    if SubResult then begin
      Result := True;
      FRecNo := i;
      Resync([rmCenter,rmExact]);
      exit;
    end;
  end;
  if NotEmpty then _ReadRecord(ActiveBuffer, FRecNo);
end;

function TzTable.Search(Farr: array of TField; SearchValue: array of Variant): boolean;
begin
  Result := Search(Farr, SearchValue, False);
end;

function TzTable.SearchNum(AField: TField; SearchValue: Variant; SearchDirection: SmallInt): boolean;
  function CheckFind(RNum: Integer): Boolean;
  begin
    Result := False;
    _ReadRecord(ActiveBuffer, RNum);
    // Check filtering.
    if IsFiltered and not FilterRecord(ActiveBuffer) then Exit;

    // Check key record equal to record.
    if (AField.AsFloat = SearchValue) {or ((AField.Value = Null) and
           (SearchValue = 0))} then begin
      Result := True;
      FRecNo := RNum;
    end;
  end;
var
   i, k, f: integer;
begin
  // Loop for all records.
  Result := False;
  if IsEmpty then Exit;
  if State in [dsEdit, dsInsert] then Post;
  UpdateCursorPos;
  case SearchDirection of
    0: begin // from top
         f := 0;
         k := TotalCount - 1;
         for i := f to k do if CheckFind(i) then begin
           Result := True;
           break;
         end;
       end;
    1: begin  // from current record to end
         f := FRecNo + 1;
         k := TotalCount - 1;
         for i := f to k do if CheckFind(i) then begin
           Result := True;
           break;
         end;
       end;
   -1: begin  // from current record to top
         f := FRecNo - 1;
         k := 0;
         for i := f downto k do if CheckFind(i) then begin
           Result := True;
           break;
         end;
       end;
   else Exit;
  end;

  _ReadRecord(ActiveBuffer, FRecNo);
  if Result then Resync([rmCenter,rmExact]);
end;

// 0 - From Bigening
// 1 - Up
// -1 - Down
function TzTable.SearchStr(AField: TField; SearchValue: string; SearchDirection: SmallInt;
                       WholeWord, MatchCase, ExactSearch: Boolean): boolean;
const
  Delimiters: string = ' ' + '!'+'"'+'('+')'+'*'+'+'+','+'-'+'.'+'/'+':'+
    ';'+'<'+'='+'>'+'?'+'['+'\'+']'+'^'+'{'+'}'+'|';

  function CheckFind(RNum: Integer; WholeWord, MatchCase, ExactSearch: Boolean): Boolean;
  var i, l: Integer;
  begin
    Result := False;
    _ReadRecord(ActiveBuffer, RNum);
    // Check filtering.
    if IsFiltered and not FilterRecord(ActiveBuffer) then Exit;

    if ExactSearch then begin
      Result := SearchValue = AField.AsString;
      if Result then FRecNo := RNum;
      Exit;
    end;

    if MatchCase then
      i := Q_PosStr(SearchValue, AField.AsString, 1)
    else i := Q_PosText(SearchValue, AField.AsString, 1);
    Result := i > 0;

    if WholeWord then begin
      l := Length(SearchValue);
      Result := ((i = 1) or Bool(Q_StrScan(Delimiters, (AField.AsString)[i - 1], 1))) and
                ((i + l > Length(AField.AsString)) or
                Bool(Q_StrScan(Delimiters, (AField.AsString)[i + l], 1)));
    end;
    if Result then FRecNo := RNum;
  end;
var
   i, k, f: integer;
begin
  // Loop for all records.
  Result := False;
  if IsEmpty then Exit;
  if State in [dsEdit, dsInsert] then Post;
  UpdateCursorPos;
  case SearchDirection of
    0: begin
         f := 0;
         k := TotalCount - 1;
         for i := f to k do if CheckFind(i, WholeWord, MatchCase, ExactSearch) then begin
           Result := True;
           break;
         end;
       end;
    1: begin
         f := FRecNo + 1;
         k := TotalCount - 1;
         for i := f to k do if CheckFind(i, WholeWord, MatchCase, ExactSearch) then begin
           Result := True;
           break;
         end;
       end;
   -1: begin
         f := FRecNo - 1;
         k := 0;
         for i := f downto k do if CheckFind(i, WholeWord, MatchCase, ExactSearch) then begin
           Result := True;
           break;
         end;
       end;
   else Exit;
  end;
  _ReadRecord(ActiveBuffer, FRecNo);
  if Result then Resync([rmCenter,rmExact]);
end;

// Locate record.
// If the keyfields are the same as sorted fields and the table is currently sorted,
// it will make a fast binary search. Otherwise it will make a sequential search.
// Binary searches dont take partial record in account.
function TzTable.LocateRecord(const KeyFields: string;
                                   const KeyValues: Variant; Options: TLocateOptions;
                                   SyncCursor: Boolean): Boolean;
var
   i:integer;
   KeyFieldsList: TList;
   Found:boolean;
   KeyRecord:PChar;
begin
  UpDateCursorPos;
     // Setup key options.
     FKeyOptions:=[];
     if loCaseInsensitive in Options then FKeyOptions:=FkeyOptions+[mtcoCaseInsensitive];
     if loPartialKey in Options then FKeyOptions:=FKeyOptions+[mtcoPartialKey];

     I := VarArrayDimCount(KeyValues);
     if I > 1 then
        raise EzTableError.Create('KeyValues VarArray invalid DimCount');
     CheckBrowseMode;
     CursorPosChanged;

     // Prepare list of fields representing the keys to search for.
     KeyFieldsList := TList.Create;
     Found:=false;
     try
        GetFieldList(KeyFieldsList, KeyFields);

        // Populate a keyrecord.
        KeyRecord:=AllocRecordBuffer;
        try
           // Fill it with values.
           PopulateRecord(KeyRecord,KeyFields,KeyValues);

           // Check if to recalc reference record before compare.
           {if FRecalcOnIndex then
           begin
                //fill calc fields part of buffer
                ClearCalcFields(KeyRecord);
                GetCalcFields(KeyRecord);
           end;}

           //DisableControls;
           try
              Found:=false;

              // Make fast binary search when the data are ordered by keyFields...
              {if FDoBinaryLocate and FSorted and (not IsFiltered) and (pos(KeyFields,FSortedOn)=1) then
              begin
                   // Do binary locate.
                   //_InternalBinarySearch(KeyFieldsList,0,FRecords.Count-1,KeyRecord,Found);
                   exit;
              end;}

              // Otherwise do a sequential search.
              //_InternalSequentialSearch(KeyFieldsList,0,FRecords.Count-1,KeyRecord,Found);
           finally
              //EnableControls;
           end;

        finally
           // Free reference record.
           FreeRecordBuffer(KeyRecord);
        end;

     finally
        KeyFieldsList.Free;
        SetFound(Found);
        Result:=Found;
     end;
end;

function TzTable.Lookup(const KeyFields: string; const KeyValues: Variant; const ResultFields: string): Variant;
var
   BM: TBookmarkStr;
begin
     Result := Null;
     BM := Bookmark;
     try
        if LocateRecord(KeyFields, KeyValues, [], False) then
        begin
             SetTempState(dsCalcFields);
             try
                CalculateFields(ActiveBuffer);
                Result := FieldValues[ResultFields];
             finally
                RestoreState(dsBrowse);
             end;
        end;
     finally
        Bookmark := BM;
     end;
end;

function TzTable.Locate(const KeyFields: string; const KeyValues: Variant; Options: TLocateOptions): Boolean;
begin
     DoBeforeScroll;
     Result := LocateRecord(KeyFields, KeyValues, Options, True);
     if Result then
     begin
          DoAfterScroll;
     end;
end;

procedure TzTable.SaveFieldsProperties;
var
   i:integer;
   fc:integer;
   DescribF : TBDescribField;
begin
  UpDateCursorPos;
  fc := FieldCount-1;
  FStream.Position := FRefreshStateHeaderSize + FNetWorkHeaderSize + FDataFileHeaderSize;

  with DescribF do for i := 0 to fc do begin
    // Copy general properties.
    FillChar(DescribF, SizeOf(DescribF), #0);

    DataType := FieldDefs[i].DataType;
    case FieldDefs[i].DataType of
      ftString: DataSize := FieldDefs[i].Size + 1;
      ftBoolean, ftSmallInt, ftWord: DataSize := 2;
      ftInteger, ftDate, ftTime: DataSize := 4;
      ftFloat, ftLargeInt, ftCurrency, ftDateTime: DataSize := 8;
      else raise EzTableError.Create ('InitFieldsDefs: Unsupported field type');
    end;
    StrCopy(Name, PChar(FieldDefs[i].Name));
    Size := FieldDefs[i].Size;
    Precision := FieldDefs[i].Precision;
    Alignment := Fields[i].Alignment;
    DisplayWidth := Fields[i].DisplayWidth;
    StrCopy(DisplayLabel, PChar(Fields[i].DisplayLabel));
    ReadOnly := Fields[i].ReadOnly;
    Tag := Fields[i].Tag;
    Visible := Fields[i].Visible;
    StrCopy(EditMask, PChar(Fields[i].EditMask));
    Required := Fields[i].Required;
    StrCopy(DefaultExpression, PChar(Fields[i].DefaultExpression));

    FStream.Write(DescribF, SizeOf(DescribF));
  end;
end;

function TzTable.IsSequenced: Boolean;
begin
     Result:=not Filtered;
end;

// Move record from one place in table to another.
function TzTable.MoveRecord(Source, Destination: Integer): Boolean;
var Buffer1, Buffer2, DeltaBuf: PChar;
    Bookmark1, Bookmark2 : TBookmarkFlag;
    i, n, nt, p: Integer;
begin
  Result := False;
  CheckActive;

  if (TotalCount <= 1) or (Source = Destination) or
     (Source = 0) or (Destination = 0) or
     (Source > FRecordCount) or (Destination > FRecordCount) then Exit;

  Buffer1 := AllocRecordBuffer;
  Buffer2 := AllocRecordBuffer;

  _ReadRecord(Buffer1, Source - 1);
  Bookmark1 := GetBookmarkFlag(Buffer1);

  if Source < Destination then begin
    Dec(Destination);
    for i := Source to Destination - 1 do begin
      _ReadRecord(Buffer2, i);

      Bookmark2 := GetBookmarkFlag(Buffer2);
      SetBookmarkFlag(Buffer1,Bookmark2);
      SetBookmarkFlag(Buffer2,Bookmark1);

      _WriteRecord(Buffer2, i - 1);
    end;
  end else begin
    nt := (Source - Destination) * FRecordSize;
    if nt > 500000 then n := 500000
    else n := nt;

    GetMem(DeltaBuf, n);

    p := FDataFileHeader.StartData + (FRecordSize * (Source - 1));
    while nt > 0 do begin
      Dec(p, n);
      FStream.Position := p;
      FStream.ReadBuffer(DeltaBuf^, n);
      FStream.Position := p + FRecordSize;
      FStream.WriteBuffer (DeltaBuf^, n);

      Dec(nt, n);
      if nt > 0 then begin
        if nt > 500000 then n := 500000
        else n := nt;
      end;
    end;
    FreeMem(DeltaBuf);

  end;
  _WriteRecord(Buffer1, Destination - 1);

  FreeMem(Buffer1);
  FreeMem(Buffer2);

  FRecNo := Destination - 1;
  Commit;

  Result := True;

  if NotEmpty then Resync([]);
  if Assigned(AfterMove) then FAfterMove(Self);
end;

function TzTable.ChangeRecords(Source, Destination: Integer): Boolean;
var Buffer1, Buffer2 : PChar;
    Bookmark1, Bookmark2 : TBookmarkFlag;
begin
  Result := False;
  if not EnableRecMove or (TotalCount <= 1) or (Source = Destination) or
     (Source <= 0) or (Destination <= 0) or
     (Source > FRecordCount) or (Destination > FRecordCount) then Exit;

  Buffer1 := AllocRecordBuffer;
  Buffer2 := AllocRecordBuffer;

  _ReadRecord(Buffer1, Source - 1);
  Bookmark1 := GetBookmarkFlag(Buffer1);

  _ReadRecord(Buffer2, Destination - 1);
  Bookmark2 := GetBookmarkFlag(Buffer2);

  SetBookmarkFlag(Buffer1,Bookmark2);
  SetBookmarkFlag(Buffer2,Bookmark1);

  _WriteRecord(Buffer2, Source - 1);
  _WriteRecord(Buffer1, Destination - 1);

  FreeMem(Buffer1);
  FreeMem(Buffer2);

  FRecNo := Destination - 1;

  Result := True;
  FAscend := 0;

  UpdateCursorPos;
  Resync([]);
  if Assigned(AfterMove) then FAfterMove(Self);
end;

// Move record to the specified destination.
function TzTable.MoveCurRecord(Destination: Integer): Boolean;
begin
  Result := MoveRecord(RecNo, Destination);
end;


// TkbmBlob.

{constructor TkbmBlob.Create;
begin
     inherited;
     FLength:=0;
     FBuffer:=nil;
end;

destructor TkbmBlob.Destroy;
begin
     FLength:=0;
     if FBuffer<>nil then
     begin
          FreeMem(FBuffer);
          FBuffer:=nil;
     end;
     inherited;
end;

procedure TkbmBlob.LoadFromStream(stream:TMemoryStream);
begin
     FLength:=stream.Size;
     FBuffer:=Allocmem(FLength);
     stream.Seek(0,soFromBeginning);
     stream.Read(FBuffer^,FLength);
end;

procedure TkbmBlob.SaveToStream(stream:TMemoryStream);
begin
     stream.SetSize(FLength);
     stream.Seek(0,soFromBeginning);
     stream.Write(FBuffer^,FLength);
     stream.Seek(0,soFromBeginning);
end;}

// TkbmBlobStream

// On create, make a stream access to the specified blobfield in the current record.
{constructor TkbmBlobStream.Create(Field:TBlobField;Mode:TBlobStreamMode);
begin
     FField:=Field;
     FFieldNo:=FField.FieldNo;
     FDataSet:=FField.DataSet as TzTable;
     FMode:=Mode;
     FPField:=FDataSet.GetFieldPointer(FDataSet.GetActiveRecordBuffer,Field);
     if FPField=nil then Exit;
     FPBlob:=PkbmBlob(FPField+1);
     if Mode<>bmRead then
     begin
          if FField.ReadOnly then DatabaseErrorFmt('Field %s is readonly.',[FField.DisplayName]);
          if not (FDataSet.State in [dsEdit, dsInsert]) then DatabaseError('Dataset is not in edit mode.');
     end;
     if Mode=bmWrite then
          Truncate
     else
        ReadBlobData;
end;}

// On destroy, update the blobfield in the current record if the blob has changed.
{destructor TkbmBlobStream.Destroy;
begin
     if FModified then
     try
        WriteBlobData;
        FField.Modified:=true;

        FDataSet.DataEvent(deFieldChange,Longint(FField));
     except
        Application.HandleException(Self);
     end;    
     inherited Destroy;
end;}

{procedure TkbmBlobStream.WriteBlobData;
var
   blob:TkbmBlob;
   Stream:TMemoryStream;
begin
     // Get old allocation if any, and free it.
     if FPBlob^<>nil then
     begin
          FPBlob^.free;
          FPBlob^:=nil;
     end;

     // Make new blob with data and enter it into the field
     blob:=TkbmBlob.Create;
     FPBlob^:=blob;

     // If to compress the blob data, do it.
     if Assigned(FDataSet.FOnCompressBlobStream) then
     begin
          Stream:=TMemoryStream.Create;
          try
             FDataSet.FOnCompressBlobStream(self,Stream);
             Stream.Position:=0;
             blob.LoadFromStream(Stream);
          finally
             Stream.free;
          end;
     end
     else
         // Otherwise just save raw data to the inmemory blob.
         blob.LoadFromStream(self);

     // Set Null flag in record.
     if Size<>0 then FPField[0]:=#1
     else FPField[0]:=#0;
end;

procedure TkbmBlobStream.ReadBlobData;
var
   blob:TkbmBlob;
   Stream:TMemoryStream;
begin
     // Get allocation.
     blob:=FPBlob^;
     if blob=nil then exit;

     // If to decompress stream, save the blob in a memory stream and decompress it.
     if Assigned(FDataSet.FOnDeCompressBlobStream) then
     begin
          Stream:=TMemoryStream.Create;
          try
             blob.SaveToStream(Stream);
             FDataSet.FOnDecompressBlobStream(Stream,self);
          finally
             Stream.free;
          end;
     end
     else
         // Copy the data to the stream.
         blob.SaveToStream(self);
     self.Position:=0;
end;

function TkbmBlobStream.Write(const Buffer;Count:Longint): Longint;
begin
     Result:=inherited Write(Buffer,Count);
     if FMode=bmWrite then FModified:=true;
end;

procedure TkbmBlobStream.Truncate;
begin
     Clear;

     // If blob allocated, remove allocation.
     if FPBlob^<>nil then
     begin
          FPBlob^.free;
          FPBlob^:=nil;
     end;

     FModified:=true;
end;}

Procedure TzTable._AppendRecord(Buffer:PChar);
BEGIN
  FStream.Seek (0, soFromEnd);
  FStream.WriteBuffer (Buffer^, FRecordSize);
  Commit;
END;

Procedure TzTable._ReadRecord(Buffer:PChar; IntRecNum:Integer);
BEGIN
  try
    FStream.Position := FDataFileHeader.StartData + (FRecordSize * IntRecNum);
    FStream.ReadBuffer (Buffer^, FRecordSize);
  except
  end;
END;

procedure TzTable.WriteRefreshState;
begin
  FStream.Position := 0;
  FStream.WriteBuffer(FRefreshStateHeader, FRefreshStateHeaderSize);
  Commit;
  UpDateCursorPos;
end;

procedure TzTable.WriteHeader;
begin
  FStream.Position := FRefreshStateHeaderSize + FNetWorkHeaderSize;
  FStream.WriteBuffer(fDataFileHeader, fDataFileHeaderSize);
  Commit;
  UpDateCursorPos;
end;

procedure TzTable.ReadHeader;
begin
  FStream.Position := FRefreshStateHeaderSize + FNetWorkHeaderSize;
  FStream.ReadBuffer(fDataFileHeader, fDataFileHeaderSize);
  UpDateCursorPos;
end;

Procedure TzTable._WriteRecord(Buffer:PChar;IntRecNum:Integer);
BEGIN
  FStream.Position := FDataFileHeader.StartData + (FRecordSize * IntRecNum);
  FStream.WriteBuffer (Buffer^, FRecordSize);
  Commit;
END;

// Quick Search through Sorted Field - Last Item
// 0 - Append
// 1..RecordCount - Founded
// -1..-RecordCount - not Founded and RecNo to Insert
function TzTable.LSearch(FindValue : Variant): Integer;
var FieldNum: Integer;
begin
  Result := FSearch(FindValue);
  if Result > 0 then begin
    FieldNum := FieldDefs.IndexOf(FSortByField);
    while not Eof and (Fields[FieldNum].Value = FindValue) do Next;
    if Eof then Result := 0
    else Result := RecNo - 1;
  end;
end;

// Quick Search through Sorted Field - First Item
// 0 - Append
// 1..RecordCount - Founded
// -1..-RecordCount - not Founded and RecNo to Insert
function TzTable.FSearch(FindValue : Variant): Integer;
  function __FindFirst(FieldNum, Rec_Num, RecStart: Integer): Integer;
  begin
    while Rec_Num >= RecStart do begin
      _ReadRecord(ActiveBuffer, Rec_Num);
      if Fields[FieldNum].Value = FindValue then Dec(Rec_Num)
      else break;
    end;
    if Rec_Num < RecStart then Result := RecStart + 1
    else Result := Rec_Num + 2;
  end;

var Rec_Num, FieldNum, RecStart, RecEnd: Integer;
begin
  Result := 0;
  FieldNum := FieldDefs.IndexOf(FSortByField);
  if (FieldNum = -1) or (RecordCount = 0) then Exit;

  try
    UpdateCursorPos;
    RecEnd := TotalCount - 1;
    RecStart := 0;

    if Fields[FieldNum].Value = FindValue then begin
      Result := __FindFirst(FieldNum, FRecNo - 1, RecStart);
      exit;
    end;

    _ReadRecord(ActiveBuffer, RecStart);
    if Fields[FieldNum].Value > FindValue then begin
      Result := -1;
      exit;
    end
    else if Fields[FieldNum].Value = FindValue then begin
      Result := 1;
      exit;
    end;

    _ReadRecord(ActiveBuffer, RecEnd);
    if Fields[FieldNum].Value < FindValue then begin
      Result := 0;
      exit;
    end
    else if Fields[FieldNum].Value = FindValue then begin
      Result := __FindFirst(FieldNum, RecEnd, RecStart);
      exit;
    end;

    while RecStart < RecEnd do begin
      if RecStart + 10 >= RecEnd then begin
        if RecStart = 0 then Rec_Num := 1
        else Rec_Num := RecStart;
        while Rec_Num <= RecEnd do begin
          _ReadRecord(ActiveBuffer, Rec_Num);
          if Fields[FieldNum].Value < FindValue then Inc(Rec_Num)
          else break;
        end;
        if (Fields[FieldNum].Value = FindValue) then
          Result := Rec_Num + 1
        else Result := -(Rec_Num + 1);
        exit;
      end;

      Rec_Num := (RecEnd + RecStart) div 2;
      _ReadRecord(ActiveBuffer, Rec_Num);
      if Fields[FieldNum].Value = FindValue then begin
        Result := __FindFirst(FieldNum, Rec_Num - 1, RecStart);
        exit;
      end;
      if Fields[FieldNum].Value > FindValue then RecEnd := Rec_Num
      else RecStart := Rec_Num;
    end;
  finally
    FRecNo := Abs(Result) - 1;
    _ReadRecord(ActiveBuffer, FRecNo);
    Resync([]);
  end;
end;

function TzTable.GetTotalCountCount: Integer;
begin
  Result := (FStream.Size - FDataFileHeader.StartData) div FRecordSize;
end;

procedure TzTable.SetExclusive(Value: Boolean);
begin
  CheckInactive;
  FExclusive := Value;
end;

procedure TzTable.SetTemporal(Value:Boolean);
Begin
  if FTemporal <> Value then begin
    CheckInactive;
    FTemporal := Value;
    if not FTemporal then SelfRefresh := False;
  end;
End;

procedure TzTable.OpenCursor(InfoQuery: Boolean);
Begin
  if FTemporal then begin
    try
      if FTmpfile = '' then CreateTmpTable;
     except
       raise;
       exit;
     end;
  end;
  inherited OpenCursor(InfoQuery);
End;

procedure TzTable.CreateTmpTable;
 function  GetUniqueFileName: string;
 var
    a, e:Pchar;
    s: string;
 begin
  a := stralloc(255);
  s := GetTempDirectory + #0;
  e := 'TMP' + #0;

   {$ifdef win32}
      getTempfilename(@s[1],e,0,a);
   {$else}
      getTempfilename(#0,e,1,a);
   {$endif}

  s := changefileext(extractfilename(StrPas(a)), '.' + StrPas(e));
  FtmpFile := Trim(StrPas(a));
  Result := s;
  strdispose(a);
 end;

begin
  CheckInactive;
  FTemporal := True;
  try
    TableName := GetTempDirectory + GetUniqueFileName;
    FTmpFile := TableName;
    CreateTable;
    FSelfRefresh := False;
  except
    raise;
  end;
end;

function TzTable.GetTempDirectory: string;
begin
  SetLength(Result, 255);
  SetLength(Result, GetTempPath(255, PChar(Result)));
end;

procedure TzTable.CloseCursor;
begin
  inherited CloseCursor;
  if FTemporal  and (FTmpFile <> '') then begin
    try
      SysUtils.deletefile(FTmpFile);
      FTmpFile := '';
    except
      raise;
    end;
  end
end;

procedure TzTable.SetFilterText(const Value: string);
const Error_s2 = 'ERROR filter operator' + #10#13;
var i: Integer;
    s, s0, s1, s2, s3: string;
begin
  s := LowerCase(Value);
  s := Trim(s);

  if (FieldDefs.Count = 0) or ((FFilterText = s) and (High(Ar0) > 0)) then begin
    FFilterText := s;
    Exit;
  end;

  Ar0 := Copy(Ar0, 0, 0);     // Operator (and, or....)
  Ar1 := Copy(Ar1, 0, 0);     // Field number 0..x
  Ar2 := Copy(Ar2, 0, 0);     // Operator (+, -, .....)
  Ar3 := Copy(Ar3, 0, 0);     // Field Value

  i := 2;
  while i <= Length(s) do begin
    if (s[i] in ['=', '>', '<']) then begin
      if not (s[i - 1] in ['=', '>', '<']) and (s[i - 1] <> ' ') then
        System.Insert(' ', s, i);
    end
    else if (s[i - 1] in ['=', '>', '<']) and (s[i] <> ' ') then
      System.Insert(' ', s, i);
    Inc(i);
  end;

  while (Length(s) > 0) do begin
    s0 := ''; s1 := ''; s2 := ''; s3 := '';

    i := Pos(' ', s);
    if i = 0 then begin
      s0 := s;
      s := '';
    end
    else begin
      s0 := Copy(s, 1, i - 1);
      System.Delete(s, 1, i);
    end;

    SetLength(Ar0, High(Ar0) + 2);
    if s0 = 'and' then begin
      Ar0[High(Ar0)] := 1;
      i := Pos(' ', s);
      if (i > 0) and (Copy(s, 1, i - 1) = 'not') then begin
        Ar0[High(Ar0)] := 2;
        System.Delete(s, 1, i);
      end;
    end
    else if s0 = 'or' then begin
      Ar0[High(Ar0)] := 3;
      i := Pos(' ', s);
      if (i > 0) and (Copy(s, 1, i - 1) = 'not') then begin
        Ar0[High(Ar0)] := 4;
        System.Delete(s, 1, i);
      end;
    end
    else if s0 = 'not' then Ar0[High(Ar0)] := 5
    else begin
      Ar0[High(Ar0)] := 0;
      s1 := s0;
      s0 := '';
    end;

    if (s0 <> '') and (s = '') then
      raise EzTableError.Create ('ERROR filter sintax.' + #10#13 +
                                        'You Can''t finish filter string by ''' + s0 + '''.');

    if s1 = '' then begin
      i := Pos(' ', s);
      if i = 0 then begin
        s1 := s;
        s := '';
      end
      else begin
        s1 := Copy(s, 1, i - 1);
        System.Delete(s, 1, i);
      end;
    end;

    i := Pos(' ', s);
    if i = 0 then begin
      s2 := s;
      s := '';
    end
    else begin
      s2 := Copy(s, 1, i - 1);
      // Check operator sintax
      if (Length(s2) > 0) and (s2[1] in ['=', '>', '<']) then begin
        System.Delete(s, 1, i);
        if (Length(s2) = 2) and (not (s2[2] in ['=', '>', '<']) or
           (s2[1] = '=') or ((s2[1] = '<') and not (s2[2] in ['>', '='])) or
           ((s2[1] = '>') and (s2[2] <> '='))) then
          raise EzTableError.Create (Error_s2 + s2);
      end
      else s2 := '';
    end;

    if s2 <> '' then begin
      if s[1] = '''' then begin
        System.Delete(s, 1, 1);
        i := Pos('''', s);
        if i = 0 then
          raise EzTableError.Create ('Missing end of string:' + #13 + Value)
        else begin
          s3 := Copy(s, 1, i - 1);
          System.Delete(s, 1, i);
          s := Trim(s);
        end;
      end
      else begin
        i := Pos(' ', s);
        if i = 0 then begin
          s3 := s;
          s := '';
        end
        else begin
          s3 := Copy(s, 1, i - 1);
          System.Delete(s, 1, i);
        end;
      end;
    end;

    i := FieldDefs.IndexOf(s1);
    // set field number
    SetLength(Ar1, High(Ar1) + 2);
    if i = -1 then
      raise EzTableError.Create (FFilterText + #10#13 + 'ERROR filter string' + #10#13 +
                                        'Can''t find field - ' + s1)
    else Ar1[High(Ar1)] := i;

    // Check string sintax
    if (Length(s3) > 0) and (Fields[i].DataType = ftString) then begin
      if s3[1] = '''' then System.Delete(s3, 1, 1);
      if s3[Length(s3)] = '''' then System.Delete(s3, Length(s3), 1);
      //if Length(s3) = 0 then s3 := '';
    end;

    // set operator
    SetLength(Ar2, High(Ar2) + 2);
    Ar2[High(Ar2)] := s2;

    // set value
    SetLength(Ar3, High(Ar3) + 2);
    try
      case Fields[Ar1[High(Ar1)]].DataType of
      ftString: Ar3[High(Ar3)] := s3;
      ftSmallint, ftInteger, ftLargeInt, ftWord: Ar3[High(Ar3)] := StrToInt(s3);
      ftFloat, ftCurrency: begin
                         if (Pos(DecimalSeparator, s3) = 0) and (Pos('.', s3) > 0) then
                            s3[Pos('.', s3)] := DecimalSeparator;
                         if (Pos(DecimalSeparator, s3) = 0) and (Pos(',', s3) > 0) then
                            s3[Pos(',', s3)] := DecimalSeparator;
                         Ar3[High(Ar3)] := StrToFloat(s3);
                         end;
      ftDateTime: begin // check syntax??
                    if s3 <> '' then
                      Ar3[High(Ar3)] := StrToDateTime(s3)
                    else Ar3[High(Ar3)] := 0.0;
                  end;
      ftDate: begin // check syntax??
                if s3 <> '' then
                  Ar3[High(Ar3)] := StrToDate(s3)
                else Ar3[High(Ar3)] := 0;
              end;
      ftTime: if s3 <> '' then
                Ar3[High(Ar3)] := StrToTime(s3)
              else Ar3[High(Ar3)] := 0;
      else Ar3[High(Ar3)] := null;
      end;
    except
      raise EzTableError.Create ('ERROR assign value' + #10#13 + s3);
      Exit;
    end;
  end;
//  FFilterText := Value;
  FFilterText := LowerCase(Trim(Value));
  if IsFiltered then SetFiltered(True);
end;

function TzTable.FilterByFilter: Boolean;
var i: Integer;
    SubResult: Boolean;
begin
  Result := True;
  SubResult := True;

  for i := 0 to High(Ar0) do begin
    if (not Result and (Ar0[i] in [1, 2])) or
       (Result and (Ar0[i] in [3, 4])) then break;
    try
      if Ar2[i] = '' then begin  // Boolean
        case Fields[Ar1[i]].DataType of
        ftBoolean: SubResult := Fields[Ar1[i]].AsBoolean;
        ftString: SubResult := Fields[Ar1[i]].AsString = '';
        ftSmallint, ftInteger, ftLargeInt, ftWord:
             SubResult := Fields[Ar1[i]].AsInteger <> 0;
        ftFloat, ftCurrency: SubResult := Fields[Ar1[i]].AsFloat <> 0;
        else SubResult := False;
        end;
      end
      else if Ar2[i] = '=' then begin  // '='
        case Fields[Ar1[i]].DataType of
        ftString: SubResult := LowerCase(Fields[Ar1[i]].AsString) = Ar3[i];
        ftSmallint, ftInteger, ftLargeInt, ftWord:
             SubResult := Fields[Ar1[i]].AsInteger = Ar3[i];
        ftFloat, ftCurrency: SubResult := Fields[Ar1[i]].AsFloat = Ar3[i];
        ftDateTime: SubResult := Fields[Ar1[i]].AsDateTime = Ar3[i];
        ftDate: SubResult := Fields[Ar1[i]].AsDateTime = Ar3[i];
        ftTime: SubResult := Fields[Ar1[i]].AsDateTime = Ar3[i];
        else SubResult := False;
        end;
      end
      else if Ar2[i][1] = '>' then begin  // '>' or '>='
        if Length(Ar2) = 1 then
          case Fields[Ar1[i]].DataType of
          ftString: SubResult := LowerCase(Fields[Ar1[i]].AsString) > Ar3[i];
          ftSmallint, ftInteger, ftLargeInt, ftWord:
               SubResult := Fields[Ar1[i]].AsInteger > Ar3[i];
          ftFloat, ftCurrency: SubResult := Fields[Ar1[i]].AsFloat > Ar3[i];
          ftDateTime: SubResult := Fields[Ar1[i]].AsDateTime > Ar3[i];
          ftDate: SubResult := Fields[Ar1[i]].AsDateTime > Ar3[i];
          ftTime: SubResult := Fields[Ar1[i]].AsDateTime > Ar3[i];
          else SubResult := False;
          end
        else
          case Fields[Ar1[i]].DataType of
          ftString: SubResult := LowerCase(Fields[Ar1[i]].AsString) >= Ar3[i];
          ftSmallint, ftInteger, ftLargeInt, ftWord:
               SubResult := Fields[Ar1[i]].AsInteger >= Ar3[i];
          ftFloat, ftCurrency: SubResult := Fields[Ar1[i]].AsFloat >= Ar3[i];
          ftDateTime: SubResult := Fields[Ar1[i]].AsDateTime >= Ar3[i];
          ftDate: SubResult := Fields[Ar1[i]].AsDateTime >= Ar3[i];
          ftTime: SubResult := Fields[Ar1[i]].AsDateTime >= Ar3[i];
          else SubResult := False;
          end;
      end
      else if Ar2[i][1] = '<' then begin  // '<' or '<=' or '<>'
        if Length(Ar2[i]) = 1 then
          case Fields[Ar1[i]].DataType of
          ftString: SubResult := LowerCase(Fields[Ar1[i]].AsString) < Ar3[i];
          ftSmallint, ftInteger, ftLargeInt, ftWord:
               SubResult := Fields[Ar1[i]].AsInteger < Ar3[i];
          ftFloat, ftCurrency: SubResult := Fields[Ar1[i]].AsFloat < Ar3[i];
          ftDateTime: SubResult := Fields[Ar1[i]].AsDateTime < Ar3[i];
          ftDate: SubResult := Fields[Ar1[i]].AsDateTime < Ar3[i];
          ftTime: SubResult := Fields[Ar1[i]].AsDateTime < Ar3[i];
          else SubResult := False;
          end
        else if Ar2[i][2] = '=' then
          case Fields[Ar1[i]].DataType of
          ftString: SubResult := LowerCase(Fields[Ar1[i]].AsString) <= Ar3[i];
          ftSmallint, ftInteger, ftLargeInt, ftWord:
               SubResult := Fields[Ar1[i]].AsInteger <= Ar3[i];
          ftFloat, ftCurrency: SubResult := Fields[Ar1[i]].AsFloat <= Ar3[i];
          ftDateTime: SubResult := Fields[Ar1[i]].AsDateTime <= Ar3[i];
          ftDate: SubResult := Fields[Ar1[i]].AsDateTime <= Ar3[i];
          ftTime: SubResult := Fields[Ar1[i]].AsDateTime <= Ar3[i];
          else SubResult := False;
          end
        else if Ar2[i][2] = '>' then
          case Fields[Ar1[i]].DataType of
          ftString: SubResult := LowerCase(Fields[Ar1[i]].AsString) <> Ar3[i];
          ftSmallint, ftInteger, ftLargeInt, ftWord:
               SubResult := Fields[Ar1[i]].AsInteger <> Ar3[i];
          ftFloat, ftCurrency: SubResult := Fields[Ar1[i]].AsFloat <> Ar3[i];
          ftDateTime: SubResult := Fields[Ar1[i]].AsDateTime <> Ar3[i];
          ftDate: SubResult := Fields[Ar1[i]].AsDateTime <> Ar3[i];
          ftTime: SubResult := Fields[Ar1[i]].AsDateTime <> Ar3[i];
          else SubResult := False;
          end;
      end;
      case Ar0[i] of
      0: Result := SubResult;
      1: Result := Result and SubResult;
      2: Result := Result and not SubResult;
      3: Result := Result or SubResult;
      4: Result := Result or not SubResult;
      5: Result := not SubResult;
      end;
    except
      raise EzTableError.Create ('ERROR filter opration');
    end;
  end;
end;

procedure TzTable.SetFiltered(Value: Boolean);
begin
  FFiltered := Value;
  FFilterCount := -1;
  if Active then Resync([]);
end;

procedure TzTable.DoBeforeOpen;
begin
  if not FIsOpen and not FTemporal then FieldDefs.Clear;
  if not FIsOpen and FTemporal and (FieldDefs.Count = 0) and Assigned(FBeforeCreate) then
      FBeforeCreate(Self);
  inherited;
end;

function TzTable.InsEdit(AValue: Variant): Integer;
var r: Integer;
begin
  if FSortByField = '' then
    raise EzTableError.Create ('Sorted Field name is empty');

  if FieldDefs.IndexOf(FSortByField) = -1 then
    raise EzTableError.Create ('Cant find Sorted Field: ' + FSortByField);

  UpdateCursorPos;
  Filtered := False;
  r := FSearch(AValue);

  if r > 0 then begin
    Result := 1;
    Edit;
  end
  else if r = 0 then begin
    Result := 2;
    Append;
  end
  else begin
    Result := 3;
    Insert;
  end;
  if Result > 1 then FieldByName(FSortByField).Value := AValue;
end;

function TzTable.CopyRecord(Destination: Integer): Boolean;
begin
  _WriteRecord(ActiveBuffer, Destination - 1);
  Result := True;
end;

procedure TzTable.DoBeforePost;
begin
  inherited;
end;

procedure TzTable.DoChanges;
begin
  ReadHeader;
  FModified := True;
  FDataFileHeader.Modified := True;
  FDataFileHeader.LastModified := Now;
  SetCompressed(False);      // writes header

  ModifiedNum := GetModifiedNum + 1;
  if ModifiedNum >= High(FRefreshStateHeader.ModifiedNum) then ModifiedNum := 1;
  FRefreshStateHeader.ModifiedNum := ModifiedNum;
  FRefreshStateHeader.Msg_Num := MsgNum;
  WriteRefreshState;

  ReadNetRecord;
  FNetRecord.LockRecNum := -1;
  WriteNetRecord;

  if MyLock <> tsLock then SetTabState(tsWait);
  FChangeTimer.Enabled := FSelfRefresh;
end;

procedure TzTable.DoAfterPost;
begin
  if BefRecNo > 0  then MoveRecord(TotalCount, BefRecNo);
  BefRecNo := 0;
  DoChanges;
  inherited;
end;

procedure TzTable.DoAfterDelete;
begin
  DoChanges;
  inherited;
end;

function TzTable.DeleteRecord(RecN: Integer): Boolean;
var WasFiltered: Boolean;
    OldRec: Integer;
begin
  Result := False;
  CheckActive;
  WasFiltered := FFiltered;
  UpDateCursorPos;
  OldRec := FRecNo;
  try
    DisableControls;
    SetFiltered(False);
    if IsEmpty then Exit;
    RecNo := RecN;
    if RecNo = RecN then begin
      InternalDelete;
      Result := True;
    end;
  finally
    SetFiltered(WasFiltered);
    FRecNo := OldRec;
    EnableControls;
  end;
end;

procedure TzTable.DeleteRecords(AField: TField; AValue: Variant);
var OldRec: Integer;
begin
  CheckActive;
  if IsEmpty then Exit;
  UpDateCursorPos;
  OldRec := FRecNo;
  if not Search(AField, AValue, True) then Exit;

  repeat
    if FRecNo < OldRec then Dec(OldRec);
    InternalDelete;
    _InternalPrior;
  until not Search(AField, AValue, False);

  if OldRec < FRecordCount then FRecNo := OldRec
  else _InternalLast;
  Resync([]);
end;

function  TzTable.NotEmpty: Boolean;
begin
  Result := not BOF or not EOF;
end;

function  TzTable.IsEmpty: Boolean;
begin
  Result := BOF and EOF
end;

procedure TzTable.DoBeforeRefresh;
var i: Integer;
begin
//  if Assigned(BeforeRefresh) then BeforeRefresh(Self);
  if ModifiedNum = GetModifiedNum then begin
    if Assigned(BeforeRefresh) then BeforeRefresh(Self);
    //inherited;
    Exit;
  end;
  FChangeTimer.Enabled := False;
  UpdateCursorPos;
  i := RecNo;

  FRecordCount := TotalCount;
  if (FRecordCount > 0) and (FSortByField <> '') then
    FBeforeRefreshValue := FieldByName(FSortByField).Value
  else FBeforeRefreshValue := null;
  //inherited;
  FRecordCount := TotalCount;
  if i > FRecordCount then _InternalLast;
  if Assigned(BeforeRefresh) then BeforeRefresh(Self);
end;

procedure TzTable.DoAfterRefresh;
begin
  if ModifiedNum = GetModifiedNum then begin
    if Assigned(AfterRefresh) then AfterRefresh(Self);
    Exit;
  end;
  UpdateCursorPos;
  FFilterCount := -1;
  if (FRecordCount > 0) and (FSortByField <> '') and
      (FBeforeRefreshValue <> null) then FSearch(FBeforeRefreshValue);
  //inherited;
  FChangeTimer.Enabled := FSelfRefresh;
  if Assigned(AfterRefresh) then AfterRefresh(Self);
end;

procedure TzTable.SetModified(Value: Boolean);
begin
  if FModified <> Value then begin
    fDataFileHeader.Modified := Value;
    fDataFileHeader.LastModified := Now;
    WriteHeader;
    FModified := Value;
  end;
end;

procedure TzTable.SetCompressed(Value: Boolean);
begin
  fDataFileHeader.Compressed := Value;
  WriteHeader;
end;

function TzTable.GetCompressed: Boolean;
begin
  ReadHeader;
  Result := fDataFileHeader.Compressed;
end;

function TzTable.GetModified: Boolean;
begin
  ReadHeader;
  Result := fDataFileHeader.Modified;
end;

procedure TzTable.Sort(AField: TField);
var FList: array of string;
begin
  if not FEnableSort or not Active or IsEmpty or not Assigned(AField) then Exit;
  SetLength(FList, 1);
  FList[0] := AField.FieldName;
  SortTable(FList);
  SortByField := AField.Name;
end;

procedure TzTable.OnChangeTimer(Sender: TObject);
var i: SmallInt;
begin
  FChangeTimer.Enabled := False;
  try
    try
      i := GetModifiedNum;
    except
      Exit;
    end;
    // Confirm User Count
    if UserConfirm <> FRefreshStateHeader.UserConfirm then begin
      ReadNetRecord;
      if not FNetRecord.Busy then begin
        FNetRecord.Busy := True;
        WriteNetRecord;
        IncreaseUserConfirm;
      end;

      // Confirm Lock
      if Boolean(MyLock) then begin
        FRefreshStateHeader.State := MyLock;
        WriteRefreshState;
      end;
      if Assigned(OnUserCountChange) then FOnUserCountChange(Self);
    end;

    // NetWork Refresh
    if (ModifiedNum <> i) and (FRefreshStateHeader.State <> tsLock) then begin
      MsgNum := FRefreshStateHeader.Msg_Num;
      //Refresh;
      ModifiedNum := i;
    end;
    if FStoredInt <> FRefreshStateHeader.Stored_Int then begin
      FStoredInt := FRefreshStateHeader.Stored_Int;
      if Assigned(OnStoredIntChange) then FOnStoredIntChange(Self);
    end;
  finally
    FChangeTimer.Enabled := FSelfRefresh;
  end;
end;

procedure TzTable.SetRefreshDelay(Value: Integer);
begin
  FChangeTimer.Interval := Value;
end;

function TzTable.GetRefreshDelay: Integer;
begin
  Result := FChangeTimer.Interval;
end;

function TzTable.GetModifiedNum: SmallInt;
begin
  FStream.Seek(0, soFromBeginning);
  FStream.ReadBuffer(FRefreshStateHeader, FRefreshStateHeaderSize);
  Result := FRefreshStateHeader.ModifiedNum;
end;

function TzTable.GetState: TTabState;
begin
  FStream.Seek(0, soFromBeginning);
  FStream.ReadBuffer(FRefreshStateHeader, FRefreshStateHeaderSize);
  Result := FRefreshStateHeader.State;
end;

function TzTable.GetComputerCount: Integer;
begin
  FStream.Seek(0, soFromBeginning);
  FStream.ReadBuffer(FRefreshStateHeader, FRefreshStateHeaderSize);
  Result := FRefreshStateHeader.Computer_Count;
end;

function TzTable.GetUserCount: Byte;
var i, cc: Byte;
begin
  Result := 0;
  cc := ComputerCount;
  for i := 0 to cc - 1 do begin
    ReadNetRecord(i);
    if FNetRecord.Busy then
      Result := Result + 1;
  end;
end;

procedure TzTable.SetSelfRefresh(Value: Boolean);
begin
  if Value <> FSelfRefresh then begin
    FSelfRefresh := Value;
    if not FSelfRefresh then FChangeTimer.Enabled := False
    else if FIsOpen then FChangeTimer.Enabled := True;
  end;
end;

procedure TzTable.CheckLock;
var i, cc: Integer;
begin
  CheckActive;
  if MyLock = tsWait then begin
    if TabState = tsLock then
      raise EzTableError.Create (ExtractFileName(TableName) + ': Table is Lock')
    else if TabState = tsBusy then
      raise EzTableError.Create (ExtractFileName(TableName) + ': Table is Busy');
  end;

  FChangeTimer.Enabled := False;
  cc := ComputerCount - 1;
  for i := 0 to cc do begin
    if i = MyNum then Continue;
    ReadNetRecord(i);
    with FNetRecord do
      if Busy and
            ((LockRecNum = FRecNo + 1) or
             ((FieldIndex < Fields.Count) and
              (LockValue = Fields[FieldIndex].AsString))) then begin
        SetTabState(tsWait);
        FChangeTimer.Enabled := SelfRefresh;
        raise EzTableError.Create ('Record is Lock');
      end;
  end;
  ReadNetRecord;
  FNetRecord.LockRecNum := FRecNo + 1;
  WriteNetRecord;
end;

procedure TzTable.DoBeforeEdit;
begin
  CheckLock;
  UpDateCursorPos;
  inherited;
end;

procedure TzTable.DoBeforeDelete;
begin
  CheckLock;
  UpDateCursorPos;
  inherited;
end;

procedure TzTable.DoBeforeInsert;
begin
  CheckLock;
  UpDateCursorPos;
  inherited;
end;

function TzTable.Lock: Boolean;
begin
  CheckActive;
  Result := SetTabState(tsLock);
end;

function TzTable.UnLock: Boolean;
begin
  CheckActive;
  Result := SetTabState(tsWait);
  FChangeTimer.Enabled := FSelfRefresh;
  if Result then Commit;
end;

function TzTable.LockByValue(FldIndex: Integer; Value: string): Boolean;
var i, cc: Integer;
begin
  Result := False;
  CheckActive;
  if FldIndex >= Fields.Count then Exit;

  cc := ComputerCount - 1;
  for i := 0 to cc do begin
    if i = MyNum then Continue;
    ReadNetRecord(i);
    with FNetRecord do
      if (FieldIndex = FldIndex) and (LockValue = Value) then begin
         if (GetUserCount > 1) then Exit
         else begin
           FieldIndex := 255;
           LockValue := '     ';
         end;
      end;
  end;
  ReadNetRecord;
  FNetRecord.FieldIndex := FldIndex;
  StrCopy(FNetRecord.LockValue, PChar(Value));
  WriteNetRecord;
  Result := True;
end;

procedure TzTable.UnLockByValue;
begin
  CheckActive;
  ReadNetRecord;
  FNetRecord.FieldIndex := 255;
  WriteNetRecord;
end;

procedure TzTable.ClearLocks;
var i, cc: Integer;
begin
  if not FIsOpen then Exit;

  FRefreshStateHeader.State := tsWait;
  IncreaseUserConfirm;

  cc := ComputerCount - 1;
  for i := 0 to cc do
    if MyNum <> i then begin
      ReadNetRecord(i);
      FNetRecord.Busy := False;
      WriteNetRecord(i);
    end;
end;

function TzTable.SetTabState(Value: TTabState): Boolean;
var ts: TTabState;
begin
  if not FTemporal and (UserConfirm <> FRefreshStateHeader.UserConfirm) then
    OnChangeTimer(nil);

  ts := TabState;
  if Value = MyLock then begin
    Result := ts = Value;
    Exit;
  end;

  if (ts = tsWait) or (MyLock = ts) then begin
    FRefreshStateHeader.State := Value;
    WriteRefreshState;
    MyLock := Value;
    Result := True;
  end
  else Result := False;
  UpDateCursorPos;
end;

procedure TzTable.SetTableKind(Value: Word);
begin
  if FTableKind <> Value then begin
    FTableKind := Value;
    if FIsOpen then begin
      FDataFileHeader.TableKind := Value;
      WriteHeader;
    end;
  end;
end;

procedure TzTable.SetPassword(Value: string);
begin
  if FPassword <> Value then begin
    FPassword := Value;
    if FIsOpen then begin
      FDataFileHeader.Password := Value;
      WriteHeader;
    end;
  end;
end;


procedure TzTable.SetOptions(Value: string);
begin
  if FOptions <> Value then begin
    FOptions := Value;
    if FIsOpen then begin
      FDataFileHeader.Options := Value;
      WriteHeader;
    end;
  end;
end;

procedure TzTable.SetStoredInt(Value: Integer);
begin
  if FStoredInt <> Value then begin
    FStoredInt := Value;
    FRefreshStateHeader.Stored_Int := Value;
    WriteRefreshState;
  end;
end;

function TzTable.GetCreated: TDateTime;
begin
  Result := FDataFileHeader.Created;
end;

function TzTable.GetLastModified: TDateTime;
begin
  ReadHeader;
  Result := FDataFileHeader.LastModified;
end;

procedure TzTable.IncreaseUserConfirm;
begin
  if FRefreshStateHeader.UserConfirm >= High(FRefreshStateHeader.UserConfirm) then
    FRefreshStateHeader.UserConfirm := 0
  else Inc(FRefreshStateHeader.UserConfirm);
  UserConfirm := FRefreshStateHeader.UserConfirm;
  WriteRefreshState;
end;

procedure TzTable.Commit;
begin
  if Active and not FTemporal and (TabState <> tsLock) then
    FlushFileBuffers(FStream.Handle);
end;

// Compare two records.  Returns -1 if REC1 < REC2, 0 if REC1 = REC2, or
// 1 if REC1 > REC2.
Function TzTable._CompareRecords(SortFields:Array of String;Rec1,Rec2:Integer):Integer; FAR;
{-Compare the records Rec1, Rec2 and return -1 if Rec1 < Rec2, 0 if Rec1 = Rec2,
  1 if Rec1 > Rec2 }
VAR
  IX : Integer;

  Function CompareHelper(KeyId:String;Rec1,Rec2:Integer):Integer;
  VAR
    SKey1, SKey2 : String;
    IKey1, IKey2 : Integer;
    fKey1, fKey2 : Double;
    dKey1, dKey2 : tDateTime;
    CompareType : tFieldType;
    KeyField : tField;
  BEGIN
    KeyField := FieldByName(KeyID);
    CompareType := KeyField.DataType;
    Case CompareType of
      ftFloat,
      ftCurrency,
      ftBCD :
        BEGIN
          _ReadRecord(ActiveBuffer,Rec1-1);
          fKey1 := KeyField.AsFloat;
          _ReadRecord(ActiveBuffer,Rec2-1);
          fKey2 := KeyField.AsFloat;
          if fKey1 < fKey2 then
            Result := -1
          else
            if fKey1 > fKey2 then
              Result := 1
            else
              Result := 0;
        END;
      ftSmallInt,
      ftInteger,
      ftLargeInt,
      ftWord :
        BEGIN
          _ReadRecord(ActiveBuffer,Rec1-1);
          IKey1 := KeyField.AsInteger;
          _ReadRecord(ActiveBuffer,Rec2-1);
          IKey2 := KeyField.AsInteger;
          if IKey1 < IKey2 then
            Result := -1
          else
            if IKey1 > IKey2 then
              Result := 1
            else
              Result := 0;
        END;
      ftDate,
      ftTime,
      ftDateTime :
        BEGIN
          _ReadRecord(ActiveBuffer,Rec1-1);
          dKey1 := KeyField.AsDateTime;
          _ReadRecord(ActiveBuffer,Rec2-1);
          dKey2 := KeyField.AsDateTime;
          if dKey1 < dKey2 then
            Result := -1
          else
            if dKey1 > dKey2 then
              Result := 1
            else
              Result := 0;
        END;
      else
        BEGIN
          _ReadRecord(ActiveBuffer,Rec1-1);
          SKey1 := KeyField.AsString;
          _ReadRecord(ActiveBuffer,Rec2-1);
          SKey2 := KeyField.AsString;
          if SKey1 < SKey2 then
            Result := -1
          else
            if SKey1 > SKey2 then
              Result := 1
            else
              Result := 0;
        END;
    END;
  END;

BEGIN
  IX := 0;
  REPEAT // Loop through all available sortfields until not equal or no more sort fiels.
    Result := CompareHelper(SortFields[IX],Rec1,Rec2);
    Inc(IX);
  UNTIL (Result <> 0) or (IX > High(SortFields));
END;

// Enhancement: Sort the table by the fields passed.
Procedure TzTable.SortTable(SortFields : Array of String; Increment: Boolean = True);

  { This is the main sorting routine. It is passed the number of elements and the
    two callback routines. The first routine is the function that will perform
    the comparison between two elements. The second routine is the procedure that
    will swap two elements if necessary } // Source: UNDU #8

  procedure QSort(uNElem: Integer);
  { uNElem - number of elements to sort }

    procedure qSortHelp(pivotP: Integer; nElem: word);
    label
      TailRecursion,
      qBreak;
    var
      leftP, rightP, pivotEnd, pivotTemp, leftTemp: word;
      lNum: Integer;
      retval: integer;
    begin
      TailRecursion:
        if (nElem <= 2) then
          begin
            if (nElem = 2) then

              begin
                rightP := pivotP +1;
                if (_CompareRecords(SortFields,pivotP, rightP) > 0) then
                  _SwapRecords(pivotP, rightP);
              end;
            exit;
          end;
        rightP := (nElem -1) + pivotP;
        leftP :=  (nElem shr 1) + pivotP;
        { sort pivot, left, and right elements for "median of 3" }
        if (_CompareRecords(SortFields,leftP, rightP) > 0) then _SwapRecords(leftP, rightP);
        if (_CompareRecords(SortFields,leftP, pivotP) > 0) then _SwapRecords(leftP, pivotP)
        else if (_CompareRecords(SortFields,pivotP, rightP) > 0) then _SwapRecords(pivotP, rightP);

        if (nElem = 3) then
          begin
            _SwapRecords(pivotP, leftP);
            exit;
          end;
        { now for the classic Horae algorithm }
        pivotEnd := pivotP + 1;
        leftP := pivotEnd;
        repeat
          retval := _CompareRecords(SortFields,leftP, pivotP);
//k          if not Increment then retval := - retval;
          while (retval <= 0) do
            begin
              if (retval = 0) then
                begin
                  _SwapRecords(LeftP, PivotEnd);
                  Inc(PivotEnd);
                end;
              if (leftP < rightP) then
                Inc(leftP)
              else
                goto qBreak;
              retval := _CompareRecords(SortFields,leftP, pivotP);
            end; {while}
          while (leftP < rightP) do
            begin

              retval := _CompareRecords(SortFields,pivotP, rightP);
              if (retval < 0) then
                Dec(rightP)
              else
                begin
                  _SwapRecords(leftP, rightP);
                  if (retval <> 0) then
                    begin
                      Inc(leftP);
                      Dec(rightP);
                    end;
                  break;
                end;
            end; {while}

        until (leftP >= rightP);
      qBreak:
        if (_CompareRecords(SortFields,leftP, pivotP) <= 0) then Inc(leftP);
        leftTemp := leftP -1;
        pivotTemp := pivotP;
        while ((pivotTemp < pivotEnd) and (leftTemp >= pivotEnd)) do
          begin
            _SwapRecords(pivotTemp, leftTemp);
            Inc(pivotTemp);
            Dec(leftTemp);
          end; {while}
        lNum := (leftP - pivotEnd);
        nElem := ((nElem + pivotP) -leftP);

        if (nElem < lNum) then
          begin
            qSortHelp(leftP, nElem);
            nElem := lNum;
          end
        else
          begin
            qSortHelp(pivotP, lNum);
            pivotP := leftP;
          end;
        goto TailRecursion;
      end; {qSortHelp }

  begin
    if (uNElem < 2) then  exit; { nothing to sort }
    qSortHelp(1, uNElem);
  end; { QSort }


BEGIN
  CheckActive;
  if fReadOnly then Exit;
  Screen.Cursor := crHourGlass;
  DisableControls;
  try
    QSort(FRecordCount {+ fDeletedCount});
    First;
  finally
    EnableControls;
    Screen.Cursor := crDefault;
  end;
END;

Procedure TzTable._SwapRecords(Rec1,REc2:Integer);
VAR
  Buffer1, Buffer2 : PChar;
  Bookmark1, BOokmark2 : TBookmarkFlag;
BEGIN
  Rec1 := Rec1 - 1;
  Rec2 := Rec2 - 1;
  if Rec1 < 0 then Exit;
  if Rec2 < 0 then Exit;
  Buffer1 := AllocRecordBuffer;
  Buffer2 := AllocRecordBuffer;
  _ReadRecord(Buffer1,Rec1);
  _ReadRecord(Buffer2,Rec2);
  Bookmark1 := GetBookmarkFlag(Buffer1);
  Bookmark2 := GetBookmarkFlag(Buffer2);
  SetBookmarkFlag(Buffer1,Bookmark2);
  SetBookmarkFlag(Buffer2,Bookmark1);
  _WriteRecord(Buffer1,Rec2);
  _WriteRecord(Buffer2,Rec1);
  FreeMem(Buffer1);
  FreeMem(Buffer2);
END;

procedure TzTable.GetComputerList(SL: TStrings);
var i, cc: Byte;
begin
  CheckActive;
  SL.Clear;
  cc := ComputerCount;
  for i := 0 to cc - 1 do begin
    ReadNetRecord(i);
    if FNetRecord.Busy then
      SL.Add(FNetRecord.ComputerName);
  end;
end;

procedure TzTable.GetUserList(SL: TStrings);
var i, cc: Byte;
begin
  CheckActive;
  SL.Clear;
  cc := ComputerCount;
  for i := 0 to cc - 1 do begin
    ReadNetRecord(i);
    if FNetRecord.Busy then
      SL.Add(FNetRecord.UserName);
  end;
end;

procedure TzTable.GetTableUserList(SL: TStrings);
var i, cc: Byte;
begin
  CheckActive;
  SL.Clear;
  cc := ComputerCount;
  for i := 0 to cc - 1 do begin
    ReadNetRecord(i);
    if FNetRecord.Busy then
      SL.Add(FNetRecord.TableUserName);
  end;
end;

procedure TzTable.GetUserListWithEntryDate(SL: TStrings);
var i, cc: Byte;
begin
  CheckActive;
  SL.Clear;
  cc := ComputerCount;
  for i := 0 to cc - 1 do begin
    ReadNetRecord(i);
    if FNetRecord.Busy then
      SL.Add(FNetRecord.UserName + '=' + DateTimeToStr(FNetRecord.EntryDate));
  end;
end;

procedure TzTable.GetTableUserListWithComputerName(SL: TStrings);
var i, cc: Byte;
begin
  CheckActive;
  SL.Clear;
  cc := ComputerCount;
  for i := 0 to cc - 1 do begin
    ReadNetRecord(i);
    if FNetRecord.Busy then
      SL.Add(FNetRecord.TableUserName + '=' + FNetRecord.ComputerName);
  end;
end;

procedure TzTable.InternalRefresh;
begin
  if not FTemporal then
    OnChangeTimer(nil);
end;

procedure TzTable.GetTableUserListWithEntryDate(SL: TStrings);
var i, cc: Byte;
begin
  CheckActive;
  SL.Clear;
  cc := ComputerCount;
  for i := 0 to cc - 1 do begin
    ReadNetRecord(i);
    if FNetRecord.Busy then
      SL.Add(FNetRecord.TableUserName + '=' + DateTimeToStr(FNetRecord.EntryDate));
  end;
end;

procedure TzTable.Refresh;
begin
  DoBeforeRefresh;
  UpdateCursorPos;
  try
    InternalRefresh;
  finally
    Resync([]);
    DoAfterRefresh;
  end;
end;

procedure TzTable.InternalAddRecord(Buffer: Pointer; Append: Boolean);
begin
//  _AppendRecord(ActiveBuffer);
end;

function Q_PosText(const FindString, SourceString: string; StartPos: Integer): Integer;
asm
        PUSH    ESI
        PUSH    EDI
        PUSH    EBX
        TEST    EAX,EAX
        JE      @@qt
        TEST    EDX,EDX
        JE      @@qt0
        MOV     ESI,EAX
        MOV     EDI,EDX
        PUSH    EDX
        MOV     EAX,[EAX-4]
        MOV     EDX,[EDX-4]
        DEC     EAX
        SUB     EDX,EAX
        DEC     ECX
        PUSH    EAX
        SUB     EDX,ECX
        JNG     @@qtx
        ADD     EDI,ECX
        MOV     ECX,EDX
        MOV     EDX,EAX
        MOVZX   EBX,BYTE PTR [ESI]
        MOV     AL,BYTE PTR [EBX+ToUpperChars]
@@lp1:  MOVZX   EBX,BYTE PTR [EDI]
        CMP     AL,BYTE PTR [EBX+ToUpperChars]
        JE      @@uu
@@fr:   INC     EDI
        DEC     ECX
        JE      @@qtx
        MOVZX   EBX,BYTE PTR [EDI]
        CMP     AL,BYTE PTR [EBX+ToUpperChars]
        JE      @@uu
        INC     EDI
        DEC     ECX
        JE      @@qtx
        MOVZX   EBX,BYTE PTR [EDI]
        CMP     AL,BYTE PTR [EBX+ToUpperChars]
        JE      @@uu
        INC     EDI
        DEC     ECX
        JE      @@qtx
        MOVZX   EBX,BYTE PTR [EDI]
        CMP     AL,BYTE PTR [EBX+ToUpperChars]
        JE      @@uu
        INC     EDI
        DEC     ECX
        JNE     @@lp1
@@qtx:  ADD     ESP,$08
@@qt0:  XOR     EAX,EAX
@@qt:   POP     EBX
        POP     EDI
        POP     ESI
        RET
@@ms:   MOVZX   EBX,BYTE PTR [ESI]
        MOV     AL,BYTE PTR [EBX+ToUpperChars]
        MOV     EDX,[ESP]
        JMP     @@fr
@@uu:   TEST    EDX,EDX
        JE      @@fd
@@lp2:  MOV     BL,BYTE PTR [ESI+EDX]
        MOV     AH,BYTE PTR [EDI+EDX]
        CMP     BL,AH
        JE      @@eq
        MOV     AL,BYTE PTR [EBX+ToUpperChars]
        MOVZX   EBX,AH
        CMP     AL,BYTE PTR [EBX+ToUpperChars]
        JNE     @@ms
@@eq:   DEC     EDX
        JNZ     @@lp2
@@fd:   LEA     EAX,[EDI+1]
        POP     ECX
        SUB     EAX,[ESP]
        POP     ECX
        POP     EBX
        POP     EDI
        POP     ESI
end;

function Q_PosStr(const FindString, SourceString: string; StartPos: Integer): Integer;
asm
        PUSH    ESI
        PUSH    EDI
        PUSH    EBX
        PUSH    EDX
        TEST    EAX,EAX
        JE      @@qt
        TEST    EDX,EDX
        JE      @@qt0
        MOV     ESI,EAX
        MOV     EDI,EDX
        MOV     EAX,[EAX-4]
        MOV     EDX,[EDX-4]
        DEC     EAX
        SUB     EDX,EAX
        DEC     ECX
        SUB     EDX,ECX
        JNG     @@qt0
        XCHG    EAX,EDX
        ADD     EDI,ECX
        MOV     ECX,EAX
        JMP     @@nx
@@fr:   INC     EDI
        DEC     ECX
        JE      @@qt0
@@nx:   MOV     EBX,EDX
        MOV     AL,BYTE PTR [ESI]
@@lp1:  CMP     AL,BYTE PTR [EDI]
        JE      @@uu
        INC     EDI
        DEC     ECX
        JE      @@qt0
        CMP     AL,BYTE PTR [EDI]
        JE      @@uu
        INC     EDI
        DEC     ECX
        JE      @@qt0
        CMP     AL,BYTE PTR [EDI]
        JE      @@uu
        INC     EDI
        DEC     ECX
        JE      @@qt0
        CMP     AL,BYTE PTR [EDI]
        JE      @@uu
        INC     EDI
        DEC     ECX
        JNE     @@lp1
@@qt0:  XOR     EAX,EAX
@@qt:   POP     ECX
        POP     EBX
        POP     EDI
        POP     ESI
        RET
@@uu:   TEST    EDX,EDX
        JE      @@fd
@@lp2:  MOV     AL,BYTE PTR [ESI+EBX]
        CMP     AL,BYTE PTR [EDI+EBX]
        JNE     @@fr
        DEC     EBX
        JE      @@fd
        MOV     AL,BYTE PTR [ESI+EBX]
        CMP     AL,BYTE PTR [EDI+EBX]
        JNE     @@fr
        DEC     EBX
        JE      @@fd
        MOV     AL,BYTE PTR [ESI+EBX]
        CMP     AL,BYTE PTR [EDI+EBX]
        JNE     @@fr
        DEC     EBX
        JE      @@fd
        MOV     AL,BYTE PTR [ESI+EBX]
        CMP     AL,BYTE PTR [EDI+EBX]
        JNE     @@fr
        DEC     EBX
        JNE     @@lp2
@@fd:   LEA     EAX,[EDI+1]
        SUB     EAX,[ESP]
        POP     ECX
        POP     EBX
        POP     EDI
        POP     ESI
end;

function Q_StrScan(const S: string; Ch: Char; StartPos: Integer): Integer;
asm
        TEST    EAX,EAX
        JE      @@qt
        PUSH    EDI
        MOV     EDI,EAX
        LEA     EAX,[ECX-1]
        MOV     ECX,[EDI-4]
        SUB     ECX,EAX
        JLE     @@m1
        PUSH    EDI
        ADD     EDI,EAX
        MOV     EAX,EDX
        POP     EDX
        REPNE   SCASB
        JNE     @@m1
        MOV     EAX,EDI
        SUB     EAX,EDX
        POP     EDI
        RET
@@m1:   POP     EDI
        XOR     EAX,EAX
@@qt:
end;




procedure Register;
begin
  RegisterComponents('zPage', [TzTable]);
end;


end.
